/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2005 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 * Copyright (c) 2009 Sun Microsystems, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_vendor_t (for umad).
 * This object represents the OpenIB vendor layer.
 * This object is part of the opensm family of objects.
 *
 * Environment:
 *    Linux User Mode
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#ifdef OSM_VENDOR_INTF_OPENIB

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/types.h>	/* for __be64 with older libibumad */

#include <iba/ib_types.h>
#include <complib/cl_qlist.h>
#include <complib/cl_math.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VENDOR_IBUMAD_C
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_mad_pool.h>
#include <opensm/osm_helper.h>
#include <vendor/osm_vendor_api.h>
#include <limits.h>
#include <sys/poll.h>
#include <sys/eventfd.h>

/****s* OpenSM: Vendor UMAD/osm_umad_bind_info_t
 * NAME
 *   osm_umad_bind_info_t
 *
 * DESCRIPTION
 *    Structure containing bind information.
 *
 * SYNOPSIS
 */
typedef struct _osm_umad_bind_info {
	osm_vendor_t *p_vend;
	void *client_context;
	osm_mad_pool_t *p_mad_pool;
	osm_vend_mad_recv_callback_t mad_recv_callback;
	osm_vend_mad_send_err_callback_t send_err_callback;
	ib_net64_t port_guid;
	int port_ids[OSM_UMAD_MAX_PORTS_PER_CA]; /* Index 0 is used as default */
	int agent_id[OSM_UMAD_MAX_PORTS_PER_CA];
	int agent_id1[OSM_UMAD_MAX_PORTS_PER_CA]; /* SMI requires two agents */
	int timeout;
	int max_retries;
} osm_umad_bind_info_t;

typedef struct _umad_receiver {
	pthread_t tid;
	osm_vendor_t *p_vend;
	osm_log_t *p_log;
} umad_receiver_t;

#define OSM_VENDOR_BIND_DEFAULT_PORT_IDX 0

typedef __typeof__(((struct umad_port *)0)->port_guid) umad_guid_t;

static void osm_vendor_close_port(osm_vendor_t * const p_vend);

static void log_send_error(osm_vendor_t * const p_vend, osm_madw_t *p_madw)
{
	if (p_madw->p_mad->mgmt_class != IB_MCLASS_SUBN_DIR) {
		/* LID routed */
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5410: "
			"Send completed with error (%s) -- dropping\n"
			"\t\t\tClass 0x%x, Method 0x%X, Attr 0x%X, "
			"TID 0x%" PRIx64 ", LID %u\n",
			ib_get_err_str(p_madw->status),
			p_madw->p_mad->mgmt_class, p_madw->p_mad->method,
			cl_ntoh16(p_madw->p_mad->attr_id),
			cl_ntoh64(p_madw->p_mad->trans_id),
			cl_ntoh16(p_madw->mad_addr.dest_lid));
	} else {
		ib_smp_t *p_smp;

		/* Direct routed SMP */
		p_smp = osm_madw_get_smp_ptr(p_madw);
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5411: "
			"DR SMP Send completed with error (%s) -- dropping\n"
			"\t\t\tMethod 0x%X, Attr 0x%X, TID 0x%" PRIx64 "\n",
			ib_get_err_str(p_madw->status),
			p_madw->p_mad->method,
			cl_ntoh16(p_madw->p_mad->attr_id),
			cl_ntoh64(p_madw->p_mad->trans_id));
		/*
		 * For planarized hca smi ports all have the same guid, just take the first one.
		 * For gsi there is single port
		 */
		osm_dump_smp_dr_path(p_vend->p_log, p_smp, p_vend->umad_port[0].port_guid,
				     OSM_LOG_ERROR);
	}
}

static void clear_madw(osm_vendor_t * p_vend)
{
	umad_match_t *m, *e, *old_m;
	ib_net64_t old_tid;
	uint8_t old_mgmt_class;

	OSM_LOG_ENTER(p_vend->p_log);
	pthread_mutex_lock(&p_vend->match_tbl_mutex);
	for (m = p_vend->mtbl.tbl, e = m + p_vend->mtbl.max; m < e; m++) {
		if (m->tid) {
			old_m = m;
			old_tid = m->tid;
			old_mgmt_class = m->mgmt_class;
			m->tid = 0;
			osm_mad_pool_put(((osm_umad_bind_info_t
					   *) ((osm_madw_t *) m->v)->h_bind)->
					 p_mad_pool, m->v);
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5401: "
				"evicting entry %p (tid was 0x%" PRIx64
				" mgmt class 0x%x)\n",
				old_m, cl_ntoh64(old_tid), old_mgmt_class);
		}
	}
	pthread_mutex_unlock(&p_vend->match_tbl_mutex);

	OSM_LOG_EXIT(p_vend->p_log);
}

static osm_madw_t *get_madw(osm_vendor_t * p_vend, ib_net64_t * tid,
			    uint8_t mgmt_class)
{
	umad_match_t *m, *e;
	ib_net64_t mtid = (*tid & CL_HTON64(0x00000000ffffffffULL));
	osm_madw_t *res;

	/*
	 * Since mtid == 0 is the empty key, we should not
	 * waste time looking for it
	 */
	if (mtid == 0 || mgmt_class == 0)
		return 0;

	pthread_mutex_lock(&p_vend->match_tbl_mutex);
	for (m = p_vend->mtbl.tbl, e = m + p_vend->mtbl.max; m < e; m++) {
		if (m->tid == mtid && m->mgmt_class == mgmt_class) {
			m->tid = 0;
			m->mgmt_class = 0;
			*tid = mtid;
			res = m->v;
			pthread_mutex_unlock(&p_vend->match_tbl_mutex);
			return res;
		}
	}

	pthread_mutex_unlock(&p_vend->match_tbl_mutex);
	return 0;
}

/*
 * If match table full, evict LRU (least recently used) transaction.
 * Maintain 2 LRUs: one for SMPs, and one for others (GS).
 * Evict LRU GS transaction if one is available and only evict LRU SMP
 * transaction if no other choice.
 */
static void
put_madw(osm_vendor_t * p_vend, osm_madw_t * p_madw, ib_net64_t tid,
	 uint8_t mgmt_class)
{
	umad_match_t *m, *e, *old_lru, *lru = 0, *lru_smp = 0;
	osm_madw_t *p_req_madw;
	osm_umad_bind_info_t *p_bind;
	ib_net64_t old_tid;
	uint32_t oldest = ~0, oldest_smp = ~0;
	uint8_t old_mgmt_class;

	pthread_mutex_lock(&p_vend->match_tbl_mutex);
	for (m = p_vend->mtbl.tbl, e = m + p_vend->mtbl.max; m < e; m++) {
		if (m->tid == 0 && m->mgmt_class == 0) {
			m->tid = tid;
			m->mgmt_class = mgmt_class;
			m->v = p_madw;
			m->version =
			    cl_atomic_inc((atomic32_t *) & p_vend->mtbl.
					  last_version);
			pthread_mutex_unlock(&p_vend->match_tbl_mutex);
			return;
		}
		if (m->mgmt_class == IB_MCLASS_SUBN_DIR ||
		    m->mgmt_class == IB_MCLASS_SUBN_LID) {
			if (oldest_smp >= m->version) {
				oldest_smp = m->version;
				lru_smp = m;
			}
		} else {
			if (oldest >= m->version) {
				oldest = m->version;
				lru = m;
			}
		}
	}

	if (oldest != ~0) {
		old_lru = lru;
		old_tid = lru->tid;
		old_mgmt_class = lru->mgmt_class;
	} else {
		CL_ASSERT(oldest_smp != ~0);
		old_lru = lru_smp;
		old_tid = lru_smp->tid;
		old_mgmt_class = lru_smp->mgmt_class;
	}
	p_req_madw = old_lru->v;
	p_bind = p_req_madw->h_bind;
	p_req_madw->status = IB_CANCELED;
	log_send_error(p_vend, p_req_madw);
	pthread_mutex_lock(&p_vend->cb_mutex);
	(*p_bind->send_err_callback) (p_bind->client_context, p_req_madw);
	pthread_mutex_unlock(&p_vend->cb_mutex);
	if (mgmt_class == IB_MCLASS_SUBN_DIR ||
	    mgmt_class == IB_MCLASS_SUBN_LID) {
		lru_smp->tid = tid;
		lru_smp->mgmt_class = mgmt_class;
		lru_smp->v = p_madw;
		lru_smp->version =
		    cl_atomic_inc((atomic32_t *) & p_vend->mtbl.last_version);
	} else {
		lru->tid = tid;
		lru->mgmt_class = mgmt_class;
		lru->v = p_madw;
		lru->version =
		    cl_atomic_inc((atomic32_t *) & p_vend->mtbl.last_version);
	}
	pthread_mutex_unlock(&p_vend->match_tbl_mutex);
	OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5402: "
		"evicting entry %p (tid was 0x%" PRIx64
		" mgmt class 0x%x)\n", old_lru,
		cl_ntoh64(old_tid), old_mgmt_class);
}

/*
 * The bind contains all the ports that make up the aport.
 * Returns the index the bind uses to represent the port numbered port_num in the aport.
 */
static uint8_t osm_vendor_get_port_idx(osm_umad_bind_info_t * p_bind, uint8_t port_num)
{
	osm_vendor_t *p_vend;

	p_vend = p_bind->p_vend;
	return p_vend->port_num_to_idx[port_num];
}

static void
ib_mad_addr_conv(ib_user_mad_t * umad, osm_mad_addr_t * osm_mad_addr,
		 uint8_t port_num, int is_smi)
{
	memset(osm_mad_addr, 0, sizeof(osm_mad_addr_t));
	ib_mad_addr_t *ib_mad_addr = umad_get_mad_addr(umad);
	osm_mad_addr->dest_lid = ib_mad_addr->lid;
	osm_mad_addr->path_bits = ib_mad_addr->path_bits;

	if (is_smi) {
		osm_mad_addr->addr_type.smi.source_lid = osm_mad_addr->dest_lid;
		osm_mad_addr->addr_type.smi.port_num = port_num;
		return;
	}

	osm_mad_addr->addr_type.gsi.remote_qp = ib_mad_addr->qpn;
	osm_mad_addr->addr_type.gsi.remote_qkey = ib_mad_addr->qkey;
	osm_mad_addr->addr_type.gsi.pkey_ix = umad_get_pkey(umad);
	osm_mad_addr->addr_type.gsi.service_level = ib_mad_addr->sl;

	if (ib_mad_addr->grh_present) {
		osm_mad_addr->addr_type.gsi.global_route = 1;
		osm_mad_addr->addr_type.gsi.grh_info.hop_limit =
			ib_mad_addr->hop_limit;
		osm_mad_addr->addr_type.gsi.grh_info.ver_class_flow =
			ib_grh_set_ver_class_flow(6,    /* GRH version */
					ib_mad_addr->traffic_class,
					ib_mad_addr->flow_label);
		memcpy(&osm_mad_addr->addr_type.gsi.grh_info.dest_gid,
				&ib_mad_addr->gid, 16);
	}
}

static void *swap_mad_bufs(osm_madw_t * p_madw, void *umad)
{
	void *old;

	old = p_madw->vend_wrap.umad;
	p_madw->vend_wrap.umad = umad;
	p_madw->p_mad = umad_get_mad(umad);

	return old;
}

static void unlock_mutex(void *arg)
{
	pthread_mutex_unlock(arg);
}

static void umad_receiver_on_port(umad_receiver_t *p_ur,
				  int index,
				  void **umad)
{
	osm_vendor_t *p_vend = p_ur->p_vend;
	osm_umad_bind_info_t *p_bind;
	osm_mad_addr_t osm_addr;
	osm_madw_t *p_madw, *p_req_madw;
	ib_mad_t *p_mad, *p_req_mad;
	ib_mad_addr_t *p_mad_addr;
	int mad_agent, length;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	length = MAD_BLOCK_SIZE;
	if ((mad_agent = umad_recv(p_vend->umad_port_id[index], *umad, &length, -1)) < 0) {
		if (length <= MAD_BLOCK_SIZE) {
			OSM_LOG(p_ur->p_log, OSM_LOG_ERROR, "ERR 5404: "
				"recv error on MAD sized umad (%m)\n");
			return;
		} else {
			umad_free(*umad);
			/* Need a larger buffer for RMPP */
			*umad = umad_alloc(1, umad_size() + length);
			if (!*umad) {
				OSM_LOG(p_ur->p_log, OSM_LOG_ERROR,
					"ERR 5405: can't alloc umad length %d\n", length);
				return;
			}

			if ((mad_agent =
				umad_recv(p_vend->umad_port_id[index], *umad, &length, -1)) < 0) {
				OSM_LOG(p_ur->p_log, OSM_LOG_ERROR,
					"ERR 5406: recv error on umad length %d (%m)\n", length);
				return;
			}
		}
	}
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
	pthread_mutex_lock(&p_vend->agents_mutex);
	if (mad_agent >= OSM_UMAD_MAX_AGENTS ||	!(p_bind = p_vend->agents[index][mad_agent])) {
		pthread_mutex_unlock(&p_vend->agents_mutex);
		OSM_LOG(p_ur->p_log, OSM_LOG_ERROR, "ERR 5407: invalid mad agent %d - dropping\n",
			mad_agent);
		return;
	}

	pthread_mutex_unlock(&p_vend->agents_mutex);

	p_mad = (ib_mad_t *) umad_get_mad(*umad);
	p_mad_addr = umad_get_mad_addr(*umad);

	OSM_LOG(p_ur->p_log, OSM_LOG_DEBUG, "class 0x%x GRH present: %d\n",
		p_mad->mgmt_class, p_mad_addr->grh_present);

	ib_mad_addr_conv(*umad, &osm_addr, p_vend->umad_port[index].portnum,
			p_mad->mgmt_class == IB_MCLASS_SUBN_LID ||
			p_mad->mgmt_class == IB_MCLASS_SUBN_DIR);

	if (!(p_madw = osm_mad_pool_get(p_bind->p_mad_pool,
					(osm_bind_handle_t) p_bind,
					MAX(length, MAD_BLOCK_SIZE),
					&osm_addr))) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5408: "
			"request for a new madw failed -- dropping packet\n");
		return;
	}

	/* Need to fix up MAD size if short RMPP packet */
	if (length < MAD_BLOCK_SIZE)
		p_madw->mad_size = length;

	/*
	 * Avoid copying by swapping mad buf pointers.
	 * Do not use umad after this line of code.
	 */
	*umad = swap_mad_bufs(p_madw, *umad);

	/* if status != 0 then we are handling recv timeout on send */
	if (umad_status(p_madw->vend_wrap.umad)) {
		if (!(p_req_madw = get_madw(p_vend, &p_mad->trans_id, p_mad->mgmt_class))) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5412: "
				"Failed to obtain request madw for timed out MAD"
				" (class=0x%X method=0x%X attr=0x%X tid=0x%"PRIx64") -- dropping\n",
				p_mad->mgmt_class, p_mad->method,
				cl_ntoh16(p_mad->attr_id),
				cl_ntoh64(p_mad->trans_id));
		} else {
			p_req_madw->status = IB_TIMEOUT;
			log_send_error(p_vend, p_req_madw);
			/* cb frees req_madw */
			pthread_mutex_lock(&p_vend->cb_mutex);
			pthread_cleanup_push(unlock_mutex, &p_vend->cb_mutex);
			(*p_bind->send_err_callback) (p_bind->client_context, p_req_madw);
			pthread_cleanup_pop(1);
		}

		osm_mad_pool_put(p_bind->p_mad_pool, p_madw);
		return;
	}

	p_req_madw = 0;
	if (ib_mad_is_response(p_mad)) {
		p_req_madw = get_madw(p_vend, &p_mad->trans_id,	p_mad->mgmt_class);
		if (PF(!p_req_madw)) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR,
				"ERR 5413: Failed to obtain request madw for received MAD "
				"(class=0x%X method=0x%X attr=0x%X tid=0x%"PRIx64") -- dropping\n",
				p_mad->mgmt_class, p_mad->method, cl_ntoh16(p_mad->attr_id),
				cl_ntoh64(p_mad->trans_id));
			/*
			 * No need to log sender error - this info can
			 * be extracted from the log message above.
			 * We do need to retire the transaction, but
			 * we don't have the request handler.
			 * Bummer!
			 * Just return the response MAD to pool.
			 */
			osm_mad_pool_put(p_bind->p_mad_pool, p_madw);
			return;
		}

		/*
		 * Make sure that attribute ID, attribute
		 * modifier, and transaction ID are the same in
		 * request and response.
		 *
		 * Exception for o15-0.2-1.11:
		 * SA response to a SubnAdmGetMulti() containing a
		 * MultiPathRecord shall have PathRecord attribute ID.
		 */
		p_req_mad = osm_madw_get_mad_ptr(p_req_madw);
		if (PF((p_mad->attr_id != p_req_mad->attr_id &&
		       !(p_mad->mgmt_class == IB_MCLASS_SUBN_ADM &&
			 p_req_mad->attr_id == IB_MAD_ATTR_MULTIPATH_RECORD &&
			 p_mad->attr_id == IB_MAD_ATTR_PATH_RECORD)) ||
			p_mad->attr_mod != p_req_mad->attr_mod ||
			p_mad->trans_id != p_req_mad->trans_id)) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 541A: "
				"Response MAD validation failed (request attr=0x%X modif=0x%X "
				"tid=0x%"PRIx64", response attr=0x%X modif=0x%X "
				"tid=0x%"PRIx64") -- dropping\n",
				cl_ntoh16(p_req_mad->attr_id), cl_ntoh32(p_req_mad->attr_mod),
				cl_ntoh64(p_req_mad->trans_id),	cl_ntoh16(p_mad->attr_id),
				cl_ntoh32(p_mad->attr_mod), cl_ntoh64(p_mad->trans_id));
			/*
			 * No need to log sender error - this info can
			 * be extracted from the log message above.
			 * What we do need to do is to call error
			 * callback to retire the transaction, and
			 * then to return the response MAD to pool.
			 */
			p_req_madw->status = IB_REMOTE_ERROR;
			pthread_mutex_lock(&p_vend->cb_mutex);
			pthread_cleanup_push(unlock_mutex, &p_vend->cb_mutex);
			(*p_bind->send_err_callback) (p_bind->client_context, p_req_madw);
			pthread_cleanup_pop(1);

			osm_mad_pool_put(p_bind->p_mad_pool, p_madw);
			return;
		}
	}

#ifndef VENDOR_RMPP_SUPPORT
	if ((p_mad->mgmt_class != IB_MCLASS_SUBN_DIR) &&
	    (p_mad->mgmt_class != IB_MCLASS_SUBN_LID) &&
	    (ib_rmpp_is_flag_set((ib_rmpp_mad_t *) p_mad, IB_RMPP_FLAG_ACTIVE))) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5414: "
			"class 0x%x method 0x%x RMPP version %d type "
			"%d flags 0x%x received -- dropping\n",
			p_mad->mgmt_class, p_mad->method,
			((ib_rmpp_mad_t *) p_mad)->rmpp_version,
			((ib_rmpp_mad_t *) p_mad)->rmpp_type,
			((ib_rmpp_mad_t *) p_mad)->rmpp_flags);
		osm_mad_pool_put(p_bind->p_mad_pool, p_madw);
		return;
	}
#endif

	/* call the CB */
	pthread_mutex_lock(&p_vend->cb_mutex);
	pthread_cleanup_push(unlock_mutex, &p_vend->cb_mutex);
	(*p_bind->mad_recv_callback) (p_madw, p_bind->client_context, p_req_madw);
	pthread_cleanup_pop(1);
}

static void *umad_receiver(void *p_ptr)
{
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
	umad_receiver_t *const p_ur = (umad_receiver_t *) p_ptr;
	osm_vendor_t *p_vend = p_ur->p_vend;
	void *umad = NULL;
	uint32_t i = 0;
	int ready;

	struct pollfd pfds[OSM_UMAD_MAX_PORTS_PER_CA + 1];

	OSM_LOG_ENTER(p_ur->p_log);

	pfds[0].fd = p_vend->eventfd;
	pfds[0].events = POLLIN;
	pfds[0].revents = 0;

	for (i = 0; i < p_vend->umad_ports_num; i++) {
		pfds[i + 1].fd = umad_get_fd(p_vend->umad_port_id[i]);
		pfds[i + 1].events = POLLIN;
		pfds[i + 1].revents = 0;
	}

	umad = umad_alloc(1, umad_size() + MAD_BLOCK_SIZE);
	if (!umad) {
		OSM_LOG(p_ur->p_log, OSM_LOG_ERROR, "ERR 5403: can't alloc MAD sized umad\n");
		goto Exit;
	}

	for (;;) {
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
		ready = poll(pfds, (p_vend->umad_ports_num + 1), -1);
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
		if (ready < 0) {
			OSM_LOG(p_ur->p_log, OSM_LOG_ERROR, "ERR 5437: "
				"poll failed %d (%d - %s)\n", ready, errno, strerror(errno));
			goto Exit;
		}

		/* check exit signal */
		if (pfds[0].revents & POLLIN) {
			OSM_LOG(p_ur->p_log, OSM_LOG_DEBUG,
				"Received message on event fd, exiting.\n");
			goto Exit;
		}

		for (i = 1; i <= p_vend->umad_ports_num; i++) {
			if (pfds[i].revents & POLLIN) {
				umad_receiver_on_port(p_ur, i - 1, &umad);
				pfds[i].revents = 0;
			}
		}
	}

Exit:
	/*
	 * For future happy debug, note that umad_receiver_on_port swaps between
	 * the original allocated buffer to an internal madw allocated, hence
	 * addresses are different between here and above allocation.
	 */
	if (umad)
		umad_free(umad);

	OSM_LOG_EXIT(p_ur->p_log);
	return NULL;
}

static int umad_receiver_start(osm_vendor_t * p_vend)
{
	umad_receiver_t *p_ur = p_vend->receiver;

	p_ur->p_vend = p_vend;
	p_ur->p_log = p_vend->p_log;

	if (pthread_create(&p_ur->tid, NULL, umad_receiver, p_ur) != 0)
		return -1;

	return 0;
}

static void umad_receiver_stop(umad_receiver_t * p_ur)
{
        uint64_t event = 1;
        write(p_ur->p_vend->eventfd, &event, sizeof(event));
	pthread_join(p_ur->tid, NULL);
	close(p_ur->p_vend->eventfd);
	p_ur->tid = 0;
	p_ur->p_vend = NULL;
	p_ur->p_log = NULL;
}

ib_api_status_t
osm_vendor_init(IN osm_vendor_t * const p_vend,
		IN osm_log_t * const p_log, IN const uint32_t timeout)
{
	char *max = NULL;
	int r, n_cas, i;

	OSM_LOG_ENTER(p_log);

	p_vend->p_log = p_log;
	p_vend->timeout = timeout;
	p_vend->max_retries = OSM_DEFAULT_RETRY_COUNT;
	pthread_mutex_init(&p_vend->cb_mutex, NULL);
	pthread_mutex_init(&p_vend->match_tbl_mutex, NULL);
	pthread_mutex_init(&p_vend->agents_mutex, NULL);
	for (i = 0; i < OSM_UMAD_MAX_PORTS_PER_CA; i++) {
		p_vend->umad_port_id[i] = -1;
	}
	p_vend->issmfd = -1;

	/*
	 * Open our instance of UMAD.
	 */
	if ((r = umad_init()) < 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR,
			"ERR 5415: Error opening UMAD\n");
	}

	if ((n_cas = umad_get_cas_names(p_vend->ca_names,
					OSM_UMAD_MAX_CAS)) < 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR,
			"ERR 5416: umad_get_cas_names failed\n");
		r = n_cas;
		goto Exit;
	}

	p_vend->ca_count = n_cas;
	p_vend->mtbl.max = DEFAULT_OSM_UMAD_MAX_PENDING;

	if ((max = getenv("OSM_UMAD_MAX_PENDING")) != NULL) {
		long tmp = strtol(max, NULL, 0);
		if (tmp > 0 && tmp <= INT_MAX)
			p_vend->mtbl.max = (int)tmp;
		else
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "Error:"
				"OSM_UMAD_MAX_PENDING=%ld is invalid\n",
				tmp);
	}

	OSM_LOG(p_vend->p_log, OSM_LOG_INFO, "%d pending umads specified\n",
		p_vend->mtbl.max);

	p_vend->mtbl.tbl = calloc(p_vend->mtbl.max, sizeof(*(p_vend->mtbl.tbl)));
	if (!p_vend->mtbl.tbl) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "Error:"
			"failed to allocate vendor match table\n");
		r = IB_INSUFFICIENT_MEMORY;
		goto Exit;
	}

Exit:
	OSM_LOG_EXIT(p_log);
	return (r);
}

osm_vendor_t *osm_vendor_new(IN osm_log_t * const p_log,
			     IN const uint32_t timeout)
{
	osm_vendor_t *p_vend = NULL;

	OSM_LOG_ENTER(p_log);

	if (!timeout) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 5433: "
			"transaction timeout cannot be 0\n");
		goto Exit;
	}

	p_vend = malloc(sizeof(*p_vend));
	if (p_vend == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 5417: "
			"Unable to allocate vendor object\n");
		goto Exit;
	}

	memset(p_vend, 0, sizeof(*p_vend));

	if (osm_vendor_init(p_vend, p_log, timeout) != IB_SUCCESS) {
		free(p_vend);
		p_vend = NULL;
	}

Exit:
	OSM_LOG_EXIT(p_log);
	return (p_vend);
}

osm_vendor_t *osm_vendor_on_port_new(IN osm_log_t * const p_log,
				     IN const uint32_t timeout,
				     IN const uint8_t port_index)
{
	osm_vendor_t *p_vend = NULL;

	OSM_LOG_ENTER(p_log);

	p_vend = osm_vendor_new(p_log, timeout);
	if (p_vend) {
		p_vend->port_index = port_index;
	}

	OSM_LOG_EXIT(p_log);
	return (p_vend);
}

uint8_t osm_vendor_get_binding_port_index(IN osm_bind_handle_t *p_bind)
{
	osm_vendor_t *p_vendor;

	p_vendor = ((osm_umad_bind_info_t *)p_bind)->p_vend;
	OSM_ASSERT(p_vendor);

	return p_vendor->port_index;
}

ib_net64_t osm_vendor_get_binding_port_guid(IN osm_bind_handle_t *h_bind)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) h_bind;

	return p_bind->port_guid;
}

void osm_vendor_close_ports(IN osm_vendor_t ** const pp_vend)
{
	osm_vendor_close_port(*pp_vend);

	clear_madw(*pp_vend);
	/* make sure all ports are closed */
	umad_done();
}

void osm_vendor_delete(IN osm_vendor_t ** const pp_vend)
{
	osm_vendor_close_ports(pp_vend);

	pthread_mutex_destroy(&(*pp_vend)->agents_mutex);
	pthread_mutex_destroy(&(*pp_vend)->cb_mutex);
	pthread_mutex_destroy(&(*pp_vend)->match_tbl_mutex);
	free((*pp_vend)->mtbl.tbl);
	free(*pp_vend);
	*pp_vend = NULL;
}

ib_api_status_t
osm_vendor_get_port_attr(IN osm_vendor_t * const p_vend,
			 IN ib_port_attr_t * const p_attr) {
	umad_port_t umad_port;
	ib_port_attr_t *attr = p_attr;
	int r = 0;

	OSM_LOG_ENTER(p_vend->p_log);

	/*
	 * For planarized hca smi ports all have the same ca name, just take the first one.
	 * Assume they all have the same state
	 * For gsi there is single port
	 */
	if (umad_get_port(p_vend->umad_port[0].ca_name, p_vend->umad_port[0].portnum, &umad_port) < 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5432: "
			"umad_get_port() failed\n");
		goto Exit;
	}

	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
		"Port GUID: 0x%" PRIx64 ", physical state: %u, logical state: %u\n",
		umad_port.port_guid, umad_port.phys_state, umad_port.state);

	attr->link_state = umad_port.state;

	umad_release_port(&umad_port);
Exit:
	OSM_LOG_EXIT(p_vend->p_log);
	return r;
}

ib_api_status_t
osm_vendor_get_all_port_attr(IN osm_vendor_t * const p_vend,
			     IN ib_port_attr_t * const p_attr_array,
			     IN uint32_t * const p_num_ports)
{
	umad_ca_t ca;
	ib_port_attr_t *attr = p_attr_array;
	unsigned done = 0;
	int r = 0, i, j, k;

	OSM_LOG_ENTER(p_vend->p_log);

	OSM_ASSERT(p_vend && p_num_ports);

	if (!*p_num_ports) {
		r = IB_INVALID_PARAMETER;
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5418: "
			"Ports in should be > 0\n");
		goto Exit;
	}

	if (!p_attr_array) {
		r = IB_INSUFFICIENT_MEMORY;
		*p_num_ports = 0;
		goto Exit;
	}

	for (i = 0; i < p_vend->ca_count && !done; i++) {
		/* For each CA, retrieve the port attributes */
		if (umad_get_ca(p_vend->ca_names[i], &ca) == 0) {
			if (ca.node_type < 1 || ca.node_type > 3)
				continue;
			for (j = 0; j <= ca.numports; j++) {
				if (!ca.ports[j])
					continue;
#ifdef HAVE_UMAD_PORT_LINK_LAYER
				if (strcmp(ca.ports[j]->link_layer, "InfiniBand") &&
				    strcmp(ca.ports[j]->link_layer, "IB"))
					continue;
#endif
				attr->port_guid = ca.ports[j]->port_guid;
				attr->lid = ca.ports[j]->base_lid;
				attr->port_num = ca.ports[j]->portnum;
				attr->sm_lid = ca.ports[j]->sm_lid;
				attr->link_state = ca.ports[j]->state;
				if (attr->num_pkeys && attr->p_pkey_table) {
					if (attr->num_pkeys > ca.ports[j]->pkeys_size)
						attr->num_pkeys = ca.ports[j]->pkeys_size;
					for (k = 0; k < attr->num_pkeys; k++)
						attr->p_pkey_table[k] =
							cl_hton16(ca.ports[j]->pkeys[k]);
				}
				attr->num_pkeys = ca.ports[j]->pkeys_size;
				if (attr->num_gids && attr->p_gid_table) {
					attr->p_gid_table[0].unicast.prefix = cl_hton64(ca.ports[j]->gid_prefix);
					attr->p_gid_table[0].unicast.interface_id = cl_hton64(ca.ports[j]->port_guid);
					attr->num_gids = 1;
				}
				attr++;
				if (attr - p_attr_array > *p_num_ports) {
					done = 1;
					break;
				}
			}
			umad_release_ca(&ca);
		}
	}

	*p_num_ports = attr - p_attr_array;

Exit:
	OSM_LOG_EXIT(p_vend->p_log);
	return r;
}

ib_net64_t osm_vendor_get_node_guid(IN osm_vendor_t * const p_vend)
{
	umad_ca_t umad_ca;
	ib_net64_t node_guid = 0;

	if (umad_get_ca(p_vend->umad_port[0].ca_name, &umad_ca) < 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 542F: "
			"umad_get_ca() failed\n");
		goto Exit;
	}

	node_guid = umad_ca.node_guid;
	umad_release_ca(&umad_ca);

Exit:
	return node_guid;
}

static int
osm_vendor_resolve_ca_port(IN osm_vendor_t * const p_vend, IN char * ca_name, OUT int * port)
{
	umad_ca_t ca;
	int active = -1, up = -1;
	int i, ret = 0;

	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "checking ca '%s'\n", ca_name);

	if (umad_get_ca(ca_name, &ca) < 0)
		return -1;

	if (ca.node_type > 3)
		goto Exit;

	if (ca.node_type == 2) {
		*port = 0;      /* switch sma port 0 */
		ret = 1;
		goto Exit;
	}

	if (*port > 0) {        /* check only the port the user wants */
		if (*port > ca.numports) {
			ret = -1;
			goto Exit;
		}
		if (!ca.ports[*port]) {
			ret = -1;
			goto Exit;
		}
		if (strcmp(ca.ports[*port]->link_layer, "InfiniBand") &&
				strcmp(ca.ports[*port]->link_layer, "IB")) {
			ret = -1;
			goto Exit;
		}
		if (ca.ports[*port]->state == 4) {
			ret = 1;
			goto Exit;
		}
		if (ca.ports[*port]->phys_state != 3)
			goto Exit;
		ret = -1;
		goto Exit;
	}

	for (i = 0; i <= ca.numports; i++) {
		if (!ca.ports[i])
			continue;

		OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "checking port %d port num %u\n", i, ca.ports[i]->portnum);

		if (strcmp(ca.ports[i]->link_layer, "InfiniBand") &&
				strcmp(ca.ports[i]->link_layer, "IB"))
			continue;

		if (up < 0 && ca.ports[i]->phys_state == 5)
			up = *port = i;

		if (ca.ports[i]->state == 4) {
			active = *port = i;
			OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "found active port %d\n", i);
			break;
		}
	}

	if (active == -1 && up == -1) { /* no active or linkup port found */
		for (i = 0; i <= ca.numports; i++) {
			OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "checking port %d\n", i);
			if (!ca.ports[i])
				continue;
			if (ca.ports[i]->phys_state != 3) {
				up = *port = i;
				break;
			}
		}
	}

	if (active >= 0) {
		ret = 1;
		goto Exit;
	}
	if (up >= 0) {
		ret = 0;
		goto Exit;
	}
	ret = -1;
Exit:
	umad_release_ca(&ca);
	return ret;
}

#define OSM_VENDOR_PAIRED_DEVICE_NUMBER		2

typedef struct osm_vendor_ca_data {
	cl_map_item_t map_item;
	ib_net64_t guid;
	/* mark if this data represents valid data */
	boolean_t is_valid;
	/* filled with the first found CAs that have the same port GUID */
	char ca_names[OSM_VENDOR_PAIRED_DEVICE_NUMBER][UMAD_CA_NAME_LEN];
	int ca_ports[OSM_VENDOR_PAIRED_DEVICE_NUMBER][IB_MAX_NUM_OF_PLANES_XDR];
	uint8_t num_found_ports_per_ca[OSM_VENDOR_PAIRED_DEVICE_NUMBER];
	/* non zero if some ports of CA at this index have the 'is sm disabled' bit set */
	int is_sm_disabled[OSM_VENDOR_PAIRED_DEVICE_NUMBER];
	uint8_t num_found_cas; /* with the same guid */
	uint8_t num_found_ports; /* with the same guid */
} osm_vendor_ca_data_t;

static osm_vendor_ca_data_t *
get_ca_data_by_guid(IN const cl_qmap_t * const p_map, IN ib_net64_t guid) {
	osm_vendor_ca_data_t *p_ca_data;

	p_ca_data = (osm_vendor_ca_data_t *) cl_qmap_get(p_map, guid);
	if (p_ca_data == (osm_vendor_ca_data_t *) cl_qmap_end(p_map))
		p_ca_data = NULL;
	return p_ca_data;
}

static void delete_ca_data(osm_vendor_ca_data_t **ca_data) {
	free(*ca_data);
	*ca_data = NULL;
}

static osm_vendor_ca_data_t *
add_ca_data_by_guid(IN osm_vendor_t * const p_vend,
		    IN OUT cl_qmap_t * const p_map,
		    IN const umad_port_t * const p_umad_port,
		    IN char * ca_name) {
	osm_vendor_ca_data_t *ca_data, *ca_data_check;

	ca_data = (osm_vendor_ca_data_t *)calloc(1, sizeof(osm_vendor_ca_data_t));

	if (!ca_data) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5439: "
			"Unable to alloc vendor ca data struct\n");
		return ca_data;
	}

	ca_data->is_valid = TRUE;
	ca_data->guid = p_umad_port->port_guid;
	strncpy(ca_data->ca_names[0], ca_name, (sizeof(ca_data->ca_names[0]) - 1));
	ca_data->ca_ports[0][0] = p_umad_port->portnum;
	ca_data->num_found_ports_per_ca[0] ++;
	ca_data->is_sm_disabled[0] |= p_umad_port->capmask & IB_PORT_CAP_SM_DISAB;
	ca_data->num_found_cas ++;
	ca_data->num_found_ports ++;

	ca_data_check =
		(osm_vendor_ca_data_t *) cl_qmap_insert(p_map,
				p_umad_port->port_guid, &ca_data->map_item);
	if (PF(ca_data_check != ca_data)) {
		/*
		 * We should never be here! Somehow, this port GUID already exists in the table.
		 */
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 543A: "
			"Failed to add GUID 0x" PRIx64 " to the ca data map, "
			"an entry with such GUID already exists.\n");
		delete_ca_data(&ca_data);
	}

	return ca_data;

}

/*
 * This function checks that the given number of found smi ports is valid number of planes.
 * Currently the supported number of planes is 2 and 4.
 * In the future we might expand it.
 */
static inline boolean_t
is_valid_number_of_smi_ports(uint8_t smi_num_ports) {
	return (smi_num_ports == IB_MIN_NUM_OF_PLANES_XDR ||
		smi_num_ports == IB_MAX_NUM_OF_PLANES_XDR);
}

static int
resolve_umad_port_from_selected_cas(IN OUT osm_vendor_t * const p_vend,
				    IN cl_qmap_t *ca_data_map,
				    boolean_t is_gsi)
{
	cl_map_item_t *p_map_item;
	osm_vendor_ca_data_t *ca_data;
	char *name;
	boolean_t is_planarized, up_port_found = FALSE, is_valid_smi_ports = FALSE;
	uint8_t sm_disabled_num;
	int smi_interface_idx, gsi_interface_idx;
	int i, j, port, port_type, prev_port_type, gsi_port_type, selected_interface_index = -1;
	int selected_ports[IB_MAX_NUM_OF_PLANES_XDR] = {0}, num_ports_to_open = 0;
	char selected_ca[UMAD_CA_NAME_LEN] = {0};
	uint8_t smi_found_ports_num;
	int ret = -1;

	for (p_map_item = cl_qmap_head(ca_data_map); p_map_item != cl_qmap_end(ca_data_map);
	     p_map_item = cl_qmap_next(p_map_item)) {
		sm_disabled_num = 0;
		port_type = smi_interface_idx = gsi_interface_idx = -1;
		ca_data = (osm_vendor_ca_data_t *)p_map_item;

		num_ports_to_open = 0;

		if (!ca_data->is_valid) {
			continue;
		}

		/* we need to merge 2 processes of getting umad_ca for type check and umad_get_port */
		for (i = 0; i < ca_data->num_found_cas; i++) {
			name = ca_data->ca_names[i];
			OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
				"inspect ca_data %u, name %s, guid 0x%" PRIx64 " is_valid %u "
				"is_sm_disabled %u num_cas %u\n",
				i, name, cl_ntoh64(ca_data->guid), ca_data->is_valid,
				ca_data->is_sm_disabled[i], ca_data->num_found_cas);

			if (ca_data->is_sm_disabled[i]) {
				sm_disabled_num ++;
				gsi_interface_idx = i;
			} else if (smi_interface_idx == -1) {
				smi_interface_idx = i;
			}
		}

		smi_found_ports_num = smi_interface_idx < 0 ?
			0 : ca_data->num_found_ports_per_ca[smi_interface_idx];
		is_valid_smi_ports = is_valid_number_of_smi_ports(smi_found_ports_num);

		if (ca_data->num_found_cas == OSM_VENDOR_PAIRED_DEVICE_NUMBER) {
			if (sm_disabled_num == 1 &&
			    smi_interface_idx >= 0 && gsi_interface_idx >=0 &&
			    is_valid_smi_ports &&
			    ca_data->num_found_ports_per_ca[gsi_interface_idx] == 1) {
				is_planarized = TRUE;
			} else {
				OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5436: "
					"interface %s has duplicate port GUID with interface %s\n",
					ca_data->ca_names[0], ca_data->ca_names[1]);
				/* go to next ca_data */
				continue;
			}
		} else {
			/* assume just one interface found */
			is_planarized = FALSE;
			port = 0;
		}

		if (is_planarized) {
			prev_port_type = -1;
			for (j = 0; j < smi_found_ports_num; j++) {
				/* try to resolve specific port */
				port_type = osm_vendor_resolve_ca_port(p_vend,
						ca_data->ca_names[smi_interface_idx],
						&ca_data->ca_ports[smi_interface_idx][j]);
				/* all plane ports have to have the same state */
				if (port_type < 0 || (j && port_type != prev_port_type))
					/* continue to next ca_data */
					break;
				prev_port_type = port_type;
			}
			/* if not all plane ports have the same state, skip to next interface */
			if (j != smi_found_ports_num)
				continue;

			gsi_port_type = osm_vendor_resolve_ca_port(p_vend,
						ca_data->ca_names[gsi_interface_idx],
						&ca_data->ca_ports[gsi_interface_idx][0]);
			/* gsi port should have the same state as smi interface's ports */
			if (port_type != gsi_port_type)
				OSM_LOG(p_vend->p_log, OSM_LOG_INFO,
					"gsi port state (%d) is different "
					"than smi port state (%d).\n", gsi_port_type, port_type);

		} else {
			/*
			 * find suitable port (has to have is_sm_disabled 0)
			 * to open on the given interface.
			 */
			if (smi_interface_idx < 0 ||
			    (port_type = osm_vendor_resolve_ca_port(
							p_vend,
							ca_data->ca_names[smi_interface_idx],
							&port)) < 0)
				/* continue to next ca_data */
				continue;
		}

		/* remember active or up port */
		if (port_type > 0 || !up_port_found) {
			if (is_planarized) {
				if (is_gsi) {
					selected_interface_index = gsi_interface_idx;
					num_ports_to_open = 1;
				} else {
					selected_interface_index = smi_interface_idx;
					num_ports_to_open = smi_found_ports_num;
				}
			} else {
				selected_interface_index = smi_interface_idx;
				num_ports_to_open = 1;
			}

			strncpy(selected_ca, ca_data->ca_names[selected_interface_index],
				(sizeof(selected_ca) - 1));
			memcpy(selected_ports, ca_data->ca_ports[selected_interface_index],
			       num_ports_to_open * sizeof(int));
		}

		if (port_type > 0)
			break;

		/* remember first up port */
		if (!up_port_found)
			up_port_found = TRUE;
	}

	/* Port found, try to get umad port */
	for (i = 0; i < num_ports_to_open; i++) {
		if (umad_get_port(selected_ca, selected_ports[i], &p_vend->umad_port[i]) < 0) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 542B: "
				"umad_get_port() failed for ca %s port %d\n",
				selected_ca, selected_ports[i]);
			break;
		}
	}

	/* successfully opened all ports */
	if (num_ports_to_open && i == num_ports_to_open) {
		p_vend->umad_ports_num = num_ports_to_open;
		ret = 0;
	}

	return ret;
}

/* need to implement here some copy of resolve_ca_name */
static int
resolve_umad_port(IN OUT osm_vendor_t * const p_vend, ib_net64_t req_guid, boolean_t is_gsi)
{
	int caidx, i, j;
	char *name;
	osm_vendor_ca_data_t *ca_data;
	cl_qmap_t ca_data_map;
	umad_ca_t umad_ca;
	ib_net64_t current_guid;
	cl_map_item_t *p_map_item;
	int ret = -1;

	memset(&umad_ca, 0, sizeof(umad_ca));
	cl_qmap_init(&ca_data_map);

	/* First find pairs of ca interface that have the same GUID */
	for (caidx = 0; caidx < p_vend->ca_count; caidx++) {
		name = p_vend->ca_names[caidx];
		if (umad_get_ca(name, &umad_ca) < 0) {
			OSM_LOG(p_vend->p_log, OSM_LOG_VERBOSE,
				"umad_get_ca() failed for ca name %s\n", name);
			continue;
		}

		/* wrong node type, continue to next interface */
		if (umad_ca.node_type > 3) {
			goto Release_umad;
		}

		for (i = 0; i <= umad_ca.numports; i++) {
			OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "Inspect interface %s port idx %d %p\n",
				name, i, umad_ca.ports[i]);
			if (!umad_ca.ports[i])
				continue;

			current_guid = umad_ca.ports[i]->port_guid;
			/* if user has given guid, select only interfaces with this guid */
			if (req_guid && current_guid != req_guid)
				continue;

			ca_data = get_ca_data_by_guid(&ca_data_map, current_guid);
			if (ca_data && ca_data->is_valid) {
				OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
					"valid ca_data found with guid 0x%" PRIx64 "\n",
					cl_ntoh64(current_guid));

				if (ca_data->num_found_ports == (IB_MAX_NUM_OF_PLANES_XDR + 1)) {
					ca_data->is_valid = FALSE;
					OSM_LOG(p_vend->p_log, OSM_LOG_ERROR,
						"ERR 543E: found ca %s port %u with GUID 0x%"
						PRIx64 ", that is already found on %u other ports."
						" Suspected duplicate GUIDs.\n",
						name, i, cl_ntoh64(current_guid),
						ca_data->num_found_ports);
					continue;
				}

				for (j = 0; j < OSM_VENDOR_PAIRED_DEVICE_NUMBER; j++) {

					int num_ports_per_ca = ca_data->num_found_ports_per_ca[j];
					/* found port of already known CA */
					if (!strncmp(ca_data->ca_names[j], name,
						     UMAD_CA_NAME_LEN)) {
						/* found extra port, duplicate GUID ? */
						if (num_ports_per_ca >= IB_MAX_NUM_OF_PLANES_XDR) {
							OSM_LOG(p_vend->p_log, OSM_LOG_ERROR,
								"ERR 543C: ca %s has %u ports "
								"with GUID 0x%" PRIx64
								" more than %u xdr planes. "
								"Suspected duplicate GUIDs.\n",
								name, (num_ports_per_ca + 1),
								cl_ntoh64(current_guid),
								IB_MAX_NUM_OF_PLANES_XDR);
							ca_data->is_valid = FALSE;
							break;
						}

						ca_data->ca_ports[j][num_ports_per_ca] =
							umad_ca.ports[i]->portnum;

						ca_data->num_found_ports_per_ca[j]++;
						ca_data->is_sm_disabled[j] |=
							umad_ca.ports[i]->capmask & IB_PORT_CAP_SM_DISAB;
						OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
							"known ca %s 0x%" PRIx64 " cap 0x%x is_sm_disabled %u"
							" is_sm_disabled[%u]=0x%x\n",
							name, cl_ntoh64(req_guid),
							umad_ca.ports[i]->capmask,
							umad_ca.ports[i]->capmask & IB_PORT_CAP_SM_DISAB,
							j, cl_ntoh32(ca_data->is_sm_disabled[j]));
						ca_data->num_found_ports++;
						/* go to next guid */
						break;
					}
				}

				int num_cas = ca_data->num_found_cas;
				if (j == OSM_VENDOR_PAIRED_DEVICE_NUMBER) {
					/*
					 * New CA with already known GUID and
					 * the expected number of CAs hasn't been found yet.
					 */
					if (num_cas < OSM_VENDOR_PAIRED_DEVICE_NUMBER) {
						/* copy to next available place */
						strncpy(ca_data->ca_names[num_cas], name,
							(sizeof(ca_data->ca_names[num_cas]) - 1));
						ca_data->ca_ports[num_cas][0] =
							umad_ca.ports[i]->portnum;
						ca_data->num_found_ports_per_ca[num_cas] = 1;
						ca_data->is_sm_disabled[num_cas] |=
							umad_ca.ports[i]->capmask & IB_PORT_CAP_SM_DISAB;
						ca_data->num_found_cas++;
						ca_data->num_found_ports++;
						OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
							"new ca %s 0x%" PRIx64 " cap 0x%x is_sm_disabled %u"
							" is_sm_disabled[%u]=0x%x\n",
							name, cl_ntoh64(req_guid),
							umad_ca.ports[i]->capmask,
							umad_ca.ports[i]->capmask & IB_PORT_CAP_SM_DISAB,
							num_cas, ca_data->is_sm_disabled[num_cas]);
					} else {
						ca_data->is_valid = FALSE;
						OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 543B: "
							"found CA %s with GUID 0x%" PRIx64
							" when %u CAs with such GUID already found."
							"Suspected duplicate GUIDs.\n",
							name, cl_ntoh64(req_guid), num_cas);
						goto Exit;
					}
				}
			} else {
				/* insert new entry, update there num_found_cas, num_found_ports_per_ca */
				add_ca_data_by_guid(p_vend, &ca_data_map, umad_ca.ports[i], name);
				OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
					"add new ca_data with guid 0x%" PRIx64 "\n",
					cl_ntoh64(current_guid));
			}

		} /* end of ports loop */
Release_umad:
		umad_release_ca(&umad_ca);
	}

	if (req_guid && cl_qmap_count(&ca_data_map) == 0) {
		/*
		 * No local CA owns this guid!
		 */
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5422: "
			"Unable to find requested CA guid 0x%" PRIx64 "\n",
			cl_ntoh64(req_guid));
		goto Exit;
	}
	
	/*
	 * Go over created map of CAs organized by guid and check validity.
	 * Find the first existing CA with an active port
	 */
	ret = resolve_umad_port_from_selected_cas(p_vend, &ca_data_map, is_gsi);
Exit:
	while (cl_qmap_count(&ca_data_map)) {
		p_map_item = cl_qmap_head(&ca_data_map);
		cl_qmap_remove_item(&ca_data_map, p_map_item);
		ca_data = (osm_vendor_ca_data_t *)p_map_item;
		delete_ca_data(&ca_data);
	}

	return ret;
}

static int
osm_vendor_open_port(IN osm_vendor_t * const p_vend,
		     IN const ib_net64_t port_guid, IN boolean_t is_gsi)
{
	int umad_port_id = -1, i, event_fd = -1;;

	OSM_ASSERT(p_vend);

	OSM_LOG_ENTER(p_vend->p_log);

	/*
	 * On planarized HCA there are number of planes ports and for gsi one port
	 * return the first port.
	 */
	if (p_vend->umad_port_id[0] >= 0) {
		umad_port_id = p_vend->umad_port_id[0];
		goto Exit;
	}

	if (resolve_umad_port(p_vend, port_guid, is_gsi) < 0) {
		goto Exit;
	}

	for (i = 0; i < p_vend->umad_ports_num; i++) {
		if ((umad_port_id = umad_open_port(p_vend->umad_port[i].ca_name,
						p_vend->umad_port[i].portnum)) < 0) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 542C: "
					"umad_open_port() failed\n");
			goto Exit;
		}
		p_vend->umad_port_id[i] = umad_port_id;
	}

	umad_port_id = p_vend->umad_port_id[0];

	/* start receiver thread */
	if (!(p_vend->receiver = calloc(1, sizeof(umad_receiver_t)))) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5423: "
			"Unable to alloc receiver struct\n");
		goto ClosePortsOnFailure;
	}

	event_fd = eventfd(0, 0);
	if (event_fd < 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5438: failed to create event fd.\n");
		goto ClosePortsOnFailure;
	}
	p_vend->eventfd = event_fd;

	if (umad_receiver_start(p_vend) != 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5420: "
			"umad_receiver_init failed\n");
		goto ClosePortsOnFailure;
	}

	/* Set up port_num_to_idx */
	memset(p_vend->port_num_to_idx, OSM_VENDOR_BIND_DEFAULT_PORT_IDX,
	       sizeof(uint8_t) * IB_ENDP_MAX_NUMBER_PORTS);
	for (i = 0; i < p_vend->umad_ports_num; i++) {
		p_vend->port_num_to_idx[p_vend->umad_port[i].portnum] = i;
	}

	/* Skip ClosePortsOnFailure */
	goto Exit;

ClosePortsOnFailure:
	for (i = 0; i < p_vend->umad_ports_num; i++) {
		umad_close_port(p_vend->umad_port_id[i]);
		umad_release_port(&p_vend->umad_port[i]);
		p_vend->umad_port[i].port_guid = 0;
		p_vend->umad_port_id[i] = umad_port_id = -1;
	}
	p_vend->umad_ports_num = 0;

Exit:
	OSM_LOG_EXIT(p_vend->p_log);
	return umad_port_id;
}

static void osm_vendor_close_port(osm_vendor_t * const p_vend)
{
	umad_receiver_t *p_ur;
	int i, j;

	p_ur = p_vend->receiver;
	p_vend->receiver = NULL;
	if (p_ur) {
		umad_receiver_stop(p_ur);
		free(p_ur);
	}

	if (p_vend->umad_ports_num > 0) {
		pthread_mutex_lock(&p_vend->agents_mutex);
		for (j = 0; j < p_vend->umad_ports_num; j++) {
			for (i = 0; i < OSM_UMAD_MAX_AGENTS; i++)
				if (p_vend->agents[j][i]) {
					umad_unregister(p_vend->umad_port_id[j], i);
					p_vend->agents[j][i] = NULL;
				}
		}
		pthread_mutex_unlock(&p_vend->agents_mutex);
		for (i = 0; i < p_vend->umad_ports_num; i++) {
			umad_close_port(p_vend->umad_port_id[i]);
			umad_release_port(&p_vend->umad_port[i]);
			p_vend->umad_port[i].port_guid = 0;
			p_vend->umad_port_id[i] = -1;
		}
		p_vend->umad_ports_num = 0;
	}

	p_vend->eventfd = -1;
}

static int set_bit(int nr, void *method_mask)
{
	long mask, *addr = method_mask;
	int retval;

	addr += nr / (8 * sizeof(long));
	mask = 1L << (nr % (8 * sizeof(long)));
	retval = (mask & *addr) != 0;
	*addr |= mask;
	return retval;
}

osm_bind_handle_t
osm_vendor_rebind(IN osm_vendor_t * const p_vend,
		  IN osm_bind_info_t * const p_user_bind,
		  IN osm_mad_pool_t * const p_mad_pool,
		  IN osm_vend_mad_recv_callback_t mad_recv_callback,
		  IN osm_vend_mad_send_err_callback_t send_err_callback,
		  IN void *context, IN osm_bind_handle_t h_bind)
{
	ib_net64_t port_guid;
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) h_bind;
	long method_mask[16 / sizeof(long)];
	int umad_port_id;
	uint8_t rmpp_version;
	boolean_t is_gsi;
	uint32_t i;

	OSM_LOG_ENTER(p_vend->p_log);

	OSM_ASSERT(p_user_bind);
	OSM_ASSERT(p_mad_pool);
	OSM_ASSERT(mad_recv_callback);
	OSM_ASSERT(send_err_callback);

	port_guid = p_user_bind->port_guid;

	OSM_LOG(p_vend->p_log, OSM_LOG_INFO,
		"Mgmt class 0x%02x binding to port GUID 0x%" PRIx64 "\n",
		p_user_bind->mad_class, cl_ntoh64(port_guid));

	is_gsi = (p_user_bind->mad_class == IB_MCLASS_SUBN_LID ||
			p_user_bind->mad_class == IB_MCLASS_SUBN_DIR) ? FALSE : TRUE;

	if ((umad_port_id = osm_vendor_open_port(p_vend, port_guid, is_gsi)) < 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5424: "
			"Unable to open port 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid));
		goto Exit;
	}

	if (!is_gsi) {
		for (i = 0; i < p_vend->umad_ports_num; i++) {
			if (umad_get_issm_path(p_vend->umad_port[i].ca_name,
						p_vend->umad_port[i].portnum,
						p_vend->issm_path,
						sizeof(p_vend->issm_path)) < 0) {
				OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 542E: "
					"Cannot resolve issm path for port %s:%u\n",
					p_vend->umad_port[i].ca_name, p_vend->umad_port[i].portnum);
				goto Exit;
			}
		}
	}
	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "is_sm_path %s\n", p_vend->issm_path);

	if (p_bind == NULL) {
		if (!(p_bind = malloc(sizeof(*p_bind)))) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5425: "
				"Unable to allocate internal bind object\n");
			goto Exit;
		}

		memset(p_bind, 0, sizeof(*p_bind));
		p_bind->p_vend = p_vend;
		p_bind->client_context = context;
		p_bind->mad_recv_callback = mad_recv_callback;
		p_bind->send_err_callback = send_err_callback;
		p_bind->p_mad_pool = p_mad_pool;
		p_bind->port_guid = port_guid;
		p_bind->timeout = p_user_bind->timeout ? p_user_bind->timeout :
				  p_vend->timeout;
		p_bind->max_retries = p_user_bind->retries;
	} else {
		pthread_mutex_lock(&p_vend->cb_mutex);
		p_bind->mad_recv_callback = mad_recv_callback;
		p_bind->send_err_callback = send_err_callback;
		pthread_mutex_unlock(&p_vend->cb_mutex);
	}
	memcpy(p_bind->port_ids, p_vend->umad_port_id, sizeof(p_bind->port_ids));

	memset(method_mask, 0, sizeof method_mask);
	if (p_user_bind->is_responder) {
		set_bit(IB_MAD_METHOD_GET, &method_mask);
		set_bit(IB_MAD_METHOD_SET, &method_mask);
		if (p_user_bind->mad_class == IB_MCLASS_SUBN_ADM) {
			set_bit(IB_MAD_METHOD_GETTABLE, &method_mask);
			set_bit(IB_MAD_METHOD_DELETE, &method_mask);
#ifdef DUAL_SIDED_RMPP
			set_bit(IB_MAD_METHOD_GETMULTI, &method_mask);
			/* Add in IB_MAD_METHOD_GETTRACETABLE */
			/* when supported by OpenSM */
#endif
		}
	}
	if (p_user_bind->is_report_processor)
		set_bit(IB_MAD_METHOD_REPORT, &method_mask);
	if (p_user_bind->is_trap_processor) {
		set_bit(IB_MAD_METHOD_TRAP, &method_mask);
		set_bit(IB_MAD_METHOD_TRAP_REPRESS, &method_mask);
	}
#ifndef VENDOR_RMPP_SUPPORT
	rmpp_version = 0;
#else
	/* If SA class, set rmpp_version */
	if (p_user_bind->mad_class == IB_MCLASS_SUBN_ADM)
		rmpp_version = 1;
	else
		rmpp_version = 0;
#endif

	for (i = 0; i < p_vend->umad_ports_num; i++) {
		if ((p_bind->agent_id[i] = umad_register(p_vend->umad_port_id[i],
						p_user_bind->mad_class,
						p_user_bind->class_version,
						rmpp_version, method_mask)) < 0) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5426: "
					"Unable to register class %u version %u\n",
					p_user_bind->mad_class, p_user_bind->class_version);
			free(p_bind);
			p_bind = 0;
			goto Exit;
		}

		pthread_mutex_lock(&p_vend->agents_mutex);
		if (p_bind->agent_id[i] >= OSM_UMAD_MAX_AGENTS ||
				p_vend->agents[i][p_bind->agent_id[i]]) {
			pthread_mutex_unlock(&p_vend->agents_mutex);
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5427: "
				"bad agent id %u or duplicate agent for class %u vers %u\n",
				p_bind->agent_id[i], p_user_bind->mad_class,
				p_user_bind->class_version);
			free(p_bind);
			p_bind = 0;
			goto Exit;
		}

		p_vend->agents[i][p_bind->agent_id[i]] = p_bind;
		pthread_mutex_unlock(&p_vend->agents_mutex);

		/* If Subn Directed Route class, register Subn LID routed class */
		if (p_user_bind->mad_class == IB_MCLASS_SUBN_DIR) {
			if ((p_bind->agent_id1[i] = umad_register(p_vend->umad_port_id[i],
							IB_MCLASS_SUBN_LID,
							p_user_bind->class_version,
							0,
							method_mask)) < 0) {
				OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5428: "
					"Unable to register class 1 version %u\n",
					p_user_bind->class_version);
				free(p_bind);
				p_bind = 0;
				goto Exit;
			}

			pthread_mutex_lock(&p_vend->agents_mutex);
			if (p_bind->agent_id1[i] >= OSM_UMAD_MAX_AGENTS ||
					p_vend->agents[i][p_bind->agent_id1[i]]) {
				pthread_mutex_unlock(&p_vend->agents_mutex);
				OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5429: "
					"bad agent id %u or duplicate agent for class 1 vers %u\n",
					p_bind->agent_id1[i], p_user_bind->class_version);
				free(p_bind);
				p_bind = 0;
				goto Exit;
			}

			p_vend->agents[i][p_bind->agent_id1[i]] = p_bind;
			pthread_mutex_unlock(&p_vend->agents_mutex);
		}
	}

Exit:
	OSM_LOG_EXIT(p_vend->p_log);
	return ((osm_bind_handle_t) p_bind);
}

osm_bind_handle_t
osm_vendor_bind(IN osm_vendor_t * const p_vend,
		IN osm_bind_info_t * const p_user_bind,
		IN osm_mad_pool_t * const p_mad_pool,
		IN osm_vend_mad_recv_callback_t mad_recv_callback,
		IN osm_vend_mad_send_err_callback_t send_err_callback,
		IN void *context)
{
	osm_bind_handle_t p_bind = 0;
	OSM_LOG_ENTER(p_vend->p_log);
	p_bind = osm_vendor_rebind(p_vend, p_user_bind, p_mad_pool, mad_recv_callback,
				   send_err_callback, context, NULL);
	OSM_LOG_EXIT(p_vend->p_log);
	return p_bind;
}

static void
__osm_vendor_recv_dummy_cb(IN osm_madw_t * p_madw,
			   IN void *bind_context, IN osm_madw_t * p_req_madw)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) osm_madw_get_bind_handle(p_madw);

	osm_mad_pool_put(p_bind->p_mad_pool, p_madw);
	if (p_req_madw)
		osm_mad_pool_put(p_bind->p_mad_pool, p_req_madw);

#ifdef _DEBUG_
	fprintf(stderr,
		"__osm_vendor_recv_dummy_cb: Ignoring received MAD after osm_vendor_unbind\n");
#endif
}

static void
__osm_vendor_send_err_dummy_cb(IN void *bind_context,
			       IN osm_madw_t * p_req_madw)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) osm_madw_get_bind_handle(p_req_madw);

	osm_mad_pool_put(p_bind->p_mad_pool, p_req_madw);

#ifdef _DEBUG_
	fprintf(stderr,
		"__osm_vendor_send_err_dummy_cb: Ignoring send error after osm_vendor_unbind\n");
#endif
}

void osm_vendor_unbind(IN osm_bind_handle_t h_bind)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) h_bind;
	osm_vendor_t *p_vend = p_bind->p_vend;

	OSM_LOG_ENTER(p_vend->p_log);

	pthread_mutex_lock(&p_vend->cb_mutex);
	p_bind->mad_recv_callback = __osm_vendor_recv_dummy_cb;
	p_bind->send_err_callback = __osm_vendor_send_err_dummy_cb;
	pthread_mutex_unlock(&p_vend->cb_mutex);

	OSM_LOG_EXIT(p_vend->p_log);
}

ib_mad_t *osm_vendor_get(IN osm_bind_handle_t h_bind,
			 IN const uint32_t mad_size,
			 IN osm_vend_wrap_t * const p_vw)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) h_bind;
	osm_vendor_t *p_vend = p_bind->p_vend;

	OSM_LOG_ENTER(p_vend->p_log);

	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
		"Acquiring UMAD for p_madw = %p, size = %u\n", p_vw, mad_size);
	OSM_ASSERT(p_vw);
	p_vw->size = mad_size;
	p_vw->umad = umad_alloc(1, mad_size + umad_size());

	/* track locally */
	p_vw->h_bind = h_bind;

	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
		"Acquired UMAD %p, size = %u\n", p_vw->umad, p_vw->size);

	OSM_LOG_EXIT(p_vend->p_log);
	return (p_vw->umad ? umad_get_mad(p_vw->umad) : NULL);
}

void
osm_vendor_put(IN osm_bind_handle_t h_bind, IN osm_vend_wrap_t * const p_vw)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) h_bind;
	osm_vendor_t *p_vend = p_bind->p_vend;
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(p_vend->p_log);

	OSM_ASSERT(p_vw);

	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "Retiring UMAD %p\n", p_vw->umad);

	/*
	 * We moved the removal of the transaction to immediately after
	 * it was looked up.
	 */

	/* free the mad but the wrapper is part of the madw object */
	umad_free(p_vw->umad);
	p_vw->umad = 0;
	p_madw = PARENT_STRUCT(p_vw, osm_madw_t, vend_wrap);
	p_madw->p_mad = NULL;

	OSM_LOG_EXIT(p_vend->p_log);
}

ib_api_status_t
osm_vendor_send(IN osm_bind_handle_t h_bind,
		IN osm_madw_t * const p_madw, IN boolean_t const resp_expected)
{
	osm_umad_bind_info_t *const p_bind = h_bind;
	osm_vendor_t *const p_vend = p_bind->p_vend;
	osm_vend_wrap_t *const p_vw = osm_madw_get_vend_ptr(p_madw);
	osm_mad_addr_t *const p_mad_addr = osm_madw_get_mad_addr_ptr(p_madw);
	ib_mad_t *const p_mad = osm_madw_get_mad_ptr(p_madw);
	ib_sa_mad_t *const p_sa = (ib_sa_mad_t *) p_mad;
	ib_mad_addr_t mad_addr;
	int ret = -1;
	int __attribute__((__unused__)) is_rmpp = 0;
	uint32_t sent_mad_size, timeout = 0;
	uint8_t port_idx;
	uint64_t tid;
	struct _gsi *p_gsi = NULL;
#ifndef VENDOR_RMPP_SUPPORT
	uint32_t paylen = 0;
#endif

	OSM_LOG_ENTER(p_vend->p_log);

	OSM_ASSERT(p_vw->h_bind == h_bind);
	OSM_ASSERT(p_mad == umad_get_mad(p_vw->umad));

	if (p_mad->mgmt_class == IB_MCLASS_SUBN_DIR) {
		umad_set_addr_net(p_vw->umad, 0xffff, 0, 0, 0);
		umad_set_grh(p_vw->umad, NULL);
		port_idx = osm_vendor_get_port_idx(p_bind, p_mad_addr->addr_type.smi.port_num);
		goto Resp;
	}
	if (p_mad->mgmt_class == IB_MCLASS_SUBN_LID) {
		umad_set_addr_net(p_vw->umad, p_mad_addr->dest_lid, 0, 0, 0);
		umad_set_grh(p_vw->umad, NULL);
		port_idx = osm_vendor_get_port_idx(p_bind, p_mad_addr->addr_type.smi.port_num);
		goto Resp;
	}
	/* GS classes */
	umad_set_addr_net(p_vw->umad, p_mad_addr->dest_lid,
			  p_mad_addr->addr_type.gsi.remote_qp,
			  p_mad_addr->addr_type.gsi.service_level,
			  IB_QP1_WELL_KNOWN_Q_KEY);

	p_gsi = &p_mad_addr->addr_type.gsi;

	/* GSI has only one port */
	port_idx = OSM_VENDOR_BIND_DEFAULT_PORT_IDX;

	if (p_mad->mgmt_class == IB_MCLASS_SUBN_ADM)
		OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "class 0x%x GRH present "
				"%d\n", p_mad->mgmt_class, p_gsi->global_route);

	if (p_gsi->global_route) {
		mad_addr.grh_present = 1;
		mad_addr.gid_index = 0;
		mad_addr.hop_limit = p_gsi->grh_info.hop_limit;
		ib_grh_get_ver_class_flow(p_gsi->grh_info.ver_class_flow, NULL,
				&mad_addr.traffic_class,
				&mad_addr.flow_label);
		memcpy(&mad_addr.gid, &p_gsi->grh_info.dest_gid, 16);
		umad_set_grh(p_vw->umad, &mad_addr);
	} else
		umad_set_grh(p_vw->umad, NULL);

	umad_set_pkey(p_vw->umad, p_mad_addr->addr_type.gsi.pkey_ix);
	if (ib_class_is_rmpp(p_mad->mgmt_class)) {	/* RMPP GS classes     FIXME: no GRH */
		if (!ib_rmpp_is_flag_set((ib_rmpp_mad_t *) p_sa,
					 IB_RMPP_FLAG_ACTIVE)) {
			/* Clear RMPP header when RMPP not ACTIVE */
			p_sa->rmpp_version = 0;
			p_sa->rmpp_type = 0;
			p_sa->rmpp_flags = 0;
			p_sa->rmpp_status = 0;
#ifdef VENDOR_RMPP_SUPPORT
		} else
			is_rmpp = 1;
		OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "RMPP %d length %d\n",
			ib_rmpp_is_flag_set((ib_rmpp_mad_t *) p_sa,
					    IB_RMPP_FLAG_ACTIVE),
			p_madw->mad_size);
#else
		} else {
			p_sa->rmpp_version = 1;
			p_sa->seg_num = cl_ntoh32(1);	/* first DATA is seg 1 */
			p_sa->rmpp_flags |= (uint8_t) 0x70;	/* RRespTime of 14 (high 5 bits) */
			p_sa->rmpp_status = 0;
			paylen = p_madw->mad_size - IB_SA_MAD_HDR_SIZE;
			paylen += (IB_SA_MAD_HDR_SIZE - MAD_RMPP_HDR_SIZE);
			p_sa->paylen_newwin = cl_ntoh32(paylen);
		}
#endif
	}

Resp:
	if (resp_expected)
		put_madw(p_vend, p_madw, p_mad->trans_id, p_mad->mgmt_class);

#ifdef VENDOR_RMPP_SUPPORT
	sent_mad_size = p_madw->mad_size;
#else
	sent_mad_size = is_rmpp ? p_madw->mad_size - IB_SA_MAD_HDR_SIZE :
	    p_madw->mad_size;
#endif
	tid = cl_ntoh64(p_mad->trans_id);
	if (resp_expected) {
		if (p_madw->timeout)
			timeout = p_madw->timeout;
		else
			timeout = p_bind->timeout;
	}

	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG,
		"Servicing p_madw = %p TID 0x%" PRIx64 " on port_guid 0x%" PRIx64
		", port_idx %u, port_id %u\n",
		p_madw, tid, cl_ntoh64(p_bind->port_guid), port_idx, p_bind->port_ids[port_idx]);
	if ((ret = umad_send(p_bind->port_ids[port_idx],
			     p_bind->agent_id[port_idx], p_vw->umad,
			     sent_mad_size, timeout,
			     p_bind->max_retries)) < 0) {
		OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5430: "
			"Send p_madw = %p of size %d, Class 0x%x, Method 0x%X, "
			"Attr 0x%X, TID 0x%" PRIx64 ", port_idx %u failed %d (%m)\n",
			p_madw, sent_mad_size, p_mad->mgmt_class,
			p_mad->method, cl_ntoh16(p_mad->attr_id), tid, port_idx, ret);
		if (resp_expected) {
			get_madw(p_vend, &p_mad->trans_id,
				 p_mad->mgmt_class);	/* remove from aging table */
			p_madw->status = IB_ERROR;
			pthread_mutex_lock(&p_vend->cb_mutex);
			(*p_bind->send_err_callback) (p_bind->client_context, p_madw);	/* cb frees madw */
			pthread_mutex_unlock(&p_vend->cb_mutex);
		} else
			osm_mad_pool_put(p_bind->p_mad_pool, p_madw);
		goto Exit;
	}

	if (!resp_expected)
		osm_mad_pool_put(p_bind->p_mad_pool, p_madw);

	OSM_LOG(p_vend->p_log, OSM_LOG_DEBUG, "Completed sending %s TID 0x%" PRIx64 "\n",
		resp_expected ? "request" : "response or unsolicited", tid);
Exit:
	OSM_LOG_EXIT(p_vend->p_log);
	return (ret);
}

ib_api_status_t osm_vendor_local_lid_change(IN osm_bind_handle_t h_bind)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) h_bind;
	osm_vendor_t *p_vend = p_bind->p_vend;

	OSM_LOG_ENTER(p_vend->p_log);
	;
	OSM_LOG_EXIT(p_vend->p_log);
	return (0);
}

void osm_vendor_get_bind_port_nums(IN osm_bind_handle_t h_bind,
				   OUT uint8_t port_nums[OSM_UMAD_MAX_PORTS_PER_CA])
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *)h_bind;
	osm_vendor_t *p_vend = p_bind->p_vend;
	uint32_t port_idx;

	memset(port_nums, IB_INVALID_PORT_NUM, OSM_UMAD_MAX_PORTS_PER_CA * sizeof(port_nums[0]));

	for (port_idx = 0; port_idx < p_vend->umad_ports_num; port_idx++) {
		if (p_vend->umad_port[port_idx].portnum >= 0)
			port_nums[port_idx] = p_vend->umad_port[port_idx].portnum;
	}
}

void osm_vendor_set_sm(IN osm_bind_handle_t h_bind, IN boolean_t is_sm_val)
{
	osm_umad_bind_info_t *p_bind = (osm_umad_bind_info_t *) h_bind;
	osm_vendor_t *p_vend = p_bind->p_vend;

	OSM_LOG_ENTER(p_vend->p_log);
	if (TRUE == is_sm_val) {
		p_vend->issmfd = open(p_vend->issm_path, O_NONBLOCK);
		if (p_vend->issmfd < 0) {
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5431: "
				"setting IS_SM capmask: cannot open file "
				"\'%s\': %s\n",
				p_vend->issm_path, strerror(errno));
			p_vend->issmfd = -1;
		}
	} else if (p_vend->issmfd != -1) {
		if (0 != close(p_vend->issmfd))
			OSM_LOG(p_vend->p_log, OSM_LOG_ERROR, "ERR 5432: "
				"clearing IS_SM capmask: cannot close: %s\n",
				strerror(errno));
		p_vend->issmfd = -1;
	}
	OSM_LOG_EXIT(p_vend->p_log);
}

void osm_vendor_set_debug(IN osm_vendor_t * const p_vend, IN int32_t level)
{
	umad_debug(level);
}

#endif				/* OSM_VENDOR_INTF_OPENIB */
