
/*
 * SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: LicenseRef-NvidiaProprietary
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

/*
 * Abstract:
 *	Implementation of ISSU.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sys/time.h>
#include <iba/ib_types.h>
#include <complib/cl_ptr_vector.h>
#include <complib/cl_hashmap.h>
#include <complib/cl_math.h>
#include <opensm/osm_base.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ISSU_MGR_C
#include <opensm/osm_issu_mgr.h>

#define sw_name(P_SW)	(P_SW)->p_node->print_desc

extern int wait_for_pending_transactions(osm_opensm_t *p_osm);
extern int osm_add_held_back_sw_issu(void *cxt, uint64_t guid);
extern int osm_restore_held_back_sw_issu(IN osm_subn_t *p_subn, IN ib_net64_t sw_guid);

/* -------------------------------------------------------------------------------------------------
 *
 * Miscellaneous definitions and functions.
 */

#define	STARTUP_WARNING_FORMAT		"startup delayed for %ld seconds to allow upgrading switches to be active\n"

#define foreach_issu_element(P_SET, VALUE, TYPE, ITER)							\
	for (ITER  = set_get_size((P_SET)) - 1;								\
	     VALUE = (TYPE)set_get(P_SET, ITER), ITER >= 1;						\
	     ITER--)											\


#define foreach_issu_switch(P_SET, P_SW, ITER)								\
	foreach_issu_element(P_SET, P_SW, osm_switch_t *, ITER)


#define foreach_issu_guid(P_SET, GUID, ITER)								\
	foreach_issu_element(P_SET, GUID, ib_net64_t, ITER)


#define foreach_issu_trap(P_MGR, P_TRAP)								\
	for (P_TRAP  = (issu_trap_t *)cl_qlist_head(&P_MGR->traps);					\
	     P_TRAP != (issu_trap_t *)cl_qlist_end(&P_MGR->traps);					\
	     P_TRAP  = (issu_trap_t *)cl_qlist_next((cl_list_item_t *)P_TRAP))


static const char *const issu_state_str[] = {
	"IDLE",
	"REQUESTED",
	"ERROR",
	"APPROVED"
};

static int cmp_guid(const void * p_lhs, const void * p_rhs)
{
	uint64_t		l_guid;
	uint64_t		r_guid;

	l_guid = *(uint64_t *)p_lhs;
	r_guid = *(uint64_t *)p_rhs;
	if (l_guid > r_guid) {
		return 1;
	} else if (l_guid < r_guid) {
		return -1;
	} else {
		return 0;
	}
}

static int cmp_addr(IN const void *p_lhs, IN const void *p_rhs)
{
	osm_switch_t		*p_lsw;
	osm_switch_t		*p_rsw;

	p_lsw = *(osm_switch_t **)p_lhs;
	p_rsw = *(osm_switch_t **)p_rhs;
	if (p_lsw > p_rsw) {
		return 1;
	} else if (p_lsw < p_rsw) {
		return -1;
	} else {
		return 0;
	}
}


/* -------------------------------------------------------------------------------------------------
 *
 * Set functions.
 */

#define	set_callback				cl_pfn_ptr_vec_apply_t

#define set_contains_guid(P_SET, P_SW)		(set_find_guid(P_SET,osm_node_get_node_guid((P_SW)->p_node)) != 0)
#define set_contains_switch(P_SET, P_SW)	(set_find_switch(P_SET, P_SW) != 0)

static void set_construct(IN issu_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(!p_set->init);

	p_set->init = FALSE;
	cl_ptr_vector_construct(&p_set->vector);
}

static boolean_t set_create(IN issu_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(!p_set->init);

	if (cl_ptr_vector_init(&p_set->vector, 1, 1024) == CL_SUCCESS) {
		cl_ptr_vector_set(&p_set->vector, 0, NULL);
		p_set->init = TRUE;
	}

	return p_set->init;
}

static inline void set_destroy(IN issu_set_t *p_set)
{
	CL_ASSERT(p_set);

	if (p_set->init) {
		cl_ptr_vector_destroy(&p_set->vector);
		p_set->init = FALSE;
	}
}

static inline boolean_t set_insert(IN issu_set_t *p_set, IN void *value)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return cl_ptr_vector_insert(&p_set->vector, value, NULL) == CL_SUCCESS;
}

static inline void set_remove(IN issu_set_t *p_set, IN size_t index)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	cl_ptr_vector_remove(&p_set->vector, index);
}

static inline void set_remove_all(IN issu_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	cl_ptr_vector_remove_all(&p_set->vector);
	cl_ptr_vector_set(&p_set->vector, 0, NULL);
}

static inline uint32_t set_get_size(IN issu_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return cl_ptr_vector_get_size(&p_set->vector);
}

static inline osm_switch_t *set_get(IN issu_set_t *p_set, IN size_t index)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return (osm_switch_t *)cl_ptr_vector_get(&p_set->vector, index);
}

static size_t set_find(IN issu_set_t *p_set, IN void *p_void, IN int (*cmp_func)(const void *p_lhs, const void *p_rhs))
{
	size_t			index;

	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	if (p_void == NULL) {
		return 0;
	}

	if (cl_ptr_vector_get_size(&p_set->vector) == 1) {
		return 0;
	}

	if (!p_set->vector.sorted) {
		cl_ptr_vector_sort_cmp(&p_set->vector, 1, 0, cmp_func);
	}

	index = cl_ptr_vector_lookup(&p_set->vector, p_void);
	if (index == cl_ptr_vector_get_size(&p_set->vector)) {
		index = 0;
	}

	return index;
}

static size_t set_find_guid(IN issu_set_t *p_set, IN uint64_t guid)
{
	return set_find(p_set, (void *)guid, cmp_guid);
}

static size_t set_find_switch(IN issu_set_t *p_set, IN osm_switch_t *p_sw)
{
	return set_find(p_set, (void *)p_sw, cmp_addr);
}

static inline boolean_t set_is_empty(IN issu_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return set_get_size(p_set) == 1;
}

static inline boolean_t set_is_not_empty(IN issu_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return !set_is_empty(p_set);
}

static inline boolean_t set_apply(IN issu_set_t *p_set, IN set_callback p_callback, IN const void *const context)
{
	osm_sm_t		*p_sm = (osm_sm_t *)context;

	cl_ptr_vector_apply_func(&p_set->vector, p_callback, context);
	return wait_for_pending_transactions(p_sm->p_subn->p_osm) == 0;
}

/* -------------------------------------------------------------------------------------------------
 *
 * Internal functions.
 */

static inline const char *issu_state_to_string(IN ib_issu_state_t state)
{
	return issu_state_str[state & 0x3];
}

static inline boolean_t issu_switch_is_capable(IN osm_switch_t *p_sw)
{
	return ib_general_info_is_extended_switchinfo_enabled(&p_sw->p_node->general_info) &&
	       ib_general_info_is_issu_enabled(&p_sw->p_node->general_info);
}

static inline ib_issu_state_t issu_get_state(IN osm_switch_t *p_sw)
{
	return ib_extended_switch_info_get_issu_state(&p_sw->subn_ext_switch_info);
}

static inline char *issu_sw_info(IN osm_switch_t *p_sw)
{
	static char	buffer[128];

	snprintf(buffer, sizeof(buffer), "%s (0x%016lx)", sw_name(p_sw), osm_node_get_node_guid_ho(p_sw->p_node));
	return buffer;
}

/* -------------------------------------------------------------------------------------------------
 *
 * Switch interface functions.
 */

static void issu_req_get_state(IN size_t index, IN void *element, IN void *context)
{
	osm_switch_t			*p_sw;
	osm_sm_t			*p_sm;
	ib_api_status_t			status;

	if (element != NULL) {
		p_sm = (osm_sm_t *)context;
		p_sw = (osm_switch_t *)element;

		status = osm_req_get_esi(p_sm, p_sw);
		if (status != IB_SUCCESS) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B902: %s Get(ESI) failed (%s)\n", issu_sw_info(p_sw), ib_get_err_str(status));
		}
	}

	return;
}

static ib_api_status_t issu_req_set_state(IN osm_sm_t *p_sm, IN osm_switch_t *p_sw, IN ib_issu_state_t state)
{
	ib_api_status_t			status;
	ib_extended_switch_info_t	*p_esi;

	p_esi = &p_sw->subn_ext_switch_info;
	ib_extended_switch_info_set_issu_state(p_esi, state);

	status = osm_req_set_esi(p_sm, p_sw, p_esi);
	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B903: %s Set(ESI) failed (%s)\n", issu_sw_info(p_sw), ib_get_err_str(status));
	}

	return status;
}

static int issu_isolate_switch(IN osm_issu_mgr_t *p_mgr, IN osm_switch_t *p_sw)
{
	ib_net64_t		guid = osm_node_get_node_guid(p_sw->p_node);
	ib_issu_state_t		issu_state = issu_get_state(p_sw);

	/*
	 * If we can't insert the switch, we just return.  We can try again next time.
	 */
	if (!set_insert(&p_mgr->isolated, (void *)guid)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B904: set_insert failed for %s\n", issu_sw_info(p_sw));
		return 0;
	}

	OSM_LOG(p_mgr->p_log, OSM_LOG_INFO, "isolate %s\n", issu_sw_info(p_sw));

	osm_add_held_back_sw_issu(p_mgr->p_subn, cl_ntoh64(guid));
	osm_report_notice_1609(p_mgr->p_subn, cl_ntoh64(guid), issu_state);
	return 0;
}

static boolean_t issu_switch_can_be_restored(IN osm_issu_mgr_t *p_mgr, IN osm_switch_t *p_sw, IN ib_net64_t guid)
{
	ib_issu_state_t		issu_state;

	/*
	 * Determine if the guid/switch can be restored.
	 *
	 * ISSU isolated switch checks:
	 * 	If the switch doesn't exist anymore, restore it.
	 * 	If the switch hasn't answered a Get(ESI) call, don't restore it.
	 * 	If the switch is NOT in the capable set, restore it.
	 * 	If the switch has a good state, restore it.
	 */
	if (p_sw == NULL) {
		return TRUE;
	} else if (!p_sw->esi_updated) {
		return FALSE;
	} else if (!set_contains_switch(&p_mgr->capable, p_sw)) {
		return TRUE;
	} else {
		issu_state = issu_get_state(p_sw);
		return ((issu_state == ISSU_IDLE) || (issu_state == ISSU_REQUESTED));
	}
}

static boolean_t issu_restore_switch(IN osm_issu_mgr_t *p_mgr, IN osm_switch_t *p_sw, IN ib_net64_t guid)
{
	char			*sw_name;
	size_t			index;
	ib_issu_state_t		issu_state;

	if (p_sw == NULL) {
                sw_name = "Unknown";
                issu_state = ISSU_IDLE;
        } else {
                sw_name = sw_name(p_sw);
                issu_state = issu_get_state(p_sw);
        }

	/*
	 * Restore the switch.
	 */
	OSM_LOG(p_mgr->p_log, OSM_LOG_INFO, "restore %s (0x%016lx)\n", sw_name, cl_ntoh64(guid));
	if ((index = set_find_guid(&p_mgr->isolated, guid)) != 0) {
		set_remove(&p_mgr->isolated, index);
	}

	osm_restore_held_back_sw_issu(p_mgr->p_subn, guid);
	osm_report_notice_1610(p_mgr->p_subn, cl_ntoh64(guid), issu_state);

	return TRUE;
}

static void issu_upgrade_complete(IN osm_issu_mgr_t *p_mgr, IN osm_switch_t *p_sw)
{
	size_t			index = set_find_switch(&p_mgr->upgrades, p_sw);

	/*
	 * Flag the switch as needing to be updated and remove it from the upgrade set.
	 */
	p_sw->need_update = 2;
	if (index > 0) {
		set_remove(&p_mgr->upgrades, index);
	}

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "upgrade complete for %s\n", issu_sw_info(p_sw));
}

/* -------------------------------------------------------------------------------------------------
 *
 * Trap functions.
 */

typedef struct {
	cl_list_item_t		item;
	uint16_t		lid_ho;
	uint32_t		state;
} issu_trap_t;


static void issu_trap_enqueue(IN osm_issu_mgr_t *p_mgr, IN uint16_t lid_ho, IN uint32_t state)
{
	issu_trap_t		*p_trap;

	/*
	 * Create a trap entry.  If we can't, the upgrade will be processed after the usual wait.
	 */
	if ((p_trap = calloc(1, sizeof(issu_trap_t))) == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B901: trap allocation error\n");
		return;
	}

	p_trap->lid_ho = lid_ho;
	p_trap->state = state;

	/*
	 * Enqueue the trap and signal the event.
	 */
	CL_PLOCK_EXCL_ACQUIRE(&p_mgr->lock);
	cl_qlist_insert_tail(&p_mgr->traps, (cl_list_item_t *)p_trap);
	CL_PLOCK_RELEASE(&p_mgr->lock);

	cl_event_signal(&p_mgr->event);
}

static issu_trap_t *issu_trap_dequeue(IN osm_issu_mgr_t *p_mgr)
{
	issu_trap_t		*p_trap;

	CL_PLOCK_EXCL_ACQUIRE(&p_mgr->lock);
	p_trap = (issu_trap_t *)cl_qlist_remove_head(&p_mgr->traps);
	if (p_trap == (issu_trap_t *)cl_qlist_end(&p_mgr->traps)) {
		p_trap = NULL;
	}
	CL_PLOCK_RELEASE(&p_mgr->lock);

	return p_trap;
}

static void issu_traps_process(IN osm_issu_mgr_t *p_mgr)
{
	int			i;
	uint16_t		lid_ho;
	uint16_t		sw_lid_ho;
	uint32_t		state;
	issu_trap_t		*p_trap;
	osm_switch_t		*p_sw;

	while ((p_trap = issu_trap_dequeue(p_mgr)) != NULL) {
		lid_ho = p_trap->lid_ho;
		state = p_trap->state;
		free(p_trap);

		/*
		 * The spec says that state == 1 means that the upgrade is complete.  Mark the switch as complete.
		 */
		if (state == 1) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "processing trap for lid %u with state %u\n", lid_ho, state);
			foreach_issu_switch(&p_mgr->upgrades, p_sw, i) {
				sw_lid_ho = osm_node_get_base_lid_ho(p_sw->p_node, 0);
				if (sw_lid_ho == lid_ho) {
					issu_upgrade_complete(p_mgr, p_sw);
					break;
				}
			}
		}
	}
}

static void issu_traps_remove_all(IN osm_issu_mgr_t *p_mgr)
{
	issu_trap_t		*p_trap;

	while ((p_trap = issu_trap_dequeue(p_mgr)) != NULL) {
		free(p_trap);
	}
}

static void issu_traps_destroy(IN osm_issu_mgr_t *p_mgr)
{
	CL_PLOCK_EXCL_ACQUIRE(&p_mgr->lock);
	cl_qlist_remove_all(&p_mgr->traps);
	CL_PLOCK_RELEASE(&p_mgr->lock);
}

static void issu_traps_init(IN osm_issu_mgr_t *p_mgr)
{
	cl_qlist_init(&p_mgr->traps);
}

/* -------------------------------------------------------------------------------------------------
 *
 * Processing functions.
 */

static inline time_t issu_time_left(IN osm_issu_mgr_t *p_mgr)
{
	time_t			time_now = time(NULL);

	return (time_now <= p_mgr->end) ? p_mgr->end - time_now : 0;
}

static inline boolean_t issu_is_timedout(IN osm_issu_mgr_t *p_mgr)
{
	return (time(NULL) >= p_mgr->end);
}

static inline boolean_t issu_is_not_timedout(IN osm_issu_mgr_t *p_mgr)
{
	return !issu_is_timedout(p_mgr);
}

static boolean_t issu_mgr_verify_upgrades(IN osm_sm_t *p_sm)
{
	int			i;
	int			status;
	time_t			time_left;
	osm_switch_t		*p_sw;
	ib_issu_state_t		issu_state;
	osm_issu_mgr_t		*p_mgr = &p_sm->issu_mgr;

	/*
	 * Loop waiting for the timeout.  We may get traps before the timeout, so we
	 * process them here.
	 */
	p_mgr->start = time(NULL);
	p_mgr->end = p_mgr->start + p_mgr->timeout + 1;
	while (set_is_not_empty(&p_mgr->upgrades) && issu_is_not_timedout(p_mgr)) {
		time_left = issu_time_left(p_mgr);

		cl_event_wait_on(&p_mgr->event, CL_SECS_TO_USECS * time_left, TRUE);
		issu_traps_process(p_mgr);
	}

	/*
	 * If we are here, we need to determine the state of outstanding upgrades.
	 */
	if (set_is_not_empty(&p_mgr->upgrades)) {
		/*
		 * Get the latest ISSU state from the outstanding switches.
		 */
		if (!set_apply(&p_mgr->upgrades, issu_req_get_state, p_sm)) {
			return FALSE;
		}

		/*
		 * Process the states.
		 */
		foreach_issu_switch(&p_mgr->upgrades, p_sw, i) {
			/*
			 * If the ESI wasn't updated, then the switch didn't answer the Get(ESI)
			 * request.  Mark it down (as unresponsive).  See below.
			 */
			if (!p_sw->esi_updated) {
				continue;
			}

			/*
			 * Process the Get(ESI) response.
			 *
			 * There are 4 possibilities here:
			 *	0) IDLE = Upgrade completed successfully or user cancelled.
			 *	1) REQUESTED = This switch wants to be approved again.  We will see it on the next light sweep.
			 *	2) ERROR - This switch is no longer usable.  (See below.)
			 *	3) APPROVED - Still waiting for the upgrade to complete.  (See below.)
			 */
			issu_state = issu_get_state(p_sw);

			switch (issu_state) {
			case ISSU_IDLE:
			case ISSU_REQUESTED:
				issu_upgrade_complete(p_mgr, p_sw);
				break;
			case ISSU_APPROVED:
			case ISSU_ERROR:
				break;
			}
		}
	}

	/*
	 * Any switch left in the upgrade set is in an error condition.  Mark it down.
	 */
	status = set_is_not_empty(&p_mgr->upgrades);

	foreach_issu_switch(&p_mgr->upgrades, p_sw, i) {
		issu_state = (p_sw->esi_updated) ? issu_get_state(p_sw) : ISSU_ERROR;
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B905: %s is in %s state - isolating\n", issu_sw_info(p_sw), issu_state_to_string(issu_state));

		set_remove(&p_mgr->upgrades, i);
		issu_isolate_switch(p_mgr, p_sw);
	}

	CL_ASSERT(set_is_empty(&p_mgr->upgrades));
	return status;
}

static void issu_preupgrade(IN osm_issu_mgr_t *p_mgr)
{
	/*
	 * This function is used to communicate upgrade state with SHARP.  We update the SMDB which
	 * SHARP reads to determine what to do.  The sleep is here to give SHARP time to notice the
	 * new SMDB and process it.
	 */
	issu_traps_remove_all(p_mgr);

	p_mgr->step = ISSU_STEP_PRE_UPGRADE;
	p_mgr->start = time(NULL);
	p_mgr->end = p_mgr->start + p_mgr->pre_upgrade_time + 1;

	osm_dump_sm_db(p_mgr->p_sm->p_subn->p_osm);
}

static void issu_start_upgrade(IN size_t index, IN void *element, void *context)
{
	osm_switch_t		*p_sw = (osm_switch_t *)element;
	osm_sm_t		*p_sm = (osm_sm_t *)context;
	osm_issu_mgr_t		*p_mgr = &p_sm->issu_mgr;

	if (p_sw != NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE, "starting upgrade for %s \n", issu_sw_info(p_sw));
		issu_req_set_state(p_sm, p_sw, ISSU_APPROVED);
	}
}

static boolean_t issu_identify_capable_switches(IN osm_issu_mgr_t *p_mgr)
{
	int			i;
	osm_switch_t		*p_sw;

	/*
	 * Get all of the switches which are ISSU capable.  If we can't put them all in the set, skip it.
	 * We will try again next sweep.
	 */
	set_remove_all(&p_mgr->capable);

	foreach_switch(p_mgr->p_subn, p_sw) {
		if (issu_switch_is_capable(p_sw)) {
			if (!set_insert(&p_mgr->capable, p_sw)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B906: set_insert failed for %s\n", issu_sw_info(p_sw));
				break;
			}
		}
	}

	/*
	 * Query the switches for their ISSU state.  If the switch doesn't respond to the Get(ESI),
	 * then remove it from the capable list since we could not get the newest ISSU status.
	 */
	if (!set_apply(&p_mgr->capable, issu_req_get_state, p_mgr->p_sm)) {
		return FALSE;
	}

	foreach_issu_switch(&p_mgr->capable, p_sw, i) {
		if (!p_sw->esi_updated) {
			set_remove(&p_mgr->capable, i);
		}
	}

	return set_is_not_empty(&p_mgr->capable);
}

static boolean_t issu_resolve_isolated_switches(IN osm_issu_mgr_t *p_mgr, IN boolean_t force)
{
	size_t			i;
	boolean_t		status;
	ib_net64_t		guid;
	osm_switch_t		*p_sw;

	status = FALSE;

	/*
	 * For each isolated switch, we will try to restore it.
	 */
	foreach_issu_guid(&p_mgr->isolated, guid, i) {
		CL_PLOCK_EXCL_ACQUIRE(p_mgr->p_sm->p_lock);
		p_sw = osm_get_switch_by_guid(p_mgr->p_subn, guid);
		CL_PLOCK_RELEASE(p_mgr->p_sm->p_lock);

		if (force || issu_switch_can_be_restored(p_mgr, p_sw, guid)) {
			status |= issu_restore_switch(p_mgr, p_sw, guid);
		}
	}

	return status;
}

static boolean_t issu_identify_upgrades(IN osm_issu_mgr_t *p_mgr)
{
	size_t			i;
	boolean_t		status;
	ib_issu_state_t		issu_state;
	osm_switch_t		*p_sw;

	status = FALSE;

	/*
	 * If the ISSU manager is not active, then it doesn't process upgrades.
	 */
	if (!issu_is_active(p_mgr)) {
		return status;
	}

	/*
	 * Find those switches which need upgrading.
	 */
	set_remove_all(&p_mgr->upgrades);

	foreach_issu_switch(&p_mgr->capable, p_sw, i) {
		if (set_contains_guid(&p_mgr->isolated, p_sw)) {
			continue;
		}

		issu_state = issu_get_state(p_sw);
		switch (issu_state) {
		case ISSU_IDLE:
			/*
			 * Nothing to do.
			 */
			break;
		case ISSU_REQUESTED:
			/*
			 * This switch wants to be approved.  If we can't insert it into the set, we return.
			 * We can try again next sweep.
			 */
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "%s requesting upgrade\n", issu_sw_info(p_sw));
			if (!set_insert(&p_mgr->upgrades, p_sw)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B907: set_insert failed for %s\n", issu_sw_info(p_sw));
			}
			break;
		case ISSU_APPROVED:
		case ISSU_ERROR:
			/*
			 * We should never see a switch in these states.  Mark it down.
			 */
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B908: %s - invalid state (%s)\n", issu_sw_info(p_sw), issu_state_to_string(issu_state));
			issu_isolate_switch(p_mgr, p_sw);
			status = TRUE;
			break;
		}
	}

	return status;
}

static boolean_t issu_perform_upgrades(IN osm_issu_mgr_t *p_mgr)
{
	boolean_t		status = FALSE;

	/*
	 * Approve the upgrade for all switches whose time has come.  We don't wait for responses
	 * because issu_mgr_verify_upgrades does the trap processing and waiting for the large
	 * timeout.
	 */
	if (set_is_not_empty(&p_mgr->upgrades)) {
		utimes(p_mgr->startup_file, NULL);
		p_mgr->step = ISSU_STEP_UPGRADING;
		osm_dump_sm_db(p_mgr->p_sm->p_subn->p_osm);

		if (set_apply(&p_mgr->upgrades, issu_start_upgrade, p_mgr->p_sm)) {
			status = issu_mgr_verify_upgrades(p_mgr->p_sm);
		}

		p_mgr->step = ISSU_STEP_IDLE;
		osm_dump_sm_db(p_mgr->p_sm->p_subn->p_osm);
	}

	return status;
}

static boolean_t issu_dealloc_resources(IN osm_issu_mgr_t *p_mgr)
{
	issu_traps_destroy(p_mgr);

	if (p_mgr->inited) {
		cl_event_destroy(&p_mgr->event);
		set_destroy(&p_mgr->capable);
		set_destroy(&p_mgr->upgrades);
		set_destroy(&p_mgr->isolated);
		cl_plock_destroy(&p_mgr->lock);
	}

	return TRUE;
}

static boolean_t issu_alloc_resources(IN osm_issu_mgr_t *p_mgr)
{
	boolean_t		status = FALSE;

	cl_event_construct(&p_mgr->event);
	cl_plock_construct(&p_mgr->lock);
	set_construct(&p_mgr->capable);
	set_construct(&p_mgr->upgrades);
	set_construct(&p_mgr->isolated);

	/*
	 * Create the trap cache.
	 */
	issu_traps_init(p_mgr);

	/*
	 * Create the timer event.
	 */
	if (cl_event_init(&p_mgr->event, FALSE) != IB_SUCCESS) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B909: error creating timer event\n");
		goto Exit;
	}

	/*
	 * Create the list lock.
	 */
	if (cl_plock_init(&p_mgr->lock) != CL_SUCCESS) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B90A: error creating lock\n");
		goto Exit;
	}

	/*
	 * Create the switch sets for upgrading.
	 */
	if (!set_create(&p_mgr->capable)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B90B: error creating capable set\n");
		goto Exit;
	}

	if (!set_create(&p_mgr->upgrades)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B90C: error creating upgrades set\n");
		goto Exit;
	}

	if (!set_create(&p_mgr->isolated)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B90D: error creating isolated set\n");
		goto Exit;
	}

	status = TRUE;

Exit:
	if (!status) {
		issu_dealloc_resources(p_mgr);
	}

	return status;
}


/* -------------------------------------------------------------------------------------------------
 *
 * API interface.
 */

const char *osm_issu_status(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw)
{
	boolean_t		issu_capable;
	int			issu_state;
	osm_issu_mgr_t		*p_mgr;
	const char		*status;

	p_mgr = &p_subn->p_osm->sm.issu_mgr;

	/*
	 * If the ISSU manager is disabled, then all statuses are inactive.
	 */
	if (issu_is_disabled(p_mgr)) {
		return "N/A";
	}

	/*
	 * Switch not ISSU enabled?
	 */
	issu_capable = ib_general_info_is_issu_enabled(&p_sw->p_node->general_info);
	if (!issu_capable) {
		return "N/A";
	}

	/*
	 * Set the status according to the switch state and the manager step.
	 *
	 * If the switch is in the isolated set and its state is APPROVED, this means that
	 * the switch didn't come out of APPROVED state.  (ie, it is stuck)
	 * If the swtich is in the upgrades set and the ISSU state is PRE_UPGRADE, this means
	 * that the switch is pending an upgrade.  This is a signal to SHARP to act.
	 * If the ISSU state is UPGRADING and the switch is in the upgrades state, it is
	 * in the process of upgrading.
	 */
	issu_state = ib_extended_switch_info_get_issu_state(&p_sw->subn_ext_switch_info);
	status = issu_state_to_string(issu_state);

	if (set_contains_guid(&p_mgr->isolated, p_sw) > 0) {
		if (issu_state == ISSU_APPROVED) {
			status = "APPROVED_TIMEOUT";
		}
	} else if (p_mgr->step == ISSU_STEP_PRE_UPGRADE) {
		if (set_contains_switch(&p_mgr->upgrades, p_sw) > 0) {
			status = "APPROVED_PENDING";
		}
	} else if (p_mgr->step == ISSU_STEP_UPGRADING) {
		if (set_contains_switch(&p_mgr->upgrades, p_sw) > 0) {
			status = "APPROVED_UPGRADING";
		}
	} else {
		// use the default string set above
	}

	return status;
}

void osm_issu_mgr_enqueue_trap(IN osm_sm_t *p_sm, IN uint16_t lid_ho, IN uint32_t state)
{
	osm_issu_mgr_t		*p_mgr;

	p_mgr = &p_sm->issu_mgr;
	if (!issu_is_working(p_mgr)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "ISSU trap received, but ISSU is disabled or not working\n");
		return;
	}

	if (state != 0) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "received trap for lid %u with state %u\n", lid_ho, state);
		issu_trap_enqueue(p_mgr, lid_ho, state);
	}
}

int osm_issu_mgr_process(IN osm_sm_t *p_sm)
{
	boolean_t		status;
	osm_issu_mgr_t		*p_mgr;

	status = FALSE;
	p_mgr = &p_sm->issu_mgr;

	/*
	 * If we aren't working, return.
	 */
	if (!issu_is_working(p_mgr)) {
		return 0;
	}

	/*
	 * Find all the ISSU capable switches and determine if we can clean up some isolated switches.
	 */
	issu_identify_capable_switches(p_mgr);
	status |= issu_resolve_isolated_switches(p_mgr, FALSE);

	/*
	 * If we found switches which are ISSU capable, perform upgrades.
	 */
	if (set_is_not_empty(&p_mgr->capable)) {
		if (p_mgr->step == ISSU_STEP_IDLE) {
			/*
			 * Look for switches which need upgrading.  If we find some, we transition to the WAIT step.
			 */
			status |= issu_identify_upgrades(p_mgr);
			if (set_is_not_empty(&p_mgr->upgrades)) {
				issu_preupgrade(p_mgr);
			}
		}

		if (p_mgr->step == ISSU_STEP_PRE_UPGRADE) {
			/*
			 * We just wait until the pre-upgrade timer expires and then start the upgrades.
			 * Note that there is a race condition here.  If the SM is starting the upgrade and
			 * just before that request the switch changes its mind and sets its state to IDLE,
			 * the switch will NOT upgrade.  When we verify the upgrade, we will see the IDLE
			 * state and think that the upgrade went smoothly.
			 */
			if (issu_is_timedout(p_mgr)) {
				status |= issu_perform_upgrades(p_mgr);
			}
		}
	}

	return status ? 1 : 0;
}


void osm_issu_mgr_pre_upgrade_time_change(IN osm_subn_t *p_subn, IN void *p_value)
{
	uint32_t		config_timer = *(uint32_t *)p_value;
	osm_issu_mgr_t		*p_mgr = &p_subn->p_osm->sm.issu_mgr;

	p_mgr->pre_upgrade_time = (config_timer != 0) ? config_timer : ISSU_PRE_UPGRADE_TIMEOUT;
	if (p_mgr->pre_upgrade_time < ISSU_PRE_UPGRADE_TIMEOUT) {
		p_mgr->pre_upgrade_time = ISSU_PRE_UPGRADE_TIMEOUT;
		OSM_LOG(p_mgr->p_log, OSM_LOG_INFO, "ISSU pre_upgrade timeout value too small, reset to default (%u)\n", p_mgr->pre_upgrade_time);
	}
}


void osm_issu_mgr_timeout_change(IN osm_subn_t *p_subn, IN void *p_value)
{
	uint32_t		config_timeout = *(uint32_t *)p_value;
	osm_issu_mgr_t		*p_mgr = &p_subn->p_osm->sm.issu_mgr;

	p_mgr->timeout = (config_timeout != 0) ? config_timeout : ISSU_UPGRADE_TIMEOUT;
	if (p_mgr->timeout < ISSU_UPGRADE_TIMEOUT) {
		p_mgr->timeout = ISSU_UPGRADE_TIMEOUT;
		OSM_LOG(p_mgr->p_log, OSM_LOG_INFO, "ISSU timeout value too small, reset to default (%u)\n", p_mgr->timeout);
	}
}


void osm_issu_mgr_mode_change(IN osm_subn_t *p_subn, IN void *p_value)
{
	uint32_t		config_mode = *(uint32_t *)p_value;
	osm_issu_mgr_t		*p_mgr = &p_subn->p_osm->sm.issu_mgr;

	/*
	 * The ISSU manager state has changed.  Get the new operational mode.
	 */
	if (config_mode >= ISSU_MODE_MAX) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B90E: invalid ISSU_mode (%u), ignoring\n", config_mode);
		return;
	}

	/*
	 * If the manager is being disabled, recover held back switches and clean up
	 * the state of the switches and traps.
	 */
	if (config_mode == ISSU_MODE_DISABLED) {
		if (issu_is_working(p_mgr)) {
			issu_resolve_isolated_switches(p_mgr, TRUE);

			set_remove_all(&p_mgr->capable);
			set_remove_all(&p_mgr->upgrades);
			set_remove_all(&p_mgr->isolated);
			issu_traps_remove_all(p_mgr);
		}
	}

	/*
	 * The manager is being enabled, make sure the resources are allocated.
	 */
	if (config_mode != ISSU_MODE_DISABLED) {
		if (!p_mgr->inited) {
			p_mgr->inited = issu_alloc_resources(p_mgr);
		}
	}

	p_mgr->mode = config_mode;
}

ib_api_status_t osm_issu_mgr_init(IN osm_issu_mgr_t *p_mgr, IN osm_sm_t *p_sm)
{
	int			fd;
	char			*dir_name;
	FILE			*stdckpt;
	time_t			now;
	time_t			wait_time;
	struct stat		statbuf;

	CL_ASSERT(sizeof(time_t) <= sizeof(void *));

	/*
	 * Initialize the issu mgr.
	 */
	memset((void *)p_mgr, 0, sizeof(osm_issu_mgr_t));
	p_mgr->p_sm = p_sm;
	p_mgr->p_log = p_sm->p_log;
	p_mgr->p_subn = p_sm->p_subn;
	p_mgr->mode = p_mgr->p_subn->opt.issu_mode;
	p_mgr->step = ISSU_STEP_IDLE;
	p_mgr->timeout = p_mgr->p_subn->opt.issu_timeout;
	p_mgr->pre_upgrade_time = p_mgr->p_subn->opt.issu_pre_upgrade_time;

	p_mgr->inited = issu_alloc_resources(p_mgr);

	/*
	 * Check the status of the checkpoint file.  Wait if the timeout period hasn't expired.
	 * This code is structured to avoid TOCTOU.
	 */
	if (p_mgr->inited) {
		errno = 0;

		/*
		 * Create the cache file name.
		 */
		dir_name = getenv("OSM_CACHE_DIR");
		if ((dir_name == NULL) || (dir_name[0] == '\0')) {
			dir_name = OSM_DEFAULT_CACHE_DIR;
		}

		snprintf(p_mgr->startup_file, PATH_MAX, "%s/%s", dir_name, ISSU_STARTUP_FILE);

		/*
		 * Check to see if we need to wait before starting.
		 */
		if ((stdckpt = fopen(p_mgr->startup_file, "r")) != NULL) {
			if (fstat(fileno(stdckpt), &statbuf) == 0) {
				wait_time = statbuf.st_mtime + p_mgr->p_subn->opt.issu_timeout;
				now = time(NULL);

				if (wait_time > now) {
					wait_time -= now;

					OSM_LOG(p_sm->p_log, OSM_LOG_INFO, STARTUP_WARNING_FORMAT, wait_time);
					printf(STARTUP_WARNING_FORMAT, wait_time);
					sleep(wait_time);
				}
			}

			fclose(stdckpt);
		} else if ((fd = open(p_mgr->startup_file, O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) >= 0) {
			close(fd);
		} else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B910: issu_mgr cannot open checkpoint file (%s)   errno=%d\n", p_mgr->startup_file, errno);
		}
	}

	/*
	 * Init is complete.
	 */
	if (p_mgr->inited) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_INFO, "issu_mgr is initialized\n");
	} else {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR B90F: issu_mgr is not initialized\n");
	}

	return (p_mgr->inited) ? IB_SUCCESS : IB_INSUFFICIENT_RESOURCES;
}


void osm_issu_mgr_destroy(IN osm_issu_mgr_t *p_mgr, IN osm_sm_t *p_sm)
{
	issu_dealloc_resources(p_mgr);
	memset((void *)p_mgr, 0, sizeof(osm_issu_mgr_t));

	OSM_LOG(p_sm->p_log, OSM_LOG_INFO, "osm_issu_mgr_destroy complete\n");
}
