/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2011 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 Sun Microsystems, Inc. All rights reserved.
 * Copyright (c) 2009-2011 ZIH, TU Dresden, Federal Republic of Germany. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_link_mgr_t.
 * This file implements the Link Manager object.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_LINK_MGR_C
#include <opensm/osm_sm.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_capability_mgr.h>

#define LINK_MGR_LEAF			TRUE
#define LINK_MGR_NON_LEAF		FALSE


static boolean_t ib_mlnx_extended_get_ame_bit(ib_mlnx_ext_port_info_t *epi)
{
	return ib_mlnx_extended_get_mepi_resvd1_bit(epi, 0, AME_BIT);
}

void ib_mlnx_extended_set_ame_bit(ib_mlnx_ext_port_info_t *epi, boolean_t enable)
{
	ib_mlnx_extended_set_mepi_resvd1_bit(epi, 0, AME_BIT, enable);
}

static boolean_t ib_mlnx_extended_get_sharp_bit(ib_mlnx_ext_port_info_t *epi)
{
	return ib_mlnx_extended_get_mepi_resvd1_bit(epi, 0, SHARP_AN_ENABLED_BIT);
}

static void ib_mlnx_extended_set_sharp_bit(ib_mlnx_ext_port_info_t *epi,
					   boolean_t enable){
	ib_mlnx_extended_set_mepi_resvd1_bit(epi, 0, SHARP_AN_ENABLED_BIT, enable);
}

static uint8_t link_mgr_get_smsl(IN osm_sm_t * sm, IN osm_physp_t * p_physp)
{
	osm_opensm_t *p_osm = sm->p_subn->p_osm;
	struct osm_routing_engine *re = p_osm->subn.routing_engine_used;
	ib_net16_t slid;
	ib_net16_t smlid;
	uint8_t sl;

	OSM_LOG_ENTER(sm->p_log);

	if (!(re && re->path_sl &&
	      (slid = osm_physp_get_base_lid(p_physp)))) {
		/*
		 * Use default SL if routing engine does not provide a
		 * path SL lookup callback.
		 */
		OSM_LOG_EXIT(sm->p_log);
		return sm->p_subn->opt.sm_sl;
	}

	smlid = sm->p_subn->sm_base_lid;

	/* Call into routing engine to find proper SL */
	sl = re->path_sl(re->context, sm->p_subn->opt.sm_sl,
			 slid, smlid);

	OSM_LOG_EXIT(sm->p_log);
	return sl;
}

static void link_mgr_set_sharp_cap(osm_sm_t *sm, osm_physp_t *p_physp)
{
	boolean_t enabled, required, supported;
	ib_mlnx_ext_port_info_t *epi = &p_physp->ext_port_info;
	osm_port_t *p_port;

	p_port = osm_get_port_by_guid(sm->p_subn, p_physp->port_guid);
	OSM_ASSERT(p_port);
	supported = ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
			IB_GENERAL_INFO_IS_SHARP_AGGR_NODE_SUPPORTED_BIT);

	if (!supported) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Sw 0x%016" PRIx64 " does "
			"not support SHArP, quit SHArP setting (port addr %p\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)), p_port);
		return;
	}

	switch (p_physp->p_node->conf_params.sharp_enabled) {
	case OSM_CAPABILITY_DISABLED:
		required = 0;
		break;
	case OSM_CAPABILITY_ENABLED:
		required = 1;
		break;
	case OSM_CAPABILITY_IGNORED:
	default:
		required = (sm->p_subn->opt.sharp_enabled == OSM_CAPABILITY_ENABLED);
		break;

	}

	enabled = ib_mlnx_extended_get_sharp_bit(epi);

	if (required == enabled)
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "No need to set SHArP support"
			" for Sw 0x%016" PRIx64 " (required: %d, enabled: %d)"
			"\n", cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			required, enabled);
	else {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Setting SHArP support for "
			"Sw 0x%016" PRIx64 " (required: %d, enabled: %d)\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			required, enabled);

		if (sm->p_subn->opt.device_configuration_file &&
		    p_physp->p_node->conf_params.sharp_enabled !=
		    OSM_CAPABILITY_IGNORED)
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Set SHARP required"
				" to %u, for Sw 0x%016" PRIx64 ", by device"
				" configuration file\n",
				required,
				cl_ntoh64(osm_physp_get_port_guid(p_physp)));

		ib_mlnx_extended_set_sharp_bit(epi, required);
		if (!required)
			/* We are about to disable sharp on the node */
			sm->sharp_nodes_disabled = TRUE;
		p_physp->ext_port_info_dirty = TRUE;
	}
}

static void link_mgr_set_router_lid_en(osm_sm_t *sm, osm_physp_t *p_physp)
{
	osm_port_t *p_port;
	boolean_t required_en, current_en;
	ib_mlnx_ext_port_info_t *epi = &p_physp->ext_port_info;

	p_port = osm_get_port_by_guid(sm->p_subn, p_physp->port_guid);

	/*
	 * Don't set the router lid enable (steering) bit on port that
	 * has no capability.
	 */
	if (!p_port ||
	    !ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
					    IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT))
		return;

	/* Enable the bit when the FLID feature is enabled and the port has FLID */
	required_en = osm_subn_is_flids_enabled(sm->p_subn) && (p_port->flid != 0);
	current_en = ib_mlnx_extended_get_router_lid_bit(epi);

	if (required_en == current_en)
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "No need to set router lid enable bit"
			" for HCA 0x%016" PRIx64 " (required: %d, enabled: %d)"
			"\n", cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			required_en, current_en);
	else {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Setting router lid enable bit for "
			"HCA 0x%016" PRIx64 " (required: %d, enabled: %d)\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			required_en, current_en);

		ib_mlnx_extended_set_router_lid_bit(epi, required_en);
		p_physp->ext_port_info_dirty = TRUE;
	}
}

uint8_t link_mgr_calc_req_hoq_life(osm_sm_t * sm, boolean_t is_leaf, osm_physp_t * p_physp)
{
	/* Prioritize Device Configuration file over SM configuration file */
	if (p_physp->conf_params.hoq_life != OSM_HOQ_LIFE_INVALID)
		return p_physp->conf_params.hoq_life;

	return is_leaf ? sm->p_subn->opt.leaf_head_of_queue_lifetime :
	    sm->p_subn->opt.head_of_queue_lifetime;
}

void link_mgr_calc_timeout_mechanism_params(osm_sm_t * sm, osm_physp_t * p_physp,
					    ib_port_info_t * p_pi)
{
	uint8_t node_type = osm_node_get_type(osm_physp_get_node_ptr(p_physp));
	uint8_t remote_node_type;
	osm_physp_t *p_remote_physp;

	if (node_type == IB_NODE_TYPE_ROUTER) {
		ib_port_info_set_hoq_lifetime(p_pi, link_mgr_calc_req_hoq_life(sm, LINK_MGR_LEAF, p_physp));
	} else if (node_type == IB_NODE_TYPE_SWITCH) {
		p_remote_physp = osm_physp_get_remote(p_physp);
		/* Is remote end CA or router (a leaf port) ? */
		remote_node_type = osm_node_get_type(osm_physp_get_node_ptr(p_remote_physp));
		if (remote_node_type != IB_NODE_TYPE_SWITCH) {
			ib_port_info_set_hoq_lifetime(p_pi,
			    link_mgr_calc_req_hoq_life(sm, LINK_MGR_LEAF, p_physp));
			ib_port_info_set_vl_stall_count(p_pi, sm->p_subn->opt.leaf_vl_stall_count);
		} else {
			ib_port_info_set_hoq_lifetime(p_pi,
			     link_mgr_calc_req_hoq_life(sm, LINK_MGR_NON_LEAF, p_physp));
			ib_port_info_set_vl_stall_count(p_pi, sm->p_subn->opt.vl_stall_count);
		}
	}
}
static int link_mgr_set_physp_pi(osm_sm_t * sm, IN osm_physp_t * p_physp,
				 IN uint8_t port_state)
{
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_port_info_t *p_pi = (ib_port_info_t *) payload;
	const ib_port_info_t *p_old_pi;
	osm_madw_t *p_madw;
	osm_madw_context_t context;
	osm_node_t *p_node;
	uint8_t port_num, op_vls, smsl = OSM_DEFAULT_SL;
	boolean_t esp0 = FALSE, send_set = FALSE;
	osm_physp_t *p_remote_physp, *physp0 = NULL;
	int ret = 0;
	boolean_t update_mkey = FALSE;
	ib_net64_t m_key;
	ib_net16_t capability_mask2;
	osm_port_t *p_port = NULL;
	boolean_t client_rereg = FALSE;
	boolean_t suppress_mc_pkey_traps_capability;
	uint8_t suppress_mc_pkey_traps = sm->p_subn->opt.suppress_mc_pkey_traps & 1;
	uint8_t is_trunk_link = 0;
	boolean_t set_op_vl_to_zero = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	p_node = osm_physp_get_node_ptr(p_physp);
	OSM_ASSERT(p_node);

	p_old_pi = &p_physp->port_info;

	port_num = osm_physp_get_port_num(p_physp);

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		physp0 = osm_node_get_physp_ptr(p_node, 0);
		p_port = osm_get_port_by_guid(sm->p_subn, osm_physp_get_port_guid(physp0));
	} else {
		p_port = osm_get_port_by_guid(sm->p_subn, osm_physp_get_port_guid(p_physp));
	}

	OSM_ASSERT(p_port);

	suppress_mc_pkey_traps_capability =
	    p_port->p_physp->port_info.capability_mask &
	    IB_PORT_CAP_HAS_MCAST_PKEY_TRAP_SUPPRESS;

	memcpy(payload, p_old_pi, sizeof(ib_port_info_t));

	if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH ||
	    port_num == 0) {
		/* Need to make sure LID and SMLID fields in PortInfo are not 0 */
		if(!p_pi->base_lid) {
			p_pi->base_lid = p_port->lid;
			sm->lid_mgr.dirty = TRUE;
			send_set = TRUE;
		}

		/* we are initializing the ports with our local sm_base_lid */
		p_pi->master_sm_base_lid = sm->p_subn->sm_base_lid;
		if (p_pi->master_sm_base_lid != p_old_pi->master_sm_base_lid)
			send_set = TRUE;

		/* Determine whether to send client reregistration */
		if (sm->p_subn->opt.client_rereg_mode == OSM_CLIENT_REREG_MODE_LINK_MGR &&
		    p_port->client_rereg &&
		    p_pi->capability_mask & IB_PORT_CAP_HAS_CLIENT_REREG &&
		    p_port->is_reachable) {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Setting client rereg on %s, port %d\n",
				p_port->p_node->print_desc, p_port->p_physp->port_num);
			ib_port_info_set_client_rereg(p_pi, 1);
			client_rereg = TRUE;
			send_set = TRUE;
		}
	}

	if (port_num == 0) {
		/*
		   CAs don't have a port 0, and for switch port 0,
		   we need to check if this is enhanced or base port 0.
		   For base port 0 the following parameters are not valid
		   (IBA 1.2.1 p.830 table 146).
		 */
		if (!p_node->sw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 4201: "
				"Cannot find switch by guid: 0x%" PRIx64 "\n",
				cl_ntoh64(p_node->node_info.node_guid));
			goto Exit;
		}

		if (ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info)
		    == FALSE) {

			/* Even for base port 0 we might have to set smsl
			   (if we are using lash routing) */
			smsl = link_mgr_get_smsl(sm, p_physp);
			if (smsl != ib_port_info_get_master_smsl(p_old_pi)) {
				send_set = TRUE;
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Setting SMSL to %d on port 0 GUID 0x%016"
					PRIx64 "\n", smsl,
					cl_ntoh64(osm_physp_get_port_guid
						  (p_physp)));
			/* Enter if base lid and master_sm_lid didn't change */
			} else if(send_set == FALSE) {
				/* This means the switch doesn't support
				   enhanced port 0 and we don't need to
				   change SMSL. Can skip it. */
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Skipping port 0, GUID 0x%016" PRIx64
					"\n",
					cl_ntoh64(osm_physp_get_port_guid
						  (p_physp)));
				/*
				 * No need to handle port0
				 * Only thing to handle is sharp enablement
				 * set m_key to be used in case that mlnx epi
				 * will need to be set
				 */
				m_key = ib_port_info_get_m_key(p_pi);
				goto Send;
			}
		} else
			esp0 = TRUE;
	}

	/*
	   Should never write back a value that is bigger then 3 in
	   the PortPhysicalState field - so can not simply copy!

	   Actually we want to write there:
	   port physical state - no change,
	   link down default state = polling
	   port state - as requested.
	 */
	p_pi->state_info2 = 0x02;
	ib_port_info_set_port_state(p_pi, port_state);

	/* Determine ports' M_Key and CapabilityMask2 */
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH &&
	    osm_physp_get_port_num(p_physp) != 0) {
		m_key = ib_port_info_get_m_key(&physp0->port_info);
		capability_mask2 = physp0->port_info.capability_mask2;
	} else {
		m_key = ib_port_info_get_m_key(p_pi);
		capability_mask2 = p_pi->capability_mask2;
	}

	/* Check whether this is base port0 smsl handling only */
	if (port_num == 0 && esp0 == FALSE) {
		ib_port_info_set_master_smsl(p_pi, smsl);
		goto Send;
	}

	/*
	   PAST THIS POINT WE ARE HANDLING EITHER A NON PORT 0 OR ENHANCED PORT 0
	 */

	if (ib_port_info_get_link_down_def_state(p_pi) !=
	    ib_port_info_get_link_down_def_state(p_old_pi))
		send_set = TRUE;

	/* didn't get PortInfo before */
	if (!ib_port_info_get_port_state(p_old_pi))
		send_set = TRUE;

	/* Link Manager also handles end-port configuration (such as FLIDs and client reregister) */

	/* we only change port fields if we do not change state */
	if (port_state == IB_LINK_NO_CHANGE) {
		/* The following fields are relevant only for CA port, router, or Enh. SP0 */
		if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH ||
		    port_num == 0) {
			p_pi->m_key = osm_physp_calc_mkey(sm->p_subn, p_physp);
			if (memcmp(&p_pi->m_key, &p_old_pi->m_key,
				   sizeof(p_pi->m_key))) {
				update_mkey = TRUE;
				send_set = TRUE;
			}

			if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA &&
			    ib_general_info_is_cap_enabled(&p_node->general_info,
						IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT)) {
				/* zero previous FLID */
				ib_port_info_set_port_flid(p_pi, p_port->flid);
				if (p_pi->subnet_prefix != p_old_pi->subnet_prefix) {
					send_set = TRUE;
					OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
						"Port guid 0x%" PRIx64 " has different FLID %u, "
						"changing to FLID %u prefix 0x%lx old prefix 0x%lx\n",
						cl_ntoh64(osm_physp_get_port_guid(p_physp)),
						ib_port_info_get_port_flid(p_old_pi),
						cl_ntoh16(p_port->flid), cl_ntoh64(p_pi->subnet_prefix),
						cl_ntoh64(p_old_pi->subnet_prefix));
				}
			} else {
				p_pi->subnet_prefix = sm->p_subn->opt.subnet_prefix;
				if (memcmp(&p_pi->subnet_prefix,
					   &p_old_pi->subnet_prefix,
					   sizeof(p_pi->subnet_prefix)))
					send_set = TRUE;
			}

			smsl = link_mgr_get_smsl(sm, p_physp);
			if (smsl != ib_port_info_get_master_smsl(p_old_pi)) {

				ib_port_info_set_master_smsl(p_pi, smsl);

				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Setting SMSL to %d on GUID 0x%016"
					PRIx64 ", port %d\n", smsl,
					cl_ntoh64(osm_physp_get_port_guid
						  (p_physp)), port_num);

				send_set = TRUE;
			}

			p_pi->m_key_lease_period =
			    sm->p_subn->opt.m_key_lease_period;
			if (memcmp(&p_pi->m_key_lease_period,
				   &p_old_pi->m_key_lease_period,
				   sizeof(p_pi->m_key_lease_period)))
				send_set = TRUE;

			p_pi->mkey_lmc = 0;
			ib_port_info_set_mpb(p_pi, sm->p_subn->opt.m_key_protect_bits);
			if (osm_lid_mgr_port_needs_rtr_lmc(sm->p_subn, p_port) ||
			    osm_lid_mgr_port_needs_prism_an_lmc(sm->p_subn, p_port))
				ib_port_info_set_lmc(p_pi, osm_subn_get_max_lmc(sm->p_subn));
			else if ((esp0 == FALSE || sm->p_subn->opt.lmc_esp0) &&
				 !osm_node_is_gpu(p_node))
				ib_port_info_set_lmc(p_pi, sm->p_subn->opt.lmc);

			if (ib_port_info_get_lmc(p_old_pi) !=
			    ib_port_info_get_lmc(p_pi) ||
			    ib_port_info_get_mpb(p_old_pi) !=
			    ib_port_info_get_mpb(p_pi))
				send_set = TRUE;

			ib_port_info_set_timeout(p_pi,
						 sm->p_subn->opt.
						 subnet_timeout);
			if (ib_port_info_get_timeout(p_pi) !=
			    ib_port_info_get_timeout(p_old_pi))
				send_set = TRUE;
		}

		/*
		   Several timeout mechanisms:
		 */
		p_remote_physp = osm_physp_get_remote(p_physp);
		if (port_num != 0 && p_remote_physp) {
			link_mgr_calc_timeout_mechanism_params(sm, p_physp, p_pi);

			if (ib_port_info_get_hoq_lifetime(p_pi) !=
			    ib_port_info_get_hoq_lifetime(p_old_pi) ||
			    ib_port_info_get_vl_stall_count(p_pi) !=
			    ib_port_info_get_vl_stall_count(p_old_pi))
				send_set = TRUE;
		}

		ib_port_info_set_phy_and_overrun_err_thd(p_pi,
							 sm->p_subn->opt.
							 local_phy_errors_threshold,
							 sm->p_subn->opt.
							 overrun_errors_threshold);
		if (p_pi->error_threshold != p_old_pi->error_threshold)
			send_set = TRUE;

		/*
		   Set the easy common parameters for all port types,
		   then determine the neighbor MTU.
		 */
		if (sm->p_subn->opt.force_link_width &&
		    (sm->p_subn->opt.force_link_width < IB_LINK_WIDTH_ACTIVE_2X ||
		     (capability_mask2 &
		      IB_PORT_CAP2_IS_LINK_WIDTH_2X_SUPPORTED)) &&
		    (sm->p_subn->opt.force_link_width != IB_LINK_WIDTH_SET_LWS ||
		     p_pi->link_width_enabled != p_pi->link_width_supported)) {
			p_pi->link_width_enabled = sm->p_subn->opt.force_link_width;
			if (p_pi->link_width_enabled != p_old_pi->link_width_enabled)
				send_set = TRUE;
		}

		/* Set parameter for suppressing mc_pkey traps */
		if (suppress_mc_pkey_traps_capability) {
			p_pi->subnet_timeout =
			    (p_pi->subnet_timeout & ~IB_PORT_MC_PKEY_TRAP_SUPPRESSION_MASK) |
				(suppress_mc_pkey_traps << IB_PORT_MC_PKEY_TRAP_SUPPRESSION_SHIFT);

			if ((p_pi->subnet_timeout & IB_PORT_MC_PKEY_TRAP_SUPPRESSION_MASK) !=
			    (p_old_pi->subnet_timeout & IB_PORT_MC_PKEY_TRAP_SUPPRESSION_MASK)) {
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Setting mc pkey trap suppression "
					"to %d on GUID 0x%016"PRIx64", port %d\n",
					suppress_mc_pkey_traps,
					cl_ntoh64(osm_physp_get_port_guid
					(p_physp)), port_num);
				send_set = TRUE;
			}
		}

		/* calc new op_vls */
		op_vls =
		    osm_physp_calc_link_op_vls(sm->p_log, sm->p_subn, p_physp,
					       ib_port_info_get_op_vls(p_old_pi));
		/*
		 * If QoS config VL is supported, setting op_vl to a non-zero value will
		 * redistribute the VL buffers equally, by that disabling the feature.
		 * Therefore if there is no change between the calculated op_vls
		 * and the old ones and the port has an active profile
		 * (that means QoSVlConfig will be sent to configure it),
		 * need to set the value for op_vls to 0 (no_change).
		 */
		ib_port_info_set_op_vls(p_pi, op_vls);
		if ((ib_port_info_get_op_vls(p_pi) != ib_port_info_get_op_vls(p_old_pi)) ||
		    (p_physp->qos_config_vl_profile == IB_QOS_CONFIG_VL_INVALID_PROFILE &&
		     (p_physp->qos_config_vl_profile != p_physp->qos_config_vl_prev_profile ||
		      sm->p_subn->first_time_master_sweep == TRUE))) {
			send_set = TRUE;
		} else if (sm->p_subn->opt.qos_config_vl_enabled &&
			   (ib_port_info_get_port_state(p_old_pi) != IB_LINK_INIT) &&
			   (p_physp->qos_config_vl_profile != IB_QOS_CONFIG_VL_INVALID_PROFILE) &&
			   ib_general_info_is_cap_enabled(&p_node->general_info,
							  IB_GENERAL_INFO_IS_QOS_CONFIG_VL_SUPPORTED_BIT)) {
			set_op_vl_to_zero = TRUE;
		}

		/* provide the vl_high_limit from the qos mgr */
		if (sm->p_subn->opt.qos &&
		    p_physp->vl_high_limit != p_old_pi->vl_high_limit) {
			send_set = TRUE;
			p_pi->vl_high_limit = p_physp->vl_high_limit;
		}
	}

Send:
	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.active_transition = FALSE;
	if (port_state != IB_LINK_NO_CHANGE &&
	    port_state != ib_port_info_get_port_state(p_old_pi)) {
		send_set = TRUE;
		if (port_state == IB_LINK_ACTIVE)
			context.pi_context.active_transition = TRUE;
	}

	context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = TRUE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.client_rereg = client_rereg;

	/* We need to send the PortInfoSet request with the new sm_lid
	   in the following cases:
	   1. There is a change in the values (send_set == TRUE)
	   2. This is a switch external port (so it wasn't handled yet by
	   osm_lid_mgr) and first_time_master_sweep flag on the subnet is TRUE,
	   which means the SM just became master, and it then needs to send at
	   PortInfoSet to every port.
	 */
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH && port_num
	    && sm->p_subn->first_time_master_sweep == TRUE)
		send_set = TRUE;

	if (p_physp->port_info_dirty)
		send_set = TRUE;

	if (!send_set)
		goto SEND_EPI;

	if (set_op_vl_to_zero)
		ib_port_info_set_op_vls(p_pi, 0);

	p_madw = osm_prepare_req_set_pi(sm, p_physp, port_num, payload, FALSE, m_key,
					&context, TRUE);
	if (!p_madw)
		ret = -1;
	else
		cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);

	/* If we sent a new mkey above, update our guid2mkey map
	   now, on the assumption that the SubnSet succeeds
	 */
	if (update_mkey)
		osm_db_guid2key_set(sm->p_subn->p_g2m,
				    cl_ntoh64(p_physp->port_guid),
				    cl_ntoh64(p_pi->m_key));
SEND_EPI:

	if (sm->p_subn->opt.sharp_enabled && (port_num == 0))
		link_mgr_set_sharp_cap(sm, p_physp);

	if ((osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH || port_num == 0) &&
	    ib_general_info_is_cap_enabled(&p_node->general_info,
					   IB_GENERAL_INFO_IS_GLOBAL_OOO_SUPPORTED_BIT)) {
		/* Set ooo_sl_mask to non-zero only when Adaptive Routing is configured */
		ib_net16_t ar_sl_mask = sm->ucast_mgr.ar_configured ?
					cl_hton16(sm->p_subn->opt.ar_sl_mask) : 0;
		if (p_physp->ext_port_info.ooo_sl_mask != ar_sl_mask) {
			p_physp->ext_port_info.ooo_sl_mask = ar_sl_mask;
			p_physp->ext_port_info_dirty = TRUE;
		}
	}

	/* Configure Adaptive Timeout SL mask */
	if ((osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH || port_num == 0) &&
	    ib_general_info_is_cap_enabled(&p_node->general_info,
					   IB_GENERAL_INFO_IS_ADAPTIVE_TO_SUPPORTED_BIT)) {
		ib_net16_t adaptive_timeout_sl_mask = cl_hton16(sm->p_subn->opt.adaptive_timeout_sl_mask);
		if (p_physp->ext_port_info.adaptive_timeout_sl_mask != adaptive_timeout_sl_mask) {
			p_physp->ext_port_info.adaptive_timeout_sl_mask = adaptive_timeout_sl_mask;
			p_physp->ext_port_info_dirty = TRUE;
		}
	}

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA)
		link_mgr_set_router_lid_en(sm, p_physp);

	/* Configure if the link is access or trunk link */
	p_remote_physp = osm_physp_get_remote(p_physp);
	if (osm_subn_is_nvlink_enabled(sm->p_subn) && port_num != 0 && p_remote_physp &&
	    (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)) {
		if (osm_node_get_type(p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH) {
			is_trunk_link = 1;
		} else {
			is_trunk_link = 0;
		}

		if (ib_mlnx_epi_get_is_trunk_port(&p_physp->ext_port_info) != is_trunk_link) {
			ib_mlnx_epi_set_access_trunk(&p_physp->ext_port_info, is_trunk_link);
			p_physp->ext_port_info_dirty = TRUE;
		}
	}

	if (p_physp->ext_port_info_dirty) {
		p_physp->ext_port_info_updated = FALSE;
		p_madw = osm_prepare_req_set(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
					     (uint8_t *)&p_physp->ext_port_info,
					     sizeof(p_physp->ext_port_info),
					     IB_MAD_ATTR_MLNX_EXTENDED_PORT_INFO,
					     cl_hton32(port_num), FALSE, m_key,
					     0, CL_DISP_MSGID_NONE,
					     osm_node_get_node_guid(p_node), &context);

		if (!p_madw)
			ret = -1;
		else
			cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);
	}

Exit:
	p_physp->port_info_dirty = FALSE;
	p_physp->ext_port_info_dirty = FALSE;
	OSM_LOG_EXIT(sm->p_log);
	return ret;
}

int link_mgr_set_port_state(osm_sm_t *sm, osm_physp_t *p_physp, uint8_t const phys_state,
			    uint8_t const port_state)
{
	uint8_t payload[IB_SMP_DATA_SIZE];
	osm_madw_context_t context;
	ib_port_info_t *pi = (ib_port_info_t *)payload;
	osm_physp_t *p_physp0 = NULL;
	osm_port_t *p_port = NULL;
	ib_net64_t m_key;
	ib_api_status_t status = IB_SUCCESS;
	osm_madw_t *p_madw;

	if (osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH) {
		p_physp0 = osm_node_get_physp_ptr(p_physp->p_node, 0);
		m_key = ib_port_info_get_m_key(&p_physp0->port_info);
		p_port = osm_get_port_by_guid(sm->p_subn, osm_physp_get_port_guid(p_physp0));
	} else {
		m_key = ib_port_info_get_m_key(&p_physp->port_info);
		p_port = osm_get_port_by_guid(sm->p_subn, osm_physp_get_port_guid(p_physp));
	}

	OSM_ASSERT(p_port);

	memcpy(payload, &p_physp->port_info, sizeof(ib_port_info_t));

	ib_port_info_set_port_state(pi, port_state);
	ib_port_info_set_port_phys_state(phys_state, pi);

	/* Issue set of PortInfo */
	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.node_guid = osm_node_get_node_guid(p_physp->p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = TRUE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;
	context.pi_context.client_rereg = FALSE;

	if (osm_node_get_type(p_physp->p_node) != IB_NODE_TYPE_SWITCH ||
	    osm_physp_get_port_num(p_physp) == 0) {
		if (!pi->base_lid)
			pi->base_lid = p_port->lid;
		pi->master_sm_base_lid = sm->p_subn->sm_base_lid;
	}

	p_madw = osm_prepare_req_set_pi(sm, p_physp, osm_physp_get_port_num(p_physp),
					payload, FALSE, m_key, &context, TRUE);
	if (!p_madw)
		status = IB_ERROR;
	else
		cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);

	return status;
}

/* returns true if the specified port and it's peer are both reachable */
static boolean_t link_mgr_is_link_reachable(osm_sm_t * sm,
					    osm_physp_t * p_physp)
{
	osm_physp_t *p_remote_physp;
	osm_node_t *p_remote_node;

	OSM_ASSERT(p_physp);

	if (lookup_isolated_port(sm, p_physp->p_node, p_physp->port_num))
		return FALSE;

	p_remote_physp = p_physp->p_remote_physp;
	if (!p_remote_physp || !osm_physp_is_valid(p_remote_physp))
		return FALSE;

	p_remote_node = p_remote_physp->p_node;
	if (lookup_isolated_port(sm, p_remote_node, p_remote_physp->port_num))
		return FALSE;

	return TRUE;
}

static boolean_t link_mgr_is_link_asymmetric(IN osm_sm_t * sm,
					     IN osm_physp_t * p_physp,
					     OUT boolean_t * p_is_port_asymmetric)
{
	osm_physp_t *p_physp_remote;
	boolean_t is_asymmetric_link = FALSE;

	OSM_ASSERT(p_physp);

	if (p_is_port_asymmetric)
		*p_is_port_asymmetric = FALSE;

	if (!sm->p_subn->is_planarized)
		goto Exit;

	if (osm_hierarchy_is_asymmetric_aport_physp(p_physp)) {
		if (p_is_port_asymmetric)
			*p_is_port_asymmetric = TRUE;

		is_asymmetric_link = TRUE;
		goto Exit;
	}

	p_physp_remote = p_physp->p_remote_physp;
	if (p_physp_remote && osm_physp_is_valid(p_physp_remote) &&
	    osm_hierarchy_is_asymmetric_aport_physp(p_physp_remote))
		is_asymmetric_link = TRUE;

Exit:
	return is_asymmetric_link;
}

boolean_t osm_link_mgr_is_init_state_expected(osm_sm_t * sm,
					      osm_physp_t * p_physp)
{
	return !link_mgr_is_link_reachable(sm, p_physp) ||
	    link_mgr_is_link_asymmetric(sm, p_physp, NULL);
}

static int link_mgr_process_node(osm_sm_t * sm, IN osm_node_t * p_node,
				 IN const uint8_t link_state,
				 IN boolean_t process_endports)
{
	osm_physp_t *p_physp, *p_physp_remote;
	uint32_t i, num_physp;
	unsigned int port_reset_required;
	int ret = 0;
	uint8_t current_state;
	boolean_t is_endport_link;
	osm_neighbors_info_t *p_neighbors_info = NULL;
	boolean_t is_asymmetric_link = FALSE, is_port_asymmetric = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	if (link_state == IB_LINK_DOWN) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Node 0x%" PRIx64 " moving unhealthy or isolated links to %s\n",
			osm_node_get_node_guid_ho(p_node),
			ib_get_port_state_str(link_state));
	}
	else {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Node 0x%" PRIx64 " attempting to go to %s\n",
			osm_node_get_node_guid_ho(p_node),
			ib_get_port_state_str(link_state));
	}

	if (sm->p_subn->topo_config_enabled)
		p_neighbors_info = osm_topoconfig_lookup(sm->p_subn->p_topoconfig_hash_tbl, osm_node_get_node_guid(p_node));

	/*
	   Set the PortInfo for every Physical Port associated
	   with this Node.  Start iterating with port 1, since the linkstate
	   is not applicable to the management port on switches.
	 */
	num_physp = osm_node_get_num_physp(p_node);
	for (i = 0; i < num_physp; i++) {
		/*
		   Don't bother doing anything if this Physical Port is not valid.
		   or if the state of the port is already better then the
		   specified state.
		 */
		p_physp = osm_node_get_physp_ptr(p_node, (uint8_t) i);
		if (!p_physp)
			continue;

		is_asymmetric_link = is_port_asymmetric = FALSE;

		current_state = osm_physp_get_port_state(p_physp);
		if (i && sm->p_subn->topo_config_enabled) {
			if (osm_topoconfig_neighbor_is_disabled(p_neighbors_info, i)) {
				if (ib_port_info_get_port_phys_state(&p_physp->port_info) != IB_PORT_PHYS_STATE_DISABLED &&
				    link_state == IB_LINK_DOWN) {
					OSM_LOG(sm->p_log, OSM_LOG_INFO,
						"Remote port of node guid 0x%" PRIx64 " port %d is marked as disabled, "
						"setting physical state to Disabled\n",
						osm_node_get_node_guid_ho(p_node), i);
					link_mgr_set_port_state(sm, p_physp, IB_PORT_PHYS_STATE_DISABLED, IB_LINK_NO_CHANGE);
				}
				continue;
			} else {
				if (ib_port_info_get_port_phys_state(&p_physp->port_info) == IB_PORT_PHYS_STATE_DISABLED &&
				    link_state == IB_LINK_DOWN) {
					OSM_LOG(sm->p_log, OSM_LOG_INFO,
						"Node guid 0x%" PRIx64 " port %d is disabled, "
						"setting physical state to Polling\n",
						osm_node_get_node_guid_ho(p_node), i);
					link_mgr_set_port_state(sm, p_physp, IB_PORT_PHYS_STATE_POLLING, IB_LINK_NO_CHANGE);
				}
			}
		}

		if (current_state == IB_LINK_DOWN)
			continue;

		p_physp_remote = osm_physp_get_remote(p_physp);

		if (i == 0 || osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH ||
		    (p_physp_remote &&
		     osm_node_get_type(osm_physp_get_node_ptr(p_physp_remote)) != IB_NODE_TYPE_SWITCH))
			is_endport_link = TRUE;
		else
			is_endport_link = FALSE;

		is_asymmetric_link = link_mgr_is_link_asymmetric(sm, p_physp, &is_port_asymmetric);

		/*
		    Set PortState to DOWN in case Remote Physical Port is
		    unreachable or port belongs to asymmetric APort. We have to check this for all ports, except
		    port zero.
		 */
		if ((i != 0) && (!link_mgr_is_link_reachable(sm, p_physp) || is_asymmetric_link)) {
			if (current_state != IB_LINK_INIT && link_state == IB_LINK_DOWN) {
				OSM_LOG(sm->p_log, OSM_LOG_INFO,
					"%sport of node guid 0x%" PRIx64
					" port %d is %s, setting state to INIT\n",
					is_port_asymmetric ? "" : "Remote ",
					osm_node_get_node_guid_ho(p_node), i,
					is_asymmetric_link ? "asymmetric APort" : "unreachable");
				link_mgr_set_physp_pi(sm, p_physp, IB_LINK_DOWN);
			}
			continue;
		}

		if (process_endports != is_endport_link)
			continue;

		/*
		 * If port is marked for logical reset, no need to send PortInfo
		 * If port is marked for physical reset, we should only send PortInfo
		 * when link_state == IB_LINK_NO_CHANGE.
		 */
		port_reset_required = p_physp->port_reset_required;
		if ((port_reset_required == OSM_PORT_RESET_TYPE_LOGICAL) ||
		    (port_reset_required == OSM_PORT_RESET_TYPE_PHYSICAL &&
		     link_state != IB_LINK_NO_CHANGE))
			continue;

		/*
		   Normally we only send state update if state is lower
		   then required state. However, we need to send update if
		   no state change required.
		 */
		if (link_state != IB_LINK_NO_CHANGE &&
		    link_state <= current_state)
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Physical port %u already %s. Skipping\n",
				p_physp->port_num,
				ib_get_port_state_str(current_state));
		else if (link_state == IB_LINK_ACTIVE && current_state == IB_LINK_INIT)
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Physical port %u is in INIT state. Skipping activation\n",
				p_physp->port_num);
		else if (link_mgr_set_physp_pi(sm, p_physp, link_state))
			ret = -1;
	}

	OSM_LOG_EXIT(sm->p_log);
	return ret;
}

int osm_link_mgr_process(osm_sm_t * sm, IN const uint8_t link_state, boolean_t process_endports)
{
	cl_qmap_t *p_node_guid_tbl;
	cl_qmap_t *p_port_guid_tbl;
	osm_node_t *p_node;
	osm_port_t *p_port;
	int ret = 0;
	int i;

	OSM_LOG_ENTER(sm->p_log);

	p_node_guid_tbl = &sm->p_subn->node_guid_tbl;
	p_port_guid_tbl = &sm->p_subn->port_guid_tbl;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	for (p_node = (osm_node_t *) cl_qmap_head(p_node_guid_tbl);
	     p_node != (osm_node_t *) cl_qmap_end(p_node_guid_tbl);
	     p_node = (osm_node_t *) cl_qmap_next(&p_node->map_item))
		if (link_mgr_process_node(sm, p_node, link_state, process_endports))
			ret = -1;

	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item))
		for (i = 0; i < sm->p_subn->opt.max_wire_smps_per_device; i++)
			osm_port_send_next_mad(sm, p_port);

	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
	return ret;
}

static boolean_t link_mgr_set_physp_speed(IN osm_sm_t * sm,
					  IN osm_physp_t * p_physp)
{
	ib_port_info_t *p_pi;
	osm_physp_t *p_rem_physp;
	osm_node_t *p_node;
	uint8_t enabled_speed, new_enabled_speed;
	uint8_t supported_speed, rem_supported_speed;
	boolean_t port_speed_changed = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	if (!sm->p_subn->opt.force_link_speed)
		goto Exit;;

	p_rem_physp = osm_physp_get_remote(p_physp);
	if (!p_rem_physp)
		goto Exit;;

	p_pi = &p_physp->port_info;
	p_node = p_physp->p_node;

	enabled_speed = ib_port_info_get_link_speed_enabled(p_pi);
	supported_speed = ib_port_info_get_link_speed_sup(p_pi);
	rem_supported_speed =
	    ib_port_info_get_link_speed_sup(&p_rem_physp->port_info);

	if (sm->p_subn->opt.force_link_speed == IB_LINK_SPEED_SET_LSS) {
		if (enabled_speed != supported_speed) {
			ib_port_info_set_link_speed_enabled(p_pi,
			    sm->p_subn->opt.force_link_speed);
			port_speed_changed = TRUE;
		}
	} else {
		new_enabled_speed = supported_speed &
				    sm->p_subn->opt.force_link_speed;
		if (!new_enabled_speed) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 4202: "
				"Can't change enabled speed of node "
				"0x%" PRIx64 ", port %d: speed is not"
				"supported by port\n",
				cl_ntoh64(p_node->node_info.node_guid),
				osm_physp_get_port_num(p_physp));
		} else if (!(new_enabled_speed & rem_supported_speed)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 4203: "
				"Can't change enabled speed of node "
				"0x%" PRIx64 ", port %d: speed is not"
				"supported by peer\n",
				cl_ntoh64(p_node->node_info.node_guid),
				osm_physp_get_port_num(p_physp));
		} else if (new_enabled_speed != enabled_speed) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Setting link_speed_enabled of "
				"of node 0x%" PRIx64 ", port %d to "
				"0x%x\n",
				cl_ntoh64(p_node->node_info.node_guid),
				osm_physp_get_port_num(p_physp),
				new_enabled_speed);
			ib_port_info_set_link_speed_enabled(p_pi,
			    new_enabled_speed);
			port_speed_changed = TRUE;
		}
	}

Exit:
	p_physp->port_info_dirty |= port_speed_changed;

	OSM_LOG_EXIT(sm->p_log);
	return port_speed_changed;
}

static boolean_t link_mgr_set_physp_ext_speed(IN osm_sm_t * sm,
					      IN osm_physp_t * p_physp)
{
	ib_port_info_t *p_pi;
	osm_physp_t *p_rem_physp, *physp0;
	osm_node_t *p_node;
	ib_net32_t cap_mask;
	uint8_t enabled_speed, new_enabled_speed;
	uint8_t supported_speed;
	boolean_t port_speed_changed = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	if (!sm->p_subn->opt.force_link_speed_ext)
		goto Exit;

	p_rem_physp = osm_physp_get_remote(p_physp);
	if (!p_rem_physp)
		goto Exit;

	p_pi = &p_physp->port_info;
	p_node = p_physp->p_node;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		physp0 = osm_node_get_physp_ptr(p_node, 0);
		cap_mask = physp0->port_info.capability_mask;
	} else
		cap_mask = p_pi->capability_mask;

	if (cap_mask & IB_PORT_CAP_HAS_EXT_SPEEDS) {
		enabled_speed = ib_port_info_get_link_speed_ext_enabled(p_pi);
		supported_speed = ib_port_info_get_link_speed_ext_sup(p_pi);

		switch (sm->p_subn->opt.force_link_speed_ext) {
		case IB_LINK_SPEED_EXT_DISABLE:
			if (enabled_speed) {
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Disabling extended speeds of node 0x%"
					PRIx64 ", port %d\n",
					cl_ntoh64(p_node->node_info.node_guid),
					osm_physp_get_port_num(p_physp));
				ib_port_info_set_link_speed_ext_enabled(p_pi,
				    IB_LINK_SPEED_EXT_DISABLE);
				port_speed_changed = TRUE;
			}
			break;
		case IB_LINK_SPEED_EXT_SET_LSES:
			if (enabled_speed != supported_speed) {
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Setting link_speed_ext_enabled of "
					"node 0x%" PRIx64 ", port %d to "
					"port's supported speed\n",
					cl_ntoh64(p_node->node_info.node_guid),
					osm_physp_get_port_num(p_physp));
				ib_port_info_set_link_speed_ext_enabled(p_pi,
				    IB_LINK_SPEED_EXT_SET_LSES);
				port_speed_changed = TRUE;
			}
			break;
		default:
			new_enabled_speed =
			    sm->p_subn->opt.force_link_speed_ext &
			    supported_speed;
			if (new_enabled_speed != enabled_speed) {
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Setting link_speed_ext_enabled of "
					"node 0x%" PRIx64 ", port %d to "
					"0x%x\n",
					cl_ntoh64(p_node->node_info.node_guid),
					osm_physp_get_port_num(p_physp),
					new_enabled_speed);
				ib_port_info_set_link_speed_ext_enabled(p_pi,
				    new_enabled_speed ? new_enabled_speed :
				    IB_LINK_SPEED_EXT_DISABLE);
				port_speed_changed = TRUE;
			}
			break;
		}
	}

Exit:
	p_physp->port_info_dirty |= port_speed_changed;

	OSM_LOG_EXIT(sm->p_log);
	return port_speed_changed;
}

static boolean_t link_mgr_set_physp_ext_speed2(IN osm_sm_t * sm,
					       IN osm_physp_t * p_physp)
{
	ib_port_info_t *p_pi;
	osm_physp_t *p_rem_physp, *physp0;
	osm_node_t *p_node;
	ib_net32_t cap_mask;
	ib_net16_t cap_mask2;
	uint8_t enabled_speed, new_enabled_speed, supported_speed;
	boolean_t port_speed_changed = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	if (!sm->p_subn->opt.force_link_speed_ext2)
		goto Exit;

	p_rem_physp = osm_physp_get_remote(p_physp);
	if (!p_rem_physp)
		goto Exit;

	p_pi = &p_physp->port_info;
	p_node = p_physp->p_node;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		physp0 = osm_node_get_physp_ptr(p_node, 0);
		cap_mask = physp0->port_info.capability_mask;
		cap_mask2 = physp0->port_info.capability_mask2;
	} else {
		cap_mask = p_pi->capability_mask;
		cap_mask2 = p_pi->capability_mask2;
	}

	if (!((cap_mask & IB_PORT_CAP_HAS_CAP_MASK2) &&
	    (cap_mask2 & IB_PORT_CAP2_IS_EXT_SPEEDS2_SUPPORTED)))
		goto Exit;

	enabled_speed = ib_port_info_get_link_speed_ext_enabled2(p_pi);
	supported_speed = ib_port_info_get_link_speed_ext_supported2(p_pi);

	switch (sm->p_subn->opt.force_link_speed_ext2) {
	case IB_LINK_SPEED_EXT2_SET_LSES2:
		if (enabled_speed != supported_speed) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Setting link_speed_ext_enabled2 of node 0x%"
				PRIx64 ", port %d to port's supported speed\n",
				cl_ntoh64(p_node->node_info.node_guid),
				osm_physp_get_port_num(p_physp));
			ib_port_info_set_link_speed_ext_enabled2(p_pi, IB_LINK_SPEED_EXT2_SET_LSES2);
			port_speed_changed = TRUE;
		}
		break;
	default:
		new_enabled_speed = sm->p_subn->opt.force_link_speed_ext2 &
				    supported_speed;
		if (new_enabled_speed != enabled_speed) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Setting link_speed_ext_enabled2 of node 0x%"
				PRIx64 ", port %d to 0x%x\n",
				cl_ntoh64(p_node->node_info.node_guid),
				osm_physp_get_port_num(p_physp),
				new_enabled_speed);
			ib_port_info_set_link_speed_ext_enabled2(p_pi, new_enabled_speed);
			port_speed_changed = TRUE;
		}
		break;
	}

Exit:
	p_physp->port_info_dirty |= port_speed_changed;

	OSM_LOG_EXIT(sm->p_log);
	return port_speed_changed;
}

static boolean_t link_mgr_set_physp_mepi_speed(IN osm_sm_t * sm,
					       IN osm_physp_t * p_physp)
{
	osm_physp_t *p_rem_physp;
	osm_node_t *p_node;
	ib_mlnx_ext_port_info_t *p_epi;
	uint8_t enabled_speed, new_enabled_speed;
	uint8_t supported_speed;
	boolean_t port_speed_changed = FALSE;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_physp->p_node;
	if (!sm->p_subn->opt.support_mepi_speeds ||
	    !is_mepi_supported(sm, p_node))
		goto Exit;

	p_rem_physp = osm_physp_get_remote(p_physp);
	if (!p_rem_physp)
		goto Exit;

	p_epi = &p_physp->ext_port_info;
	supported_speed = p_epi->link_speed_supported;

	if (supported_speed) {
		enabled_speed = p_physp->ext_port_info.link_speed_enabled;
		new_enabled_speed = sm->p_subn->opt.mepi_enabled_speeds &
				    supported_speed;
		if (new_enabled_speed != enabled_speed) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Setting Mellanox link_speed_enabled of node "
				"0x%" PRIx64 ", port %d to 0x%x\n",
				cl_ntoh64(p_node->node_info.node_guid),
				osm_physp_get_port_num(p_physp),
				new_enabled_speed);
			p_epi->state_change_enable = 0x01;
			p_epi->link_speed_enabled = new_enabled_speed;
			port_speed_changed = TRUE;
		}
	} else {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Node 0x%" PRIx64 ", port %d does not share common "
			"Mellanox link speeds with peer port\n",
			cl_ntoh64(p_node->node_info.node_guid),
			osm_physp_get_port_num(p_physp));
	}

Exit:
	p_physp->ext_port_info_dirty |= port_speed_changed;

	OSM_LOG_EXIT(sm->p_log);;
	return port_speed_changed;
}

static boolean_t link_mgr_set_physp_mtu(IN osm_sm_t * sm,
					IN osm_physp_t * p_physp)
{
	ib_port_info_t *p_pi;
	uint8_t mtu, new_mtu;
	boolean_t dirty = FALSE;

	p_pi = &p_physp->port_info;

	mtu = ib_port_info_get_neighbor_mtu(p_pi);
	new_mtu = osm_physp_calc_link_mtu(sm->p_log, p_physp, mtu);
	if (sm->p_subn->opt.max_neighbor_mtu &&
	    sm->p_subn->opt.max_neighbor_mtu < new_mtu)
		new_mtu = sm->p_subn->opt.max_neighbor_mtu;

	if (new_mtu != mtu) {
		ib_port_info_set_neighbor_mtu(p_pi, new_mtu);
		dirty = TRUE;
	}

	p_physp->port_info_dirty |= dirty;
	return dirty;
}

static boolean_t link_mgr_set_physp_vl(IN osm_sm_t * sm,
				       IN osm_physp_t * p_physp)
{
	ib_port_info_t *p_pi;
	uint8_t op_vls, new_op_vls;
	boolean_t dirty = FALSE;

	p_pi = &p_physp->port_info;

	op_vls = ib_port_info_get_op_vls(p_pi);
	new_op_vls = osm_physp_calc_link_op_vls(sm->p_log, sm->p_subn,
						p_physp, op_vls);
	if (new_op_vls != op_vls) {
		ib_port_info_set_op_vls(p_pi, new_op_vls);
		dirty = TRUE;
	}

	p_physp->port_info_dirty |= dirty;
	return dirty;
}

static boolean_t link_mgr_set_physp_ame_bit(IN osm_sm_t * sm,
					    IN osm_physp_t * p_physp)
{
	ib_mlnx_ext_port_info_t *p_epi;
	boolean_t ame_configured, ame_enabled;
	boolean_t dirty = FALSE;

	p_epi = &p_physp->ext_port_info;
	ame_configured = sm->p_subn->opt.mlnx_enhanced_link_enable;
	ame_enabled = ib_mlnx_extended_get_ame_bit(p_epi);
	if (ame_configured != ame_enabled){
		ib_mlnx_extended_set_ame_bit(p_epi, ame_configured);
		dirty = TRUE;
	}

	p_physp->ext_port_info_dirty |= dirty;
	return dirty;
}

void osm_link_mgr_preprocess_links(osm_sm_t * sm)
{
	cl_qmap_t *p_node_guid_tbl;
	osm_node_t *p_node;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_physp_t *p_rem_physp;
	osm_physp_t *p_sm_physp = NULL;
	osm_port_reset_reason_enum port_reset_reason;
	osm_port_reset_type_enum port_reset_action;
	osm_port_reset_type_enum port_speed_change_action;
	osm_port_reset_type_enum port_ext_speed_change_action;
	osm_port_reset_type_enum port_ext_speed2_change_action;
	osm_port_reset_type_enum port_mepi_speed_change_action;
	osm_port_reset_type_enum port_mtu_change_action;
	osm_port_reset_type_enum port_vl_change_action;
	boolean_t invalidate_cache = FALSE;
	osm_port_reset_type_enum port_ame_bit_change_action;
	boolean_t ame_supported;
	ib_general_info_smp_t *p_gi;
	uint8_t i, num_physp;

	OSM_LOG_ENTER(sm->p_log);

	port_speed_change_action = sm->p_subn->opt.port_speed_change_action;
	port_ext_speed_change_action =
	    sm->p_subn->opt.port_ext_speed_change_action;
	port_ext_speed2_change_action =
	    sm->p_subn->opt.port_ext_speed2_change_action;
	port_mepi_speed_change_action =
	    sm->p_subn->opt.port_mepi_speed_change_action;
	port_mtu_change_action = sm->p_subn->opt.port_mtu_change_action;
	port_vl_change_action = sm->p_subn->opt.port_vl_change_action;
	port_ame_bit_change_action = sm->p_subn->opt.port_ame_bit_change_action;


	/* Find SM port to avoid resetting its port when option not enabled */
	if (!sm->p_subn->opt.allow_sm_port_reset) {
		p_port = osm_get_port_by_guid(sm->p_subn, sm->p_subn->sm_port_guid);
		OSM_ASSERT(p_port && p_port->p_node);
		if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_SWITCH)
			p_sm_physp = osm_port_get_physp(p_port);
	}

	p_node_guid_tbl = &sm->p_subn->node_guid_tbl;
	for (p_node = (osm_node_t *) cl_qmap_head(p_node_guid_tbl);
	     p_node != (osm_node_t *) cl_qmap_end(p_node_guid_tbl);
	     p_node = (osm_node_t *) cl_qmap_next(&p_node->map_item)) {
		num_physp = osm_node_get_num_physp(p_node);
		for (i = 1; i < num_physp; i++) {
			p_physp = osm_node_get_physp_ptr(p_node, i);
			if (!p_physp ||
			    osm_physp_get_port_state(p_physp) == IB_LINK_DOWN)
				continue;

			p_rem_physp = osm_physp_get_remote(p_physp);
			if (osm_hierarchy_is_asymmetric_aport_physp(p_physp) ||
			    (p_rem_physp && osm_hierarchy_is_asymmetric_aport_physp(p_rem_physp)))
				continue;

			port_reset_action =  OSM_PORT_RESET_TYPE_NONE;
			p_physp->port_reset_required = OSM_PORT_RESET_TYPE_NONE;
			port_reset_reason = OSM_PORT_RESET_REASON_NONE;
			p_physp->port_reset_reason = OSM_PORT_RESET_REASON_NONE;

			/* Update port's enabled speeds */
			if (link_mgr_set_physp_speed(sm, p_physp) &&
			    port_speed_change_action > port_reset_action) {
				port_reset_action = port_speed_change_action;
				port_reset_reason = OSM_PORT_RESET_REASON_SPEED;
			}

			/* Update port's enabled extended speeds */
			if (link_mgr_set_physp_ext_speed(sm, p_physp) &&
			    port_ext_speed_change_action > port_reset_action) {
				port_reset_action = port_ext_speed_change_action;
				port_reset_reason = OSM_PORT_RESET_REASON_EXTSPEED;
			}

			/* Update port's enabled extended speeds 2 */
			if (link_mgr_set_physp_ext_speed2(sm, p_physp) &&
			    port_ext_speed2_change_action > port_reset_action) {
				port_reset_action = port_ext_speed2_change_action;
				port_reset_reason = OSM_PORT_RESET_REASON_EXTSPEED2;
			}

			/* Update port's enabled mepi speeds */
			if (link_mgr_set_physp_mepi_speed(sm, p_physp) &&
			    port_mepi_speed_change_action > port_reset_action) {
				port_reset_action = port_mepi_speed_change_action;
				port_reset_reason = OSM_PORT_RESET_REASON_MLNXSPEED;
			}

			/* Update port's MTU */
			if (link_mgr_set_physp_mtu(sm, p_physp) &&
			    port_mtu_change_action > port_reset_action) {
				port_reset_action = port_mtu_change_action;
				port_reset_reason = OSM_PORT_RESET_REASON_MTU;
			}

			/* Update port's Operational VLs */
			if (link_mgr_set_physp_vl(sm, p_physp) &&
			    port_vl_change_action > port_reset_action) {
				port_reset_action = port_vl_change_action;
				port_reset_reason = OSM_PORT_RESET_REASON_OPVLS;
			}

			p_gi = &p_node->general_info;
			ame_supported = ib_general_info_is_cap_enabled(p_gi,
					    IB_GENERAL_INFO_AME_SUPPORTED_BIT);


			if (sm->p_subn->opt.support_mlnx_enhanced_link &&
			    ame_supported &&
			    link_mgr_set_physp_ame_bit(sm, p_physp) &&
			    port_ame_bit_change_action > port_reset_action) {
				port_reset_action = port_ame_bit_change_action;
				port_reset_reason = OSM_PORT_RESET_REASON_AME;
			}

			/*
			 *  Avoid performing logical reset to ports in INIT
			 *  state
			 */
			if (port_reset_action == OSM_PORT_RESET_TYPE_LOGICAL &&
			    osm_physp_get_port_state(p_physp) == IB_LINK_INIT)
				continue;

			/* Update p_physp->port_reset_required */
			if (port_reset_action != OSM_PORT_RESET_TYPE_NONE) {

				if (p_sm_physp && (p_sm_physp == p_physp || p_sm_physp == p_rem_physp))
					continue;

				p_physp->port_reset_required = port_reset_action;
				p_physp->port_reset_reason = port_reset_reason;
				sm->p_subn->found_link_for_reset = TRUE;
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Node 0x%" PRIx64 ", port %d marked "
					"for %s reset due to change in port's "
					"%s\n",
					cl_ntoh64(p_node->node_info.node_guid),
					i, osm_port_reset_type_str(port_reset_action),
					osm_port_reset_reason_str(port_reset_reason));

				if (p_rem_physp && p_node->sw &&
				    p_rem_physp->p_node->sw)
					invalidate_cache = TRUE;
			}
		}
	}

	if (invalidate_cache) {
		OSM_LOG(sm->p_log, OSM_LOG_INFO,
			"Invalidating ucast cache due to switch-to-switch "
			"links reset\n");
		sm->ucast_mgr.variables_invalidate_cache = invalidate_cache;
	}

	OSM_LOG_EXIT(sm->p_log);
}
