/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2011 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_mlnx_epi_rcv_t.
 * This object represents the MLNX ExtendedPortInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MLNX_EXT_PORT_INFO_RCV_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_mad_pool.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_unhealthy_ports.h>

extern void pi_rcv_get_remote_node_info(IN osm_sm_t *sm,
					IN osm_node_t *p_node,
					IN osm_port_t *p_port,
					IN osm_physp_t *p_physp);

static boolean_t mlnx_epi_get_cond_unhealthy_reason(uint8_t unhealthy_reason,
						    osm_hm_cond_type_enum * p_cond)
{
	switch (unhealthy_reason) {
	case IB_MEPI_UNHEALTHY_REASON_CREDIT_WATCHDG:
		*p_cond = OSM_HM_COND_CREDIT_WATCHDOG;
		return TRUE;
	case IB_MEPI_UNHEALTHY_REASON_RAW_BER:
		*p_cond = OSM_HM_COND_RAW_BER;
		return TRUE;
	case IB_MEPI_UNHEALTHY_REASON_EFFECTIVE_BER:
		*p_cond = OSM_HM_COND_EFFECTIVE_BER;
		return TRUE;
	case IB_MEPI_UNHEALTHY_REASON_SYMBOL_BER:
		*p_cond = OSM_HM_COND_SYMBOL_BER;
		return TRUE;
	case IB_MEPI_UNHEALTHY_REASON_CONTAIN_AND_DRAIN:
		/*
		 * For NVLink Single Node, avoid adding a port that was
		 * triggered by ContainAndDrain as an unhealthy port
		 */
		return FALSE;
	default:
		/* Should not get here */
		return FALSE;
	}
}

static boolean_t mlnx_epi_is_unhealthy_reason_applicable(osm_sm_t * sm,
							 osm_node_t * p_node)
{
	boolean_t fast_recovery_support, contain_and_drain_support;

	/*
	 * unhealthy_reason applicable only when feature that uses this field
	 * is enabled from SM side, and the feature is supported on FW.
	 */

	if (!p_node->sw)
		return FALSE;

	fast_recovery_support = sm->p_subn->opt.fast_recovery_enabled == OSM_CAPABILITY_ENABLED &&
	    ib_general_info_is_cap_enabled(&p_node->general_info,
					   IB_GENERAL_INFO_IS_FAST_RECOVERY_SUPPORTED_BIT);

	contain_and_drain_support = osm_subn_is_nvlink_enabled(sm->p_subn) &&
	   ib_general_info_is_cap_enabled(&p_node->general_info,
					  IB_GENERAL_INFO_IS_CONTAIN_AND_DRAIN_SUPPORTED_BIT);

	if (!fast_recovery_support && !contain_and_drain_support)
		return FALSE;

	return TRUE;
}

static void mlnx_epi_check_unhealthy_reason(osm_sm_t * sm,
					    ib_mlnx_ext_port_info_t * p_pi,
					    osm_node_t * p_node,
					    osm_physp_t * p_physp,
					    uint8_t port_num)
{
	osm_hm_cond_type_enum cond;
	uint8_t unhealthy_reason;
	ib_net64_t node_guid = osm_node_get_node_guid(p_node);
	osm_physp_t *p_remote_physp;

	if (!mlnx_epi_is_unhealthy_reason_applicable(sm, p_node))
		return;

	/*
	 * For marking port as unhealthy, FW change its state to Init,
	 * and update unhealthy reason field accordingly.
	 */
	unhealthy_reason = ib_mlnx_epi_get_unhealthy_reason(p_pi);

	/*
	 * For Contain and Drain unhealthy reason, dump message and return.
	 * Avoid marking the port as unhealthy on SM DBs (next steps of this funciton).
	 */
	if (unhealthy_reason == IB_MEPI_UNHEALTHY_REASON_CONTAIN_AND_DRAIN) {
		OSM_LOG(sm->p_log, OSM_LOG_INFO,
			"Switch 0x%016" PRIx64 " (port %u) triggered by "
			"ContainAndDrain event, according to unhealthy reason (%u) "
			"of ExtendedPortInfo\n",
			cl_ntoh64(node_guid), port_num,	unhealthy_reason);
		return;
	}

	if (ib_port_info_get_port_state(&p_physp->port_info) != IB_LINK_INIT ||
	    unhealthy_reason <= IB_MEPI_UNHEALTHY_REASON_NONE)
		return;

	/* Avoid isolating SM port due to Fast Recovery reasons */
	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_physp->port_guid == sm->p_subn->sm_port_guid ||
	    (p_remote_physp && p_remote_physp->port_guid == sm->p_subn->sm_port_guid)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Switch 0x%016" PRIx64 " port %u connected to SM is marked as"
			" unhealthy by switch (unhealthy reason %u)."
			" Ignores this switch decision, keep as healthy.\n",
			cl_ntoh64(node_guid), port_num, unhealthy_reason);
		return;
	}

	/*
	 * Prioritize 'healthy' manual setting of
	 * unhealthy configuration file over unhealthy_reason
	 */
	if (osm_hm_is_marked_healthy_port(sm, node_guid, port_num)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Switch 0x%016" PRIx64 " (port %u) marked  as healthy"
			" by unhealthy configuration file."
			" Ignores unhealthy reason (%u) of ExtendedPortInfo\n",
			cl_ntoh64(node_guid), port_num,
			unhealthy_reason);
	} else if (!osm_hm_is_unhealthy_by_sw_decision(sm, node_guid, port_num)) {
		if (mlnx_epi_get_cond_unhealthy_reason(unhealthy_reason, &cond) &&
		    osm_hm_set_by_peer_physp(sm, p_physp, cond)) {
			/* Marking port as unhealthy requires additional heavy sweep */
			sm->p_subn->force_heavy_sweep = TRUE;
		}
	}
}

void osm_mlnx_epi_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_mlnx_ext_port_info_t *p_pi;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_node_t *p_node;
	osm_pi_context_t *p_context;
	ib_net64_t port_guid, node_guid;
	uint8_t port_num, portnum, start_port = 1;
	boolean_t verbose_bypass = FALSE;
	ib_special_port_type_t special_port_type;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_pi_context_ptr(p_madw);
	p_pi = ib_smp_get_payload_ptr(p_smp);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_MLNX_EXTENDED_PORT_INFO);

	port_num = (uint8_t) cl_ntoh32(p_smp->attr_mod);

	port_guid = p_context->port_guid;
	node_guid = p_context->node_guid;

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	verbose_bypass = osm_verbose_bypass_handle_mepi(sm, p_port);

	osm_dump_mlnx_ext_port_info_v2(sm->p_log, node_guid, port_guid, port_num,
				       p_pi, FILE_ID, OSM_LOG_DEBUG);

	if (!p_port) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F01: "
			"No port object for port with GUID 0x%" PRIx64
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid),
			cl_ntoh64(node_guid), cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	special_port_type = ib_mlnx_epi_get_special_port_type(p_pi);
	if (special_port_type >= IB_SPECIAL_PORT_TYPE_MAX) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F02: "
			"Invalid special port type 0x%x for port with GUID 0x%" PRIx64
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			special_port_type, cl_ntoh64(port_guid),
			cl_ntoh64(node_guid), cl_ntoh64(p_smp->trans_id));
		special_port_type = IB_SPECIAL_PORT_TYPE_RESERVED;
	}

	if (!ib_smp_get_status(p_smp)) {
		if (port_num != 255) {
			p_physp = osm_node_get_physp_ptr(p_node, port_num);
			OSM_ASSERT(p_physp);
			p_physp->ext_port_info = *p_pi;
			p_physp->ext_port_info_updated = TRUE;

			if (p_node->sw &&
			    (special_port_type == IB_SPECIAL_PORT_TYPE_AN)) {
				p_node->sw->an_port_num = port_num;
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Setting sharp aggregation node "
					"for switch 0x%016" PRIx64 " (port %u)\n",
					cl_ntoh64(node_guid), port_num);
			}

			/* Continue discovery if needed */
			if (p_context->continue_discovery) {
				pi_rcv_get_remote_node_info(sm, p_node, p_port, p_physp);
			}

			/* Check whether port is marked to be unhealthy by FW */
			mlnx_epi_check_unhealthy_reason(sm, p_pi, p_node, p_physp, port_num);
		} else {
			/* Handle all ports on set/set resp */
			if (p_node->sw &&
			    ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info))
				start_port = 0;

			for (portnum = start_port;
			     portnum < osm_node_get_num_physp(p_node);
			     portnum++) {
				p_physp = osm_node_get_physp_ptr(p_node, portnum);
				OSM_ASSERT(p_physp);
				p_physp->ext_port_info = *p_pi;
				p_physp->ext_port_info_updated = TRUE;
			}
		}
	} else {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		if (p_context->set_method) {
			osm_hm_set_by_node(sm, p_node, OSM_HM_COND_SET_ERR);
			sm->p_subn->subnet_initialization_error = TRUE;
		}
	}

	osm_port_send_next_mad(sm, p_port);

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	OSM_LOG_EXIT(sm->p_log);
}
