/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2012 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_pi_rcv_t.
 * This object represents the PortInfo Receiver object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdlib.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PORT_INFO_RCV_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_mad_pool.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_pkey.h>
#include <opensm/osm_remote_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_inform.h>

static void pi_rcv_check_and_fix_lid(osm_sm_t * sm, ib_port_info_t * pi,
				     osm_port_t * p_port)
{
	osm_physp_t *p_physp;

	p_physp = p_port->p_physp;

	if (PF(cl_ntoh16(pi->base_lid) > IB_LID_UCAST_END_HO)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F04: "
			"Got invalid base LID %u from the network. "
			"Corrected to %u\n", cl_ntoh16(pi->base_lid),
			cl_ntoh16(p_physp->port_info.base_lid));
		pi->base_lid = p_physp->port_info.base_lid;
	}

	if (!p_port->is_new &&
	     p_physp->port_info.base_lid != pi->base_lid)
			sm->ucast_mgr.variables_invalidate_cache = TRUE;
}

static void pi_rcv_set_min_mtu(IN osm_sm_t * sm,
			       IN const ib_port_info_t * p_pi,
			       IN ib_net64_t guid,
			       IN uint8_t port_num,
			       IN boolean_t is_switch,
			       OUT uint8_t * p_mtu)
{
	uint8_t mtu = ib_port_info_get_mtu_cap(p_pi);
	if (mtu < *p_mtu) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Setting minimal MTU to:%u defined by %s guid: 0x%"
			PRIx64 " port: %u\n", mtu, is_switch ? "switch" : "endport",
			cl_ntoh64(guid), port_num);
		*p_mtu = mtu;
	}
}

static void pi_rcv_set_min_rate(IN osm_sm_t * sm,
				IN osm_physp_t * p_physp,
				IN const ib_port_info_t * p_pi,
				IN ib_net64_t guid,
				IN uint8_t port_num,
				IN boolean_t is_switch,
				OUT uint8_t * p_rate)
{
	uint8_t rate;

	rate = ib_port_info_compute_rate_v2(p_pi, p_pi);
	if (ib_path_compare_rates(rate, *p_rate) < 0) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Setting minimal rate to:%u defined by %s guid: 0x%"
			PRIx64 " port: %u\n", rate, is_switch ? "switch" : "endport",
			cl_ntoh64(guid), port_num);
		*p_rate = rate;
	}
}

static void pi_rcv_process_endport(IN osm_sm_t * sm, IN osm_physp_t * p_physp,
				   IN const ib_port_info_t * p_pi)
{
	osm_madw_context_t context;
	ib_api_status_t status;
	ib_net64_t port_guid;
	uint8_t port_num, mpb;
	unsigned data_vls;
	cl_qmap_t *p_sm_tbl;
	osm_remote_sm_t *p_sm;
	uint8_t max_op_vls;
	osm_dr_path_t *p_dr_path;

	OSM_LOG_ENTER(sm->p_log);

	port_guid = osm_physp_get_port_guid(p_physp);
	port_num = osm_physp_get_port_num(p_physp);
	p_dr_path = osm_physp_get_dr_path_ptr(sm, p_physp);
	max_op_vls = osm_get_max_op_vls_by_node(sm->p_subn, p_physp->p_node);

	/* HACK extended port 0 should be handled too! */
	if (port_num != 0 && ib_port_info_get_port_state(p_pi) != IB_LINK_DOWN) {
		/* track the minimal endport MTU, rate, and operational VLs */
		pi_rcv_set_min_mtu(sm, p_pi, port_guid, port_num, FALSE, &sm->p_subn->min_ca_mtu);
		pi_rcv_set_min_rate(sm, p_physp, p_pi, port_guid, port_num,
				    FALSE, &sm->p_subn->min_ca_rate);
		data_vls = 1U << (ib_port_info_get_vl_cap(p_pi) - 1);

		if (data_vls > 1U << (max_op_vls - 1))
			data_vls = 1U << (max_op_vls - 1);

		if (data_vls >= IB_MAX_NUM_VLS)
			data_vls = IB_MAX_NUM_VLS - 1;

		if ((uint8_t)data_vls < sm->p_subn->min_data_vls) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Setting endport minimal data VLs to:%u defined by port:0x%"
				PRIx64 "\n", data_vls, cl_ntoh64(port_guid));
			sm->p_subn->min_data_vls = data_vls;
		}
	}

	/* Check M_Key vs M_Key protect, can we control the port ? */
	mpb = ib_port_info_get_mpb(p_pi);
	if (mpb > 0 && p_pi->m_key == 0) {
		OSM_LOG(sm->p_log, OSM_LOG_INFO,
			"Port 0x%" PRIx64 " has unknown M_Key, protection level %u\n",
			cl_ntoh64(port_guid), mpb);
	}

	if (port_guid != sm->p_subn->sm_port_guid) {
		p_sm_tbl = &sm->p_subn->sm_guid_tbl;
		if (p_pi->capability_mask & IB_PORT_CAP_IS_SM) {
			/*
			 * Before querying the SM - we want to make sure we
			 * clean its state, so if the querying fails we
			 * recognize that this SM is not active.
			 */
			p_sm =
			    (osm_remote_sm_t *) cl_qmap_get(p_sm_tbl,
							    port_guid);
			if (p_sm != (osm_remote_sm_t *) cl_qmap_end(p_sm_tbl))
				/* clean it up */
				p_sm->smi.pri_state =
				    0xF0 & p_sm->smi.pri_state;

			if (sm->p_subn->opt.ignore_other_sm) {
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Ignoring SM on port 0x%" PRIx64 "\n",
					cl_ntoh64(port_guid));
			} else if (!osm_is_sm_guid_allowed(sm->p_subn, port_guid)) {
				if (p_sm != (osm_remote_sm_t *) cl_qmap_end(p_sm_tbl)) {
					p_sm =
					    (osm_remote_sm_t *) cl_qmap_remove(p_sm_tbl,
									       port_guid);
					free(p_sm);
				}
				osm_report_notice_1306_1307(sm->p_subn, p_pi->base_lid,
							    port_guid, TRUE, p_dr_path,
							    SM_VS_TRAP_NUM_UNALLOWED_SM_GUID_FOUND);
				OSM_LOG(sm->p_log, OSM_LOG_ERROR,
					"ERR 0F09: Ignoring SM on port 0x%" PRIx64 ", "
					"is not part of the allowed port GUIDs\n",
					cl_ntoh64(port_guid));
			} else {
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Detected another SM. Requesting SMInfo"
					"\n\t\t\t\tPort 0x%" PRIx64 "\n",
					cl_ntoh64(port_guid));

				/*
				   This port indicates it's an SM and
				   it's not our own port.
				   Acquire the SMInfo Attribute.
				 */
				memset(&context, 0, sizeof(context));
				context.smi_context.set_method = FALSE;
				context.smi_context.port_guid = port_guid;
				status = osm_req_get(sm,
						     p_dr_path,
						     p_physp,
						     IB_MAD_ATTR_SM_INFO, 0,
						     FALSE,
						     ib_port_info_get_m_key(&p_physp->port_info),
						     0, CL_DISP_MSGID_NONE,
						     &context);

				if (status != IB_SUCCESS)
					OSM_LOG(sm->p_log, OSM_LOG_ERROR,
						"ERR 0F05: "
						"Failure requesting SMInfo (%s)\n",
						ib_get_err_str(status));
			}
		} else {
			p_sm =
			    (osm_remote_sm_t *) cl_qmap_remove(p_sm_tbl,
							       port_guid);
			if (p_sm != (osm_remote_sm_t *) cl_qmap_end(p_sm_tbl))
				free(p_sm);
		}
	}

	OSM_LOG_EXIT(sm->p_log);
}

boolean_t check_force_reroute(IN osm_sm_t *sm,
			      IN osm_node_t *p_node,
			      IN uint8_t port_num,
			      OUT boolean_t *p_is_no_discover_port)
{
	boolean_t found_ignored_port = FALSE;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_node_t *p_remote_node;
	boolean_t is_isolated_port = FALSE;
	boolean_t is_no_discover_port = FALSE;

	switch (p_node->physp_ignored[port_num]) {
	case OSM_PHYSP_NORMAL:
		is_isolated_port = lookup_isolated_port(sm, p_node, port_num);
		is_no_discover_port = lookup_no_discover_port(sm, p_node, port_num);
		if (is_isolated_port || is_no_discover_port ||
		    lookup_held_back_switch(sm->p_subn, p_node)) {
			p_node->physp_ignored[port_num] = OSM_PHYSP_IGNORE_EXISTING;
			sm->p_subn->ignore_existing_lfts = TRUE;
			found_ignored_port = TRUE;
			if (sm->p_subn->opt.use_ucast_cache) {
				OSM_LOG(sm->p_log, OSM_LOG_INFO,
					"Invalidate cache as "
					"new isolated/no discover/held back "
					"port found 0x%016" PRIx64 " port %d\n",
					osm_node_get_node_guid_ho(p_node),
					port_num);
				sm->ucast_mgr.variables_invalidate_cache = TRUE;
			}
			/*
			 * Check if this is a switch to switch link, and if it
			 * does, mark multicast trees that uses the port for
			 * recalculation.
			 */
			if (port_num && is_isolated_port && p_node->sw) {
				p_remote_node =
				    osm_node_get_remote_node(p_node, port_num,
							     NULL);
				if (p_remote_node && p_remote_node->sw)
					osm_subn_mark_mbox_by_sw(sm->p_subn,
								 p_node->sw,
								 port_num);
			}
		}
		break;
	case OSM_PHYSP_IGNORE_NEW:
		is_no_discover_port = lookup_no_discover_port(sm, p_node, port_num);
		if (lookup_isolated_port(sm, p_node, port_num) ||
		    is_no_discover_port ||
		    lookup_held_back_switch(sm->p_subn, p_node)) {
			p_node->physp_ignored[port_num] = OSM_PHYSP_IGNORE_EXISTING;
			found_ignored_port = TRUE;
		} else {
			sm->p_subn->ignore_existing_lfts = TRUE;
			if (sm->p_subn->opt.use_ucast_cache) {
				OSM_LOG(sm->p_log, OSM_LOG_INFO,
					"Invalidate cache as "
					"isolated/no discover/heldback "
					"port recovered 0x%016" PRIx64 " port %d\n",
					osm_node_get_node_guid_ho(p_node),
					port_num);
				sm->ucast_mgr.variables_invalidate_cache = TRUE;
			}
			p_node->physp_ignored[port_num] = OSM_PHYSP_NORMAL;
		}
		break;
	case OSM_PHYSP_IGNORE_EXISTING:
		found_ignored_port = TRUE;
		break;
	default:
		break;
	}

	if (found_ignored_port) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (p_physp) {
			p_port = osm_get_port_by_guid(sm->p_subn,
						      p_physp->port_guid);
			if (p_port) {
				if (!cl_is_object_in_list(&sm->p_subn->ignored_nodes_list,
							  p_port)) {
					if (cl_list_insert_tail(&sm->p_subn->ignored_nodes_list,
								p_port) != CL_SUCCESS)
						OSM_LOG(sm->p_log,
							OSM_LOG_ERROR,
							"ERR 0F30: Out of memory\n");
				}
			}
		}
	}

	if (p_is_no_discover_port)
		*p_is_no_discover_port = is_no_discover_port;

	return found_ignored_port;
}

static boolean_t is_spst_supported(IN ib_port_info_t * p_pi)
{
	return ((p_pi->capability_mask & IB_PORT_CAP_HAS_CAP_MASK2) &&
		(p_pi->capability_mask2 &
		 IB_PORT_CAP2_IS_SWITCH_PORT_STATE_TBL_SUPP));
}

static void pi_rcv_get_spst(IN osm_sm_t * sm,
			    IN osm_physp_t * p_physp,
			    IN uint8_t num_ports)
{
	uint32_t blocks_num, block;
	uint32_t attr_mod_ho;
	osm_madw_context_t context;
	ib_api_status_t status;

	OSM_LOG_ENTER(sm->p_log);

	memset(&context, 0, sizeof(osm_madw_context_t));
	context.sps_context.node_guid = osm_node_get_node_guid(p_physp->p_node);
	context.sps_context.port_guid = osm_physp_get_port_guid(p_physp);

	blocks_num = (uint32_t)(num_ports/IB_SPST_BLOCK_SIZE + 1);
	for (block = 0 ; block < blocks_num ; block++) {
		attr_mod_ho = OSM_SPST_PORT_STATE_MASK_ATTR_MOD_HO & block;
		status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
				     IB_MAD_ATTR_SWITCH_PORT_STATE_TBL,
				     cl_hton32(attr_mod_ho),
				     FALSE,
				     ib_port_info_get_m_key(&p_physp->port_info),
				     0, CL_DISP_MSGID_NONE, &context);
		if (status != IB_SUCCESS)
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F0B: "
				"Failure initiating SwitchPortStateTbl "
				"request (%s)\n",
				ib_get_err_str(status));
	}
	OSM_LOG_EXIT(sm->p_log);
}

void pi_rcv_process_switch_external_port(IN osm_sm_t * sm,
					 IN osm_node_t * p_node,
					 IN osm_port_t * p_port,
					 IN uint8_t port,
					 IN osm_physp_t * p_physp,
					 IN osm_madw_context_t *p_context)
{
	osm_madw_t *p_madw = NULL;
	boolean_t is_prev_physp_state_normal = FALSE;
	boolean_t is_no_discover_port = FALSE;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Querying external port number %u "
		"of node GUID 0x%" PRIx64 "\n",
		port, osm_node_get_node_guid_ho(p_node));

	/*
	 * If found new isolated port, stop the discovery and
	 * mark it. We'll continue the discovery through it
	 * during state_mgr_sweep_ignored stage.
	 */
	if (p_port->disc_through != OSM_DISC_IGNORED) {
		if (p_node->physp_ignored[port] == OSM_PHYSP_NORMAL)
			is_prev_physp_state_normal = TRUE;

		if (check_force_reroute(sm, p_node, port, &is_no_discover_port)) {
			sm->p_subn->found_ignored_port = TRUE;

			/*
			 * Send PI the first time a port is discovered
			 * in unhealthy no_discover state
			 */
			if (is_no_discover_port && is_prev_physp_state_normal)
				goto Send;

			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Skipping "
				"PI for node 0x%" PRIx64 " port %d\n",
				osm_node_get_node_guid_ho(p_node),
				port);

			return;
		}
	}

Send:
	p_madw = osm_prepare_req_get_pi(sm, p_physp, port, FALSE,
					ib_port_info_get_m_key(&p_physp->port_info),
					p_context, TRUE);
	if (!p_madw)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F16: "
			"Failure initiating PortInfo request\n");
	else
		cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);
}

void osm_update_discovery(osm_sm_t *sm, osm_node_t *p_node,
			  osm_port_t *p_port, uint8_t port_num,
			  boolean_t check_force)
{
	if (!p_node->physp_discovered[port_num]) {
		p_port->discovery_count++;
		p_node->physp_discovered[port_num] = 1;
		if (check_force && check_force_reroute(sm, p_node, port_num, NULL))
			sm->p_subn->found_ignored_port = TRUE;
	}
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
static void pi_rcv_process_switch_port0(IN osm_sm_t * sm,
					IN osm_node_t * p_node,
					IN osm_port_t * p_port,
					IN osm_physp_t * p_physp,
					IN ib_port_info_t * p_pi)
{
	uint8_t port, num_ports;
	osm_madw_context_t context;

	OSM_LOG_ENTER(sm->p_log);

	if (p_physp->need_update)
		sm->p_subn->ignore_existing_lfts = TRUE;

	pi_rcv_check_and_fix_lid(sm, p_pi, p_port);

	/* Update the PortInfo attribute */
	osm_physp_set_port_info(p_physp, p_pi, sm, FALSE);

	/* Determine if base switch port 0 */
	if (p_node->sw &&
	    !ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info))
		/* PortState is not used on BSP0 but just in case it is DOWN */
		p_physp->port_info = *p_pi;

	num_ports = osm_node_get_num_physp(p_node);

	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = FALSE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;
	context.pi_context.client_rereg = FALSE;

	/*
	 * If LID is 0, it indicates switch was reset, hence need to mark it for update.
	 *
	 * Marking the switch during discovery is needed in order to re-query cached
         * information such as GeneralInfo-SMP during discovery.
         */
	if (!p_pi->master_sm_base_lid && p_node->sw) {
		p_node->sw->need_update = 2;
	}

	/*
	 * If supported, send SwitchPortStateTable MAD in order to determine
	 * which external ports needs to be queried with PortInfo.
	 *
	 * If one of the following conditions apply, query all ports of the
	 * device with PortInfo in order to update SM's DB:
	 * 1. Device does not support SwitchPortStateTable MAD.
	 * 2. Newly discovered switch.
	 * 3. SM in discovering state.
	 * 4. First time master sweep.
	 * 5. Master SM LID of the device has changed.
	 * 6. SPST MAD got a timeout/bad status
	 */
	if (!is_spst_supported(p_pi) || p_port->is_new ||
	    sm->p_subn->sm_state_info.sm_state == IB_SMINFO_STATE_DISCOVERING ||
	    sm->p_subn->first_time_master_sweep ||
	    sm->p_subn->master_sm_base_lid != p_pi->master_sm_base_lid ||
	    (p_node->sw && p_node->sw->send_port_info)) {
		for (port = 1; port < num_ports; port ++) {
			pi_rcv_process_switch_external_port(sm, p_node, p_port,
							    port, p_physp,
							    &context);
		}

		osm_port_send_next_mads(sm, p_port);

		if (p_node->sw)
			p_node->sw->send_port_info = FALSE;
	} else {
		/*
		 * Mark the sw so that OpenSM will send portInfo to external ports
		 * in case of bad status/timeout in the SPST MAD
		 */
		if (p_node->sw)
			p_node->sw->send_port_info = TRUE;

		pi_rcv_get_spst(sm, p_physp, num_ports);
	}

	pi_rcv_process_endport(sm, p_physp, p_pi);
	OSM_LOG_EXIT(sm->p_log);
}

static int pi_rcv_get_mlnx_ext_port_info(IN osm_sm_t * sm,
					 IN osm_node_t * p_node,
					 IN osm_port_t * p_port,
					 IN osm_physp_t * p_physp,
					 IN ib_port_info_t * p_pi,
					 IN boolean_t continue_discovery)
{
	osm_physp_t *p_physp0;
	osm_madw_t *p_madw;
	osm_madw_context_t context;
	ib_net64_t m_key;
	int send = 0;
	uint8_t port_num;
	boolean_t updated_by_state;

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH) {
		p_physp0 = osm_node_get_physp_ptr(p_node, 0);
		m_key = ib_port_info_get_m_key(&p_physp0->port_info);
	} else
		m_key = ib_port_info_get_m_key(&p_physp->port_info);

	port_num = osm_physp_get_port_num(p_physp);

	if (sm->p_subn->opt.fast_recovery_enabled &&
	    (osm_physp_get_port_state(p_physp) != IB_LINK_INIT &&
	     ib_port_info_get_port_state(p_pi) == IB_LINK_INIT))
		updated_by_state = FALSE;
	else
		updated_by_state = (osm_physp_get_port_state(p_physp) == IB_LINK_ACTIVE);

	p_physp->ext_port_info_updated =
	    p_physp->ext_port_info_updated &&
	    sm->p_subn->opt.mepi_cache_enabled &&
	    !sm->p_subn->coming_out_of_standby &&
	    updated_by_state;

	/*
	 * INIT state check is required for Fast Recovery and Contain And Drain
	 * features, in which when state changes to Init state, need to check
	 * MEPI->unhealthy_reason
	 */

	if (p_physp->p_remote_physp &&
	    (sm->p_subn->opt.fast_recovery_enabled || osm_subn_is_nvlink_enabled(sm->p_subn))) {
		/* In case state is not active but remote is unhealthy, should skip sending MEPI */
		if (osm_physp_is_unhealthy_by_sw_decision(sm, p_physp) &&
		    (osm_physp_get_port_state(p_physp) != IB_LINK_ACTIVE)) {
			p_physp->ext_port_info_updated = TRUE;
		}
	}

	if (!p_physp->ext_port_info_updated && is_mepi_supported(sm, p_node)) {
		context.pi_context.node_guid = osm_node_get_node_guid(p_node);
		context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
		context.pi_context.set_method = FALSE;
		context.pi_context.light_sweep = FALSE;
		context.pi_context.active_transition = FALSE;
		context.pi_context.client_rereg = FALSE;
		context.pi_context.continue_discovery = continue_discovery;

		p_madw = osm_prepare_req_get(sm,
					     osm_physp_get_dr_path_ptr(sm, p_physp),
					     IB_MAD_ATTR_MLNX_EXTENDED_PORT_INFO,
					     cl_hton32(port_num), FALSE, m_key,
					     0, CL_DISP_MSGID_NONE,
					     osm_node_get_node_guid(p_node), &context);
		if (!p_madw) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F11: "
				"Failure initiating MLNX ExtPortInfo request\n");
		} else {
			cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);
			send = 1;
		}
	}

	return send;
}

void pi_rcv_get_remote_node_info(IN osm_sm_t *sm,
				 IN osm_node_t *p_node,
				 IN osm_port_t *p_port,
				 IN osm_physp_t *p_physp)
{
	osm_neighbor_info_t *p_neighbor_info = NULL;
	osm_madw_context_t context;
	osm_dr_path_t path;
	osm_physp_t *p_remote_physp;
	osm_unhealthy_port_t *p_unhealthy_port;
	ib_net64_t guid;
	boolean_t unresponsive = FALSE;
	boolean_t unresp_cond = FALSE;
	boolean_t no_discover = FALSE;
	boolean_t answered = FALSE;
	ib_api_status_t status = IB_SUCCESS;
	uint8_t port_num;

	port_num = osm_physp_get_port_num(p_physp);

	p_remote_physp = osm_physp_get_remote(p_physp);

	/*
	 * Request node info from the other end of this link:
	 * 1) Copy the current path from the parent node.
	 * 2) Extend the path to the next hop thru this port.
	 * 3) Request node info with the new path
	 */

	path = *osm_physp_get_dr_path_ptr(sm, p_physp);
	if (osm_dr_path_extend(&path, osm_physp_get_port_num(p_physp))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 0F08: "
			"DR path with hop count %d couldn't be extended\n",
			path.hop_count);
		return;
	}

	p_unhealthy_port = osm_hm_get_unhealthy_port(sm, p_node, port_num);
	if (p_unhealthy_port) {
		unresp_cond = osm_hm_is_cond_set(p_unhealthy_port, OSM_HM_COND_UNRESPONSIVE);
		answered = p_unhealthy_port->answered_last_ls;
		unresponsive = (unresp_cond && !p_remote_physp && !answered);
		no_discover = p_unhealthy_port->no_discover;

		if (!answered) {
			/*
			 * Mark the port as responsive
			 * in order to force OpenSM
			 * querying it on the next sweep
			 */
			p_unhealthy_port->answered_last_ls = TRUE;
		}

		if (unresponsive || no_discover) {
			guid = osm_node_get_node_guid(p_node);
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"Node 0x%" PRIx64 ": Skipping "
				"NI for unhealthy port %u\n",
				cl_ntoh64(guid), port_num);
			return;
		}
	}

	if (sm->p_subn->topo_config_enabled &&
	    !ib_mlnx_epi_get_special_port_type(&p_physp->ext_port_info)) {
		p_neighbor_info = osm_topoconfig_lookup_port(sm->p_subn->p_topoconfig_hash_tbl, osm_node_get_node_guid(p_node),
							     osm_physp_get_port_num(p_physp));
		if (!p_neighbor_info) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR,
				"ERR 0F32: No entry for the node guid: 0x%" PRIx64 "\n",
				osm_node_get_node_guid(p_node));
			return;
		}
		if (p_neighbor_info->state != OSM_TOPOCONFIG_ACTIVE) {
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Node 0x%" PRIx64 ": Skipping NI for Disabled/No-Discover port %u\n",
				osm_node_get_node_guid_ho(p_node), port_num);
			return;
		}
	}

	memset(&context, 0, sizeof(context));
	context.ni_context.node_guid = osm_node_get_node_guid(p_node);
	context.ni_context.port_num = osm_physp_get_port_num(p_physp);
	context.ni_context.node_type = osm_node_get_type(p_node);
	context.ni_context.is_special_port =
	    ib_mlnx_epi_get_special_port_type(&p_physp->ext_port_info) ? TRUE : FALSE;

	if (sm->p_subn->topo_config_enabled && p_neighbor_info) {
		context.ni_context.expected_neighbor_node_guid = p_neighbor_info->guid;
		context.ni_context.expected_neighbor_port_num = p_neighbor_info->port;
		context.ni_context.expected_neighbor_node_type = p_neighbor_info->node_type;
	}

	if (p_port->disc_through == OSM_DISC_IGNORED ||
	    p_node->physp_ignored[port_num] != OSM_PHYSP_NORMAL)
		context.ni_context.disc_through = OSM_DISC_IGNORED;
	else
		context.ni_context.disc_through = OSM_DISC_NORMAL;

	p_physp->nodeinfo_timeout_on_link = TRUE;
	status = osm_req_get(sm, &path, p_remote_physp,
			     IB_MAD_ATTR_NODE_INFO, 0,
			     TRUE, 0, 0,
			     CL_DISP_MSGID_NONE,
			     &context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR 0F02: "
			"Failure initiating NodeInfo request (%s)\n",
			ib_get_err_str(status));
}

/**********************************************************************
 The plock must be held before calling this function.
**********************************************************************/
void pi_rcv_process_switch_ext_port(IN osm_sm_t * sm,
				    IN osm_node_t * p_node,
				    IN osm_port_t * p_port,
				    IN osm_physp_t * p_physp,
				    IN ib_port_info_t * p_pi)
{
	osm_physp_t *p_remote_physp;
	osm_node_t *p_remote_node;
	unsigned data_vls;
	uint8_t port_num, max_op_vls;
	boolean_t continue_discovery;

	OSM_LOG_ENTER(sm->p_log);

	/*
	   Check the state of the physical port.
	   If there appears to be something on the other end of the wire,
	   then ask for NodeInfo.  Ignore the switch management port.
	 */
	port_num = osm_physp_get_port_num(p_physp);

	/* if in_sweep_hop_0 is TRUE, then this means the SM is on the switch,
	   and we got switchInfo of our local switch. Do not continue
	   probing through the switch. */
	switch (ib_port_info_get_port_state(p_pi)) {
	case IB_LINK_DOWN:
		break;

	case IB_LINK_INIT:
	case IB_LINK_ARMED:
	case IB_LINK_ACTIVE:
		/*
		 * Don't continue discovery in case OpenSM is in sweep hop 0.
		 * Also to avoid looping forever, only probe the port if it
		 * is NOT the port that responded to the SMP.
		 */
		continue_discovery = (sm->p_subn->in_sweep_hop_0 == FALSE) &&
				     (p_pi->local_port_num != port_num);

		/*
		 * If SM decides to send MEPI, it should continue discovery when
		 * handling  MEPI response.
		 * Otherwise SM should continue discovery from here.
		 */
		if (!pi_rcv_get_mlnx_ext_port_info(sm, p_node, p_port, p_physp, p_pi, continue_discovery) &&
		    sm->p_subn->in_sweep_hop_0 == FALSE) {
			if (p_pi->local_port_num == port_num) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Skipping SMP responder port %u\n",
					p_pi->local_port_num);
				break;
			}

			pi_rcv_get_remote_node_info(sm, p_node, p_port, p_physp);
		}
		break;

	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F03: "
			"Unknown link state = %u, port = %u\n",
			ib_port_info_get_port_state(p_pi),
			p_pi->local_port_num);
		break;
	}

	if (ib_port_info_get_port_state(p_pi) > IB_LINK_INIT && p_node->sw &&
	    !ib_switch_info_get_state_change(&p_node->sw->switch_info) &&
	    p_node->sw->need_update == 1)
		p_node->sw->need_update = 0;

	if (p_physp->need_update)
		sm->p_subn->ignore_existing_lfts = TRUE;

	/*
	   Update the PortInfo attribute.
	 */
	osm_physp_set_port_info(p_physp, p_pi, sm, FALSE);

	if (ib_port_info_get_port_state(p_pi) == IB_LINK_DOWN)
		goto Exit;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_remote_physp) {
		p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
		max_op_vls = osm_get_max_op_vls_by_node(sm->p_subn, p_node);
		if (p_remote_node->sw) {
			data_vls = 1U << (ib_port_info_get_vl_cap(p_pi) - 1);

			if (data_vls > 1U << (max_op_vls - 1))
				data_vls = 1U << (max_op_vls - 1);

			if (data_vls >= IB_MAX_NUM_VLS)
				data_vls = IB_MAX_NUM_VLS - 1;

			if ((uint8_t)data_vls < sm->p_subn->min_sw_data_vls) {
				OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
					"Setting switch port minimal data VLs to:%u defined by node:0x%"
					PRIx64 ", port:%u\n", data_vls,
					osm_node_get_node_guid_ho(p_node),
					port_num);
				sm->p_subn->min_sw_data_vls = data_vls;
			}
		}
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

static void pi_rcv_process_ca_or_router_port(IN osm_sm_t * sm,
					     IN osm_port_t * p_port,
					     IN osm_physp_t *p_physp,
					     IN ib_port_info_t * p_pi)
{
	OSM_LOG_ENTER(sm->p_log);

	pi_rcv_check_and_fix_lid(sm, p_pi, p_port);

	osm_physp_set_port_info(p_physp, p_pi, sm, FALSE);

	pi_rcv_process_endport(sm, p_physp, p_pi);

	OSM_LOG_EXIT(sm->p_log);
}

#define IBM_VENDOR_ID  (0x5076)
static void get_pkey_table(IN osm_log_t * p_log, IN osm_sm_t * sm,
			   IN osm_node_t * p_node, IN osm_physp_t * p_physp)
{

	osm_madw_context_t context;
	ib_api_status_t status;
	osm_dr_path_t path;
	osm_physp_t *physp0;
	ib_net64_t m_key;
	uint8_t port_num;
	uint16_t block_num, max_blocks;
	uint32_t attr_mod_ho;

	OSM_LOG_ENTER(p_log);

	path = *osm_physp_get_dr_path_ptr(sm, p_physp);

	context.pkey_context.node_guid = osm_node_get_node_guid(p_node);
	context.pkey_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pkey_context.set_method = FALSE;

	port_num = p_physp->port_num;
	context.pkey_context.port_num = port_num;

	if (!p_node->sw || port_num == 0)
		/* The maximum blocks is defined by the node info partition cap
		   for CA, router, and switch management ports. */
		max_blocks =
		    (cl_ntoh16(p_node->node_info.partition_cap) +
		     IB_NUM_PKEY_ELEMENTS_IN_BLOCK - 1)
		    / IB_NUM_PKEY_ELEMENTS_IN_BLOCK;
	else {
		/* This is a switch, and not a management port. The maximum blocks
		   is defined in the switch info partition enforcement cap. */

		/* Check for IBM eHCA firmware defect in reporting partition enforcement cap */
		if (cl_ntoh32(ib_node_info_get_vendor_id(&p_node->node_info)) ==
		    IBM_VENDOR_ID)
			p_node->sw->switch_info.enforce_cap = 0;

		/* Bail out if this is a switch with no partition enforcement capability */
		if (cl_ntoh16(p_node->sw->switch_info.enforce_cap) == 0)
			goto Exit;

		max_blocks = (cl_ntoh16(p_node->sw->switch_info.enforce_cap) +
			      IB_NUM_PKEY_ELEMENTS_IN_BLOCK -
			      1) / IB_NUM_PKEY_ELEMENTS_IN_BLOCK;
	}

	p_physp->pkeys.rcv_blocks_cnt = max_blocks;
	for (block_num = 0; block_num < max_blocks; block_num++) {
		if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH ||
		    osm_physp_get_port_num(p_physp) == 0) {
			attr_mod_ho = block_num;
			m_key = ib_port_info_get_m_key(&p_physp->port_info);
		} else {
			attr_mod_ho = block_num | (port_num << 16);
			physp0 = osm_node_get_physp_ptr(p_node, 0);
			m_key = ib_port_info_get_m_key(&physp0->port_info);
		}
		status = osm_req_get(sm, &path, p_physp, IB_MAD_ATTR_P_KEY_TABLE,
				     cl_hton32(attr_mod_ho), FALSE,
				     m_key, 0, CL_DISP_MSGID_NONE, &context);

		if (status != IB_SUCCESS) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 0F12: "
				"Failure initiating PKeyTable request (%s)\n",
				ib_get_err_str(status));
			goto Exit;
		}
	}

Exit:
	OSM_LOG_EXIT(p_log);
}

static void pi_rcv_get_pkey_slvl_vla_tables(IN osm_sm_t * sm,
					    IN osm_node_t * p_node,
					    IN osm_physp_t * p_physp)
{
	OSM_LOG_ENTER(sm->p_log);

	get_pkey_table(sm->p_log, sm, p_node, p_physp);

	OSM_LOG_EXIT(sm->p_log);
}

static int osm_pi_rcv_update_self(IN osm_sm_t *sm, IN osm_physp_t *p_physp,
				  IN ib_port_info_t *p_pi)
{
	boolean_t not_standalone_mode = !(sm->p_subn->sm_port_guid == p_physp->port_guid &&
                                    sm->p_subn->opt.stand_alone != OSM_STANDALONE_NA &&
				    ib_port_info_get_port_state(p_pi) == IB_LINK_DOWN);
	if (ib_port_info_get_port_state(p_pi) == IB_LINK_DOWN && not_standalone_mode)
		return 0;

	if (sm->p_subn->need_update ||
	    ib_port_info_get_port_state(p_pi) == IB_LINK_INIT)
		return 1;

	return p_physp->need_update;
}

/*
 * This function works for planarized port.
 * Copies common to APort parameters in Port Info across all planes of the port.
 */
static void set_aport_common_data(IN const ib_port_info_t *p_orig_pi, IN osm_physp_t *p_orig_physp)
{
	osm_physp_t *p_physp;
	osm_aport_t *p_aport = osm_physp_get_aport(p_orig_physp);
	ib_port_info_t *p_pi;
	uint8_t plane = 0, min_plane = 0, max_plane = 0;

	/* Only aports of planarized hosts copy parameters. */
	if (osm_node_get_type(osm_physp_get_node_ptr(p_orig_physp)) != IB_NODE_TYPE_CA ||
	   !p_aport || !osm_hierarchy_is_planarized_physp(p_orig_physp))
		return;

	min_plane = p_aport->min_found_plane_number;
	max_plane = p_aport->max_found_plane_number;
	for (plane = min_plane; plane <= max_plane; plane ++) {
		p_physp = p_aport->plane_ports[plane];
		/* Skip the plane which received the PortInfo, since it was already updated */
		if (!p_physp || p_physp == p_orig_physp || !osm_physp_is_valid(p_physp))
			continue;

		p_pi = &p_physp->port_info;

		/*
		 * Copy both fields that are read/write and fields that are read-only.
		 *
		 * The following fields are shared in the APort, but are skipped due to various
		 * reasons.
		 *
		 * Fields that are not implemented in the SM:
		 * M_KeyProtectBitsExt, PartitionTopEnabled, EnhancedQoSArbiterEnabled.
		 *
		 * Fields that are writeable, but writing to them just resets their value and
		 * doesn't actually set it:
		 * M_KeyViolations, P_KeyViolations, Q_KeyViolations, ClientReregister.
		 */

		ib_port_info_set_m_key(p_pi, ib_port_info_get_m_key(p_orig_pi));
		p_pi->subnet_prefix = p_orig_pi->subnet_prefix;
		p_pi->base_lid = p_orig_pi->base_lid;
		p_pi->master_sm_base_lid = p_orig_pi->master_sm_base_lid;
		p_pi->capability_mask = p_orig_pi->capability_mask;
		p_pi->diag_code = p_orig_pi->diag_code;
		p_pi->m_key_lease_period = p_orig_pi->m_key_lease_period;
		ib_port_info_set_mpb(p_pi, ib_port_info_get_mpb(p_orig_pi));
		ib_port_info_set_lmc(p_pi, ib_port_info_get_lmc(p_orig_pi));
		ib_port_info_set_neighbor_mtu(p_pi, ib_port_info_get_neighbor_mtu(p_orig_pi));
		ib_port_info_set_master_smsl(p_pi, ib_port_info_get_master_smsl(p_orig_pi));
		ib_port_info_set_vl_cap(p_pi, ib_port_info_get_vl_cap(p_orig_pi));
		p_pi->vl_high_limit = p_orig_pi->vl_high_limit;
		p_pi->vl_arb_high_cap = p_orig_pi->vl_arb_high_cap;
		p_pi->vl_arb_low_cap = p_orig_pi->vl_arb_low_cap;
		ib_port_info_set_mtu_cap(p_pi, ib_port_info_get_mtu_cap(p_orig_pi));
		ib_port_info_set_op_vls(p_pi, ib_port_info_get_op_vls(p_orig_pi));
		p_pi->guid_cap = ib_port_info_get_guid_cap(p_orig_pi);
		ib_port_info_set_mcast_pkey_trap_suppress(p_pi, ib_port_info_get_mcast_pkey_trap_suppress(p_orig_pi));
		ib_port_info_set_timeout(p_pi, ib_port_info_get_timeout(p_orig_pi));
		ib_port_info_set_resp_time_value(p_pi, ib_port_info_get_resp_time_value(p_orig_pi));
		p_pi->capability_mask2 = p_orig_pi->capability_mask2;
	}
}

static void pi_rcv_process_set(IN osm_sm_t * sm,
			       IN uint8_t port_num, IN osm_port_t *p_port, IN osm_madw_t * p_madw)
{
	osm_node_t * p_node;
	osm_physp_t *p_physp;
	ib_net64_t port_guid;
	ib_smp_t *p_smp;
	ib_port_info_t *p_pi;
	osm_pi_context_t *p_context;
	osm_log_level_t level;

	OSM_LOG_ENTER(sm->p_log);

	p_context = osm_madw_get_pi_context_ptr(p_madw);

	OSM_ASSERT(p_port && p_port->p_node);

	p_node = p_port->p_node;

	p_physp = osm_node_get_physp_ptr(p_node, port_num);
	OSM_ASSERT(p_physp);

	port_guid = osm_physp_get_port_guid(p_physp);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_pi = ib_smp_get_payload_ptr(p_smp);

	if (ib_smp_get_status(p_smp)) {
		/* If port already ACTIVE, don't treat status 7 as error */
		if (p_context->active_transition &&
		    (ib_smp_get_status(p_smp) == IB_MAD_STATUS_INVALID_FIELD)) {
			level = OSM_LOG_INFO;
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Received error status 0x%x for SetResp() during ACTIVE transition\n",
				cl_ntoh16(ib_smp_get_status(p_smp)));
			/* Should there be a subsequent Get to validate that port is ACTIVE ? */
		} else {
			level = OSM_LOG_ERROR;
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F10: "
				"Received error status for SetResp()\n");
			osm_hm_set_by_node(sm, p_node, OSM_HM_COND_SET_ERR);
		}
		osm_dump_port_info_v2(sm->p_log, osm_node_get_node_guid(p_node),
				      port_guid, port_num, p_pi, FILE_ID, level);
	} else {
		osm_physp_set_port_info(p_physp, p_pi, sm, p_context->client_rereg);
		/* update common APort data in port info accross all planes of the same port */
		set_aport_common_data(p_pi, p_physp);

		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Received logical SetResp() for GUID 0x%" PRIx64
			", port num %u"
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			" TID 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid), port_num,
			osm_node_get_node_guid_ho(p_node),
			cl_ntoh64(p_smp->trans_id));
	}

	if (p_context->active_transition && port_num &&
		ib_port_info_get_port_state(p_pi) != IB_LINK_ACTIVE) {
			/* Failed attempt to change state to ACTIVE */	
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F31: "
				"Failed to change state of port %u, GUID: 0x%" PRIx64 " to ACTIVE."
				" Current state is %s.\n",
				port_num, cl_ntoh64(port_guid),
				ib_get_port_state_str(ib_port_info_get_port_state(p_pi)));
	}

	OSM_LOG_EXIT(sm->p_log);
}

static int osm_pi_rcv_update_neighbor(IN osm_physp_t *p_physp)
{
	osm_physp_t *p_rem_physp = p_physp->p_remote_physp;
	osm_node_t *p_node;

	/*
	 * Our own port - this is the only case where CA port
	 * is discovered before its' neighbor port
	 */
	if (!p_rem_physp)
		return p_physp->need_update;

	p_node = osm_physp_get_node_ptr(p_rem_physp);
	OSM_ASSERT(p_node);

	/* CA/RTR to CA/RTR connection */
	if (!p_node->sw)
		return p_physp->need_update;

	return (ib_switch_info_get_state_change(&p_node->sw->switch_info) ? 1 : p_physp->need_update);
}

static ib_api_status_t validate_port_info_existing(osm_sm_t *sm,
						   ib_port_info_t *p_pi,
						   osm_physp_t *p_physp)
{
	ib_port_info_t *p_ex_pi = &p_physp->port_info;

	/* If this port wasn't discovered earlier */
	if (osm_physp_is_valid(p_physp) == FALSE)
		return IB_SUCCESS;

	if (ib_port_info_get_port_state(p_pi) == IB_LINK_DOWN ||
	    ib_port_info_get_port_state(p_ex_pi) == IB_LINK_DOWN)
		return IB_SUCCESS;

	/*
	 * If SM doesn't aware of a peer of this port,
	 * don't validate it and overwrite its' PortInfo
	 */
	if (!osm_physp_get_remote(p_physp))
		return IB_SUCCESS;

	/* Avoid validating port when coming out of standby */
	if (PF(sm->p_subn->coming_out_of_standby))
		return IB_SUCCESS;

	/* Don't validate GUIDCap for ext. switch ports */
	if (PF((!p_physp->p_node->sw || !p_physp->port_num) &&
	       ib_port_info_get_guid_cap(p_pi) !=
	       ib_port_info_get_guid_cap(p_ex_pi))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F22: "
			"Received PortInfo for port GUID 0x%" PRIx64 " port %u "
			"is illegal and is being ignored since the "
			"GUIDCap %u doesn't match previously discovered %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)), p_physp->port_num,
			ib_port_info_get_guid_cap(p_pi),
			ib_port_info_get_guid_cap(p_ex_pi));
		return IB_ERROR;
	}

	/* Skip validations if it's base port 0 */
	if (p_physp->p_node->sw && !p_physp->port_num &&
	    !ib_switch_info_is_enhanced_port0(&p_physp->p_node->sw->switch_info))
		return IB_SUCCESS;

	if (PF(ib_port_info_get_vl_cap(p_pi) !=
	       ib_port_info_get_vl_cap(p_ex_pi))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F21: "
			"Received PortInfo for port GUID 0x%" PRIx64 " port %u "
			"is illegal and is being ignored since the "
			"VLCap %u doesn't match previously discovered %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)), p_physp->port_num,
			ib_port_info_get_vl_cap(p_pi),
			ib_port_info_get_vl_cap(p_ex_pi));
		return IB_ERROR;
	}

	if (PF(ib_port_info_get_mtu_cap(p_pi) !=
	       ib_port_info_get_mtu_cap(p_ex_pi))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F23: "
			"Received PortInfo for port GUID 0x%" PRIx64 " port %u "
			"is illegal and is being ignored since the "
			"MTUCap %u doesn't match previously discovered %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)), p_physp->port_num,
			ib_port_info_get_mtu_cap(p_pi),
			ib_port_info_get_mtu_cap(p_ex_pi));
		return IB_ERROR;
	}

	if (PF(p_ex_pi->vl_arb_high_cap != p_pi->vl_arb_high_cap)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F27: "
			"Received PortInfo for port GUID 0x%" PRIx64 " port %u "
			"is illegal and is being ignored since the "
			"VLArbitrationHighCap %u doesn't match previously "
			"discovered %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)), p_physp->port_num,
			p_pi->vl_arb_high_cap, p_ex_pi->vl_arb_high_cap);
		return IB_ERROR;
	}

	if (PF(p_ex_pi->vl_arb_low_cap != p_pi->vl_arb_low_cap)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F28: "
			"Received PortInfo for port GUID 0x%" PRIx64 " port %u "
			"is illegal and is being ignored since the "
			"VLArbitrationLowCap %u doesn't match previously "
			"discovered %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)), p_physp->port_num,
			p_pi->vl_arb_low_cap, p_ex_pi->vl_arb_low_cap);
		return IB_ERROR;
	}

	if (PF(p_pi->base_lid && cl_ntoh16(p_pi->base_lid) <= IB_LID_UCAST_END_HO &&
	       p_ex_pi->base_lid &&
	       ((p_ex_pi->base_lid != p_pi->base_lid) ||
		(p_physp->expected_lmc != OSM_PORT_LMC_NOT_SET &&
		 (p_physp->expected_lmc != ib_port_info_get_lmc(p_pi)))))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F29: "
			"Received PortInfo for port GUID 0x%" PRIx64 " port %u "
			"is illegal and is being ignored since the "
			"base_lid %u lmc %u don't match previously "
			"configured base_lid %u lmc %u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)), p_physp->port_num,
			cl_ntoh16(p_pi->base_lid), ib_port_info_get_lmc(p_pi),
			cl_ntoh16(p_ex_pi->base_lid), p_physp->expected_lmc);
		return IB_ERROR;
	}

	return IB_SUCCESS;
}

static ib_api_status_t validate_port_info(osm_sm_t *sm,
					  ib_port_info_t *p_pi,
					  osm_node_t *p_node,
					  uint8_t port_num)
{
	if (ib_port_info_get_port_state(p_pi) == IB_LINK_DOWN)
		return IB_SUCCESS;

	if (PF(port_num > osm_node_get_num_physp(p_node))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F17: "
			"Received PortInfo for node GUID 0x%" PRIx64 " is "
			"non-compliant and is being ignored since the "
			" port num %u > num ports %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			osm_node_get_num_physp(p_node));
		return IB_ERROR;
	}

	if (PF(p_pi->local_port_num > osm_node_get_num_physp(p_node))) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F15: "
			"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
			"is non-compliant and is being ignored since the "
			"local port num %u > num ports %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			p_pi->local_port_num, p_node->node_info.num_ports);
		return IB_ERROR;
	}

	if (PF(ib_port_info_get_link_speed_sup(p_pi) > IB_LINK_SPEED_MAX_VALUE)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F33: "
			"Received PortInfo for node GUID 0x%" PRIx64 " port %u is non-compliant and"
			" is being ignored since the LinkSpeedSupported is %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			ib_port_info_get_link_speed_sup(p_pi));
		return IB_ERROR;
	}

	if (PF(ib_port_info_get_link_speed_ext_sup(p_pi) > IB_LINK_SPEED_EXT_MAX_VALUE)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F34: "
			"Received PortInfo for node GUID 0x%" PRIx64 " port %u is non-compliant and"
			" is being ignored since the LinkSpeedExtSupported is %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			ib_port_info_get_link_speed_ext_sup(p_pi));
		return IB_ERROR;
	}

	if (PF(ib_port_info_get_link_speed_ext_supported2(p_pi) > IB_LINK_SPEED_EXT2_MAX_VALUE)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F35: "
			"Received PortInfo for node GUID 0x%" PRIx64 " port %u is non-compliant and"
			" is being ignored since the LinkSpeedExtSupported2 is %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			ib_port_info_get_link_speed_ext_supported2(p_pi));
		return IB_ERROR;
	}

	/* Skip validations if it's base port 0 */
	if (p_node->sw && !port_num &&
	    !ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info))
		return IB_SUCCESS;

	if (PF(ib_port_info_get_port_phys_state(p_pi) > IB_PORT_PHYS_STATE_PHYTEST)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F22: "
			"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
			"is non-compliant and is being ignored since the "
			"PortPhysicalState is %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			ib_port_info_get_port_phys_state(p_pi));
		return IB_ERROR;
	}

	if (PF(ib_port_info_get_link_down_def_state(p_pi) > IB_PORT_PHYS_STATE_POLLING)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F23: "
			"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
			"is non-compliant and is being ignored since the "
			"LinkDownDefaultState is %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			ib_port_info_get_link_down_def_state(p_pi));
		return IB_ERROR;

	}

	if (PF(ib_port_info_get_port_state(p_pi) >= IB_LINK_ACT_DEFER)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F21: "
			"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
			"is non-compliant and is being ignored since the "
			"PortState is %u\n",
			osm_node_get_node_guid_ho(p_node), port_num,
			ib_port_info_get_port_state(p_pi));
		return IB_ERROR;
	}

	if (sm->p_subn->opt.validate_smps) {
		if (PF(ib_port_info_get_mtu_cap(p_pi) < IB_MTU_LEN_256 ||
		       ib_port_info_get_mtu_cap(p_pi) > IB_MTU_LEN_4096)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F24: "
				"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
				"is non-compliant and is being ignored since the "
				"MTUCap is %u\n",
				osm_node_get_node_guid_ho(p_node), port_num,
				ib_port_info_get_mtu_cap(p_pi));
			return IB_ERROR;
		}

		if (PF(ib_port_info_get_vl_cap(p_pi) < IB_MIN_VL ||
		       ib_port_info_get_vl_cap(p_pi) > IB_MAX_VL)) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F26: "
				"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
				"is non-compliant and is being ignored since the "
				"VLCap is %u\n",
				osm_node_get_node_guid_ho(p_node), port_num,
				ib_port_info_get_vl_cap(p_pi));
			return IB_ERROR;
		}
	} else {
		if (PF(ib_port_info_get_mtu_cap(p_pi) < IB_MTU_LEN_256 ||
		       ib_port_info_get_mtu_cap(p_pi) > IB_MTU_LEN_4096))
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
				"is non-compliant since the MTUCap is %u\n",
				osm_node_get_node_guid_ho(p_node), port_num,
				ib_port_info_get_mtu_cap(p_pi));

		if (PF(ib_port_info_get_vl_cap(p_pi) < IB_MIN_VL ||
		       ib_port_info_get_vl_cap(p_pi) > IB_MAX_VL))
			OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
				"Received PortInfo for node GUID 0x%" PRIx64 " port %u "
				"is non-compliant since the VLCap is %u\n",
				osm_node_get_node_guid_ho(p_node), port_num,
				ib_port_info_get_vl_cap(p_pi));
	}
	return IB_SUCCESS;
}

void pi_rcv_process_get(IN osm_sm_t * sm, IN osm_node_t *p_node,
			IN osm_port_t *p_port, IN osm_physp_t *p_physp,
			IN const osm_madw_t *p_madw, IN ib_port_info_t *p_pi)
{
	osm_dr_path_t *p_dr_path;
	ib_net64_t node_guid = osm_node_get_node_guid(p_node);
	uint8_t port_num = osm_physp_get_port_num(p_physp);

	/*
	 * Update the directed route path to this port
	 * in case the old path is no longer usable.
	 */
	p_dr_path = osm_physp_get_dr_path_ptr(sm, p_physp);
	osm_madw_get_smp_dr(p_madw, p_dr_path);

	p_physp->need_update = osm_pi_rcv_update_self(sm, p_physp, p_pi);

	switch (osm_node_get_type(p_node)) {
	case IB_NODE_TYPE_CA:
	case IB_NODE_TYPE_ROUTER:
		osm_update_discovery(sm, p_node, p_port, port_num, FALSE);

		p_physp->need_update = osm_pi_rcv_update_neighbor(p_physp);
		/* pass the physp for which the port info was received */
		pi_rcv_process_ca_or_router_port(sm, p_port, p_physp,
						 p_pi);
		pi_rcv_get_mlnx_ext_port_info(sm, p_node, p_port, p_physp, p_pi, FALSE);
		break;
	case IB_NODE_TYPE_SWITCH:
		osm_update_discovery(sm, p_node, p_port, port_num, TRUE);

		if (port_num == 0)
			pi_rcv_process_switch_port0(sm, p_node, p_port,
						    p_physp, p_pi);
		else
			pi_rcv_process_switch_ext_port(sm, p_node, p_port,
						       p_physp, p_pi);

		if (ib_port_info_get_port_state(p_pi) == IB_LINK_DOWN) {
			return;
		}

		break;
	default:
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F07: "
			"Unknown node type %u with GUID 0x%" PRIx64
			"\n", osm_node_get_type(p_node),
			cl_ntoh64(node_guid));
		break;
	}

	/*
	   Get the tables on the physp.
	 */
	if (p_physp->need_update || (p_node->sw &&
				     p_node->sw->need_update))
		pi_rcv_get_pkey_slvl_vla_tables(sm, p_node, p_physp);
}

void osm_pi_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_port_info_t *p_pi;
	ib_smp_t *p_smp;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_node_t *p_node;
	osm_pi_context_t *p_context;
	ib_net64_t port_guid, node_guid;
	uint8_t context_port_num;
	uint8_t port_num;
	uint8_t port_state;
	uint8_t old_state;
	boolean_t verbose_bypass = FALSE;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = osm_madw_get_pi_context_ptr(p_madw);
	p_pi = ib_smp_get_payload_ptr(p_smp);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_PORT_INFO);

	port_guid = p_context->port_guid;
	node_guid = p_context->node_guid;

	verbose_bypass = osm_verbose_bypass_handle_pi(sm, p_pi->base_lid,
						      port_guid, node_guid);

	/*
	 * Attribute modifier has already been validated upon MAD receive,
	 * which means that port_num has to be valid - it originated from
	 * the request attribute modifier.
	 */
	port_num = (uint8_t) cl_ntoh32(p_smp->attr_mod);

	context_port_num = p_context->port_num;

	osm_dump_port_info_v2(sm->p_log, node_guid, port_guid, port_num, p_pi,
			      FILE_ID, OSM_LOG_DEBUG);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	p_port = osm_get_port_by_guid(sm->p_subn, port_guid);
	if (PF(!p_port)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F06: "
			"No port object for port with GUID 0x%" PRIx64
			"\n\t\t\t\tfor parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(port_guid),
			cl_ntoh64(node_guid), cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	/* On receipt of client reregister, clear the reregister bit so
	   reregistering won't be sent again and again */
	if (p_context->set_method &&
	    (ib_port_info_get_client_rereg(p_pi) || p_context->client_rereg)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Client reregister received on response\n");
		ib_port_info_set_client_rereg(p_pi, 0);
		p_port->client_rereg = FALSE;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	/*
	   We might get a response during a light sweep looking for a change in
	   the status of a remote port that did not respond in earlier sweeps.
	   So if the context of the Get was light_sweep - we do not need to
	   do anything with the response - just flag that we need a heavy sweep
	 */
	if (p_context->light_sweep == TRUE) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Got light sweep response from remote port of parent node "
			"GUID 0x%" PRIx64 " port 0x%016" PRIx64
			", Commencing heavy sweep\n",
			cl_ntoh64(node_guid),
			cl_ntoh64(port_guid));
		sm->p_subn->force_heavy_sweep = TRUE;
		sm->p_subn->ignore_existing_lfts = TRUE;
		osm_hm_set_port_answer_ls(sm, p_node, context_port_num, TRUE);
		goto Exit;
	}

	if (PF(validate_port_info(sm, p_pi, p_node, port_num) != IB_SUCCESS)) {
		osm_hm_set_by_smp_dr_path(sm, p_madw, OSM_HM_COND_ILLEGAL);
		goto Exit;
	}

	p_physp = osm_node_get_physp_ptr(p_node, port_num);

	OSM_ASSERT(p_physp);

	if (PF(validate_port_info_existing(sm, p_pi, p_physp) !=
	    IB_SUCCESS)) {
		sm->p_subn->subnet_initialization_error = TRUE;
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 0F0A: "
			"Failed to validate PortInfo for 0x%016" PRIx64
			" [%d]\n", cl_ntoh64(port_guid), port_num);
		p_node->force_drop_node = TRUE;
		goto Exit;
	}

	/*
	   If we were setting the PortInfo, then receiving
	   this attribute was not part of sweeping the subnet.
	   In this case, just update the PortInfo attribute.

	   In an unfortunate blunder, the IB spec defines the
	   return method for Set() as a GetResp().  Thus, we can't
	   use the method (what would have been SetResp()) to determine
	   our course of action.  So, we have to carry this extra
	   boolean around to determine if we were doing Get() or Set().
	 */
	if (p_context->set_method) {
		pi_rcv_process_set(sm, port_num, p_port, p_madw);
	} else {
		/*
		   This PortInfo arrived because we did a Get() method,
		   most likely due to a subnet sweep in progress.
		 */
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Discovered port num %u with GUID 0x%" PRIx64
			" for parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			port_num, cl_ntoh64(port_guid),
			cl_ntoh64(node_guid), cl_ntoh64(p_smp->trans_id));

		/*
		 * Log port state changes.
		 * However, avoid reporting state changes for new devices, and
		 * in the first sweep,
		 */
		old_state = ib_port_info_get_port_state(&p_physp->port_info);
		port_state = ib_port_info_get_port_state(p_pi);

		if (p_node->sw && !p_port->is_new &&
		    !sm->p_subn->first_time_master_sweep &&
		    sm->p_subn->sm_state_info.sm_state != IB_SMINFO_STATE_DISCOVERING &&
		    port_state != old_state) {
			OSM_LOG(sm->p_log, OSM_LOG_INFO,
				"Switch 0x%" PRIx64 " %s port %u%s%s%s"
				" changed state from %s to %s \n",
				cl_ntoh64(port_guid), p_node->print_desc, port_num,
				*p_physp->hierarchy_info.port_label ? "(" : "",
				p_physp->hierarchy_info.port_label,
				*p_physp->hierarchy_info.port_label ? ")" : "",
				ib_get_port_state_str(old_state),
				ib_get_port_state_str(port_state));
		}

		pi_rcv_process_get(sm, p_node, p_port, p_physp, p_madw, p_pi);
	}

	osm_port_send_next_mad(sm, p_port);


Exit:
	CL_PLOCK_RELEASE(sm->p_lock);
	osm_verbose_bypass_end(verbose_bypass, sm->p_log);
	/*
	   Release the lock before jumping here!!
	 */
	OSM_LOG_EXIT(sm->p_log);
}
