/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2005,2015 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_router_t.
 * This object represents an Infiniband router.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_C
#include <opensm/osm_router.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>

void osm_router_tbl_construct(IN osm_router_tbl_t * p_router_tbl)
{
	OSM_ASSERT(p_router_tbl);
	cl_ptr_vector_construct(&p_router_tbl->blocks);
}

void osm_router_tbl_delete(IN osm_router_tbl_t * p_router_tbl)
{
	ib_rtr_adj_table_block_t *p_block;
	size_t num_blocks;
	int i;

	OSM_ASSERT(p_router_tbl);
	num_blocks = cl_ptr_vector_get_size(&p_router_tbl->blocks);
	if (num_blocks) {
		for (i = num_blocks - 1; i >= 0; i--) {
			if ((p_block =
				cl_ptr_vector_remove(&p_router_tbl->blocks, i)))
				free(p_block);
		}
	}
}

void osm_router_tbl_destroy(IN osm_router_tbl_t * p_router_tbl)
{
	OSM_ASSERT(p_router_tbl);
	osm_router_tbl_delete(p_router_tbl);
	cl_ptr_vector_destroy(&p_router_tbl->blocks);
}

ib_api_status_t osm_router_tbl_init(IN osm_router_tbl_t * p_router_tbl)
{
	OSM_ASSERT(p_router_tbl);
	if (cl_ptr_vector_init(&p_router_tbl->blocks, 0, 1) == CL_SUCCESS)
		return IB_SUCCESS;
	return IB_UNKNOWN_ERROR;
}

ib_api_status_t osm_adj_router_tbl_set(IN osm_router_tbl_t * p_router_tbl,
				       IN uint16_t block,
				       IN ib_rtr_adj_table_block_t * p_tbl)
{
	ib_rtr_adj_table_block_t *p_router_block;

	OSM_ASSERT(p_router_tbl);
	OSM_ASSERT(p_tbl);

	/* make sure the block is allocated */
	if (cl_ptr_vector_get_size(&p_router_tbl->blocks) > block)
		p_router_block = (ib_rtr_adj_table_block_t *)
				  cl_ptr_vector_get(&p_router_tbl->blocks,
						    block);
	else
		p_router_block = NULL;

	if (!p_router_block) {
		p_router_block = (ib_rtr_adj_table_block_t *)
			calloc(1, sizeof(ib_rtr_adj_table_block_t));
		if (!p_router_block)
			return IB_ERROR;
		cl_ptr_vector_set(&p_router_tbl->blocks, block, p_router_block);
	}

	/* sets the block values */
	memcpy(p_router_block, p_tbl, sizeof(ib_rtr_adj_table_block_t));


	return IB_SUCCESS;
}

ib_api_status_t osm_next_hop_router_tbl_set(IN osm_router_tbl_t *p_router_tbl,
					    IN uint16_t block_num,
					    IN ib_rtr_next_hop_table_block_t *p_block)
{
	ib_rtr_next_hop_table_block_t *p_router_block;

	OSM_ASSERT(p_router_tbl);
	OSM_ASSERT(p_block);

	/* make sure the block is allocated */
	if (cl_ptr_vector_get_size(&p_router_tbl->blocks) > block_num)
		p_router_block = (ib_rtr_next_hop_table_block_t *)
				  cl_ptr_vector_get(&p_router_tbl->blocks,
						    block_num);
	else
		p_router_block = NULL;

	if (!p_router_block) {
		p_router_block = (ib_rtr_next_hop_table_block_t *)
			calloc(1, sizeof(ib_rtr_next_hop_table_block_t));
		if (!p_router_block)
			return IB_ERROR;
		cl_ptr_vector_set(&p_router_tbl->blocks, block_num, p_router_block);
	}

	/* sets the block values */
	memcpy(p_router_block, p_block, sizeof(ib_rtr_next_hop_table_block_t));


	return IB_SUCCESS;
}

void osm_adj_router_lid_info_set(IN osm_router_t *p_router,
				 IN uint32_t block,
				 IN ib_rtr_adj_subnets_lid_info_block_t * p_block) {
	memcpy(p_router->adj_router_lid_info + block, p_block, sizeof(*p_block));
}

uint32_t osm_router_lid_tbl_offset_to_block(IN osm_router_t *p_router, IN uint32_t offset)
{
	return offset + RTR_TBL_ENTRY_IDX_TO_BLOCK(p_router->global_flid_start,
						   IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
}

void osm_router_lid_table_set(IN osm_router_t *p_router,
			      IN uint32_t block,
			      IN ib_rtr_router_lid_tbl_block_t * p_block)
{
	uint32_t offset;

	/*
	 * router_lid_tbl keeps number of blocks according to the global range.
	 * it starts not from block 0, but from the first block that contains
	 * global range lid start. So the 'offset' needs to be calculated respectively.
	 */
	offset = osm_router_lid_tbl_block_to_global_offset(p_router, block);
	memcpy(p_router->router_lid_tbl + offset, p_block, sizeof(*p_block));
}

/********************** END OF ROUTER TABLE **********************************/
void osm_router_delete(IN OUT osm_router_t ** pp_rtr)
{
	cl_fmap_item_t *p_item, *p_next_item;
	osm_subnet_prefix_pkey_t *p_key;

	osm_router_tbl_destroy(&(*pp_rtr)->adj_router_tbl);
	osm_router_tbl_destroy(&(*pp_rtr)->next_hop_router_tbl);
	p_next_item = cl_fmap_head(&(*pp_rtr)->subnet_prefix_tbl);
	while (p_next_item != cl_fmap_end(&(*pp_rtr)->subnet_prefix_tbl)) {
		p_item = p_next_item;
		p_next_item = cl_fmap_next(p_item);
		cl_fmap_remove_item(&(*pp_rtr)->subnet_prefix_tbl, p_item);
		p_key = (osm_subnet_prefix_pkey_t *)cl_fmap_key(p_item);
		free(p_key);
		free(p_item);
	}
	if ((*pp_rtr)->adj_router_lid_info)
		free((*pp_rtr)->adj_router_lid_info);

	if ((*pp_rtr)->router_lid_tbl)
		free((*pp_rtr)->router_lid_tbl);

	if ((*pp_rtr)->router_lid_tbl_locals)
		free((*pp_rtr)->router_lid_tbl_locals);

	if ((*pp_rtr)->new_router_lid_tbl)
		free((*pp_rtr)->new_router_lid_tbl);

	if ((*pp_rtr)->enabled_flids)
		free((*pp_rtr)->enabled_flids);

	free((*pp_rtr)->subn_ar_group_to_lid);

	free(*pp_rtr);
	*pp_rtr = NULL;
}

osm_router_t *osm_router_new(osm_sm_t * sm, IN osm_port_t * p_port)
{
	osm_router_t *p_rtr;
	uint32_t num_blocks = 0;

	OSM_ASSERT(p_port);

	p_rtr = malloc(sizeof(*p_rtr));
	if (!p_rtr)
		return NULL;

	memset(p_rtr, 0, sizeof(*p_rtr));
	p_rtr->p_port = p_port;

	osm_router_tbl_construct(&p_rtr->adj_router_tbl);
	osm_router_tbl_init(&p_rtr->adj_router_tbl);
	osm_router_tbl_construct(&p_rtr->next_hop_router_tbl);
	osm_router_tbl_init(&p_rtr->next_hop_router_tbl);
	cl_fmap_init(&p_rtr->subnet_prefix_tbl, ib_compare_prefix_pkey);

	p_rtr->global_flid_start = sm->p_subn->opt.global_flid_start;
	p_rtr->global_flid_end = sm->p_subn->opt.global_flid_end;
	p_rtr->local_flid_start = sm->p_subn->opt.local_flid_start;
	p_rtr->local_flid_end = sm->p_subn->opt.local_flid_end;

	num_blocks = RTR_TBL_NUM_BLOCKS_RANGE(p_rtr->global_flid_start,
					      p_rtr->global_flid_end,
					      IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
	/* if the FLIDs feature is configured for use */
	if (p_rtr->global_flid_start && num_blocks) {
		p_rtr->router_lid_tbl = calloc(num_blocks, sizeof(ib_rtr_router_lid_tbl_block_t));

		if (!p_rtr->router_lid_tbl) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8002: "
				"Failed to allocate memory for router lid tables "
				"for router node GUID 0x%" PRIx64 ".\n",
				osm_node_get_node_guid_ho(p_rtr->p_port->p_node));
			free(p_rtr);
			return NULL;
		}

		p_rtr->router_lid_tbl_locals = calloc(num_blocks, sizeof(ib_rtr_router_lid_tbl_block_t));

		if (!p_rtr->router_lid_tbl_locals) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 8003: "
				"Failed to allocate memory for local flids router lid table "
				"for router node GUID 0x%" PRIx64 ".\n",
				osm_node_get_node_guid_ho(p_rtr->p_port->p_node));
			free(p_rtr->router_lid_tbl);
			free(p_rtr);
			return NULL;
		}

		p_rtr->adj_router_lid_info_top = 0;
		/* assign calculated blocks for the local range */
		num_blocks = RTR_TBL_NUM_BLOCKS_RANGE(p_rtr->local_flid_start,
						      p_rtr->local_flid_end,
						      IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);

		if (num_blocks) {
			p_rtr->new_router_lid_tbl = calloc(num_blocks,
							   sizeof(ib_rtr_router_lid_tbl_block_t));
			if (!p_rtr->new_router_lid_tbl) {
				free(p_rtr->router_lid_tbl);
				free(p_rtr->router_lid_tbl_locals);
				free(p_rtr);
				return NULL;
			}
		}

		p_rtr->is_router_info_set_error = FALSE;
	}

	return p_rtr;
}

ib_api_status_t osm_router_find_pkey(
	IN osm_subn_t *p_subn,
	IN const osm_alias_guid_t *p_router_alias_guid,
	IN const ib_gid_t *p_remote_gid,
	IN const osm_pkey_tbl_t *pkey_tbl_src,
	IN const osm_pkey_tbl_t *pkey_tbl_dest,
	OUT ib_net16_t *pkey)
{
	ib_api_status_t status = IB_SUCCESS;
	cl_map_item_t *p_item;
	cl_fmap_item_t *p_fitem;
	osm_router_t *p_rtr;
	osm_router_tbl_data_t *p_subn_pkey_data;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* find router by p_router_alias_guid */
	p_item = cl_qmap_get(&p_subn->rtr_guid_tbl,
			     p_router_alias_guid->base_port_guid);
	if (p_item == cl_qmap_end(&p_subn->rtr_guid_tbl))
	{
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 8000: No router found with guid 0x%016" PRIx64
			"\n", cl_ntoh64(p_router_alias_guid->base_port_guid));
		status = IB_NOT_FOUND;
		goto Exit;
	}
	p_rtr = (osm_router_t *)p_item;

	/*
	 * go over subnets of this router and take the first matching subnet
	 * p_remote_gid.subnet pair pkey.
	 */
	for (p_fitem = cl_fmap_head(&p_rtr->subnet_prefix_tbl);
	     p_fitem != cl_fmap_end(&p_rtr->subnet_prefix_tbl);
             p_fitem = cl_fmap_next(p_fitem)) {
		p_subn_pkey_data = (osm_router_tbl_data_t *)p_fitem;

		/* the subnet prefixes are of site local type */
		if (is_equal_subnet_prefix(p_subn_pkey_data->subnet_prefix,
					   ib_gid_get_subnet_prefix(p_remote_gid))) {
			*pkey = p_subn_pkey_data->pkey;
			/*
			 * take the first pkey that both src and
			 * destination share
			 */
			if (osm_pkey_tbl_share_this_pkey(pkey_tbl_src,
							 pkey_tbl_dest,
							 *pkey,
							 p_subn->opt.
							     allow_both_pkeys))
				break;
		}
	}
	/* didn't find matching subnet in our data base */
	if (p_fitem == cl_fmap_end(&p_rtr->subnet_prefix_tbl))
		status = IB_NOT_FOUND;
Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

uint32_t osm_router_rtr_lid_tbl_get_set(IN OUT osm_router_t *p_router,
					IN uint16_t flid,
					IN osm_rtr_lid_tbl_type_enum rtr_lid_tbl_type,
					IN osm_rtr_lid_tbl_action_enum rtr_rtr_lid_tbl_action) {
	ib_rtr_router_lid_tbl_block_t *p_block;
	uint32_t result_record_in_block;

	/* find the right block */
	uint32_t block_idx = RTR_TBL_ENTRY_IDX_TO_BLOCK(flid, IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
	uint32_t offset;

	uint8_t record_in_block = RTR_RTR_LID_TBL_BLOCK_RECORD_IDX(flid);
	uint8_t idx_in_record = RTR_RTR_LID_TBL_BLOCK_IDX_IN_RECORD(flid);

	/* get the subnet block, set the calculated block */
	switch (rtr_lid_tbl_type) {
		case OSM_RTR_LID_TBL_SUBNET:
			offset = osm_router_lid_tbl_block_to_global_offset(p_router, block_idx);
			p_block = p_router->router_lid_tbl + offset;
			break;
		case OSM_RTR_LID_TBL_LOCAL_CALCULATION:
			offset = osm_router_lid_tbl_block_to_local_offset(p_router, block_idx);
			p_block = p_router->new_router_lid_tbl + offset;
			break;
		case OSM_RTR_LID_TBL_SUBNET_LOCALS_ONLY:
			offset = osm_router_lid_tbl_block_to_global_offset(p_router, block_idx);
			p_block = p_router->router_lid_tbl_locals + offset;
			break;
		default:
			return 0;
	}

	uint32_t lids = cl_ntoh32(p_block->lids[record_in_block]);

	if (rtr_rtr_lid_tbl_action == OSM_RTR_LID_TBL_GET)
		result_record_in_block = lids & (1 << idx_in_record);
	else {
		/* update calculated blocks */
		if (rtr_rtr_lid_tbl_action == OSM_RTR_LID_TBL_SET_ENABLE)
			lids |= (1 << idx_in_record);
		else /* OSM_RTR_LID_TBL_SET_DISABLE */
			lids &= ~(1 << idx_in_record);
		p_block->lids[record_in_block] = cl_hton32(lids);
		result_record_in_block = lids;
	}
	return result_record_in_block;
}

void osm_router_set_flid(IN osm_subn_t *p_subn, IN OUT osm_router_t *p_router, IN uint16_t flid)
{
	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* flid is in the right local range */
	OSM_ASSERT(flid >= p_router->local_flid_start &&
		   flid <= p_router->local_flid_end);

	osm_router_rtr_lid_tbl_get_set(p_router, flid,
			OSM_RTR_LID_TBL_LOCAL_CALCULATION, OSM_RTR_LID_TBL_SET_ENABLE);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

boolean_t osm_router_is_router_in_the_list(osm_router_t *p_router, cl_list_t *p_rtrs_list)
{
	cl_list_iterator_t list_iter;

	for (list_iter = cl_list_head(p_rtrs_list);
	     list_iter != cl_list_end(p_rtrs_list);
	     list_iter = cl_list_next(list_iter)) {
		if (p_router == (osm_router_t *)cl_list_obj(list_iter))
			return TRUE;
	}
	return FALSE;
}

void osm_router_ar_group_to_lid_tbl_clear(IN osm_router_t *p_router)
{
	int num_blocks;

	if (!p_router->num_ar_group_to_lid_entries)
		return;

	num_blocks = RTR_TBL_NUM_BLOCKS(p_router->num_ar_group_to_lid_entries, IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS);
	memset(p_router->subn_ar_group_to_lid, 0, num_blocks * sizeof(ib_rtr_ar_group_to_lid_tbl_block_t));
}

void osm_router_ar_group_to_lid_tbl_reinit(IN osm_subn_t *p_subn, IN osm_router_t *p_router, IN uint16_t num_entries)
{
	int num_blocks;

	/* Number of entires hasnt chagned, just flush tables */
	if (num_entries == p_router->num_ar_group_to_lid_entries) {
		osm_router_ar_group_to_lid_tbl_clear(p_router);
		return;
	}

	/* Free tables and realloacte */

	free(p_router->subn_ar_group_to_lid);
	p_router->subn_ar_group_to_lid = NULL;
	p_router->num_ar_group_to_lid_entries = 0;

	if (!num_entries)
		return;

	num_blocks = RTR_TBL_NUM_BLOCKS(num_entries, IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS);
	p_router->subn_ar_group_to_lid = calloc(num_blocks, sizeof(ib_rtr_ar_group_to_lid_tbl_block_t));
	if (!p_router->subn_ar_group_to_lid) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 8004: Failed to allocate AR group to LID table\n");
		return;
	}

	p_router->num_ar_group_to_lid_entries = num_entries;
}
