/*
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_router_mgr_t.
 * This object implements the GUID manager object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <complib/cl_qlist.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_MGR_C
#include <opensm/osm_log.h>
#include <opensm/osm_router.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>

/* Growth size for prefix to routers table */
#define GROWTH_SIZE 8

extern int wait_for_pending_transactions(osm_opensm_t * p_osm);

typedef struct osm_neighborhood_flid {
	uint16_t flid;
	uint16_t count;
} osm_neighborhood_flid_t;

typedef struct osm_neighborhood_flids_data {
	uint16_t id; /* neighborhood id */
	cl_ptr_vector_t flids_to_count;
} osm_neighborhood_flids_data_t;

typedef struct osm_flid_neighborhood {
	uint16_t neighborhood_id;
	uint16_t count; /* how many times flid appears in this neighborhood */
} osm_flid_neighborhood_t;

/* stores number of appearances of flid in each neighborhood */
typedef struct osim_flid_neighborhoods {
	uint16_t flid;
	cl_ptr_vector_t neighborhood_to_count;
} osm_flid_neighborhoods_t;

static void router_mgr_clear_flids_not_in_subnet(IN osm_router_mgr_t * p_mgr)
{
	memset(p_mgr->flids_not_in_subnet, 0,
		(p_mgr->p_subn->opt.global_flid_end + 1) * sizeof(*(p_mgr->flids_not_in_subnet)));
}

ib_api_status_t osm_router_mgr_init(IN osm_router_mgr_t * p_mgr, IN osm_sm_t * sm)
{
	OSM_LOG_ENTER(sm->p_log);

	p_mgr->p_log = sm->p_log;
	p_mgr->p_subn = sm->p_subn;
	cl_fmap_init(&p_mgr->prefix_rtr_table, ib_compare_prefix_pkey);
	cl_fmap_init(&p_mgr->flid_to_rtrs, osm_compare_uint16);

	p_mgr->flid_compression = sm->p_subn->opt.flid_compression;

	cl_u64_vector_init(&p_mgr->switches_with_flids, 0, TRUE);
	cl_ptr_vector_init(&p_mgr->flid_to_neighborhoods, GROWTH_SIZE, GROWTH_SIZE);
	cl_ptr_vector_init(&p_mgr->neighborhood_to_flids, GROWTH_SIZE, GROWTH_SIZE);
	cl_fmap_init(&p_mgr->local_flid_to_routers, osm_compare_uint16);

	p_mgr->flids_not_in_subnet = NULL;

	/* if flids are enabled */
	if (osm_subn_is_flids_enabled(p_mgr->p_subn)) {
		uint32_t global_flids = p_mgr->p_subn->opt.global_flid_end + 1;
		p_mgr->flids_not_in_subnet =
			(uint8_t *)calloc(global_flids, sizeof(*p_mgr->flids_not_in_subnet));
		if (!p_mgr->flids_not_in_subnet) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7721: "
				"Failed to allocate memory for FLIDs not in subnet db.\n");
		}
	}

	OSM_LOG_EXIT(sm->p_log);
	return IB_SUCCESS;
}

static void router_mgr_destroy_flids_to_count(osm_neighborhood_flids_data_t *p_flids_data)
{
	uint16_t flids_num = cl_ptr_vector_get_size(&p_flids_data->flids_to_count);
	int flid;
	osm_neighborhood_flid_t *p_neighborhood_flid;

	if (flids_num) {
		for (flid = flids_num - 1; flid >= 0; flid --) {
			if ((p_neighborhood_flid =
				cl_ptr_vector_remove(&p_flids_data->flids_to_count, flid))) {
				free(p_neighborhood_flid);
			}
		}
	}
}

static void
router_mgr_destroy_flids_to_neighborhoods(IN osm_flid_neighborhoods_t *p_flid_neighborhoods)
{
	uint16_t neighborhoods_num =
		cl_ptr_vector_get_size(&p_flid_neighborhoods->neighborhood_to_count);
	int n_id;
	osm_flid_neighborhood_t *p_neighborhood;

	if (neighborhoods_num) {
		for (n_id = neighborhoods_num - 1; n_id >= 0; n_id--) {
			if ((p_neighborhood = cl_ptr_vector_remove(
				&p_flid_neighborhoods->neighborhood_to_count, n_id))) {
				free(p_neighborhood);
			}
		}
	}
}

static void router_mgr_destroy_flid_compression_db(IN osm_router_mgr_t * p_mgr)
{
	uint16_t num_neighborhoods = cl_ptr_vector_get_size(&p_mgr->neighborhood_to_flids);
	osm_neighborhood_flids_data_t *p_data;
	uint16_t flid_neighborhoods_num =
		cl_ptr_vector_get_size(&p_mgr->flid_to_neighborhoods);
	osm_flid_neighborhoods_t *p_flid_data;
	int n_id, flid_id;
	cl_fmap_item_t *p_item;
	osm_flid_rtrs_t *p_flid_to_rtrs;

	if (num_neighborhoods) {
		for (n_id = num_neighborhoods - 1; n_id >= 0; n_id--) {
			if ((p_data = cl_ptr_vector_remove(&p_mgr->neighborhood_to_flids, n_id))) {
				router_mgr_destroy_flids_to_count(p_data);
				cl_ptr_vector_destroy(&p_data->flids_to_count);
				free(p_data);
			}
		}
	}

	if (flid_neighborhoods_num) {
		for (flid_id = flid_neighborhoods_num - 1; flid_id >= 0; flid_id--) {
			if ((p_flid_data =
			      cl_ptr_vector_remove(&p_mgr->flid_to_neighborhoods, flid_id))) {
				router_mgr_destroy_flids_to_neighborhoods(p_flid_data);
				cl_ptr_vector_destroy(&p_flid_data->neighborhood_to_count);
				free(p_flid_data);
			}
		}
	}

	while (cl_fmap_count(&p_mgr->local_flid_to_routers)) {
		p_item = cl_fmap_head(&p_mgr->local_flid_to_routers);
		cl_fmap_remove_item(&p_mgr->local_flid_to_routers, p_item);
		p_flid_to_rtrs = (osm_flid_rtrs_t *)p_item;
		cl_ptr_vector_destroy(&p_flid_to_rtrs->routers);
		free(p_item);
	}
}

static void router_mgr_destroy_on_flid_sweep(IN osm_router_mgr_t * p_mgr) {
	/* destroy flid router mapping */
	osm_router_mgr_destroy_flid_router_db(p_mgr);
	cl_u64_vector_remove_all(&p_mgr->switches_with_flids);
}

static void router_mgr_destroy_internal(IN osm_router_mgr_t * p_mgr) {
	/* remove router db subnets prefixes table */
	osm_router_mgr_destroy_prefix_router_db(p_mgr);

	/* destroy flid router mapping */

	router_mgr_destroy_on_flid_sweep(p_mgr);
	router_mgr_destroy_flid_compression_db(p_mgr);
}

void osm_router_mgr_destroy(IN osm_router_mgr_t * p_mgr)
{
	OSM_LOG_ENTER(p_mgr->p_log);

	router_mgr_destroy_internal(p_mgr);

	cl_ptr_vector_destroy(&p_mgr->neighborhood_to_flids);
	cl_ptr_vector_destroy(&p_mgr->flid_to_neighborhoods);
	cl_u64_vector_destroy(&p_mgr->switches_with_flids);

	if (p_mgr->flids_not_in_subnet) {
		free(p_mgr->flids_not_in_subnet);
		p_mgr->flids_not_in_subnet = NULL;
	}

	OSM_LOG_EXIT(p_mgr->p_log);
	return;
}

void osm_router_mgr_destroy_flid_router_db(IN osm_router_mgr_t * p_mgr)
{
	cl_fmap_item_t *p_item;
	osm_flid_rtrs_list_t *p_flid_rtrs_list;

	while (cl_fmap_count(&p_mgr->flid_to_rtrs)) {
		p_item = cl_fmap_head(&p_mgr->flid_to_rtrs);
		cl_fmap_remove_item(&p_mgr->flid_to_rtrs, p_item);
		p_flid_rtrs_list = (osm_flid_rtrs_list_t *)p_item;
		cl_list_remove_all(&p_flid_rtrs_list->rtrs_list);
		cl_list_destroy(&p_flid_rtrs_list->rtrs_list);
		free(p_item);
	}

	free(p_mgr->p_ar_group_to_lid_tbl);
	p_mgr->p_ar_group_to_lid_tbl = NULL;
}

void osm_router_mgr_destroy_prefix_router_db(IN osm_router_mgr_t * p_mgr)
{
	cl_fmap_item_t *p_item;
	osm_subnet_prefix_pkey_t *p_key;
	osm_router_prefix_rtrs_map_t *p_rtrs_map;

	while (cl_fmap_count(&p_mgr->prefix_rtr_table)) {
		p_item = cl_fmap_head(&p_mgr->prefix_rtr_table);
		cl_fmap_remove_item(&p_mgr->prefix_rtr_table, p_item);
		/* get cl map of the router guids */
		p_rtrs_map = (osm_router_prefix_rtrs_map_t *)p_item;

		cl_map_remove_all(&p_rtrs_map->rtrs);
		cl_map_destroy(&p_rtrs_map->rtrs);
		p_key = (osm_subnet_prefix_pkey_t *)cl_fmap_key(p_item);
		free(p_key);
		free(p_rtrs_map->rtr_guid_arr);
		free(p_item);
	}
}

static void router_mgr_router_info_get(IN cl_map_item_t * p_object, IN void *context)
{
        ib_api_status_t status = IB_SUCCESS;
        osm_madw_context_t mad_context;
	osm_router_t *const p_router = (osm_router_t *) p_object;
	osm_sm_t *sm = context;
	osm_physp_t *p_physp = p_router->p_port->p_physp;

        OSM_LOG_ENTER(sm->p_log);

        memset(&mad_context, 0, sizeof(mad_context));
        mad_context.ri_context.port_guid = osm_physp_get_port_guid(p_physp);
        mad_context.ri_context.set_method = FALSE;
        mad_context.ri_context.sweep_type = OSM_SWEEP_CONTEXT_HEAVY;

        /* Request a RouterInfo attribute */
        status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
                             IB_MAD_ATTR_ROUTER_INFO, 0, TRUE, 0,
                             0, CL_DISP_MSGID_NONE, &mad_context);
        if (status != IB_SUCCESS)
                /* continue despite error */
                OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 7701: "
                        "Failure initiating RouterInfo request (%s) "
			"in heavy sweep\n",
                        ib_get_err_str(status));
        OSM_LOG_EXIT(sm->p_log);
}

static ib_api_status_t router_mgr_subnet_add(IN osm_router_t * p_rtr,
			   IN OUT osm_router_mgr_t *p_mgr,
			   IN osm_subnet_prefix_pkey_t *p_prefix_pkey)
{
	ib_api_status_t status = IB_SUCCESS;
	cl_fmap_t *p_prefix_rtr_tbl;
	osm_router_prefix_rtrs_map_t *p_prefix_2_routers;
	osm_subnet_prefix_pkey_t *p_prefix_pkey_loc;
	ib_net64_t rtr_guid;

	OSM_LOG_ENTER(p_mgr->p_log);
	p_prefix_rtr_tbl = &p_mgr->prefix_rtr_table;
	rtr_guid = osm_port_get_guid(p_rtr->p_port);

	p_prefix_2_routers = (osm_router_prefix_rtrs_map_t *)cl_fmap_get(
							     p_prefix_rtr_tbl,
							     p_prefix_pkey);
	if (p_prefix_2_routers !=
	    (osm_router_prefix_rtrs_map_t *)cl_fmap_end(p_prefix_rtr_tbl)) {
		/*
		 * the map holds routers guids as key,
		 * the value is not important
		 * */
		if (NULL == cl_map_insert(&p_prefix_2_routers->rtrs,
					  rtr_guid,
					  p_rtr)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 7702: Failed to add router 0x%016" PRIx64
				" to subnets prefixes table for"
				" subnet prefix 0x%016" PRIx64
				" and pkey 0x%04" PRIx16 ".\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Updating db for router 0x%016" PRIx64
			", subnet prefix 0x%016" PRIx64
			" with pkey 0x%04" PRIx16 " were found\n",
			cl_ntoh64(rtr_guid),
			cl_ntoh64(p_prefix_pkey->subnet_prefix),
			cl_ntoh16(p_prefix_pkey->pkey));
	} else {
		/* need to create one */
		p_prefix_2_routers = (osm_router_prefix_rtrs_map_t *)
					calloc(1,
					sizeof(osm_router_prefix_rtrs_map_t));
		if (!p_prefix_2_routers) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7703: "
                                "Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 " could not allocate "
				"new entry in subnet prefixes table\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			status = IB_INSUFFICIENT_MEMORY;
			goto Exit;
		}
		cl_map_construct(&p_prefix_2_routers->rtrs);
		if (CL_SUCCESS != cl_map_init(&p_prefix_2_routers->rtrs,
			GROWTH_SIZE)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 7704: Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 " failed to init new"
				" routers map\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			osm_router_mgr_free_rtr_map_item(p_prefix_2_routers);
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}
		if (NULL == cl_map_insert(&p_prefix_2_routers->rtrs, rtr_guid,
					  p_rtr)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 7705: Failed to add router 0x%016" PRIx64
				" to subnets prefixes table for"
				" subnet prefix 0x%016" PRIx64
				" and pkey 0x%04" PRIx16 ".\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			osm_router_mgr_free_rtr_map_item(p_prefix_2_routers);
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}

		p_prefix_pkey_loc = (osm_subnet_prefix_pkey_t *)
				      calloc(1,
				             sizeof(osm_subnet_prefix_pkey_t));
		if (!p_prefix_pkey_loc) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 7706: Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 ", could not allocate "
				"new key in subnets prefixes table\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			osm_router_mgr_free_rtr_map_item(p_prefix_2_routers);
			status = IB_INSUFFICIENT_MEMORY;
			goto Exit;
		}
		memcpy(p_prefix_pkey_loc, p_prefix_pkey,
			sizeof(osm_subnet_prefix_pkey_t));
		if (NULL == cl_fmap_insert(p_prefix_rtr_tbl, p_prefix_pkey_loc,
					   &p_prefix_2_routers->map_item)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 7707: Router: 0x%016" PRIx64
				", subnet prefix: 0x%016" PRIx64
				", pkey: 0x%04" PRIx16 ", failed to add "
				"new entry in subnets prefixes table\n",
				cl_ntoh64(rtr_guid),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
			osm_router_mgr_free_rtr_map_item(p_prefix_2_routers);
			free(p_prefix_pkey_loc);
			status = IB_UNKNOWN_ERROR;
			goto Exit;
		}
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Updating db for router 0x%016" PRIx64
			", subnet prefix 0x%016" PRIx64
			" with pkey 0x%04" PRIx16 " weren't found\n",
			cl_ntoh64(rtr_guid),
			cl_ntoh64(p_prefix_pkey->subnet_prefix),
			cl_ntoh16(p_prefix_pkey->pkey));
	}
Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

static boolean_t is_pkey_valid(osm_router_mgr_t *p_mgr, osm_router_t *p_router,
			       ib_net16_t pkey, uint32_t block_num,
			       uint32_t record_num)
{
	const osm_pkey_tbl_t *p_pkey_tbl;
	boolean_t is_valid = TRUE;
	ib_net64_t rtr_guid = osm_port_get_guid(p_router->p_port);
	ib_net16_t *found_pkey;


	p_pkey_tbl = osm_physp_get_pkey_tbl(
			osm_port_get_physp(p_router->p_port));

	/* check that we have full membership */
	if (!(pkey & IB_PKEY_TYPE_MASK)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 770D: router 0x%016" PRIx64
			" router table block %u, entry %u has invalid"
			" pkey 0x%04" PRIx16 " with partial"
			" membership.\n",
			cl_ntoh64(rtr_guid),
			block_num, record_num, cl_ntoh16(pkey));
		is_valid = FALSE;
		goto Exit;
	}
	/* check that the same pkey is configured on port */
	found_pkey = osm_pkey_tbl_get_pkey(p_mgr->p_log, pkey, p_pkey_tbl,
					   p_mgr->p_subn->opt.allow_both_pkeys);
	if (!found_pkey) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 770E: router 0x%016" PRIx64
			", router table block %u, entry %u has invalid"
			" pkey 0x%04" PRIx16 " that is not configured"
			" on the port.\n",
			cl_ntoh64(rtr_guid),
			block_num, record_num, cl_ntoh16(pkey));
		is_valid = FALSE;
		goto Exit;
	}

	/* let's compare the pkeys' membership */
	if (*found_pkey != pkey) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 770F: router 0x%016" PRIx64
			" router table block %u, entry %u has invalid"
			" pkey (0x%04" PRIx16 ") that has different"
			" membership than pkey configured on"
			" the port (0x%04" PRIx16 ").\n",
			cl_ntoh64(rtr_guid), block_num, record_num,
			cl_ntoh16(pkey), cl_ntoh16(*found_pkey));
		is_valid = FALSE;
	}
Exit:
	return is_valid;
}

static boolean_t router_mgr_table_record_validate(osm_router_mgr_t *p_mgr,
				uint32_t block_num, uint32_t record_num,
				rtr_adj_site_local_subn_record_t *p_rt_record,
				osm_router_t *p_router)
{
	uint16_t sm_lid;
	boolean_t is_valid;

	OSM_LOG_ENTER(p_mgr->p_log);
	OSM_ASSERT(p_rt_record);
	OSM_ASSERT(p_router);

	is_valid = is_pkey_valid(p_mgr, p_router, p_rt_record->pkey, block_num,
				 record_num);

	if (!is_valid)
		goto Exit;

	/* SM LID verification */
	sm_lid = cl_ntoh16(p_rt_record->master_sm_lid);
	if (sm_lid == 0 || sm_lid > IB_LID_MCAST_START_HO) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 7710: router 0x%016" PRIx64
			" router table block %u, entry %u has invalid"
			" sm lid 0x%04" PRIx16 ".\n",
			osm_port_get_guid_ho(p_router->p_port),
			block_num, record_num, sm_lid);
		is_valid = FALSE;
		goto Exit;
	}
Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return is_valid;
}

static osm_subnet_prefix_pkey_t *create_entry_key(osm_router_mgr_t *p_mgr, ib_net16_t pkey,
						  ib_net64_t subnet_prefix,
						  ib_net64_t router_guid)
{
	osm_subnet_prefix_pkey_t *p_entry_key;

	p_entry_key = (osm_subnet_prefix_pkey_t *)calloc(1,
				sizeof(osm_subnet_prefix_pkey_t));
	if (!p_entry_key) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 7709: Failed to create key for router 0x%"
			PRIx64 " for subnet 0x%016" PRIx64
			", pkey 0x%04" PRIx16"\n",
			cl_ntoh64(router_guid), cl_ntoh64(subnet_prefix), cl_ntoh16(pkey));
		goto Exit;
	}
	p_entry_key->subnet_prefix = subnet_prefix;
	p_entry_key->pkey = pkey;
Exit:
	return p_entry_key;
}

static osm_router_tbl_data_t *create_router_data(osm_router_mgr_t *p_mgr,
						 osm_subnet_prefix_pkey_t *p_entry_key,
						 boolean_t is_adjacent,
						 ib_net16_t master_sm_lid,
						 ib_net64_t rtr_guid)
{
	osm_router_tbl_data_t *p_rtr_data;

	p_rtr_data = (osm_router_tbl_data_t *)calloc(1,
			sizeof(osm_router_tbl_data_t));
	if (!p_rtr_data) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 770A: Failed to allocate"
			" data for router 0x%" PRIx64 "\n", cl_ntoh64(rtr_guid));
		goto Exit;
	}
	p_rtr_data->subnet_prefix = p_entry_key->subnet_prefix;
	p_rtr_data->pkey = p_entry_key->pkey;
	p_rtr_data->is_adjacent = is_adjacent;
	p_rtr_data->master_sm_lid = master_sm_lid;

Exit:
	return p_rtr_data;
}

static void add_router_to_db(osm_router_mgr_t *p_mgr, osm_router_t *p_router,
			     ib_net16_t pkey, ib_net64_t subnet_prefix,
			     uint32_t block_num, uint32_t record_num,
			     boolean_t adjacent, ib_net16_t master_sm_lid)
{
	osm_subnet_prefix_pkey_t *p_entry_key;
	osm_router_tbl_data_t *p_rtr_data;
	ib_net64_t rtr_guid;
	cl_fmap_t *p_router_prefixes;
	cl_fmap_item_t *p_check_item;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_router_prefixes = &p_router->subnet_prefix_tbl;
	rtr_guid = osm_port_get_guid(p_router->p_port);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Router 0x%016" PRIx64
		": adding subnet 0x%016" PRIx64 " pkey 0x%04" PRIx16
		"\n", cl_ntoh64(rtr_guid), cl_ntoh64(subnet_prefix), cl_ntoh16(pkey));

	p_entry_key = create_entry_key(p_mgr, pkey, subnet_prefix, rtr_guid);

	if (!p_entry_key)
		goto Exit;

	if (adjacent)
		p_rtr_data = create_router_data(p_mgr, p_entry_key, TRUE,
						master_sm_lid, rtr_guid);
	else
		p_rtr_data = create_router_data(p_mgr, p_entry_key, FALSE, 0,
						rtr_guid);

	if (!p_rtr_data) {
		free(p_entry_key);
		goto Exit;
	}

	p_check_item = cl_fmap_insert(p_router_prefixes,
				      p_entry_key,
				      &p_rtr_data->map_item);
	/*
	 * TODO: use <subnet, pkey> in p_rtr_data, as a key for insert,
	 * so won't need to allocate/free separate key.
	 * */
	if (p_check_item != &p_rtr_data->map_item) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 770B: router 0x%016" PRIx64
			", duplicate entry in block %u, "
			"entry %u for subnet 0x%016" PRIx64
			", pkey 0x%04" PRIx16 ".\n",
			cl_ntoh64(rtr_guid), block_num, record_num,
			cl_ntoh64(p_entry_key->subnet_prefix),
			cl_ntoh16(p_entry_key->pkey));
		free(p_entry_key);
		free(p_rtr_data);
		goto Exit;
	}

	/* update subnets to routers db */
	if (router_mgr_subnet_add(p_router, p_mgr, p_entry_key) != IB_SUCCESS) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 770C: router 0x%016" PRIx64
			", failed to update subnets prefixes table "
			"for subnet 0x%016" PRIx64
			", pkey 0x%04" PRIx16 ".\n",
			cl_ntoh64(rtr_guid),
			cl_ntoh64(p_entry_key->subnet_prefix),
			cl_ntoh16(p_entry_key->pkey));
		/* update failed, revert changes and free memory */
		cl_fmap_remove(p_router_prefixes, p_entry_key);
		free(p_entry_key);
		free(p_rtr_data);
	}
Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

static void router_mgr_update_rtr_arrays(osm_router_mgr_t *p_mgr)
{
	cl_fmap_item_t *p_item;
	osm_router_prefix_rtrs_map_t * cur_item;
	cl_map_iterator_t rtr_iter;
	size_t size;
	ib_net64_t *guid_arr;
	int guid_idx;

	/* iterate over prefix + pkey map */
	for (p_item = cl_fmap_head(&p_mgr->prefix_rtr_table);
	     p_item != cl_fmap_end(&p_mgr->prefix_rtr_table);
	     p_item = cl_fmap_next(p_item))
	{
		cur_item = ((osm_router_prefix_rtrs_map_t *) p_item);

		size = cl_map_count((cl_map_t *)&cur_item->rtrs);
		guid_arr = (ib_net64_t *) realloc(cur_item->rtr_guid_arr,
						  size * sizeof(ib_net64_t));
		if (!guid_arr) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 7712: could not allocate memory for router "
				"guids array\n");
			free(cur_item->rtr_guid_arr);
			cur_item->num_valid_routers = 0;
			continue;
		}

		cur_item->rtr_guid_arr = guid_arr;
		guid_idx = 0;
		/* create router guid array - iterate over each routers map item */
		for (rtr_iter = cl_map_head((cl_map_t *)&cur_item->rtrs);
		     rtr_iter != cl_map_end((cl_map_t *)&cur_item->rtrs);
		     rtr_iter = cl_map_next(rtr_iter))
		{
			osm_router_t *p_router = (osm_router_t *)cl_map_obj(rtr_iter);
			OSM_ASSERT(cl_map_key(rtr_iter) == osm_port_get_guid(p_router->p_port));

			if (p_router->p_port->disc_through != OSM_DISC_NORMAL)
				continue;
			cur_item->rtr_guid_arr[guid_idx++] = cl_map_key(rtr_iter);
		}

		cur_item->num_valid_routers = guid_idx;
	}
}

static void router_mgr_update_db(IN OUT osm_router_t * p_rtr, IN osm_router_mgr_t *p_mgr)
{
	uint32_t i, num_entries, idx = 0, block_num = 0;
	ib_rtr_adj_table_block_t *p_adj_block = NULL;
	ib_rtr_next_hop_table_block_t *p_nh_block = NULL;
	ib_net64_t subnet_prefix, rtr_guid;

	OSM_LOG_ENTER(p_mgr->p_log);
	OSM_ASSERT(p_rtr);

	rtr_guid = osm_port_get_guid(p_rtr->p_port);
	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Updating router 0x%016" PRIx64	" in db.\n",
		cl_ntoh64(rtr_guid));

	/* Update adjacent routers */
	num_entries = p_rtr->router_info.adjacent_site_local_subnets_tbl_top;

	/* go over router table blocks and build db */
	for (i = 0; i < num_entries; i++) {
		/* get the next block, initialized already to first */
		if ((i % IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK) == 0) {
			block_num = i / IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK;
			if (block_num >=
			    osm_router_tbl_get_num_blocks(&p_rtr->adj_router_tbl)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 7708: Router 0x%016" PRIx64
					" adjacent block number %u in "
					"subnets prefixes table is not found, "
					"inconsistent to table top value %u.\n",
					cl_ntoh64(rtr_guid), block_num, num_entries);
				break;
			}

			p_adj_block = osm_router_tbl_adj_block_get(&p_rtr->adj_router_tbl,
								   block_num);
		}
		/* skip NULL blocks */
		if (!p_adj_block) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_INFO,
				"Router 0x%016" PRIx64 " encountered NULL "
				"adjacent router table block, block number %u.\n",
				cl_ntoh64(rtr_guid), block_num);
			/* promote i to first element on the next block */
			i += IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK;
			continue;
		}

		idx = i % IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK;

		/* skip adding invalid record to db */
		if (!router_mgr_table_record_validate(p_mgr, block_num, idx,
						 &p_adj_block->record[idx], p_rtr))
			continue;

		/* use site local prefix */
		subnet_prefix = IB_SITE_LOCAL_SUBNET_PREFIX_HO;
		/* complete received 16 bits to site local subnet prefix */
		subnet_prefix |= cl_ntoh16(p_adj_block->record[idx].subnet_prefix);

		add_router_to_db(p_mgr, p_rtr, p_adj_block->record[idx].pkey,
				 cl_hton64(subnet_prefix), block_num, idx, TRUE,
				 p_adj_block->record[idx].master_sm_lid);

	}

	/* Update next hop routers */
	num_entries = cl_ntoh32(p_rtr->router_info.next_hop_table_top);

	/* go over router table blocks and build db */
	for (i = 0; i < num_entries; i++) {
		/* get the next block, initialized already to first */
		if ((i % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK) == 0) {
			block_num = i / IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;
			if (block_num >=
			    osm_router_tbl_get_num_blocks(&p_rtr->next_hop_router_tbl)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 7711: Router 0x%016" PRIx64
					" next hop block number %u in "
					"subnets prefixes table is not found, "
					"inconsistent to table top value %u",
					cl_ntoh64(rtr_guid), block_num, num_entries);
				break;
			}
			p_nh_block = osm_router_tbl_nh_block_get(&p_rtr->next_hop_router_tbl,
								 block_num);
		}
		/* skip NULL blocks */
		if (!p_nh_block) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_INFO,
				"Router 0x%016" PRIx64 " encountered NULL "
				"next hop router table block, block number %u.\n",
				cl_ntoh64(rtr_guid), block_num);
			/* promote i to first element on the next block */
			i += IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;
			continue;
		}

		idx = i % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;

		/* skip adding invalid record to db */
		if (!is_pkey_valid(p_mgr, p_rtr, p_nh_block->record[idx].pkey,
				   block_num, idx))
			continue;

		add_router_to_db(p_mgr, p_rtr, p_nh_block->record[idx].pkey,
				 p_nh_block->record[idx].subnet_prefix, block_num,
				 idx, FALSE, 0);

	}
	OSM_LOG_EXIT(p_mgr->p_log);
}

static void clear_subnet_prefix_tbl(cl_fmap_t *p_prefix_map)
{
	osm_subnet_prefix_pkey_t *p_key;
	cl_fmap_item_t *p_item;

	p_item = cl_fmap_head(p_prefix_map);

	while (p_item != cl_fmap_end(p_prefix_map))
	{
		p_key = (osm_subnet_prefix_pkey_t *)cl_fmap_key(p_item);
		cl_fmap_remove_item(p_prefix_map, p_item);
		free(p_item);
		free(p_key);
		p_item = cl_fmap_head(p_prefix_map);
	}

}

static void router_mgr_destroy_db(IN osm_router_mgr_t *p_mgr)
{
	cl_map_item_t *p_map_item;
	osm_router_t *p_router;

	router_mgr_destroy_internal(p_mgr);
	/*
	 * remove the data base content
         * without removing routing table blocks
         */
	for (p_map_item = cl_qmap_head(&p_mgr->p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&p_mgr->p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_router = (osm_router_t *) p_map_item;
		clear_subnet_prefix_tbl(&p_router->subnet_prefix_tbl);
	}
}

void osm_router_mgr_set_rtr_lid_tables(IN osm_router_mgr_t *p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_sm_t *sm = &p_subn->p_osm->sm;
	cl_map_item_t *p_map_item;
	osm_router_t *p_router;
	uint32_t start_block, end_block, block_idx;
	uint32_t global_offset, local_offset;
	ib_rtr_router_lid_tbl_block_t *p_subn_block, *p_calc_block;
	osm_madw_context_t context;
	osm_physp_t       *p_physp;
	ib_api_status_t status;

	OSM_LOG_ENTER(p_mgr->p_log);

	for (p_map_item = cl_qmap_head(&p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {

		p_router = (osm_router_t *) p_map_item;
		/*
		 * Skip the router that doesn't support FLID or
		 * wasn't successfully configured with FLIDs ranges in router info set.
		 */
		if (!ib_general_info_is_cap_enabled(
					&p_router->p_port->p_node->general_info,
					IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT) ||
		    !osm_router_is_router_flid_enabled(p_subn, p_router))
			continue;

		/* local lids blocks */
		start_block = RTR_TBL_ENTRY_IDX_TO_BLOCK(p_router->local_flid_start,
							 IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
		end_block = RTR_TBL_ENTRY_IDX_TO_BLOCK(p_router->local_flid_end,
						       IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
		for (block_idx = start_block; block_idx <= end_block; block_idx++) {
			global_offset = osm_router_lid_tbl_block_to_global_offset(p_router, block_idx);
			local_offset = osm_router_lid_tbl_block_to_local_offset(p_router, block_idx);
			/* look at the local flids that were read from the subnet */
			p_subn_block = p_router->router_lid_tbl_locals + global_offset;
			p_calc_block = p_router->new_router_lid_tbl + local_offset;

			p_physp = p_router->p_port->p_physp;

			/*
			 * Compare only relevant bits in the block.
			 * In case foreign FLIDs share block with local FLIDs,
			 * it is important to compare only the local FLIDs setting,
			 * and ignore the foreign FLIDs change.
			 */
			if (!p_subn->need_update && !p_router->p_port->is_new &&
			    !p_physp->need_update &&
			    !memcmp(p_subn_block, p_calc_block,
				    sizeof(ib_rtr_router_lid_tbl_block_t)))
				continue;

			memset(&context, 0, sizeof(context));
			context.ri_context.port_guid = p_physp->port_guid;
			context.ri_context.set_method = TRUE;

			/*
			 * Zero current block in order to ensure we resend it in case of failure.
			 */
			memset(p_subn_block, 0, sizeof(ib_rtr_router_lid_tbl_block_t));

			status = osm_req_set(sm, osm_physp_get_dr_path_ptr(sm, p_physp),
					p_physp, (uint8_t *)p_calc_block,
					sizeof(ib_rtr_router_lid_tbl_block_t),
					IB_MAD_ATTR_RTR_ROUTER_LID_TBL, cl_hton32(block_idx),
					FALSE, ib_port_info_get_m_key(&p_physp->port_info),
					0, CL_DISP_MSGID_NONE, &context);

			if (status != IB_SUCCESS)
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7713: "
					"Setting router LID table configuration (block %u) "
					"on router 0x%" PRIx64 " failed (%s)\n",
					block_idx,
					osm_node_get_node_guid_ho(p_router->p_port->p_node),
					ib_get_err_str(status));
		}
	}
	OSM_LOG_EXIT(p_mgr->p_log);
}

static int add_to_flid_enable_routers(osm_router_mgr_t *p_mgr,
				      uint16_t flid,
				      osm_router_t *p_router)
{
	cl_fmap_item_t *p_item;
	osm_flid_rtrs_list_t *p_rtrs_list;
	cl_fmap_t *flid_to_rtrs = &p_mgr->flid_to_rtrs;

	p_item = cl_fmap_get(flid_to_rtrs, &flid);

	if (p_item == cl_fmap_end(flid_to_rtrs)) {
		p_rtrs_list = (osm_flid_rtrs_list_t *)calloc(1, sizeof(osm_flid_rtrs_list_t));

		if (!p_rtrs_list) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR AA15: Not enough memory to add router GUID 0x%" PRIx64
				" to flid %u list\n",
				osm_node_get_node_guid_ho(p_router->p_port->p_node),
				flid);
			return -1;
		}
		if (cl_list_init(&p_rtrs_list->rtrs_list, 1)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR AA16: Failed to initialize list\n");
			free(p_rtrs_list);
			return -1;
		}
		p_rtrs_list->flid = flid;
		cl_fmap_insert(flid_to_rtrs, &p_rtrs_list->flid, &p_rtrs_list->map_item);
	} else {
		p_rtrs_list = (osm_flid_rtrs_list_t *)p_item;
	}

	/* check that the router is not in the list already */
	if (osm_router_is_router_in_the_list(p_router, &p_rtrs_list->rtrs_list))
		return 0;

	if (cl_list_insert_tail(&p_rtrs_list->rtrs_list, p_router)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR AA17: Failed to add router 0x%" PRIx64 " to "
			"routers list for flid %u.\n",
			osm_node_get_node_guid_ho(p_router->p_port->p_node),
			flid);
		return -1;
	}
	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Added router 0x%" PRIx64 " to the list of routers that enable FLID %u "
		"(%lu in list).\n", osm_node_get_node_guid_ho(p_router->p_port->p_node),
		flid, cl_list_count(&p_rtrs_list->rtrs_list));

	return 0;
}

static void router_mgr_mark_foreign_flids(IN OUT osm_router_mgr_t *p_mgr)
{
	osm_router_t  *p_router;
	osm_subn_t *p_subn = p_mgr->p_subn;
	cl_map_item_t *p_map_item;
	uint32_t lids;
	uint32_t num_blocks = 0, block_idx;
	uint16_t block_start_lid, block_end_lid, flid;
	int start_lid, end_lid, flid_index;
	uint16_t max_flid = 0;
	uint8_t record_in_block;
	unsigned int reset_bytes_num;
	int foreign_flid_count = osm_router_get_foreign_flid_count(p_subn);
	int total_flid_capable_routers = 0;

	OSM_LOG_ENTER(p_mgr->p_log);
	/* find all the foreign FLIDs */
	for (p_map_item = cl_qmap_head(&p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {

		p_router = (osm_router_t *) p_map_item;
		/*
		 * Skip the router that doesn't support FLID or
		 * wasn't successfully configured with FLIDs ranges in router info set.
		 */
		if (!ib_general_info_is_cap_enabled(
					&p_router->p_port->p_node->general_info,
					IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT))
			continue;

		/* This router is FLID capability enabled. Increase the counter. */
		total_flid_capable_routers++;

		if (!osm_router_is_router_flid_enabled(p_subn, p_router))
			continue;

		/*
		 * for every foreign FLID that is enabled on this router
		 * mark in FLID -> routers map
		 */
		num_blocks = RTR_TBL_NUM_BLOCKS_RANGE(p_router->global_flid_start,
						      p_router->global_flid_end,
						      IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);

		/*
		 * Allocate memory for this router's enabled_flids array or set to zero
		 * if already allocated.
		 * Allocate array with one extra so that it can end with 0.
		 */
		if (!p_router->enabled_flids) {
			p_router->enabled_flids = (uint16_t *)calloc(foreign_flid_count + 1, sizeof(uint16_t));
			if (!p_router->enabled_flids) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AA19: "
						"Failed to allocate memory for "
						"enabled_flids on router\n");
				continue;
				}
		}
		else {
			memset(p_router->enabled_flids, 0, foreign_flid_count*sizeof(uint16_t));
		}

		for (block_idx = 0, flid_index = 0; block_idx < num_blocks; block_idx++) {
			/* find the range of this block */
			block_start_lid = osm_router_lid_tbl_offset_to_block(p_router, block_idx) *
					IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM;
			block_end_lid = block_start_lid + IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM - 1;

			/*
			 * Copy the block that was read from subnet
			 * for later comparison of only local parts of router lid table before
			 * deciding if to send it.
			 */
			memcpy(p_router->router_lid_tbl_locals + block_idx,
			       p_router->router_lid_tbl + block_idx,
			       sizeof(ib_rtr_router_lid_tbl_block_t));

			/* skip the block that contains only our local lids */
			if (p_router->local_flid_start <= block_start_lid &&
			    p_router->local_flid_end >= block_end_lid) {
				continue;
			}

			start_lid = end_lid = -1;
			/* foreign lids to look at inside this block */

			/* current block has only foreign FLIDs */
			if (p_router->local_flid_end < block_start_lid ||
			    p_router->local_flid_start > block_end_lid) {

				start_lid = block_start_lid;
				end_lid = block_end_lid;
			}
			/* the local range starts before block start and ends inside the block */
			else if (p_router->local_flid_start <= block_start_lid) {

				start_lid = p_router->local_flid_end + 1;
				end_lid = block_end_lid;
			}
			/*
			 * the local range starts inside block and ends after its end,
			 * or the local range is fully inside this block.
			 */
			else if (p_router->local_flid_start > block_start_lid) {

				start_lid = block_start_lid;
				end_lid = p_router->local_flid_start - 1;
			}

			/* make sure that we inspect only flids inside global flid range */
			if (start_lid < p_router->global_flid_start) {
				/*
				 * The global range starts in the middle of the block.
				 * The global range start is 256 aligned. So it's exactly the middle of the block.
				 * Erase any bits set before the start of global range in the block,
				 * that was acquired from the subnet.
				 */
				start_lid = p_router->global_flid_start;
				record_in_block = RTR_RTR_LID_TBL_BLOCK_RECORD_IDX((start_lid));
				reset_bytes_num = sizeof(uint32_t) * record_in_block;
				memset(((uint8_t *)(p_router->router_lid_tbl_locals + block_idx)),
				       0,
				       reset_bytes_num);
			}
			if (end_lid > p_router->global_flid_end) {
				/*
				 * The global range ends in the middle of the block.
				 * The global range end is 256 aligned. So it's exactly the middle of the block.
				 * Erase any bits set after the end of global range in the block,
				 * that was acquired from the subnet.
				 */
				end_lid = p_router->global_flid_end;
				record_in_block = RTR_RTR_LID_TBL_BLOCK_RECORD_IDX((end_lid + 1));
				reset_bytes_num = sizeof(uint32_t) * record_in_block;
				memset(((uint8_t *)(p_router->router_lid_tbl_locals + block_idx)) +
					reset_bytes_num,
				       0,
				       (sizeof(ib_rtr_router_lid_tbl_block_t) - reset_bytes_num));
			}

			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "On router 0x%" PRIx64
				" block_start_lid %u block_end_lid %u start_lid %u end_lid %u\n",
				osm_node_get_node_guid_ho(p_router->p_port->p_node),
				block_start_lid, block_end_lid, start_lid, end_lid);

			OSM_ASSERT(start_lid != -1 && end_lid != -1);

			while ((unsigned)end_lid <= block_end_lid) {
				for (flid = (unsigned)start_lid; flid <= (unsigned)end_lid; flid++) {
					/* get if this flid is enabled */
					lids = osm_router_rtr_lid_tbl_get_set(p_router, flid,
							OSM_RTR_LID_TBL_SUBNET,
							OSM_RTR_LID_TBL_GET);
					/* this flid is enabled, mark it in db */
					if (lids) {
						add_to_flid_enable_routers(p_mgr, flid, p_router);
						p_router->enabled_flids[flid_index++] = flid;
						if (max_flid < flid)
							max_flid = flid;

						OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
							"On router 0x%" PRIx64
							" foreign FLID %u is enabled. "
							"max flid in the fabric %u\n",
							osm_node_get_node_guid_ho(p_router->p_port->p_node),
							flid, max_flid);
						/*
						 * Clear global enabled flid from
						 * subnet with locals only enabled block.
						 */
						osm_router_rtr_lid_tbl_get_set(p_router, flid,
							OSM_RTR_LID_TBL_SUBNET_LOCALS_ONLY,
							OSM_RTR_LID_TBL_SET_DISABLE);
					}
				}
				/*
				 * The whole local range is inside block,
				 * need to check the second global part of the block.
				 */
				if (p_router->local_flid_start > block_start_lid &&
				    p_router->local_flid_end < block_end_lid &&
				    /* first pass */
				    end_lid == (p_router->local_flid_start - 1)) {
					start_lid = p_router->local_flid_end + 1;
					/* make sure that we inspect only flids inside global flid range */
					end_lid = block_end_lid > p_router->global_flid_end ?
						p_router->global_flid_end : block_end_lid;
					OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "On router 0x%" PRIx64
						" second part: block_start_lid %u block_end_lid %u "
						"start_lid %u end_lid %u\n",
						osm_node_get_node_guid_ho(p_router->p_port->p_node),
						block_start_lid, block_end_lid, start_lid, end_lid);
				} else {
					/* finish the check of the block */
					end_lid = block_end_lid + 1;
				}
			}
		}
	}

	OSM_LOG(p_mgr->p_log, OSM_LOG_INFO,
		"There are a total of %d routers that support FLID capability."
		"FLID ranges: global %u-%u local %u-%u compression ratio %u\n",
		total_flid_capable_routers,
		p_subn->opt.global_flid_start, p_subn->opt.global_flid_end,
		p_subn->opt.local_flid_start, p_subn->opt.local_flid_end,
		p_mgr->flid_compression);

	OSM_LOG_EXIT(p_mgr->p_log);
}

boolean_t osm_router_mgr_is_foreign_flid_enabled(IN osm_router_mgr_t *p_mgr, uint16_t flid)
{
	cl_fmap_t *flid_to_rtrs = &p_mgr->flid_to_rtrs;

	/* return TRUE if the fiven flid is currently enabled on any router */
	return (cl_fmap_get(flid_to_rtrs, &flid) != cl_fmap_end(flid_to_rtrs));
}

/*
 * This function is called due to some change in the enabled FLIDs on another subnet
 * causing a change in some router lid tables.
 */
void osm_router_mgr_process_changes(IN osm_router_mgr_t * p_mgr)
{
	osm_sm_t *sm = &p_mgr->p_subn->p_osm->sm;
	struct osm_routing_engine *p_routing_engine_used = p_mgr->p_subn->routing_engine_used;
	cl_map_item_t *p_item;
	osm_router_t *p_router;
	osm_madw_context_t mad_context;
	osm_physp_t *p_physp;
        ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "Begin FLID sweep.\n");

	/* No calculate_flids function in routing engine */
	if (p_routing_engine_used->calculate_flids == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 7715: No calculate_flids function in routing engine. "
			"Exiting FLID sweep.\n");
		goto Exit;
	}

	/* FLIDs are not enabled */
	if (!osm_subn_is_flids_enabled(p_mgr->p_subn)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 7716: FLIDs are not enabled. Exiting FLID sweep.\n");
		goto Exit;
	}

	/* No routers */
	if (cl_is_qmap_empty(&p_mgr->p_subn->rtr_guid_tbl)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Router guid list is empty. Exiting FLID sweep.\n");
		goto Exit;
	}

	p_mgr->ri_routers_changes_bitmask = 0;

	/* Send Gets */
	for (p_item = cl_qmap_head(&p_mgr->p_subn->rtr_guid_tbl);
	     p_item != cl_qmap_end(&p_mgr->p_subn->rtr_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {

		p_router = (osm_router_t *) p_item;
		p_physp = p_router->p_port->p_physp;

		memset(&mad_context, 0, sizeof(mad_context));
		mad_context.ri_context.port_guid = osm_physp_get_port_guid(p_physp);
		mad_context.ri_context.set_method = FALSE;
		mad_context.ri_context.sweep_type = OSM_SWEEP_CONTEXT_FLID;

		/* Request a RouterInfo attribute */
		status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
				     IB_MAD_ATTR_ROUTER_INFO, 0, TRUE, 0,
				     0, CL_DISP_MSGID_NONE, &mad_context);
		if (status != IB_SUCCESS)
			/* continue despite error */
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7714: "
				"Failure initiating RouterInfo request (%s) "
				"in FLID sweep\n",
				ib_get_err_str(status));
	}

	/* Wait */
	if (wait_for_pending_transactions(p_mgr->p_subn->p_osm)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Failed on waiting for transactions. Stopping FLID sweep.\n");
		goto Exit;
	}

	/*
	 * No routers reported a change. FLID sweep complete.
	 */
	if (!p_mgr->ri_routers_changes_bitmask) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"No routers reported any change. Stopping FLID sweep.\n");
		goto Exit;
	}

	/*
	 * Not only FLIDs have changed. Heavy sweep signaled in trap_rcv
	 */
	if (p_mgr->ri_routers_changes_bitmask != IB_ROUTER_RLTC){
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"Not only FLIDs have been changed. "
			"Stopping FLID sweep and executing Heavy Sweep.\n");
		goto Exit;
	}

	/*
	 * The exclusive lock for writing should be held
	 * to prevent inconsistency in reply to Path Record queries,
	 * as we change the osm subnet db related to routers.
	 */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	/* clean previous data base */
	router_mgr_destroy_on_flid_sweep(p_mgr);

	CL_PLOCK_RELEASE(sm->p_lock);

	CL_PLOCK_EXCL_ACQUIRE(&p_mgr->p_subn->p_osm->routing_lock);
	/*
	 * Calculate FLIDs and set forwarding tables on the switches.
	 */
	router_mgr_mark_foreign_flids(p_mgr);

	p_routing_engine_used->calculate_flids(p_routing_engine_used->context);

	CL_PLOCK_RELEASE(&p_mgr->p_subn->p_osm->routing_lock);

	osm_router_mgr_set_rtr_lid_tables(&sm->router_mgr);

	osm_ucast_mgr_set_fwd_tables(&sm->ucast_mgr, OSM_NODE_FILTER_ALL);

	osm_db_store(sm->ucast_mgr.p_flid2groupid, sm->p_subn->opt.fsync_high_avail_files);

	osm_dump_subn_all(sm->p_subn->p_osm, sm->p_subn, "", OSM_DUMP_FLAG_ROUTING);

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE, "FLID sweep completed.\n");

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

void osm_router_mgr_process(IN osm_router_mgr_t *p_mgr)
{
	osm_sm_t *sm = &p_mgr->p_subn->p_osm->sm;
	cl_map_item_t *p_map_item;
	osm_router_t  *p_router;

	OSM_LOG_ENTER(p_mgr->p_log);

	/*
	 * Clean previous data base, that might hold old data from previous sweeps
	 * when there were routers.
	 */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);
	router_mgr_destroy_db(p_mgr);
	CL_PLOCK_RELEASE(sm->p_lock);

	/* nothing to do */
	if (cl_is_qmap_empty(&p_mgr->p_subn->rtr_guid_tbl)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Router guid list is empty\n");
		return;
	}

	cl_qmap_apply_func(&p_mgr->p_subn->rtr_guid_tbl, router_mgr_router_info_get, sm);

	/* barrier */
	if (wait_for_pending_transactions(p_mgr->p_subn->p_osm)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Failed on waiting for transactions.\n");
	}

	/*
	 * The exclusive lock for writing should be held
	 * to prevent inconsistency in reply to Path Record queries,
	 * as we change the osm subnet db related to routers.
	 */
	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	/* go over routers and update the FIB*/
	for (p_map_item = cl_qmap_head(&p_mgr->p_subn->rtr_guid_tbl);
	     p_map_item != cl_qmap_end(&p_mgr->p_subn->rtr_guid_tbl);
	     p_map_item = cl_qmap_next(p_map_item)) {
		p_router = (osm_router_t *) p_map_item;
		if (!p_router->is_router_info_set_error) {
			router_mgr_update_db(p_router, p_mgr);
		} else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Skip updating db for router 0x%" PRIx64
				" that is in router info set error.\n",
				osm_node_get_node_guid_ho(p_router->p_port->p_node));
		}
	}

	router_mgr_update_rtr_arrays(p_mgr);
	/* TODO: check if to hold the lock also for flids update, in case it is used in PRs too */
	CL_PLOCK_RELEASE(sm->p_lock);

	if (osm_subn_is_flids_enabled(p_mgr->p_subn))
		router_mgr_mark_foreign_flids(p_mgr);

	OSM_LOG_EXIT(p_mgr->p_log);
}

osm_router_prefix_rtrs_map_t *
osm_get_routers_by_subnet_prefix(IN osm_router_mgr_t * p_mgr,
				 IN osm_subnet_prefix_pkey_t * p_subnet_pkey)
{

	cl_fmap_item_t *p_item = cl_fmap_get(&p_mgr->prefix_rtr_table,
					     p_subnet_pkey);
	if (p_item == cl_fmap_end(&p_mgr->prefix_rtr_table))
		return NULL;

	return (osm_router_prefix_rtrs_map_t *)p_item;
}

void osm_router_mgr_free_rtr_map_item(IN osm_router_prefix_rtrs_map_t * p_item)
{
	if (p_item && p_item->rtr_guid_arr)
		free(p_item->rtr_guid_arr);

	free(p_item);
}

void osm_router_mgr_prefix_router_tbl_remove_rtr(IN osm_router_mgr_t * p_mgr,
						 IN struct osm_router *p_rtr)
{
	cl_fmap_item_t *p_item, *p_rtr_mgr_item;
	osm_subnet_prefix_pkey_t *p_prefix_pkey;
	osm_router_prefix_rtrs_map_t *routers;

	/* remove this router from all the subnets it appears in */
	for (p_item = cl_fmap_head(&p_rtr->subnet_prefix_tbl);
	     p_item != cl_fmap_end(&p_rtr->subnet_prefix_tbl);
	     p_item = cl_fmap_next(p_item)) {
		p_prefix_pkey = (osm_subnet_prefix_pkey_t *)cl_fmap_key(p_item);
		routers =
		    osm_get_routers_by_subnet_prefix(p_mgr, p_prefix_pkey);
		if (routers) {
			cl_map_remove(&routers->rtrs,
				      osm_port_get_guid(p_rtr->p_port));
			/*
			 * remove the whole entry if it was the last router
			 * for this subnet and pkey
			 */
			if (cl_map_count(&routers->rtrs) == 0) {
				cl_map_destroy(&routers->rtrs);
				p_rtr_mgr_item = cl_fmap_remove(
						&p_mgr->prefix_rtr_table,
						p_prefix_pkey);
				free((osm_subnet_prefix_pkey_t *)cl_fmap_key(
								p_rtr_mgr_item));

				osm_router_mgr_free_rtr_map_item((osm_router_prefix_rtrs_map_t *)p_rtr_mgr_item);
			}
		} else {
			/* it means that the data base are not synchronized */
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR 7532: router 0x%016" PRIx64
				", subnets prefixes table "
				"doesn't have entry for subnet 0x%016" PRIx64
				", pkey 0x%04" PRIx16 ".\n",
				osm_port_get_guid_ho(p_rtr->p_port),
				cl_ntoh64(p_prefix_pkey->subnet_prefix),
				cl_ntoh16(p_prefix_pkey->pkey));
		}
	}

	router_mgr_update_rtr_arrays(p_mgr);
}

static
cl_status_t neighborhoods_find_flid(IN const size_t index, IN const void *const element, IN void *context)
{
	osm_neighborhood_flid_t *p_flid_element = (osm_neighborhood_flid_t *)element;
	uint16_t *p_flid = (uint16_t *)context;

	if (p_flid_element && p_flid && p_flid_element->flid == *p_flid)
		return CL_SUCCESS;
	return CL_NOT_FOUND;
}

static
cl_status_t flids_find_neighborhood(IN const size_t index, IN const void *const element, IN void *context)
{
	osm_flid_neighborhood_t *p_neighborhood = (osm_flid_neighborhood_t *)element;
	uint16_t *p_neighborhood_id = (uint16_t *)context;

	if (p_neighborhood && p_neighborhood_id && p_neighborhood->neighborhood_id == *p_neighborhood_id)
		return CL_SUCCESS;
	return CL_NOT_FOUND;
}

static boolean_t router_mgr_is_switch_processed(IN osm_router_mgr_t * p_mgr, IN ib_net64_t sw_guid)
{
	uint32_t switches_with_flids_size = cl_u64_vector_get_size(&p_mgr->switches_with_flids);

	if (switches_with_flids_size &&
	    cl_u64_vector_lookup(&p_mgr->switches_with_flids, sw_guid) != switches_with_flids_size)
		return TRUE;

	return FALSE;
}

static void router_mgr_mark_switch_processed(IN osm_router_mgr_t * p_mgr, IN ib_net64_t sw_guid)
{
	cl_u64_vector_insert(&p_mgr->switches_with_flids, sw_guid, NULL);
	cl_u64_vector_sort(&p_mgr->switches_with_flids);
}

void osm_router_mgr_update_local_flid_use_in_routers(IN osm_router_mgr_t * p_mgr,
						     IN osm_port_t *p_sw_port,
						     IN osm_sw_rtrs_list_t *p_rtr_list)
{
	cl_fmap_t *flid_to_rtrs_map = &p_mgr->local_flid_to_routers;
	osm_flid_rtrs_t *p_flid_to_rtrs;
	osm_router_t *p_router;
	int num_routers = 0, i;
	cl_list_iterator_t list_iter;
	ib_net64_t router_port_guid;
	uint64_t sw_guid = osm_port_get_guid_ho(p_sw_port);
	uint16_t flid = cl_ntoh16(p_sw_port->flid);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Update sw 0x%" PRIx64 " routers with FLID %u p_rtrs_list %p size %lu\n",
		sw_guid, flid, p_rtr_list,
		p_rtr_list ? cl_list_count(&p_rtr_list->rtrs_list) : 0);

	/* switch was not assigned flid, nothing to update */
	if (!flid)
		return;

	p_flid_to_rtrs = (osm_flid_rtrs_t *)cl_fmap_get(flid_to_rtrs_map, &flid);

	/* calculate intersection of the existing list and an input list */
	if (p_flid_to_rtrs != (osm_flid_rtrs_t *)cl_fmap_end(flid_to_rtrs_map)) {
		/* given switch has no reachable routers for its FLID */
		if (!p_rtr_list) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"SW 0x%" PRIx64 " with FLID %u has no reachable routers. "
				"Empty the routers list for this FLID.\n",
				sw_guid, flid);
			/* the common list of routers for this FLID should be empty */
			cl_ptr_vector_remove_all(&p_flid_to_rtrs->routers);
			return;
		}

		/* check if router in existing list is not present in given list */
		num_routers = cl_ptr_vector_get_size(&p_flid_to_rtrs->routers);
		for (i = num_routers - 1; i >= 0; i--) {
			p_router = cl_ptr_vector_get(&p_flid_to_rtrs->routers, i);
			if (!p_router)
				continue;
			/*
			 * Check if the router in the existing FLID's list
			 * is in the switch's list.
			 */
			if (!osm_router_is_router_in_the_list(p_router, &p_rtr_list->rtrs_list)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"Router 0x%" PRIx64 " is not present in the given list "
					"of sw 0x%" PRIx64 " for flid %u. "
					"Remove it at index %u.\n",
					osm_port_get_guid_ho(p_router->p_port),
					sw_guid, flid, i);
				cl_ptr_vector_remove(&p_flid_to_rtrs->routers, i);
			}
		}
	} else {
		p_flid_to_rtrs = (osm_flid_rtrs_t *)calloc(1, sizeof(osm_flid_rtrs_t));

		if (!p_flid_to_rtrs) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7718: "
				"could not allocate memory for local FLID (%u) to routers db.\n",
				flid);
			return;
		}
		p_flid_to_rtrs->flid = flid;
		cl_ptr_vector_init(&p_flid_to_rtrs->routers, 0, GROWTH_SIZE);
		cl_fmap_insert(flid_to_rtrs_map, &p_flid_to_rtrs->flid, &p_flid_to_rtrs->map_item);

		/* given switch has no reachable routers for its FLID, the list would be empty */
		if (!p_rtr_list)
			return;
		/*
		 * The first switch with FLID 'flid' sets the list of routers,
		 * on which 'flid' should be set, to its own list of reachable routers.
		 * Other switches assigned with this 'flid' can only reduce this list,
		 * if one of the routers in the initial list is not in their list.
		 */
		for (list_iter = cl_list_head(&p_rtr_list->rtrs_list);
				list_iter != cl_list_end(&p_rtr_list->rtrs_list);
				list_iter = cl_list_next(list_iter)) {
			p_router = (osm_router_t *)cl_list_obj(list_iter);
			if (p_router &&
			    cl_ptr_vector_insert(
				    &p_flid_to_rtrs->routers, p_router, NULL) != CL_SUCCESS) {
				router_port_guid = osm_port_get_guid(p_router->p_port);

				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 771F: "
					"Failed to add router 0x%" PRIx64 " for FLID %u "
					"to local FLID to routers db.\n",
					cl_ntoh64(router_port_guid), flid);
			}
		}
	}
}

static void
router_mgr_update_flid_use(IN osm_router_mgr_t * p_mgr, IN ib_net64_t sw_guid, IN uint16_t n_id,
			   IN uint16_t flid, IN osm_neighborhood_flid_t * p_exist_flid)
{
	osm_neighborhood_flid_t *p_flid;
	osm_neighborhood_flids_data_t *p_data;
	osm_flid_neighborhoods_t *p_flid_data;
	uint32_t found_flid_index;
	uint32_t found_neighborhood_index;
	osm_flid_neighborhood_t *p_neighborhood;
	osm_lid_mgr_t *p_lid_mgr;

	if (!flid && !p_exist_flid)
		return;

	p_lid_mgr = &p_mgr->p_subn->p_osm->sm.lid_mgr;

	if (cl_ptr_vector_at(&p_mgr->neighborhood_to_flids, n_id, (void *)&p_data) != CL_SUCCESS ||
	    !p_data) {
		/* need to add one */
		p_data = (osm_neighborhood_flids_data_t *)calloc(1, sizeof(osm_neighborhood_flids_data_t));
		if (!p_data) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7718: "
				"Failed to allocate memory for neighborhood (%u) FLIDs data.\n",
				n_id);
			return;
		}
		p_data->id = n_id;
		cl_ptr_vector_init(&p_data->flids_to_count, 0, GROWTH_SIZE);
		/* set at index n_id */
		cl_ptr_vector_set(&p_mgr->neighborhood_to_flids, n_id, p_data);
	}

	/* flid to search was given */
	if (flid) {
		found_flid_index = cl_ptr_vector_find_from_start(&p_data->flids_to_count,
				neighborhoods_find_flid, (void *)&flid);
		/* search for his flid, if not found allocate new one and insert */
		if (found_flid_index == cl_ptr_vector_get_size(&p_data->flids_to_count)) {
			p_flid = (osm_neighborhood_flid_t *)calloc(1, sizeof(osm_neighborhood_flid_t));
			if (!p_flid) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7720: "
					"Failed to allocate memory for neighborhood (%u) FLID.\n",
					n_id);
				return;
			}
			p_flid->flid = flid;
			p_flid->count = 0;
			if (cl_ptr_vector_insert(&p_data->flids_to_count, p_flid, NULL) != CL_SUCCESS) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 771A: "
					"Failed to add FLID %u to FLID to count data.\n",
					p_flid->flid);
			}
		} else {
			p_flid = (osm_neighborhood_flid_t *)cl_ptr_vector_get(&p_data->flids_to_count,
					found_flid_index);
			if (!p_flid) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 771B: "
					"Failed to get number of times FLID %u "
					"was assigned in neighborhood %u.\n", flid, n_id);
				return;
			}
		}
	}
	/* the object in the list was given */
	else {
		p_flid = p_exist_flid;
	}
	/* update number of times flid appeared in neighborhood with n_id in neighborhood to flids */
	p_flid->count ++;

	/* update flid in flid_to_neighborhoods */
	if (cl_ptr_vector_at(&p_mgr->flid_to_neighborhoods, p_flid->flid, (void *)&p_flid_data) != CL_SUCCESS ||
	    p_flid_data == NULL) {
		p_flid_data = (osm_flid_neighborhoods_t *)calloc(1, sizeof(osm_flid_neighborhoods_t));
		if (!p_flid_data) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 771C: "
				"Failed to allocate memory for FLID (%u) to neighborhoods data.\n",
				p_flid->flid);
			return;
		}
		p_flid_data->flid = p_flid->flid;
		cl_ptr_vector_init(&p_flid_data->neighborhood_to_count, 0, GROWTH_SIZE);
		cl_ptr_vector_set(&p_mgr->flid_to_neighborhoods, p_flid->flid, p_flid_data);
	}
	found_neighborhood_index = cl_ptr_vector_find_from_start(&p_flid_data->neighborhood_to_count,
				flids_find_neighborhood, (void *)&n_id);
	if (found_neighborhood_index == cl_ptr_vector_get_size(&p_flid_data->neighborhood_to_count)) {
		p_neighborhood = (osm_flid_neighborhood_t *)calloc(1, sizeof(osm_flid_neighborhood_t));
		if (!p_neighborhood) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 771D: "
				"Failed to allocate memory for FLID's (%u) neighborhoods data.\n",
				p_flid->flid);
			return;
		}
		p_neighborhood->neighborhood_id = n_id;
		p_neighborhood->count = 0;
		if (cl_ptr_vector_insert(&p_flid_data->neighborhood_to_count, p_neighborhood, NULL) != CL_SUCCESS) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 771E: "
				"Failed to add neighborhood's FLID's (%u) neighborhoods data.\n",
				p_flid->flid);
			return;
		}
	} else {
		p_neighborhood = cl_ptr_vector_get(&p_flid_data->neighborhood_to_count,
						   found_neighborhood_index);
	}
	/* update number of times flid appeared in neighborhood with flid in flid to neighborhoods */
	p_neighborhood->count ++;

	osm_lid_mgr_set_guid_flid(p_lid_mgr, cl_ntoh64(sw_guid), p_flid->flid);

	router_mgr_mark_switch_processed(p_mgr, sw_guid);
}

static boolean_t router_mgr_is_flid_in_another_neighborhood(IN osm_router_mgr_t * p_mgr,
							    IN uint16_t flid,
							    IN uint16_t n_id)
{
	osm_flid_neighborhoods_t *p_flid_data;
	uint32_t flid_num_of_neighborhoods = 0;
	osm_flid_neighborhood_t *p_flid_neighborhood_data;

	if (cl_ptr_vector_at(&p_mgr->flid_to_neighborhoods,
			     flid,
			     (void *)&p_flid_data) != CL_SUCCESS ||
	    !p_flid_data) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 7717: "
				"Failed to find flid %u from neighborhood id %u "
				"in flids to neighborhoods db\n", flid, n_id);
		return TRUE;
	}

	flid_num_of_neighborhoods =
		cl_ptr_vector_get_size(&p_flid_data->neighborhood_to_count);
	/* check that this flid is not in another neighborhood */
	if (flid_num_of_neighborhoods > 1 ||
	    (flid_num_of_neighborhoods == 1 &&
	     cl_ptr_vector_at(&p_flid_data->neighborhood_to_count, 0,
				 (void *)&p_flid_neighborhood_data) == CL_SUCCESS &&
	     p_flid_neighborhood_data &&
	     p_flid_neighborhood_data->neighborhood_id != n_id)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Can't assign flid %u "
			"from neighborhood %u, since it appears in other "
			"neighborhoods\n", p_flid_data->flid, n_id);
		return TRUE;
	}

	return FALSE;
}

static boolean_t
router_mgr_is_persistent_flid_not_in_subnet(IN osm_router_mgr_t * p_mgr, uint16_t flid)
{
	OSM_ASSERT(flid <= p_mgr->p_subn->opt.global_flid_end);
	return p_mgr->flids_not_in_subnet[flid];
}

static osm_neighborhood_flid_t *
router_mgr_get_neighborhood_flid(IN osm_router_mgr_t * p_mgr, IN uint16_t n_id)
{
	osm_neighborhood_flids_data_t *p_data;
	osm_neighborhood_flid_t *p_flid = NULL, *p_min_count_flid = NULL;
	uint16_t flids_num;
	uint16_t i, min_count_flid = 0, min_count = 0;

	if (cl_ptr_vector_at(&p_mgr->neighborhood_to_flids, n_id, (void *)&p_data) != CL_SUCCESS ||
	    !p_data) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"neighborhood id %u is not found in db\n", n_id);
		goto Exit;
	}

	flids_num = (uint16_t)cl_ptr_vector_get_size(&p_data->flids_to_count);
	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"neighborhood id %u has %u flids\n", n_id, flids_num);

	for (i = 0; i < flids_num; i++) {
		p_flid = (osm_neighborhood_flid_t *)cl_ptr_vector_get(&p_data->flids_to_count, i);
		/*
		 * Find the least used FLID in the neighborhood among FLIDs that
		 * are used by less than 'flid_compression' switches.
		 * in addition check that this flid is not in another neighborhood
		 */
		if (!p_flid || p_flid->count >= p_mgr->flid_compression)
			continue;

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "neighborhood id %u flid %u "
			"has %u instances, less than compression %u\n",
			n_id, p_flid->flid, p_flid->count, p_mgr->flid_compression);

		if (router_mgr_is_persistent_flid_not_in_subnet(p_mgr, p_flid->flid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Skip assigning FLID %u "
				"from neighborhood id %u, since FLID is persistent to GUID, "
				"that is currently not in the subnet.\n",
				p_flid->flid, n_id);
			continue;
		}

		/* check that this flid is not in another neighborhood */
		if (router_mgr_is_flid_in_another_neighborhood(p_mgr, p_flid->flid, n_id))
			continue;

		if (!min_count_flid || p_flid->count < min_count) {
			min_count = p_flid->count;
			p_min_count_flid = p_flid;
			min_count_flid = p_flid->flid;
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Updated to use flid %u "
				"(neighborhood %u), that was assigned %u times so far.\n",
				p_flid->flid, n_id, p_flid->count);
		}
	}

Exit:
	return p_min_count_flid;
}

uint16_t osm_router_mgr_make_flid(IN osm_router_mgr_t * p_mgr,
				  IN osm_port_t * p_sw_port,
				  IN boolean_t preprocess)
{
	osm_lid_mgr_t *p_lid_mgr;
	osm_switch_t *p_sw;
	uint16_t flid = 0;
	uint16_t n_id;
	ib_net64_t guid;
	osm_neighborhood_flid_t *p_min_count_flid = NULL;

	OSM_ASSERT(p_sw_port && p_sw_port->p_node && p_sw_port->p_node->sw);

	p_sw = p_sw_port->p_node->sw;
	guid = osm_node_get_node_guid(p_sw->p_node);

	if (p_sw->coord == OSM_SW_NO_COORD) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Switch 0x%" PRIx64 " has no neighborhood id.\n", cl_ntoh64(guid));
		goto Exit;
	}

	n_id = p_sw->coord;
	p_lid_mgr = &p_mgr->p_subn->p_osm->sm.lid_mgr;

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Switch 0x%" PRIx64 " neighborhood id %u wants FLID.\n", cl_ntoh64(guid), n_id);

	/*
	 * This switch has already assigned flid and it's updated in router manager internal db,
	 * just return it. Note that the assigned lid can be 0
	 * (ex: when ran out of available FLIDs)
	 */
	if (router_mgr_is_switch_processed(p_mgr, guid)) {
		flid = cl_ntoh16(p_sw_port->flid);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Switch 0x%" PRIx64 " with neighborhood id %u has FLID already %u.\n",
			cl_ntoh64(guid), n_id, flid);
		return flid;
	}

	flid = osm_lid_mgr_query_switch_flid(p_lid_mgr, p_sw_port);

	if (flid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Switch 0x%" PRIx64 " n_id %u got flid %u according to guid2flid db.\n",
			cl_ntoh64(guid), n_id, flid);
	}

	/*
	 * switch doesn't have flid assigned in guid to flid db,
	 * try to assign flid from already present in the same neighborhood.
	 * During preprocess stage, skip switches that don't have flid assigned in guid 2 flid db
	 */
	if (!flid && !preprocess) {
		/* go over flids and find one that has less than compression appearances */
		p_min_count_flid = router_mgr_get_neighborhood_flid(p_mgr, n_id);
		if (p_min_count_flid) {
			flid = p_min_count_flid->flid;
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Switch 0x%" PRIx64 " n_id %u got flid %u according to neighborhood.\n",
				cl_ntoh64(guid), n_id, flid);
		}
		/* allocate a new flid */
		else {
			/* gets new flid */
			flid = osm_lid_mgr_get_switch_flid(p_lid_mgr, p_sw_port);
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Switch 0x%" PRIx64 " n_id %u got new flid 0x%x from the local range.\n",
				cl_ntoh64(guid), n_id, flid);
		}
	}

	/* update internal router manager data base */
	if (flid) {
		router_mgr_update_flid_use(p_mgr, guid, n_id, p_min_count_flid ? 0 : flid, p_min_count_flid);
		p_sw_port->flid = cl_hton16(flid);
	}

Exit:
	return flid;
}

void osm_router_mgr_preprocess_local_flids(IN osm_router_mgr_t * p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	cl_list_iterator_t leaf_switch_iter;
	osm_switch_t *p_sw;
	osm_port_t *p_sw_port0;
	osm_lid_mgr_t *p_lid_mgr = &p_mgr->p_subn->p_osm->sm.lid_mgr;

	router_mgr_clear_flids_not_in_subnet(p_mgr);
	osm_lid_mgr_get_flids_not_in_subnet(p_lid_mgr);

	foreach_leaf_iter(p_subn, leaf_switch_iter) {
		p_sw = leaf_iter_to_switch(leaf_switch_iter);
		p_sw_port0 = osm_switch_get_port0(p_sw, p_subn);
		osm_router_mgr_make_flid(p_mgr, p_sw_port0, TRUE);
	}
}

void osm_router_mgr_set_local_flid_on_routers(IN osm_router_mgr_t *p_mgr)
{
	cl_fmap_item_t *p_item;
	osm_flid_rtrs_t *flid_to_rtrs;
	uint32_t num_routers, i;
	osm_router_t *p_router;

	for (p_item = cl_fmap_head(&p_mgr->local_flid_to_routers);
	     p_item != cl_fmap_end(&p_mgr->local_flid_to_routers);
	     p_item = cl_fmap_next(p_item)) {
		flid_to_rtrs = (osm_flid_rtrs_t *)p_item;
		num_routers = cl_ptr_vector_get_size(&flid_to_rtrs->routers);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"FLID %u will be set on %u routers\n",
			flid_to_rtrs->flid, num_routers);
		for (i = 0; i < num_routers; i++) {
			p_router = cl_ptr_vector_get(&flid_to_rtrs->routers, i);
			if (p_router) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"Set local FLID %u on router 0x%" PRIx64 "\n",
					flid_to_rtrs->flid, osm_port_get_guid_ho(p_router->p_port));
				osm_router_set_flid(p_mgr->p_subn, p_router, flid_to_rtrs->flid);
			}
		}
	}
}


/*
 * Return maximal ar group to be used for the router
 */
static int router_mgr_pfrn_calc_max_ar_group(osm_router_mgr_t *p_mgr, osm_router_t *p_router)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_ucast_mgr_t	*p_ucast_mgr = &p_sm->ucast_mgr;

	if (p_router->num_ar_group_to_lid_entries < p_ucast_mgr->num_reserved_local_ar_groups)
		return p_router->num_ar_group_to_lid_entries;
	else
		return p_ucast_mgr->num_reserved_local_ar_groups;
}


/*
 * Prepare RouterInfo MAD to configure.
 * Returns 1 if configuration is need, 0 if not.
 */
static int router_mgr_pfrn_calc_router_info(osm_router_mgr_t *p_mgr, osm_router_t *p_router,
					    ib_router_info_t *p_ri,
					    boolean_t pfrn_enabled)

{
	*p_ri = p_router->router_info;
	osm_subn_t *p_subn = p_mgr->p_subn;
	int num_entries;
	int result = 0;

	/* Zero the change bits in order to ensure we do not miss changes */
	p_ri->table_changes_bitmask = 0;

	if (!pfrn_enabled) {
		if (!ib_router_info_get_pfrn_rtr_en(p_ri)) {
			/* pFRN already disabled */
			result = 0;
			goto exit;
		} else {
			/* Need to disable PFRN */
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "router 0x%" PRIx64 " disabling pFRN required\n",
				osm_port_get_guid_ho(p_router->p_port));
			ib_router_info_set_pfrn_rtr_en(p_ri, 0);
			result = 1;
			goto exit;
		}
	}

	/* pFRN feature is enabled */

	if (!ib_router_info_get_pfrn_rtr_en(p_ri)) {
		/* pFRN currently disabled on the device, need to enable it */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "router 0x%" PRIx64 " enabling pFRN required\n",
			osm_port_get_guid_ho(p_router->p_port));
		ib_router_info_set_pfrn_rtr_en(p_ri, 1);
		result = 1;
	} else if (!osm_router_is_router_flid_enabled(p_subn, p_router)) {
		/* pFRN currenlty enabled in the device, but since device is misconfigured, disable it */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "router 0x%" PRIx64 " disabling pFRN required since flid not enabled on it\n",
			osm_port_get_guid_ho(p_router->p_port));
		ib_router_info_set_pfrn_rtr_en(p_ri, 0);
		result = 1;
	}

	num_entries = router_mgr_pfrn_calc_max_ar_group(p_mgr, p_router);
	if (p_router->num_ar_group_to_lid_entries &&
	    (ib_router_info_get_max_ar_group_id(p_ri) != num_entries - 1)) {
		/* Incorrect number of AR Groups */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "router 0x%" PRIx64 " setting max_ar_group_id required\n",
			osm_port_get_guid_ho(p_router->p_port));
		ib_router_info_set_max_ar_group_id(p_ri, num_entries - 1);
		result = 1;
	}

exit:
	return result;
}


static void router_mgr_pfrn_set_router_info(IN osm_router_mgr_t *p_mgr, boolean_t pfrn_enabled)
{
	ib_router_info_t payload;
	osm_madw_context_t context;
	osm_subn_t *p_subn = p_mgr->p_subn;
	cl_qmap_t *p_router_guid_tbl = &p_subn->rtr_guid_tbl;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_router_t *p_router;
        ib_api_status_t status = IB_SUCCESS;

	for (p_router = (osm_router_t *) cl_qmap_head(p_router_guid_tbl);
	     p_router != (osm_router_t *) cl_qmap_end(p_router_guid_tbl);
	     p_router = (osm_router_t *) cl_qmap_next(&p_router->map_item)) {

		osm_port_t *p_port = p_router->p_port;
		osm_physp_t *p_physp = osm_port_get_physp(p_router->p_port);

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"router 0x%" PRIx64 " pfrn support cap %d\n",
			osm_node_get_node_guid_ho(p_port->p_node),
			!!(p_router->router_info.cap_mask & IB_ROUTER_INFO_CAP_IS_PFRN_SUPPORTED));

		/* If router does not support pFRN over router, skip it */
		if (!(p_router->router_info.cap_mask & IB_ROUTER_INFO_CAP_IS_PFRN_SUPPORTED))
			continue;

		if (!router_mgr_pfrn_calc_router_info(p_mgr, p_router, &payload, pfrn_enabled) &&
		    !p_subn->need_update && !p_router->p_port->is_new &&
		    !p_physp->need_update)
			continue;

		context.ri_context.port_guid = p_physp->port_guid;
		context.ri_context.set_method = TRUE;
		context.ri_context.sweep_type = 0;

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "router 0x%" PRIx64 " sending RouterInfo set request:\n",
			osm_port_get_guid_ho(p_router->p_port));

		osm_dump_router_info(p_mgr->p_log,
			     osm_node_get_node_guid(p_port->p_node),
                             p_port->guid, p_port->p_physp->port_num,
                             &payload, FILE_ID, OSM_LOG_DEBUG);

		status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp), p_physp,
				     (uint8_t *)&payload,
				     sizeof(payload), IB_MAD_ATTR_ROUTER_INFO, 0,
				     FALSE, ib_port_info_get_m_key(&p_physp->port_info),
				     0, CL_DISP_MSGID_NONE, &context);
		if (status != IB_SUCCESS) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 7719: "
				"Failed to Send RouterInfo set for 0x%" PRIx64
				" (status = %s)\n",
				cl_ntoh64(p_port->guid), ib_get_err_str(status));
		}
	}
}


/*
 * Calculate AR group to router LID tables
 * Currently support only singe AR group for switch.
 */
static void router_mgr_pfrn_calc_ar_group_to_lid_tables(IN osm_router_mgr_t *p_mgr, boolean_t pfrn_enabled)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_ucast_mgr_t	*p_ucast_mgr = &p_sm->ucast_mgr;
	cl_qmap_t *p_port_guid_tbl = &p_subn->port_guid_tbl;
	osm_port_t *p_port;
	int num_blocks;
	int num_entries;
	int block_idx;
	int block_offset;
	uint16_t group_id;

	/*
	 * This function calculate AR group to router LID tables.
	 * Since the table is identical for all routers, the calculated
	 * tables are stored in the router manager.
	 */

	/* Clear previously configured table */
	free(p_mgr->p_ar_group_to_lid_tbl);
	p_mgr->p_ar_group_to_lid_tbl = NULL;
	p_mgr->num_ar_group_to_lid_tbl_blocks = 0;


	/* Get number of AR groups assigned for local leaf switches */
	num_entries = p_ucast_mgr->num_reserved_local_ar_groups;
	if (!num_entries) {
		/* No reserved AR groups, nothing to do */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "no reserved AR groups for local switches\n");
		return;
	}


	/* Allocate memory */
	num_blocks = RTR_TBL_NUM_BLOCKS(num_entries, IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS);
	p_mgr->p_ar_group_to_lid_tbl = calloc(num_blocks, sizeof(ib_rtr_ar_group_to_lid_tbl_block_t));
	if (!p_mgr->p_ar_group_to_lid_tbl) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
			"ERR 7717: Failed to allocate memory for pFRN over router\n");
		return;
	} else {
		p_mgr->num_ar_group_to_lid_tbl_blocks = num_blocks;
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "number of AR groups to lid blocks %u\n",
			p_mgr->num_ar_group_to_lid_tbl_blocks);
	}


	if (!pfrn_enabled) {
		/* Noting to do */
		return;
	}


	/* Fill in AR group ID to lid table */

	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
			p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
			p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		if (!p_port->flid || !p_port->p_node->sw)
			continue;

		group_id = osm_ucast_mgr_get_groupid(p_ucast_mgr,
				osm_port_get_base_lid_ho(p_port),
				0);
		if (!group_id || group_id > num_entries)
			continue;

		block_idx = group_id / IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS;
		block_offset = group_id % IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS;
		p_mgr->p_ar_group_to_lid_tbl[block_idx].lids[block_offset] = p_port->flid;
	}
}


static void router_mgr_pfrn_allocate_ar_group_to_lid_tables(osm_router_mgr_t *p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	cl_qmap_t *p_router_guid_tbl = &p_subn->rtr_guid_tbl;
	osm_router_t *p_router;
	int num_entries;

	/*
	 * For each router, check if there is need to allocate or flush AR group
	 * to LID tables
	 */
	for (p_router = (osm_router_t *) cl_qmap_head(p_router_guid_tbl);
	     p_router != (osm_router_t *) cl_qmap_end(p_router_guid_tbl);
	     p_router = (osm_router_t *) cl_qmap_next(&p_router->map_item)) {

		num_entries = cl_ntoh16(p_router->router_info.ar_group_to_router_lid_table_cap);
		if (p_router->num_ar_group_to_lid_entries != num_entries) {
			/* Reallocate tables */
			osm_router_ar_group_to_lid_tbl_reinit(p_subn, p_router, num_entries);
		} else if (num_entries && !ib_router_info_get_pfrn_rtr_en(&p_router->router_info)) {
			/* Flush tables */
			osm_router_ar_group_to_lid_tbl_clear(p_router);	
		}

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "router 0x%" PRIx64 " set number of ar group entires to %u\n",
			osm_port_get_guid_ho(p_router->p_port),
			p_router->num_ar_group_to_lid_entries);
	}
}


static void router_mgr_pfrn_set_ar_group_to_lid_tables(osm_router_mgr_t *p_mgr)
{
	const ib_rtr_ar_group_to_lid_tbl_block_t *p_calc_block = NULL;
	ib_rtr_ar_group_to_lid_tbl_block_t *p_subn_block = NULL;
	osm_madw_context_t context;
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	cl_qmap_t *p_router_guid_tbl = &p_subn->rtr_guid_tbl;
	osm_router_t *p_router;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	osm_madw_t *p_madw;
	ib_net64_t m_key;
	size_t block_size;
	int num_blocks;
	int num_entries;
	int force_send;
	int block_idx;
	int count;

	/*
	 * When the feature is not enabled there is no need to send
	 * ARGroupToLID MADs.
	 *
	 * Disabling the feature is done by resting the enablement bit in RouterInfo
	 * which results with the router zeroing the tables automatically.
	 */
	for (p_router = (osm_router_t *) cl_qmap_head(p_router_guid_tbl);
	     p_router != (osm_router_t *) cl_qmap_end(p_router_guid_tbl);
	     p_router = (osm_router_t *) cl_qmap_next(&p_router->map_item)) {

		/* If router does not support pFRN over router, skip it */
		if (!(p_router->router_info.cap_mask & IB_ROUTER_INFO_CAP_IS_PFRN_SUPPORTED))
			continue;

		/* If pFRN is disabled on this device, skip it */
		if (!ib_router_info_get_pfrn_rtr_en(&p_router->router_info))
			continue;

		/* If router is misconfigured, skip it */
		if (!osm_router_is_router_flid_enabled(p_subn, p_router))
			continue;

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "router 0x%" PRIx64 " calculating AR group to LID tables\n",
			osm_port_get_guid_ho(p_router->p_port));

		p_port = p_router->p_port;

		OSM_ASSERT(p_port);
		OSM_ASSERT(!cl_qlist_count(&p_port->mads_queue));

		p_physp = osm_port_get_physp(p_port);
		m_key = ib_port_info_get_m_key(&p_physp->port_info);
		force_send = p_subn->need_update || p_router->p_port->is_new ||
			     p_physp->need_update;

		/*
		 * Number of entries is the minimum of switch capability and number of AR groups
		 * reserved for the local leaf switches.
		 */
		num_entries = router_mgr_pfrn_calc_max_ar_group(p_mgr, p_router);
		num_blocks = RTR_TBL_NUM_BLOCKS(num_entries, IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS);
		context.ri_context.port_guid = osm_port_get_guid(p_port);
		/*
		 * Configure the first bulk according to information stored in router manager.
		 * Zero the rest of the blocks.
		 */
		for (block_idx = 0; block_idx < num_blocks; block_idx++) {
			OSM_ASSERT(block_idx <= p_mgr->num_ar_group_to_lid_tbl_blocks);
			p_subn_block = &p_router->subn_ar_group_to_lid[block_idx];
			p_calc_block = &p_mgr->p_ar_group_to_lid_tbl[block_idx];
			block_size =
			    sizeof(p_calc_block->lids[0]) *
			    (block_idx == num_blocks - 1 ?
			     num_entries % IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS + 1:
			     IB_ROUTER_AR_GROUP_TO_LID_TBL_ELEMENTS);

			if (!force_send && !memcmp(p_subn_block, p_calc_block, block_size))
				continue;

			/*
			 * Zero current block in order to ensure we resend it in case of failure.
			 */
			memset(p_subn_block, 0, sizeof(*p_subn_block));

			p_madw = osm_prepare_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp),
						     (uint8_t *)p_calc_block, block_size,
						     IB_MAD_ATTR_RTR_AR_GROUP_TO_LID_TBL, cl_hton32(block_idx),
						     FALSE, m_key,
						     0, CL_DISP_MSGID_NONE,
						     osm_node_get_node_guid(osm_port_get_node(p_port)),
						     &context);
			if (!p_madw) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 7718: Failed to allocate memory for MADs\n");
				break;

			}
			cl_qlist_insert_tail(&p_port->mads_queue, &p_madw->list_item);
		}
	}

	/*
	 * Start sending the MADs
	 * Send number of MADs according to max_wire_smps_per_device.
	 */
	for (count = 0; count < p_subn->opt.max_wire_smps_per_device; count++) {
		for (p_router = (osm_router_t *) cl_qmap_head(p_router_guid_tbl);
		     p_router != (osm_router_t *) cl_qmap_end(p_router_guid_tbl);
		     p_router = (osm_router_t *) cl_qmap_next(&p_router->map_item)) {
			osm_port_send_next_mad(p_sm, p_router->p_port);
		}
	}
}


void osm_router_mgr_pfrn_process(IN osm_router_mgr_t *p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	boolean_t pfrn_enabled;

	OSM_LOG_ENTER(p_mgr->p_log);
	LOGPERF_START(p_subn->p_osm, "ROUTER_MGR_PFRN_PROCESS", "");

	if (p_subn->opt.pfrn_over_router_enabled == OSM_CAPABILITY_IGNORED)
		goto exit;

	/* pFRN over router should be enabled only when pFRN is enabled */
	pfrn_enabled =
	    osm_subn_is_flids_enabled(p_subn) &&
	    p_subn->opt.pfrn_over_router_enabled == OSM_CAPABILITY_ENABLED &&
	    p_subn->pfrn_supported;

	router_mgr_pfrn_allocate_ar_group_to_lid_tables(p_mgr);

	router_mgr_pfrn_set_router_info(p_mgr, pfrn_enabled);
	if (wait_for_pending_transactions(p_subn->p_osm)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Failed on waiting for transactions.\n");
	}

	router_mgr_pfrn_calc_ar_group_to_lid_tables(p_mgr, pfrn_enabled);

	if (!pfrn_enabled)
		goto exit;

	router_mgr_pfrn_set_ar_group_to_lid_tables(p_mgr);
	if (wait_for_pending_transactions(p_subn->p_osm)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Failed on waiting for transactions.\n");
	}
exit:

	LOGPERF_END(p_subn->p_osm, "ROUTER_MGR_PFRN_PROCESS", "");
	OSM_LOG_EXIT(p_mgr->p_log);
	return;
}

void osm_router_mgr_add_flid_not_in_subnet(IN osm_router_mgr_t * p_mgr, uint16_t flid)
{
	OSM_ASSERT(flid <= p_mgr->p_subn->opt.global_flid_end);
	p_mgr->flids_not_in_subnet[flid] = 1;
}

void osm_router_mgr_remove_flid_not_in_subnet(IN osm_router_mgr_t * p_mgr, uint16_t flid)
{
	OSM_ASSERT(flid <= p_mgr->p_subn->opt.global_flid_end);
	p_mgr->flids_not_in_subnet[flid] = 0;
}
