/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of AsymTree Routing Engine.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <math.h>
#include <float.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_types_osd.h>
#include <complib/cl_types.h>
#include <complib/cl_byteswap.h>
#include <complib/cl_ptr_vector.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_fraction.h>

#include <iba/ib_types.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_ASYMTREE_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_root_detection.h>
#include <opensm/osm_max_flow_algorithm.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_helper.h>

/* ==============================================================================================================
 *
 * Miscellaneous definitions.
 */

#define asym_min(X,Y)		((X) < (Y)) ? (X) : (Y)
#define asym_max(X,Y)		((X) > (Y)) ? (X) : (Y)

/*
 * Macro for clean code.
 */
#define	asym_snprintf(format, ...)	bufused += snprintf(buffer+bufused, bufsize-bufused, format, ## __VA_ARGS__)

typedef	ssize_t			asym_iter_t;

typedef enum {
	DIR_DOWN = 0,
	DIR_UP   = 1,
} direction_t;

typedef enum {
	NODE_NA = 0,
	NODE_UP = 1,
} node_state_t;

typedef enum {
	LINK_DOWN    = 0,
	LINK_UP      = 1,
	LINK_REMOVED = 2,
	LINK_NA      = 3,
} link_state_t;

typedef enum {
	ASYM_TASK_SWITCHES        = 0,
	ASYM_TASK_PODS            = 1,
	ASYM_TASK_NON_LEAVES      = 2,
	ASYM_TASK_NON_POD_LEAVES  = 3,
} task_type_t;

/*
 * Layer definitions.
 */
#define NO_LAYER		128
#define MAX_LAYERS		16

/*
 * Debugging variables.
 */
__attribute__((unused))
static int	asym_debug = 0;

__attribute__((unused))
static char	asym_buffer[64];

/* ==============================================================================================================
 *
 * Set section.
 */

typedef struct {
	int			init;
	cl_ptr_vector_t		vector;
} asym_set_t;

static cl_status_t asym_set_cmp(IN const size_t index, const void *const element, IN void *context)
{
	uintptr_t		a;
	uintptr_t		b;

	a = (uintptr_t)element;
	b = (uintptr_t)context;

	return (a == b) ? CL_SUCCESS : CL_NOT_FOUND;
}

static inline boolean_t asym_set_is_active(IN asym_set_t *p_set)
{
	CL_ASSERT(p_set);

	return p_set->init;
}

static boolean_t asym_set_create(IN asym_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(!p_set->init);

	p_set->init = FALSE;
	cl_ptr_vector_construct(&p_set->vector);
	if (cl_ptr_vector_init(&p_set->vector, 1, 1024) == CL_SUCCESS) {
		cl_ptr_vector_set(&p_set->vector, 0, NULL);
		p_set->init = TRUE;
	}

	return p_set->init;
}

static void asym_set_destroy(IN asym_set_t *p_set)
{
	CL_ASSERT(p_set);

	if (p_set->init) {
		cl_ptr_vector_remove_all(&p_set->vector);
		cl_ptr_vector_destroy(&p_set->vector);
	}

	memset((void *)p_set, 0, sizeof(asym_set_t));
}

static inline uint32_t asym_set_get_size(IN asym_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return cl_ptr_vector_get_size(&p_set->vector);
}

static inline void *asym_set_get(IN asym_set_t *p_set, IN size_t index)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return cl_ptr_vector_get(&p_set->vector, index);
}

static inline size_t asym_set_find(IN asym_set_t *p_set, IN void *value)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return cl_ptr_vector_find_from_start(&p_set->vector, asym_set_cmp, value);
}

static inline boolean_t asym_set_contains(IN asym_set_t *p_set, IN void *value)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return asym_set_find(p_set, value) != asym_set_get_size(p_set);
}

static inline void asym_set_insert(IN asym_set_t *p_set, IN void *value)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	cl_ptr_vector_insert(&p_set->vector, value, NULL);
}

static inline void asym_set_remove(IN asym_set_t *p_set, IN void *value)
{
	size_t		index;

	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	if (asym_set_contains(p_set, value)) {
		index = asym_set_find(p_set, value);
		cl_ptr_vector_remove(&p_set->vector, index);
	}
}

static inline void asym_set_remove_all(IN asym_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	cl_ptr_vector_remove_all(&p_set->vector);
	cl_ptr_vector_set(&p_set->vector, 0, NULL);
}

static inline void asym_set_sort(IN asym_set_t *p_set, IN void *p_func)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	cl_ptr_vector_sort_cmp(&p_set->vector, 1, 0, p_func);
}

static inline boolean_t asym_set_is_empty(IN asym_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return asym_set_get_size(p_set) == 1;
}

static inline boolean_t asym_set_is_not_empty(IN asym_set_t *p_set)
{
	CL_ASSERT(p_set);
	CL_ASSERT(p_set->init);

	return !asym_set_is_empty(p_set);
}

#define foreach_set_element(P_SET, P_VOID, TYPE, ITER)								\
	for (ITER = (asym_iter_t)asym_set_get_size((P_SET)) - 1;						\
	     P_VOID = (TYPE)asym_set_get(P_SET, ITER), ITER >= 1;						\
	     ITER--)

/*
 * This function is not a primitive which maps into the underlying data structure.
 */
static void asym_set_union(IN asym_set_t *p_dst_set, IN asym_set_t *p_src_set)
{
	void			*p_void;
	asym_iter_t		iter;

	foreach_set_element(p_src_set, p_void, void *, iter) {
		if (!asym_set_contains(p_dst_set, p_void)) {
			asym_set_insert(p_dst_set, p_void);
		}
	}
}

/* ==============================================================================================================
 *
 * Memory section.
 */

static void *asym_calloc(size_t count, size_t length)
{
	return calloc(count, length);
}

static void asym_free(IN void *address)
{
	free(address);
}

static int count_nulls(IN int count, va_list ptrs)
{
	int			i;
	boolean_t		num_nulls;

	num_nulls = 0;

	for (i = 0; i < count; i++) {
		if (va_arg(ptrs, void *) == NULL) {
			num_nulls++;
			break;
		}
	}

	va_end(ptrs);
	return num_nulls;
}

static boolean_t any_null(IN int count, ...)
{
	va_list			ptrs;
	boolean_t		status;

	va_start(ptrs, count);
	status = count_nulls(count, ptrs);
	va_end(ptrs);

	return status != 0;
}

__attribute__((unused))
static boolean_t all_non_null(IN int count, ...)
{
	va_list			ptrs;
	boolean_t		status;

	va_start(ptrs, count);
	status = count_nulls(count, ptrs);
	va_end(ptrs);

	return status == 0;
}

/* ==============================================================================================================
 *
 * Fraction section.
 */

#define F(a,b)			cl_fraction(a,b)

#define F_add(a, b)		cl_fraction_add(a, b)
#define F_sub(a, b)		cl_fraction_sub(a, b)
#define F_mul(a, b)		cl_fraction_mul(a, b)
#define F_div(a, b)		cl_fraction_div(a, b)

#define F_floor(a)		cl_fraction_floor(a)
#define F_ceil(a)		cl_fraction_ceil(a)

#define F_eq(a, b)		cl_fraction_eq(a, b)
#define F_ne(a, b)		cl_fraction_ne(a, b)

#define F_gt(a, b)		cl_fraction_gt(a, b)
#define F_lt(a, b)		cl_fraction_lt(a, b)

#define F_ge(a, b)		cl_fraction_ge(a, b)
#define F_le(a, b)		cl_fraction_le(a, b)

#define F_to_string(f)		cl_fraction_to_string(f, asym_buffer)

/* ==============================================================================================================
 *
 * Weight and threshold section.
 */

typedef cl_fraction_t		weight_t;

#define W(W)			F(W,1)

#define W_add(W1, W2)		F_add(W1, W2)
#define W_sub(W1, W2)		F_sub(W1, W2)
#define W_mul(W1, W2)		F_mul(W1, W2)
#define W_div(W1, v)		F_div(W1, F(v,1))

#define W_floor(W)		F_floor(W)
#define W_ceil(W)		F_ceil(W)

#define W_eq(W1, W2)		F_eq(W1, W2)
#define W_ne(W1, W2)		F_ne(W1, W2)

#define W_gt(W1, W2)		F_gt(W1, W2)
#define W_lt(W1, W2)		F_lt(W1, W2)

#define W_ge(W1, W2)		F_ge(W1, W2)
#define W_le(W1, W2)		F_le(W1, W2)

#define W_to_string(W)		F_to_string(W)


typedef cl_fraction_t		threshold_t;

#define T(A,B)			F(A,B)
#define T_enabled(T)		F_ne(T, F(0,1))
#define T_disabled(T)		F_eq(T, F(0,1))

#define T_add(T1, T2)		F_add(T1, T2)
#define T_sub(T1, T2)		F_sub(T1, T2)
#define T_mul(T1, T2)		F_mul(T1, T2)
#define T_div(T1, V)		F_div(T1, F(V,1))
#define T_inc(T1)		F_add(T1, F(1,1))
#define T_dec(T1)		F_sub(T1, F(1,1))

#define T_floor(T)		F_floor(T)
#define T_ceil(T)		F_ceil(T)

#define T_eq(T1, T2)		F_eq(T1, T2)
#define T_ne(T1, T2)		F_ne(T1, T2)

#define T_gt(T1, T2)		F_gt(T1, T2)
#define T_lt(T1, T2)		F_lt(T1, T2)

#define T_ge(T1, T2)		F_ge(T1, T2)
#define T_le(T1, T2)		F_le(T1, T2)

#define T_to_string(T)		F_to_string(T)

/* ==============================================================================================================
 *
 * Looping and mapping macros.
 */

#define switch_to_index(P_SW)		((P_SW)->re_index)
#define switch_to_vertex(P_SW)		index_to_vertex(switch_to_index(P_SW))

#define index_to_anode(INDEX)		&p_asym->nodes[(INDEX)]
#define index_to_vertex(INDEX)		&p_asym->p_graph->vertices[(INDEX)]
#define index_to_distance(INDEX)	&p_asym->distances[(INDEX) * p_asym->num_switches]

#define foreach_asym_node(P_ASYM, P_NODE)									\
	for (P_NODE = P_ASYM->nodes;										\
	     P_NODE < &P_ASYM->nodes[P_ASYM->num_switches];							\
	     P_NODE++)

#define foreach_asym_port(P_NODE, P_PORT, ITER)									\
	foreach_set_element(&P_NODE->ports, P_PORT, asym_port_t *, ITER)

#define foreach_asym_pod(P_ASYM,P_POD)										\
	for (P_POD  = (asym_pod_t *)cl_qmap_head(&((P_ASYM)->pod_map));      					\
	     P_POD != (asym_pod_t *)cl_qmap_end(&((P_ASYM)->pod_map));       					\
	     P_POD  = (asym_pod_t *)cl_qmap_next((cl_map_item_t*)P_POD))

#define foreach_pod_leaf(P_POD, P_SW, ITER)									\
	foreach_set_element(&((P_POD)->leaves), P_SW, osm_switch_t *, ITER)					\
		if (!asym_switch_is_isolated(p_asym->p_subn, P_SW))

#define foreach_pod_spine(P_POD, P_SW, ITER)									\
	foreach_set_element(&((P_POD)->spines), P_SW, osm_switch_t *, ITER)					\
		if (!asym_switch_is_isolated(p_asym->p_subn, P_SW))

#define foreach_set_switch(P_SET, P_SW, ITER)									\
	foreach_set_element(P_SET, P_SW, osm_switch_t *, ITER)

#define foreach_set_adata(P_SET, P_ADATA, ITER)									\
	foreach_set_element(P_SET, P_ADATA, asym_adata_t *, ITER)

#define foreach_set_alink(P_SET, P_ALINK, ITER)									\
	foreach_set_element(P_SET, P_ALINK, asym_alink_t *, ITER)

#define foreach_set_removal(P_SET, P_ALINK, ITER)								\
	foreach_set_element(P_SET, P_ALINK, asym_alink_t *, ITER)

#define foreach_aport_lid(P_APORT, LID)										\
	for (LID = P_APORT->min_lid; LID <= P_APORT->max_lid; LID++)

#define foreach_active_switch(P_SUBN, P_SW)									\
	foreach_switch(P_SUBN, P_SW)										\
		if (!asym_switch_is_isolated(P_SUBN, P_SW))

#define foreach_active_vertex(P_ASYM, P_VERTEX)									\
	foreach_vertex((P_ASYM)->p_graph, P_VERTEX)								\
		if (!asym_switch_is_isolated((P_ASYM)->p_subn, (P_VERTEX)->p_sw))

#define foreach_layer(P_LAYER, ITER)										\
	for (ITER = 0, P_LAYER = p_algo->layers; ITER < p_algo->num_layers; ITER++, P_LAYER++)

#define foreach_active_layer(P_LAYER, ITER)									\
	for (ITER = 0, P_LAYER = p_algo->layers; ITER < p_algo->num_layers; ITER++, P_LAYER++)

#define foreach_quasi_layer(P_QUASI, P_LAYER, ITER)								\
	for (ITER = 0, P_LAYER = P_QUASI->layers; ITER < P_QUASI->num_layers; ITER++, P_LAYER++)

/* ==============================================================================================================
 *
 * These data structs are shared between AR_ASYM and AR_ASYM_ALGORITHM.
 */

typedef asym_set_t		asym_layer_t;

typedef struct {
	cl_map_item_t		map_item;
	int			index;
	int			leaf_rank;
	osm_switch_t		*p_fc_sw;		// a switch which is fully connected in the pod
	asym_layer_t		leaves;
	asym_layer_t		spines;
} asym_pod_t;

typedef struct {
	int			index;
	int			num_layers;
	asym_layer_t		*layers;
} asym_quasi_t;

typedef struct {
	ib_special_port_type_t	type;
	uint16_t		min_lid;
	uint16_t		max_lid;
} asym_port_t;

typedef struct {
	int			index;
	int			q_max;
	int			quasi;
	uint16_t		group_id;
	boolean_t		has_hcas;
	boolean_t		is_fully_connected;
	uint16_t		max_group_id;
	int			group;
	int			pod;
	osm_switch_t		*p_sw;
	asym_set_t		ports;
	int			num_links;
} asym_node_t;

typedef struct {
	osm_opensm_t		*p_osm;
	osm_subn_t		*p_subn;
	asym_node_t		*nodes;			// list of node attributes which are read-only
	asym_port_t		*ports;			// list of port attributes which are read-only
	asym_set_t		excluded_spines;
	cl_plock_t		lock;
	void			*re_context;
	osm_log_t		*p_log;
	osm_ucast_mgr_t		*p_mgr;
	flow_graph_t		*p_graph;
	cl_qmap_t		pod_map;
	int			num_pods;
	int			num_quasis;
	int			num_switches;
	int			num_ports;
	int			num_links;
	int			max_layers;
	uint8_t			ar_state;
	int			ar_enable;
	int			ar_mode;
	int			max_rank;
	int			core_rank;
	threshold_t		threshold;
	void			*p_algo;
} asym_t;

/* ==============================================================================================================
 *
 * These data structs are private to AR_ASYM_ALGORITHM.
 */

/* ==============================================================================================================
 * Macros.
 */
#define sw_name(P_SW)		(P_SW->p_node->print_desc)
#define adata_name(P_ADATA)	sw_name(P_ADATA->p_sw)
#define anode_name(P_ADATA)	sw_name(P_ANODE->p_sw)
#define index_name(INDEX)	sw_name(index_to_switch((INDEX)))

/*
 * Mapping macros.
 */
#define index_to_alink(INDEX)		 (&p_algo->p_alink[(INDEX)])
#define index_to_adata(INDEX)		 (&p_algo->p_adata[(INDEX)])
#define index_to_switch(INDEX)		((&p_asym->nodes[(INDEX)])->p_sw)
#define index_to_layer(INDEX)	 	 (&p_algo->layers[INDEX])

#define switch_to_adata(P_SW)		index_to_adata(switch_to_index(P_SW))
#define switch_to_anode(P_SW)		index_to_anode(switch_to_index(P_SW))

#define adata_to_anode(P_ADATA)		index_to_anode((P_ADATA)->index)
#define adata_to_switch(P_ADATA)	((P_ADATA)->p_sw)

#define anode_to_adata(P_ANODE)		index_to_adata((P_ANODE)->index)
#define anode_to_switch(P_ANODE)	((P_ANODE)->p_sw)

#define edge_to_layer(P_EDGE, INDEX)										\
	(index_to_adata((P_EDGE)->INDEX))->layer

#define alink_to_switches(P_ALINK, P_SW1, P_SW2)								\
	P_SW1 = index_to_switch((P_ALINK)->p_edge->index);							\
	P_SW2 = index_to_switch((P_ALINK)->p_edge->adj_index)

#define alink_to_adata(P_ALINK, P_ADATA1, P_ADATA2)								\
	P_ADATA1 = index_to_adata((P_ALINK)->p_edge->index);							\
	P_ADATA2 = index_to_adata((P_ALINK)->p_edge->adj_index)

#define alink_to_incoming_adata(P_ALINK)									\
	((edge_to_layer((P_ALINK)->p_edge, index) > edge_to_layer((P_ALINK)->p_edge, adj_index))		\
		? index_to_adata((P_ALINK)->p_edge->index)							\
		: index_to_adata((P_ALINK)->p_edge->adj_index))

#define alink_to_outgoing_adata(P_ALINK)									\
	((edge_to_layer((P_ALINK)->p_edge, index) < edge_to_layer((P_ALINK)->p_edge, adj_index))		\
		? index_to_adata((P_ALINK)->p_edge->index)							\
		: index_to_adata((P_ALINK)->p_edge->adj_index))

#define alink_to_incoming_switch(P_ALINK)									\
	((edge_to_layer((P_ALINK)->p_edge, index) > edge_to_layer((P_ALINK)->p_edge, adj_index))		\
		? index_to_switch((P_ALINK)->p_edge->index)							\
		: index_to_switch((P_ALINK)->p_edge->adj_index))

#define alink_to_outgoing_switch(P_ALINK)									\
	((edge_to_layer((P_ALINK)->p_edge, index) < edge_to_layer((P_ALINK)->p_edge, adj_index))		\
		? index_to_switch((P_ALINK)->p_edge->index)							\
		: index_to_switch((P_ALINK)->p_edge->adj_index))

#define alink_to_incoming_port(P_ALINK)										\
	((edge_to_layer((P_ALINK)->p_edge, index) > edge_to_layer((P_ALINK)->p_edge, adj_index))		\
		? (P_ALINK)->p_edge->port_num									\
		: (P_ALINK)->p_edge->adj_port_num)

#define alink_to_outgoing_port(P_ALINK)										\
	((edge_to_layer((P_ALINK)->p_edge, index) < edge_to_layer((P_ALINK)->p_edge, adj_index))		\
		? (P_ALINK)->p_edge->port_num									\
		: (P_ALINK)->p_edge->adj_port_num)

#define alink_to_ports(P_ALINK, PORT1, PORT2)									\
	PORT1 = ((P_ALINK)->p_edge->port_num);									\
	PORT2 = ((P_ALINK)->p_edge->adj_port_num)

/*
 * Looping macros.
 */
#define foreach_anode(P_ASYM, P_ANODE)										\
	for (P_ANODE = P_ASYM->nodes; P_ANODE < &P_ASYM->nodes[P_ASYM->num_switches]; P_ANODE++)		\
		if (P_ANODE->p_sw != NULL)

#define foreach_adata(P_ALGO, P_ADATA)										\
	for (P_ADATA = P_ALGO->p_adata; P_ADATA < &P_ALGO->p_adata[P_ALGO->num_adata]; P_ADATA++)		\
		if (P_ADATA->p_sw != NULL)

#define foreach_alink(P_ALGO, P_ALINK)										\
	for (P_ALINK = P_ALGO->p_alink; P_ALINK < &P_ALGO->p_alink[P_ALGO->num_alink]; P_ALINK++)

#define foreach_layer_step(P_ALGO, P_LAYER1, P_LAYER2, P_FUNCTION, ITER)					\
	for (ITER  = 0, P_ALGO->layer_number = ITER;								\
	     P_LAYER1 = index_to_layer(ITER),									\
	     P_LAYER2 = index_to_layer(ITER+1),									\
	     P_FUNCTION = P_ALGO->layer_functions[ITER],							\
	     ITER < P_ALGO->num_layers;										\
	     ITER++, P_ALGO->layer_number++)

#define foreach_layer_switch(P_LAYER, P_SW, ITER)								\
	foreach_set_element(P_LAYER, P_SW, osm_switch_t *, ITER)

#define foreach_link_removal(P_REMS, P_ALINK, ITER)								\
	foreach_set_element(P_REMS, P_ALINK, asym_alink_t *, ITER)

#define foreach_switch_removal(P_REMS, P_SW, ITER)								\
	foreach_set_element(P_REMS, P_SW, osm_switch_t *, ITER)

/*
 * Node macros.
 */
#define anode_has_attached_hcas(P_ANODE)	((P_ANODE)->has_hcas)

/*
 * Data macros.
 */
#define node_is_na(P_ADATA)			(P_ADATA->state == NODE_NA)
#define node_is_up(P_ADATA)			(P_ADATA->state == NODE_UP)

/*
 * Link macros.
 */
#define link_is_up(P_ALINK)			(P_ALINK->state == LINK_UP)
#define link_is_down(P_ALINK)			(P_ALINK->state == LINK_DOWN)
#define link_is_removed(P_ALINK)		(P_ALINK->state == LINK_REMOVED)
#define link_is_na(P_ALINK)			(P_ALINK->state == LINK_NA)

#define alink_test(P_ADATA, P_ALINK)										\
	(P_ALINK->p_edge->index == P_ADATA->index)

#define alink_is_connected_to_adata(P_ADATA, P_ALINK)								\
	((P_ALINK->p_edge->index == P_ADATA->index) || (P_ALINK->p_edge->adj_index == P_ADATA->index))

#define alink_local_index(P_ADATA, P_ALINK)									\
	((alink_test(P_ADATA, P_ALINK)) ? P_ALINK->p_edge->index : P_ALINK->p_edge->adj_index)

#define alink_local_port_num(P_ADATA, P_ALINK)									\
	((alink_test(P_ADATA, P_ALINK)) ? P_ALINK->p_edge->port_num : P_ALINK->p_edge->adj_port_num)

#define alink_local_adata(P_ADATA, P_ALINK)									\
	(index_to_adata(alink_local_index(P_ADATA, P_ALINK)))

#define alink_local_switch(P_ADATA, P_ALINK)									\
	(index_to_switch(alink_local_index(P_ADATA, P_ALINK)))

#define alink_remote_index(P_ADATA, P_ALINK)									\
	((alink_test(P_ADATA, P_ALINK)) ? P_ALINK->p_edge->adj_index : P_ALINK->p_edge->index)

#define alink_remote_port_num(P_ADATA, P_ALINK)									\
	((alink_test(P_ADATA, P_ALINK)) ? P_ALINK->p_edge->adj_port_num : P_ALINK->p_edge->port_num)

#define alink_remote_anode(P_ADATA, P_ALINK)									\
	(index_to_anode(alink_remote_index(P_ADATA, P_ALINK)))

#define alink_remote_adata(P_ADATA, P_ALINK)									\
	(index_to_adata(alink_remote_index(P_ADATA, P_ALINK)))

#define alink_remote_switch(P_ADATA, P_ALINK)									\
	(index_to_switch(alink_remote_index(P_ADATA, P_ALINK)))

#define link_is_incoming(P_ADATA, P_ALINK)									\
	(P_ADATA->layer == alink_remote_adata(P_ADATA, P_ALINK)->layer + 1)

#define link_is_outgoing(P_ADATA, P_ALINK)									\
	(P_ADATA->layer == alink_remote_adata(P_ADATA, P_ALINK)->layer - 1)

#define foreach_link(P_ADATA, P_ALINK, ITER)									\
	foreach_set_element(&P_ADATA->links, P_ALINK, asym_alink_t *, ITER)

#define foreach_incoming_link(P_ADATA, P_ALINK, ITER)								\
	foreach_link(P_ADATA, P_ALINK, ITER)									\
		if (link_is_up(P_ALINK) && link_is_incoming(P_ADATA,P_ALINK))

#define foreach_outgoing_link(P_ADATA, P_ALINK, ITER)								\
	foreach_link(P_ADATA, P_ALINK, ITER)									\
		if (link_is_up(P_ALINK) && link_is_outgoing(P_ADATA,P_ALINK))

/* ----------------------------------------------------------------------------------------------------------- */

typedef struct _asym_algorithm asym_algorithm_t;
typedef void (*asym_layer_function_t)(asym_algorithm_t *p_algo, asym_set_t *curr_layer, asym_set_t *next_layer);

/*
 * Algorithm typdefs.
 */
typedef struct {
	void			*p_algo;
	flow_edge_t		*p_edge;

	link_state_t		state;
	weight_t		lw;
} asym_alink_t;

typedef struct {
	int			index;
	node_state_t		state;
	osm_switch_t		*p_sw;

	asym_set_t		links;
	int			port_index;

	int			layer;
	weight_t		nw;
	threshold_t		threshold;
	uint8_t			hops;
	uint8_t			in_count;
	uint8_t			out_count;
	weight_t		node_weight;
	weight_t		link_weight;
	boolean_t		balanced;
} asym_adata_t;

typedef struct _asym_algorithm {
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;

	asym_t			*p_asym;
	int			layer_number;
	int			num_layers;
	int			max_layers;
	int			retries;
	int			dst_distance;

	asym_set_t		excluded_spines;

	asym_pod_t		*p_src_pod;		// current source pod
	asym_pod_t		*p_dst_pod;		// current destination pod

	asym_set_t		removals;		// destination specific removals
	asym_set_t		fc_removals;		// cache for performance shortcut
	asym_set_t		src_removals;		// src specific removals
	asym_set_t		ar_removals;		// AR removals

	size_t			alink_size;
	size_t			num_alink;
	asym_alink_t		*p_alink;

	size_t			adata_size;
	size_t			num_adata;
	asym_adata_t		*p_adata;		// array of nodes attributes which are writeable

	size_t			layer_size;
	asym_layer_t		layers[MAX_LAYERS];
	asym_layer_function_t	layer_functions[MAX_LAYERS];

	size_t			distances_size;
	uint8_t			*distances;

	size_t			visited_size;
	uint8_t			*visited;

	size_t			queue_size;
	uint16_t		*queue;
	int			q_max;

	asym_set_t		scratch;		// for small functions to use

	boolean_t		fully_connected_done;
} asym_algorithm_t;

/* ==============================================================================================================
 *
 * Debug section.
 */

#ifdef ASYM_DEBUG
// include debugging files here
#endif // ASYM_DEBUG

static void asym_log_by_line(osm_log_t *p_log, int log_level, char *buffer)
{
	char			*cp;
	char			*ep;

	cp = buffer;
	while ((*cp != '\0') && ((ep = strchr(cp, '\n')) != NULL)) {
		*ep = '\0';

		switch (log_level) {
		case OSM_LOG_VERBOSE:
			OSM_LOG(p_log, OSM_LOG_VERBOSE, "%s\n", cp);
			break;
		default:
			OSM_LOG(p_log, OSM_LOG_DEBUG, "%s\n", cp);
			break;
		}

		*ep = '\n';
		cp = ep+1;
	}
}

__attribute__((unused))
static void asym_print_removals(IN asym_algorithm_t *p_algo, IN asym_set_t *p_removals)
{
	size_t			count;
	uint8_t			port_num;
	osm_switch_t		*p_sw;
	asym_t			*p_asym = p_algo->p_asym;
	asym_iter_t		iter1;
	asym_alink_t		*p_alink;

	count = asym_set_get_size(p_removals);
	if (count > 1) {
		printf("%s REMOVALS ", sw_name(p_algo->p_dst_sw));
		foreach_link_removal(p_removals, p_alink, iter1) {
			p_sw = alink_to_incoming_switch(p_alink);
			port_num = alink_to_incoming_port(p_alink);
			printf("%s/%u ", sw_name(p_sw), port_num);
		}

		printf("\n");
	}
}

__attribute__((unused))
static void asym_print_remove(IN asym_algorithm_t *p_algo, IN asym_alink_t *p_alink, IN char *reason)
{
	int			aport;
	int			bport;
	asym_adata_t		*p_adata;
	asym_adata_t		*p_bdata;

	alink_to_adata(p_alink, p_adata, p_bdata);
	alink_to_ports(p_alink, aport, bport);
	printf("%-10s   %s REMOVE %s/%d:%s/%d\n",
		reason,
		sw_name(p_algo->p_dst_sw),
		adata_name(p_adata), aport,
		adata_name(p_bdata), bport);
}

/* ==============================================================================================================
 *
 * UFM section.
 */

__attribute__((unused))
static void asym_process_excluded_spines(IN asym_t *p_asym)
{
	asym_iter_t		iter;
	asym_set_t		*p_set;
	osm_subn_t		*p_subn;
	uint64_t		guid;

	p_subn = p_asym->p_subn;
	p_set = (asym_set_t *)p_asym->re_context;

	/*
	 * Determine if any of the old excluded spines are NOT in the new set.
	 */
	foreach_set_element(p_set, guid, uint64_t, iter) {
		if (!asym_set_contains(&p_asym->excluded_spines, (void *)guid)) {
			asym_set_remove(p_set, (void *)guid);
			osm_report_notice_1606(p_subn, guid);
		}
	}

	/*
	 * Determine if any of the new excluded spines are NOT in the old set.
	 */
	foreach_set_element(&p_asym->excluded_spines, guid, uint64_t, iter) {
		if (!asym_set_contains(p_set, (void *)guid)) {
			asym_set_insert(p_set, (void *)guid);
			osm_report_notice_1605(p_subn, guid);
		}
	}

	return;
}

static void asym_update_excluded_spines(IN asym_algorithm_t *p_algo)
{
	asym_t			*p_asym;
	asym_iter_t		iter;
	uint64_t		guid;

	p_asym = p_algo->p_asym;

	CL_PLOCK_EXCL_ACQUIRE(&p_asym->lock);
	foreach_set_element(&p_algo->excluded_spines, guid, uint64_t, iter) {
		if (!asym_set_contains(&p_asym->excluded_spines, (void *)guid)) {
			asym_set_insert(&p_asym->excluded_spines, (void *)guid);
		}
	}
	CL_PLOCK_RELEASE(&p_asym->lock);
}

/* ==============================================================================================================
 *
 * Wrapper section.
 */

static ib_special_port_type_t asym_get_special_port_type(IN osm_port_t *p_port)
{
	osm_physp_t		*p_physp;

	p_physp = p_port->p_physp;

	if (p_port->p_node->sw != NULL) {
		return IB_SPECIAL_PORT_TYPE_MAX;
	} else if (p_physp == NULL) {
		return IB_SPECIAL_PORT_TYPE_MAX;
	} else if (p_physp->p_remote_physp == NULL) {
		return IB_SPECIAL_PORT_TYPE_MAX;
	}

	return osm_physp_get_special_port_type(p_physp);
}

static osm_switch_t *asym_get_switch_by_port(IN osm_port_t *p_port)
{
	osm_physp_t		*p_physp;
	osm_physp_t		*p_remote_physp;

	p_physp = osm_port_get_physp(p_port);
	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_remote_physp && p_remote_physp->p_node) {
		return p_remote_physp->p_node->sw;
	}

	return NULL;
}

static inline boolean_t asym_switch_is_isolated(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw)
{
	return ((p_sw->rank == OSM_SW_NO_RANK) ||
	        ((p_sw->rank == 0) && (lookup_held_back_switch(p_subn, p_sw->p_node))));
}

static inline boolean_t asym_link_is_isolated(IN asym_t *p_asym, IN asym_alink_t *p_alink)
{
	osm_sm_t		*p_sm;
	osm_switch_t		*p_sw1;
	osm_switch_t		*p_sw2;
	osm_unhealthy_port_t	*p_unhealthy_port;
	flow_edge_t		*p_edge;

	p_sm = &p_asym->p_osm->sm;
	p_edge = p_alink->p_edge;
	p_sw1 = p_asym->p_graph->vertices[p_edge->index].p_sw;
	p_sw2 = p_asym->p_graph->vertices[p_edge->adj_index].p_sw;

	if (asym_switch_is_isolated(p_asym->p_subn, p_sw1)) {
		return TRUE;
	}

	if (asym_switch_is_isolated(p_asym->p_subn, p_sw2)) {
		return TRUE;
	}


	if ((p_unhealthy_port = osm_hm_get_unhealthy_port(p_sm, p_sw1->p_node, p_edge->port_num)) != NULL) {
		if (p_unhealthy_port->isolate) {
			return TRUE;
		}
	}

	if ((p_unhealthy_port = osm_hm_get_unhealthy_port(p_sm, p_sw2->p_node, p_edge->adj_port_num)) != NULL) {
		if (p_unhealthy_port->isolate) {
			return TRUE;
		}
	}

	return FALSE;
}

static boolean_t asym_subgroup_is_empty(IN osm_ar_subgroup_t *p_group)
{
	int			i = 0;
	uint8_t			value = 0;

	if (p_group != NULL) {
		for (i = 0; i < sizeof(p_group->mask); i++) {
			value |= p_group->mask[i];
		}
	}

	return value == 0;
}

/* ==============================================================================================================
 *
 * Comparison function section.
 */

static asym_adata_t *node_choose(IN asym_adata_t *p_adata, IN asym_adata_t *p_bdata)
{
	osm_switch_t	*p_asw = adata_to_switch(p_adata);
	osm_switch_t	*p_bsw = adata_to_switch(p_bdata);

	return (strcmp(sw_name(p_asw), sw_name(p_bsw)) <= 0) ? p_adata : p_bdata;
}

static asym_alink_t *link_choose(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata, IN asym_alink_t *p_alink, IN asym_alink_t *p_blink)
{
	char		aname[256];
	char		bname[256];
	asym_t		*p_asym = p_algo->p_asym;
	osm_switch_t	*p_asw = alink_remote_switch(p_adata, p_alink);
	osm_switch_t	*p_bsw = alink_remote_switch(p_adata, p_blink);

	sprintf(aname, "%s/%d", sw_name(p_asw), alink_remote_port_num(p_adata, p_alink));
	sprintf(bname, "%s/%d", sw_name(p_bsw), alink_remote_port_num(p_adata, p_blink));

	return (strcmp(aname, bname) < 0) ? p_alink : p_blink;
}

static int asym_highest_weight_cmp(IN const void *p_lhs, IN const void *p_rhs)
{
	asym_adata_t		*p_adata;
	asym_adata_t		*p_bdata;

	p_adata = *(asym_adata_t **)p_lhs;
	p_bdata = *(asym_adata_t **)p_rhs;

	if (W_lt(p_adata->nw, p_bdata->nw)) {
		return -1;
	} else if (W_gt(p_adata->nw, p_bdata->nw)) {
		return 1;
	} else {
		return (node_choose(p_adata, p_bdata) == p_adata) ? 1 : -1;
	}
}

static int asym_threshold_cmp(IN const void *p_lhs, IN const void *p_rhs)
{
	asym_adata_t		*p_adata;
	asym_adata_t		*p_bdata;

	p_adata = *(asym_adata_t **)p_lhs;
	p_bdata = *(asym_adata_t **)p_rhs;

	if (T_lt(p_adata->threshold, p_bdata->threshold)) {
		return -1;
	} else if (T_gt(p_adata->threshold, p_bdata->threshold)) {
		return 1;
	} else {
		return (node_choose(p_adata, p_bdata) == p_adata) ? -1 : 1;
	}
}

static int asym_fini_cmp(IN const void *p_lhs, IN const void *p_rhs)
{
	asym_adata_t		*p_adata;
	asym_adata_t		*p_bdata;

	p_adata = *(asym_adata_t **)p_lhs;
	p_bdata = *(asym_adata_t **)p_rhs;

	if (W_lt(p_adata->nw, p_bdata->nw)) {
		return 1;
	} else if (W_gt(p_adata->nw, p_bdata->nw)) {
		return -1;
	} else {
		return (node_choose(p_adata, p_bdata) == p_adata) ? 1 : -1;
	}
}

static int asym_fini_leaf_cmp(IN const void *p_lhs, IN const void *p_rhs)
{
	asym_adata_t		*p_adata;
	asym_adata_t		*p_bdata;

	p_adata = *(asym_adata_t **)p_lhs;
	p_bdata = *(asym_adata_t **)p_rhs;

	if (W_lt(p_adata->nw, p_bdata->nw)) {
		return 1;
	} else if (W_gt(p_adata->nw, p_bdata->nw)) {
		return -1;
	} else {
		return (node_choose(p_adata, p_bdata) == p_adata) ? 1 : -1;
	}
}

static int asym_sw_name_cmp(IN const void *p_lhs, IN const void *p_rhs)
{
	osm_switch_t		*p_asw;
	osm_switch_t		*p_bsw;

	p_asw = *(osm_switch_t **)p_lhs;
	p_bsw = *(osm_switch_t **)p_rhs;

	return -strcmp(sw_name(p_asw), sw_name(p_bsw));
}

static int asym_int_cmp(IN const void *p_lhs, IN const void *p_rhs)
{
	int		a;
	int		b;

	a = *(int *)p_lhs;
	b = *(int *)p_rhs;

	return (a - b);
}

static int asym_removal_cmp(IN const void *p_lhs, IN const void *p_rhs)
{
	asym_adata_t		*p_adata;
	asym_adata_t		*p_bdata;
	asym_alink_t		*p_alink;
	asym_alink_t		*p_blink;
	asym_algorithm_t	*p_algo;

	p_alink = *(asym_alink_t **)p_lhs;
	p_blink = *(asym_alink_t **)p_rhs;

	p_algo = (asym_algorithm_t *)p_alink->p_algo;

	p_adata = alink_to_incoming_adata(p_alink);
	p_bdata = alink_to_incoming_adata(p_blink);

	if (p_adata->layer != p_bdata->layer) {
		return p_adata->layer - p_bdata->layer;
	} else if (p_adata->index != p_bdata->index) {
		return p_adata->index - p_bdata->index;
	} else {
		return alink_local_port_num(p_adata, p_alink) - alink_local_port_num(p_adata, p_blink);
	}

	return 0;
}

/* ==============================================================================================================
 *
 * ARLFT section.
 */

static void asym_check_group_top(IN asym_node_t *p_anode,
				 IN osm_advanced_routing_data_t *p_ar_data,
				 IN uint16_t group_id,
				 IN uint16_t group_top)
{

	if (group_id > p_anode->max_group_id) {
		p_anode->max_group_id = group_id;
	}

	if (group_id > group_top) {
		group_top = group_id;
		osm_ar_set_required_group_top(p_ar_data, group_top);
	}
}
static uint8_t asym_find_free_group_port(IN osm_switch_t *p_src_sw,
					 IN uint16_t dst_lid,
					 IN OUT uint8_t *new_port,
					 IN osm_ar_subgroup_t *p_group)
{
	uint8_t                         port;
	uint8_t                         lft_port;
	int                             attempts;

	if ((lft_port = osm_switch_get_port_by_lid(p_src_sw, dst_lid, OSM_NEW_LFT)) == OSM_NO_PATH) {
		return OSM_NO_PATH;
	}

	/*
	 * If the current static port is not in the AR group, find the next port that is.
	 */
	if (!osm_ar_subgroup_check_port(p_group, lft_port)) {
		lft_port = OSM_NO_PATH;
		port = *new_port;

		for (attempts = 0; attempts < 256; attempts++) {
			port++;
			if (osm_ar_subgroup_check_port(p_group, port)) {
				p_src_sw->new_lft[dst_lid] = port;
				lft_port = port;
				break;
			}
		}

		*new_port = port;
	}

	return lft_port;
}

static void inline asym_clear_group0(IN osm_switch_t *p_sw)
{
	osm_ar_subgroup_t		*p_subgroup;
	osm_advanced_routing_data_t	*p_ar_data;

	/*
	 * Clear the switch AR group 0 ports.
	 */
	p_ar_data = &p_sw->ar_data;
	p_subgroup = osm_ar_subgroup_get(p_ar_data, 0, 0);
	memset(p_subgroup, 0, sizeof(osm_ar_subgroup_t));
}

static void asym_set_lft_entry(IN asym_t *p_asym,
			       IN osm_switch_t *p_sw,
			       IN uint16_t lid,
			       IN uint8_t port,
			       IN int state,
			       IN int id)
{
	if ((lid != 0) && (port != OSM_NO_PATH)) {
		osm_ar_lft_set(&p_sw->ar_data, lid, 0, port, state, id);
	} else {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "%s - no port found for lid %d\n", sw_name(p_sw), lid);
	}
}

/* ----------------------------------------------------------------------------------------------------------- */

static void asym_set_static_lft_entry(IN asym_t *p_asym,
				      IN osm_switch_t *p_sw,
				      IN uint16_t lid)
{
	uint8_t			port;

	port = osm_switch_get_port_by_lid(p_sw, lid, OSM_NEW_LFT);
	asym_set_lft_entry(p_asym, p_sw, lid, port, OSM_AR_IB_LID_STATE_STATIC, 0);
}

static void asym_set_static_lft_entry_by_lid_range(IN asym_t *p_asym,
					           IN osm_switch_t *p_sw,
					           IN uint16_t min_lid,
					           IN uint16_t max_lid)
{
	uint16_t		lid;

	for (lid = min_lid; lid <= max_lid; lid++) {
		asym_set_static_lft_entry(p_asym, p_sw, lid);
	}
}

static void asym_set_static_lft_entry_by_port(IN asym_t *p_asym,
					    IN osm_switch_t *p_sw,
					    IN asym_port_t *p_aport)
{
	asym_set_static_lft_entry_by_lid_range(p_asym, p_sw, p_aport->min_lid, p_aport->max_lid);
}

/* ----------------------------------------------------------------------------------------------------------- */

static void asym_set_free_lft_entry(IN asym_t *p_asym,
				    IN osm_switch_t *p_sw,
				    IN uint16_t lid,
				    IN uint8_t *new_port,
				    IN uint16_t group_id,
				    IN osm_ar_subgroup_t *p_ar_group)
{
	uint8_t			lft_port;

	/*
	 * Set the subgroup with the remote switch's AR group.
	 */
	if ((lft_port = asym_find_free_group_port(p_sw, lid, new_port, p_ar_group)) != OSM_NO_PATH) {
		asym_set_lft_entry(p_asym, p_sw, lid, lft_port, OSM_AR_IB_LID_STATE_FREE, group_id);
	}
}

static void asym_set_free_lft_entry_by_lid_range(IN asym_t *p_asym,
						 IN osm_switch_t *p_sw,
						 IN uint16_t min_lid,
						 IN uint16_t max_lid,
						 IN uint8_t *new_port,
						 IN uint16_t group_id,
						 IN osm_ar_subgroup_t *p_ar_group)
{
	uint16_t		lid;

	for (lid = min_lid; lid <= max_lid; lid++) {
		asym_set_free_lft_entry(p_asym, p_sw, lid, new_port, group_id, p_ar_group);
	}
}

static void asym_set_free_lft_entry_by_port(IN asym_t *p_asym,
					    IN osm_switch_t *p_sw,
					    IN asym_port_t *p_aport,
					    IN uint8_t *new_port,
					    IN uint16_t group_id,
					    IN osm_ar_subgroup_t *p_ar_group)
{
	asym_set_free_lft_entry_by_lid_range(p_asym, p_sw, p_aport->min_lid, p_aport->max_lid, new_port, group_id, p_ar_group);
}

/* ----------------------------------------------------------------------------------------------------------- */

static void asym_set_lft_entry_port0(IN asym_t *p_asym,
				     IN osm_switch_t *p_src_sw,
				     IN osm_switch_t *p_dst_sw,
				     IN osm_ar_subgroup_t *p_group,
				     IN boolean_t use_static_entry)
{
	uint8_t				new_port;
	uint16_t			min_lid;
	uint16_t			max_lid;
	uint16_t			group_id;
	uint16_t			group_top;
	asym_node_t			*p_anode;
	asym_node_t			*p_dst_anode;
	osm_subn_t			*p_subn;
	osm_port_t			*p_dst_port0;
	osm_ar_subgroup_t		dst_ar_group;
        osm_advanced_routing_data_t	*p_ar_data;

	p_subn = p_asym->p_subn;
	p_anode = switch_to_anode(p_src_sw);

	p_ar_data = &p_src_sw->ar_data;
	group_top = ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info);

	/*
	 * Set a free route to dst switch port 0.
	 */
	new_port = 0;

	p_dst_port0 = osm_switch_get_port0(p_dst_sw, p_subn);
	osm_port_get_lid_range_ho(p_dst_port0, &min_lid, &max_lid);

	p_dst_anode = switch_to_anode(p_dst_sw);
	group_id = p_dst_anode->group_id;

	if ((!use_static_entry) && (p_group == NULL)) {
		osm_switch_clear_and_get_minhop_ports(p_src_sw, min_lid, &dst_ar_group, NULL);
		p_group = &dst_ar_group;
	}

	/*
	 * Set static routes.
	 */
	if (use_static_entry) {
		asym_set_static_lft_entry_by_lid_range(p_asym, p_src_sw, min_lid, max_lid);
	} else {
		asym_check_group_top(p_anode, p_ar_data, group_id, group_top);
		asym_set_free_lft_entry_by_lid_range(p_asym, p_src_sw, min_lid, max_lid, &new_port, group_id, p_group);
	}
}

/* ==============================================================================================================
 *
 * Routing section.
 */

static void asym_log_removals(IN asym_algorithm_t *p_algo, asym_set_t *p_removals)
{
	size_t			count;
	size_t			bufsize;
	size_t			bufused;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_alink_t		*p_alink;
	flow_edge_t		*p_edge;
	char			*buffer;

        if (asym_debug) {
		asym_print_removals(p_algo, p_removals);
	}

        if (!osm_log_is_active_v2(p_algo->p_asym->p_log, OSM_LOG_VERBOSE, FILE_ID)) {
		return;
	}

	/*
	 * If there are no removals, then return.
	 */
	count = asym_set_get_size(p_removals);
	if (count <= 1) {
		return;
	}

	bufused = 0;
	bufsize = (256 * count);

	if ((buffer = asym_calloc(1, bufsize)) != NULL) {
		p_asym = p_algo->p_asym;

		asym_snprintf("\n%s REMOVALS\n", sw_name(p_algo->p_dst_sw));
		foreach_link_removal(p_removals, p_alink, iter1) {
			p_edge = p_alink->p_edge;

			asym_snprintf("\t%s/%d <-> %s/%d\n",
				      index_name(p_edge->index),     p_edge->port_num,
				      index_name(p_edge->adj_index), p_edge->adj_port_num);
		}

		asym_log_by_line(p_asym->p_log, OSM_LOG_VERBOSE, buffer);
		asym_free(buffer);
	}
}

__attribute__((unused))
static void asym_set_routes(IN asym_t *p_asym, IN osm_switch_t *p_src_sw, IN osm_switch_t *p_dst_sw, IN osm_ar_subgroup_t *p_group)
{
	uint8_t 			new_port = 0;
	uint16_t			dst_lid;
	uint16_t			group_id;
	uint16_t			group_cap;
	uint16_t			group_top;
	boolean_t			use_static_entry;
	asym_node_t			*p_anode;
	asym_node_t			*p_dst_anode;
	asym_port_t			*p_aport;
	asym_iter_t			iter;
	osm_subn_t			*p_subn;
	osm_ar_subgroup_t		*p_subgroup;
	osm_ar_subgroup_t		dst_ar_group;
	osm_advanced_routing_data_t	*p_ar_data;

	p_subn = p_asym->p_subn;

	p_ar_data = &p_src_sw->ar_data;
	group_top = ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info);
	group_cap = ib_ar_info_get_group_cap_ho(&p_ar_data->ar_info);

	p_anode = switch_to_anode(p_src_sw);
	p_dst_anode = switch_to_anode(p_dst_sw);
	group_id  = p_dst_anode->group_id;

	/*
	 * Is this switch AR capable?
	 */
	if (!osm_switch_is_ar_supported(p_src_sw)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "%s is not AR enabled\n", sw_name(p_src_sw));
	        return;
	}

	if (asym_switch_is_isolated(p_subn, p_src_sw) || asym_switch_is_isolated(p_subn, p_dst_sw)) {
		return;
	}

	/*
	 * We previously processed the ports_list and cached the lids which were attached to
	 * each switch.  We now take that list and use it to create the AR entries.
	 */

	/*
	 * If we are only using static entries, do them and return.  This is the case if
	 *	1) The fabric AR state is static.
	 *	2) The group id is invalid.
	 *	3) The group id is too large for this device.
	 */
	use_static_entry = ((p_asym->ar_state == OSM_AR_IB_LID_STATE_STATIC) ||
			    (group_id == UCAST_MGR_INVALID_GROUP_ID) ||
			    (group_id >= group_cap));

	/*
	 * Get the ar group for this switch if needed.
	 */
	if (!use_static_entry && anode_has_attached_hcas(p_dst_anode) && (p_group == NULL)) {
		dst_lid = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);
		osm_switch_clear_and_get_minhop_ports(p_src_sw, dst_lid, &dst_ar_group, NULL);
		p_group = &dst_ar_group;
	}

	/*
	 * Set the entries.
	 */
	if (!use_static_entry && (p_group != NULL)) {
		p_subgroup = osm_ar_subgroup_get(p_ar_data, group_id, 0);
		memcpy(p_subgroup, p_group, sizeof(osm_ar_subgroup_t));
	}

	/*
	 * If the subgroup is empty, use static route only.
	 */
	if (asym_subgroup_is_empty(p_group)) {
		use_static_entry = TRUE;
	}

	/*
	 * Port 0.
	 */
	asym_set_lft_entry_port0(p_asym, p_src_sw, p_dst_sw, p_group, use_static_entry);

	/*
	 * All other attached ports.
	 */
	foreach_asym_port(p_dst_anode, p_aport, iter) {
		if (use_static_entry) {
			asym_set_static_lft_entry_by_port(p_asym, p_src_sw, p_aport);
		} else if (p_aport->type != IB_SPECIAL_PORT_TYPE_RESERVED) {
			asym_set_static_lft_entry_by_port(p_asym, p_src_sw, p_aport);
		} else if (p_group == NULL) {
			asym_set_static_lft_entry_by_port(p_asym, p_src_sw, p_aport);
		} else {
			asym_check_group_top(p_anode, p_ar_data, group_id, group_top);
			asym_set_free_lft_entry_by_port(p_asym, p_src_sw, p_aport, &new_port, group_id, p_group);
		}
	}

}

static void asym_set_ar_groups(IN asym_algorithm_t *p_algo, IN asym_set_t *p_removals)
{
	uint8_t			port_num;
	uint16_t		dst_lid;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_alink_t		*p_alink;
	asym_adata_t		*p_adata;
	asym_adata_t		*p_curr_adata;
	asym_adata_t		*p_prev_adata;
	osm_switch_t		*p_curr_sw;
	osm_switch_t		*p_prev_sw;
	osm_switch_t		*p_dst_sw;
	osm_switch_t		*p_sw;
	osm_ar_subgroup_t	dst_ar_group;

	p_asym = p_algo->p_asym;

	p_dst_sw = p_algo->p_dst_sw;
	dst_lid = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);

	/*
	 * Sort the removals so we can combine them easily.
	 */
	asym_set_sort(p_removals, asym_removal_cmp);
	asym_log_removals(p_algo, p_removals);

	asym_set_remove_all(&p_algo->scratch);
	foreach_set_removal(p_removals, p_alink, iter1) {
		p_sw = alink_to_incoming_switch(p_alink);
		if (!asym_set_contains(&p_algo->scratch, p_sw)) {
			asym_set_insert(&p_algo->scratch, p_sw);
		}
	}

	/*
	 * Combine the removals for each switch.
	 */
	p_prev_adata = NULL;
	p_curr_adata = NULL;
	foreach_set_removal(p_removals, p_alink, iter1) {
		p_curr_adata = alink_to_incoming_adata(p_alink);

		if (p_curr_adata != p_prev_adata) {
			if (p_prev_adata != NULL) {
				p_prev_sw = adata_to_switch(p_prev_adata);
				asym_set_routes(p_asym, p_prev_sw, p_dst_sw, &dst_ar_group);
			}

			p_curr_sw = adata_to_switch(p_curr_adata);
			osm_switch_clear_and_get_minhop_ports(p_curr_sw, dst_lid, &dst_ar_group, NULL);
			p_prev_adata = p_curr_adata;
		}

		port_num = alink_local_port_num(p_curr_adata, p_alink);
		osm_ar_subgroup_remove_port(&dst_ar_group, port_num);
	}

	/*
	 * Update the last switch.
	 */
	if (p_curr_adata != NULL) {
		p_curr_sw = adata_to_switch(p_curr_adata);
		asym_set_routes(p_asym, p_curr_sw, p_dst_sw, &dst_ar_group);
	}

	/*
	 * Update all switches without removals.  These don't include the leaf and spine switches
	 * in the same pod.  Those are done in the pod_to_pod_short() function.
	 */
	foreach_active_switch(p_asym->p_subn, p_sw) {
		p_adata = switch_to_adata(p_sw);
		if ((1 < p_adata->layer) && (p_adata->layer < NO_LAYER)) {
			if (!asym_set_contains(&p_algo->scratch, p_sw)) {
				osm_switch_clear_and_get_minhop_ports(p_sw, dst_lid, &dst_ar_group, NULL);
				asym_set_routes(p_asym, p_sw, p_dst_sw, &dst_ar_group);
			}
		}
	}
}

/* ==============================================================================================================
 *
 * Pod section.
 */

__attribute__((unused))
static void asym_pod_dump(IN asym_t *p_asym, IN asym_pod_t *p_pod)
{
	size_t			count;
	size_t			bufsize;
	size_t			bufused;
	asym_iter_t		iter1;
	asym_node_t		*p_anode;
	osm_switch_t		*p_sw;
	char			*buffer;
	boolean_t		log_active;

	/*
	 * Only dump if we are at the proper log level or debugging.
	 */
        log_active = osm_log_is_active_v2(p_asym->p_log, OSM_LOG_DEBUG, FILE_ID);
        if (!asym_debug && !log_active) {
		return;
	}

	count = asym_set_get_size(&p_pod->leaves) + asym_set_get_size(&p_pod->spines);

	bufused = 0;
	bufsize = (128 * count) + 256;

	if ((buffer = asym_calloc(1, bufsize)) != NULL) {
		asym_snprintf("POD %d  (leaf_rank=%d)\n", p_pod->index, p_pod->leaf_rank);

		/*
		 * Dump the leaves.
		 */
		asym_snprintf("\tLEAVES\n");
		foreach_pod_leaf(p_pod, p_sw, iter1) {
			p_anode = switch_to_anode(p_sw);
			asym_snprintf("\t\t%-64s %s\n", sw_name(p_sw), p_anode->is_fully_connected ? "*" : " ");
		}

		/*
		 * Dump the spines.
		 */
		asym_snprintf("\tSPINES\n");
		foreach_pod_spine(p_pod, p_sw, iter1) {
			p_anode = switch_to_anode(p_sw);
			asym_snprintf("\t\t%-64s\n", sw_name(p_sw));
		}

		if (asym_debug) {
			printf("%s\n", buffer);
		}

		if (log_active) {
			asym_log_by_line(p_asym->p_log, OSM_LOG_DEBUG, buffer);
		}

		asym_free(buffer);
	}
}

static asym_pod_t *asym_pod_create(IN asym_t *p_asym, IN osm_switch_t *p_base_sw)
{
	int			i;
	int			j;
	int			last;
	int			count;
	int			max_links;
	int			max_count;
	int			num_spines;
	int			*scratch;
	boolean_t		leaves_status;
	boolean_t		spines_status;
	asym_pod_t		*p_pod;
	asym_iter_t		iter1;
	asym_node_t		*p_anode;
	flow_edge_t		*p_edge;
	flow_vertex_t		*p_vertex;
	osm_switch_t		*p_sw;

	/*
	 * Create the pod.
	 */
	if ((p_pod = asym_calloc(1, sizeof(asym_pod_t))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B103: allocation failed for pods\n");
		return NULL;
	}

	p_asym->num_pods++;
	p_pod->index = p_asym->num_pods;
	p_pod->leaf_rank = p_base_sw->rank;

	/*
	 * Create the leaf and spine sets.
	 */
	leaves_status = asym_set_create(&p_pod->leaves);
	spines_status = asym_set_create(&p_pod->spines);
	if (!leaves_status || !spines_status) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B104: pod set creation failed\n");

		if (!leaves_status) {
			asym_set_destroy(&p_pod->leaves);
		}

		if (!spines_status) {
			asym_set_destroy(&p_pod->spines);
		}

		asym_free(p_pod);
		return NULL;
	}

	/*
	 * Get all of the nodes with the same coordinate.
	 */
	foreach_active_switch(p_asym->p_subn, p_sw) {
		p_anode = switch_to_anode(p_sw);
		if (p_sw->coord == p_base_sw->coord) {
			p_anode->pod = p_pod->index;

			if (p_sw->rank == p_base_sw->rank) {
				asym_set_insert(&p_pod->leaves, p_sw);
			} else {
				asym_set_insert(&p_pod->spines, p_sw);
			}
		}
	}

	/*
	 * Allocate scratch space for counting links.  If the allocation fails,
	 * we assume that there are no fully connected leaves and return.
	 */
	max_links = 0;
	foreach_pod_leaf(p_pod, p_sw, iter1) {
		p_anode = switch_to_anode(p_sw);

		if (p_anode->num_links > max_links) {
			max_links = p_anode->num_links;
		}
	}

	if ((scratch = asym_calloc(1, max_links * sizeof(int))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B105: allocation failure for scratch area\n");
		return p_pod;
	}

	/*
	 * Find the fully connected leaves.
	 */
	max_count = 0;
	foreach_pod_leaf(p_pod, p_sw, iter1) {
		p_vertex = switch_to_vertex(p_sw);

		i = 0;
		memset(scratch, 0, max_links * sizeof(int));
		foreach_vertex_edge(p_asym->p_graph, p_vertex, p_edge) {
			scratch[i++] = p_edge->adj_index;
		}

		qsort(scratch, i, sizeof(int), asym_int_cmp);

		count = 0;
		last = UINT_MAX;
		for (j = 0; j < i; j++) {
			count = (scratch[j] == last) ? count + 1 : 1;
			last = scratch[j];
			max_count = (count > max_count) ? count : max_count;
		}
	}

	asym_free(scratch);

	/*
	 * We know the maximum count of links between a leaf and spine.
	 * To be fully connected, the number of connections needs to be
	 * (max_count * num_spines).
	 */
	num_spines = asym_set_get_size(&p_pod->spines) - 1;

	foreach_pod_leaf(p_pod, p_sw, iter1) {
		p_anode = switch_to_anode(p_sw);
		p_anode->is_fully_connected = p_anode->num_links == (max_count * num_spines);
		if (p_anode->is_fully_connected) {
			p_pod->p_fc_sw = p_sw;
		}
	}

	asym_pod_dump(p_asym, p_pod);
	return p_pod;
}

static int asym_pods_create(IN asym_t *p_asym)
{
	int			rank;
	int			status;
	osm_switch_t		*p_sw;
	flow_vertex_t		*p_vertex;
	asym_node_t		*p_anode;
	asym_pod_t		*p_pod;

	status = 0;
	for (rank = p_asym->max_rank; rank > 0; rank--) {
		foreach_active_vertex(p_asym, p_vertex) {
			p_sw = p_vertex->p_sw;
			if ((p_sw->rank != rank) || (p_sw->coord == OSM_SW_NO_COORD)) {
				continue;
			}

			p_anode = switch_to_anode(p_sw);
			if (p_anode->pod == 0) {
				if ((p_pod = asym_pod_create(p_asym, p_sw)) == NULL) {
					status = -1;
					goto exit;
				}

				cl_qmap_insert(&p_asym->pod_map, p_pod->index, (cl_map_item_t *)p_pod);
			}
		}
	}

exit:
	return status;
}

static void asym_pods_destroy(IN asym_t *p_asym)
{
	cl_map_item_t		*p_item;
	cl_qmap_t		*p_map;
	asym_pod_t		*p_pod;
	asym_set_t		*p_rank;

	p_map = &p_asym->pod_map;
	while ((p_item = cl_qmap_head(p_map)) != cl_qmap_end(p_map)) {
		p_pod = (asym_pod_t *)p_item;

		asym_set_destroy(p_rank = &p_pod->leaves);
		asym_set_destroy(p_rank = &p_pod->spines);

		cl_qmap_remove(p_map, p_pod->index);
		asym_free(p_pod);
	}

	cl_qmap_remove_all(&p_asym->pod_map);
}

/* ==============================================================================================================
 *
 * Utility section.
 */

__attribute__((unused))
static void asym_bfs_dump(IN asym_algorithm_t *p_algo)
{
	int			i;
	int			v;
	size_t			bufsize;
	size_t			bufused;
	uint8_t			d;
	uint8_t			d0;
	uint8_t			*distances;
	uint16_t		*queue;
	flow_vertex_t		*p_vertex;
	osm_switch_t		*p_sw;
	asym_node_t		*p_anode;
	asym_t			*p_asym;
	char			*buffer;
	boolean_t		log_active;

	/*
	 * Only dump if we are at the proper log level or debugging.
	 */
        log_active = osm_log_is_active_v2(p_algo->p_asym->p_log, OSM_LOG_DEBUG, FILE_ID);
        if (!asym_debug && !log_active) {
		return;
	}

	d0 = OSM_NO_PATH;
	bufused= 0;
	bufsize = (192 * p_algo->q_max) + 256;

	if ((buffer = asym_calloc(1, bufsize)) != NULL) {
		p_asym = p_algo->p_asym;
		distances = p_algo->distances;
		queue = p_algo->queue;

		asym_snprintf("BFS from POD %d -> POD %d\n", p_algo->p_dst_pod->index, p_algo->p_src_pod->index);
		asym_snprintf("         i    v  d :  rank   isolated  switch\n");

		for (i = 0; i < p_algo->q_max; i++) {
			v = queue[i];
			d = distances[v];
			p_vertex = &p_asym->p_graph->vertices[v];
			p_sw = p_vertex->p_sw;
			p_anode = switch_to_anode(p_sw);

			if (d != d0) {
				d0 = d;
				asym_snprintf("------------------------------------------------------------------------\n");
			}

			asym_snprintf("      %4d %4d %2d :   %d        %d      %s %s\n",
				i,
				v,
				d,
				p_sw->rank,
				asym_switch_is_isolated(p_asym->p_subn, p_sw),
				sw_name(p_sw),
				p_anode->has_hcas ? "***" : "");
		}

		if (asym_debug) {
			printf("%s\n", buffer);
		}

		if (log_active) {
			asym_log_by_line(p_asym->p_log, OSM_LOG_DEBUG, buffer);
		}

		asym_free(buffer);
	}
}

static void asym_ar_bfs(IN asym_algorithm_t *p_algo)
{
	int			i;
	int			v;
	int			up;
	int			l_adj;
	int			q_max;
	int			src_rank;
	int			dst_rank;
	int			dir_rank;
	int			hop_count;
	int			curr_rank;
	int			next_rank;
	uint8_t			*distances;
	uint8_t			*visited;
	uint8_t			hop0;
	uint8_t			hop1;
	uint8_t			hop2;
	uint16_t		src_lid;
	uint16_t		dst_lid;
	uint16_t		leaf_lid;
	uint16_t		remote_lid;
	uint16_t		*queue;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_src_leaf;
	osm_switch_t		*p_dst_leaf;
	osm_switch_t		*p_curr_sw;
	osm_switch_t		*p_remote_sw;
	flow_vertex_t		*p_vertex;
	flow_edge_t		*p_edge;
	asym_adata_t		*p_adata;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_pod_t		*p_src_pod;
	asym_pod_t		*p_dst_pod;
	asym_t			*p_asym;

	p_src_pod = p_algo->p_src_pod;
	p_dst_pod = p_algo->p_dst_pod;

	p_asym = p_algo->p_asym;
	distances = p_algo->distances;
	visited = p_algo->visited;
	queue = p_algo->queue;

	memset(distances, 0x00, p_algo->distances_size);
	memset(visited, 0x00, p_algo->visited_size);
	memset(queue, 0x00, p_algo->queue_size);

	q_max = 0;

	foreach_pod_leaf(p_algo->p_dst_pod, p_sw, iter1) {
		p_adata = switch_to_adata(p_sw);
		p_adata->state = NODE_UP;
		p_adata->hops = 0;

		distances[switch_to_index(p_sw)] = 0;
		visited[switch_to_index(p_sw)] = 1;
		queue[q_max++] = switch_to_index(p_sw);
	}

	/*
	 * Calculate the rank of the turn between going up and going down.
	 */
	hop_count = OSM_NO_PATH;
	if ((p_src_pod->p_fc_sw != NULL) && (p_dst_pod->p_fc_sw != NULL)) {
		dst_lid = osm_node_get_base_lid_ho(p_dst_pod->p_fc_sw->p_node, 0);
		hop_count = osm_switch_get_least_hops(p_src_pod->p_fc_sw, dst_lid);
	} else if (p_src_pod->p_fc_sw != NULL) {
		foreach_pod_leaf(p_algo->p_dst_pod, p_dst_leaf, iter2) {
			dst_lid = osm_node_get_base_lid_ho(p_dst_leaf->p_node, 0);
			hop0 = osm_switch_get_least_hops(p_src_pod->p_fc_sw, dst_lid);
			hop_count = asym_min(hop_count, hop0);
		}
	} else if (p_dst_pod->p_fc_sw != NULL) {
		foreach_pod_leaf(p_algo->p_src_pod, p_src_leaf, iter1) {
			src_lid = osm_node_get_base_lid_ho(p_src_leaf->p_node, 0);
			hop0 = osm_switch_get_least_hops(p_dst_pod->p_fc_sw, src_lid);
			hop_count = asym_min(hop_count, hop0);
		}
	} else {
		foreach_pod_leaf(p_algo->p_src_pod, p_src_leaf, iter1) {
			foreach_pod_leaf(p_algo->p_dst_pod, p_dst_leaf, iter2) {
				dst_lid = osm_node_get_base_lid_ho(p_dst_leaf->p_node, 0);
				hop0 = osm_switch_get_least_hops(p_src_leaf, dst_lid);
				hop_count = asym_min(hop_count, hop0);
			}
		}
	}

	src_rank = p_algo->p_src_pod->leaf_rank;
	dst_rank = p_algo->p_dst_pod->leaf_rank;
	dir_rank = ((src_rank + dst_rank) - hop_count) / 2;

	up = TRUE;
	for (i = 0; i < q_max; i++) {
		v = queue[i];
		hop_count = distances[v];
		p_vertex = &p_asym->p_graph->vertices[v];
		p_curr_sw = p_vertex->p_sw;
		curr_rank = p_curr_sw->rank;
		if (curr_rank == dir_rank) {
			up = FALSE;
		}

		next_rank = (up) ? curr_rank - 1 : curr_rank + 1;

		/*
		 * Run the tree.
		 */
		foreach_vertex_edge(p_asym->p_graph, p_vertex, p_edge) {
			l_adj = p_edge->adj_index;

			if (visited[l_adj] != 0) {
				continue;
			}

			p_remote_sw = p_asym->p_graph->vertices[l_adj].p_sw;

			/*
			 * If the remote switch is isolated or the rank isn't appropriate
			 * for the distance from the source, drop it.
			 */
			if (asym_switch_is_isolated(p_asym->p_subn, p_remote_sw)) {
				continue;
			}

			if (p_remote_sw->rank != next_rank) {
				continue;
			}

			remote_lid = osm_node_get_base_lid_ho(p_remote_sw->p_node, 0);

			if ((hop1 = osm_switch_get_least_hops(p_curr_sw, remote_lid)) == OSM_NO_PATH) {
				continue;
			}

			/*
			 * Determine if the current switch can reach ANY src leaf in less hops.
			 */
			foreach_pod_leaf(p_src_pod, p_src_leaf, iter1) {
				leaf_lid = osm_node_get_base_lid_ho(p_src_leaf->p_node, 0);

				/*
				 * Check if the node is one closer to the src leaf.
				 */
				if ((hop0 = osm_switch_get_least_hops(p_curr_sw, leaf_lid)) == OSM_NO_PATH) {
					continue;
				}

				if ((hop2 = osm_switch_get_least_hops(p_remote_sw, leaf_lid)) == OSM_NO_PATH) {
					continue;
				}

				if (hop2 != hop0-1) {
					continue;
				}

				/*
				 * Add this node to the queue.  The current node under consideration has a path
				 * to at least one src pod leaf.
				 */
				visited[l_adj] = 1;
				distances[l_adj] = hop_count + 1;
				queue[q_max++] = l_adj;
				break;
			}
		}
	}

	p_algo->q_max = q_max;
	asym_bfs_dump(p_algo);
}

/* ==============================================================================================================
 *
 * Link section.
 */

static void asym_link_remove(IN asym_algorithm_t *p_algo, IN asym_alink_t *p_alink, IN char *reason)
{
	if (asym_debug) {
		asym_print_remove(p_algo, p_alink, reason);
	}

	/*
	 * Remove the link in question.
	 */
	p_alink->state = LINK_REMOVED;
	if (!asym_set_contains(&p_algo->removals, p_alink)) {
		asym_set_insert(&p_algo->removals, p_alink);
	}
}

static void asym_link_fini_remove_from_spine(IN asym_algorithm_t *p_algo, IN asym_layer_t *curr_layer, IN asym_layer_t *next_layer, IN asym_adata_t *p_adata)
{
	int64_t			ceil_weight;
	weight_t		incoming_weight;
	weight_t		cand_weight;
	boolean_t		flag;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_adata_t		*p_cdata;
	asym_adata_t		*p_ndata;
	asym_adata_t		*p_acand;
	asym_alink_t		*p_alink;
	osm_switch_t		*p_sw;

	p_acand = NULL;
	cand_weight = W(0);

	/*
	 * Find the node in the src_pod to remove a link from.
	 */
	foreach_layer_switch(next_layer, p_sw, iter1) {
		p_ndata = switch_to_adata(p_sw);

		incoming_weight = W(0);

		flag = FALSE;
		foreach_incoming_link(p_ndata, p_alink, iter2) {
			if (link_is_up(p_alink)) {
				p_cdata = alink_remote_adata(p_ndata, p_alink);
				flag |= (p_cdata == p_adata);
				incoming_weight = W_add(incoming_weight, p_cdata->node_weight);
			}
		}

		if (!flag) {
			continue;
		}


		ceil_weight = W_ceil(incoming_weight);
		if (ceil_weight < (int64_t)p_ndata->in_count) {
			if (p_acand == NULL) {
				p_acand = p_ndata;
				cand_weight = incoming_weight;
			} else if (W_gt(incoming_weight, cand_weight)) {
				p_acand = p_ndata;
				cand_weight = incoming_weight;
			} else if ((W_eq(incoming_weight, cand_weight)) && (node_choose(p_acand, p_ndata) != p_acand)) {
				p_acand = p_ndata;
				cand_weight = incoming_weight;
			}
		}
	}

	/*
	 * We have the node, now find the link to remove.
	 */
	if (p_acand != NULL) {
		foreach_outgoing_link(p_adata, p_alink, iter2) {
			p_ndata = alink_remote_adata(p_adata, p_alink);
			if (p_ndata == p_acand) {
				asym_link_remove(p_algo, p_alink, "fini_spine");
				break;
			}
		}
	}
}

static void asym_link_fini_remove_from_leaf(IN asym_algorithm_t *p_algo,
					    IN asym_layer_t *curr_layer,
					    IN asym_layer_t *next_layer,
					    IN asym_adata_t *p_adata,
					    IN int quasi)
{
	char			port_num[8];
	char			port_num_cand[8];
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_t			*p_asym;
	asym_node_t		*p_cnode;
	asym_adata_t		*p_cdata;
	asym_adata_t		*p_acand;
	asym_alink_t		*p_alink;
	asym_alink_t		*p_clink;
	osm_switch_t		*p_sw;

	p_asym = p_algo->p_asym;
	p_acand = NULL;

	/*
	 * Sort the spines in order of node weights (least first)
	 */
	asym_set_remove_all(&p_algo->scratch);

	foreach_layer_switch(curr_layer, p_sw, iter1) {
		p_cnode = switch_to_anode(p_sw);
		p_cdata = switch_to_adata(p_sw);
		if ((p_cnode->quasi == quasi) && node_is_up(p_cdata)) {
			asym_set_insert(&p_algo->scratch, p_cdata);
		}
	}

	asym_set_sort(&p_algo->scratch, asym_fini_leaf_cmp);

	/*
	 * Find the first node and link which is connected to this node and remove the link.
	 */
	foreach_set_adata(&p_algo->scratch, p_cdata, iter1) {
		p_clink = NULL;
		memset(port_num, '\0', sizeof(port_num));
		foreach_incoming_link(p_adata, p_alink, iter2) {
			p_acand = alink_remote_adata(p_adata, p_alink);
			sprintf(port_num_cand, "%d", alink_local_port_num(p_adata, p_alink));

			if (p_acand == p_cdata) {
				if (p_clink == NULL) {
					p_clink = p_alink;
					strcpy(port_num, port_num_cand);
				} else if (strcmp(port_num_cand, port_num) < 0) {
					p_clink = p_alink;
					strcpy(port_num, port_num_cand);
				}
			}
		}

		if (p_clink != NULL) {
			asym_link_remove(p_algo, p_clink, "fini_leaf");
			break;
		}
	}
}

static weight_t asym_link_weight(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	weight_t		incoming_weight;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	incoming_weight = W(0);

	foreach_link(p_adata, p_alink, iter) {
		if (link_is_incoming(p_adata, p_alink) && link_is_up(p_alink)) {
			incoming_weight = W_add(incoming_weight, p_alink->lw);
		}
	}

	return incoming_weight;
}

static void asym_link_counts_and_weight(IN asym_algorithm_t *p_algo,
					IN asym_adata_t *p_adata,
					OUT int *in_count,
					OUT int *out_count,
					IN weight_t *in_weight)
{
	int			incoming_count;
	int			outgoing_count;
	weight_t		incoming_weight;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	incoming_count = 0;
	outgoing_count = 0;
	incoming_weight = W(0);

	foreach_link(p_adata, p_alink, iter) {
		if (link_is_incoming(p_adata, p_alink) && link_is_up(p_alink)) {
			incoming_count++;
			incoming_weight = W_add(incoming_weight, p_alink->lw);
		} else if (link_is_outgoing(p_adata, p_alink) && link_is_up(p_alink)) {
			outgoing_count++;
		}
	}

	*in_count = incoming_count;
	*out_count = outgoing_count;
	*in_weight = incoming_weight;
}

static void asym_links_reset(IN asym_algorithm_t *p_algo)
{
	asym_alink_t		*p_alink;

	foreach_alink(p_algo, p_alink) {
		p_alink->state = LINK_NA;
	}
}

/* ==============================================================================================================
 *
 * Node section.
 */

static boolean_t asym_node_is_balanced(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	weight_t		link_weight;
	weight_t		zero_weight;
	boolean_t		balanced;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	zero_weight = W(0);
	link_weight = W(-1);
	balanced = TRUE;
	foreach_incoming_link(p_adata, p_alink, iter) {
		if (W_lt(link_weight, zero_weight)) {
			link_weight = p_alink->lw;
		}

		if (W_gt(p_alink->lw, zero_weight)) {
			balanced &= W_eq(link_weight, p_alink->lw);
		}
	}

	return balanced;
}

__attribute__((unused))
static boolean_t asym_node_is_not_balanced(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	return !asym_node_is_balanced(p_algo, p_adata);
}

static void asym_node_link_counts(IN asym_algorithm_t *p_algo,
				  IN asym_adata_t *p_adata,
				  OUT int *in_count,
				  OUT int *out_count)
{
	int			incoming_count;
	int			outgoing_count;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	incoming_count = 0;
	outgoing_count = 0;

	foreach_link(p_adata, p_alink, iter) {
		if (link_is_incoming(p_adata, p_alink) && link_is_up(p_alink)) {
			incoming_count++;
		} else if (link_is_outgoing(p_adata, p_alink) && link_is_up(p_alink)) {
			outgoing_count++;
		}
	}

	*in_count = incoming_count;
	*out_count = outgoing_count;
}

static void asym_node_calculate_link_weight(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	int			incoming_count;
	int			outgoing_count;
	weight_t		incoming_weight;
	weight_t		link_weight;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	asym_link_counts_and_weight(p_algo, p_adata, &incoming_count, &outgoing_count, &incoming_weight);

	link_weight = (outgoing_count != 0) ? W_div(incoming_weight, outgoing_count) : W(0);

	foreach_outgoing_link(p_adata, p_alink, iter) {
		p_alink->lw = link_weight;
	}
}

static void asym_node_calculate_weight(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	int			incoming_count;
	int			outgoing_count;
	weight_t		incoming_weight;

	/*
	 * Get the counts and weights.  If there are no outgoing links (due to removals),
	 * then the node weight is set to 0.  This node is now dropped from consideration for
	 * the algorithm.
	 */
	asym_link_counts_and_weight(p_algo, p_adata, &incoming_count, &outgoing_count, &incoming_weight);
	p_adata->nw = (outgoing_count != 0) ? W_div(incoming_weight, outgoing_count) : W(0);

}

static void asym_node_set_counts_and_weight(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	int			incoming_count;
	int			outgoing_count;
	weight_t		incoming_weight;

	asym_link_counts_and_weight(p_algo, p_adata, &incoming_count, &outgoing_count, &incoming_weight);

	p_adata->in_count = incoming_count;
	p_adata->out_count = outgoing_count;
	p_adata->node_weight = incoming_weight;
	p_adata->link_weight = (outgoing_count != 0) ? W_div(incoming_weight, outgoing_count) : W(0);
}

static void asym_node_calculate_threshold(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	if (p_adata->out_count != 0) {
		p_adata->threshold = T(p_adata->in_count, p_adata->out_count);
	} else {
		p_adata->threshold = T(0,100);
	}
}

static void asym_node_exclude(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	asym_iter_t		iter;
	asym_alink_t		*p_alink;
	osm_switch_t		*p_sw;
	uint64_t		guid;

	/*
	 * We exclude the node by removing all of its outgoing links.
	 */
	OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_VERBOSE, "destination %s : excluding switch %s from AR group\n",
		sw_name(p_algo->p_dst_sw),
		adata_name(p_adata));

	/*
	 * Remove all of the links from this node.
	 */
	foreach_link(p_adata, p_alink, iter) {
		if (link_is_up(p_alink)) {
			asym_link_remove(p_algo, p_alink, "threshold");
		} else if (link_is_na(p_alink)) {
			asym_link_remove(p_algo, p_alink, "threshold");
		}
	}

	p_sw = adata_to_switch(p_adata);
	guid = osm_node_get_node_guid_ho(p_sw->p_node);
	if (!asym_set_contains(&p_algo->excluded_spines, (void *)guid)) {
		asym_set_insert(&p_algo->excluded_spines, (void *)guid);
	}

}

/* ==============================================================================================================
 *
 * Layer section.
 */

static int asym_layer_initialize(IN asym_algorithm_t *p_algo, IN asym_layer_t *p_layer)
{
	int			count;
	int			incoming_count;
	size_t			layer_number;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_adata_t		*p_adata;
	asym_adata_t		*p_rdata;
	asym_alink_t		*p_alink;
	osm_switch_t		*p_sw;

	p_asym = p_algo->p_asym;
	asym_set_sort(p_layer, asym_sw_name_cmp);

        layer_number = p_layer - p_algo->layers;

	/*
	 * Only the destination switch is UP in the first layer.  Set its outgoing
	 * links to UP to feed the next layer.
	 */
	if (layer_number == 0) {
		count = 0;

		foreach_layer_switch(p_layer, p_sw, iter1) {
			p_adata = switch_to_adata(p_sw);
			p_adata->state = (p_sw == p_algo->p_dst_sw) ? NODE_UP : NODE_NA;
			p_adata->nw = W(0);

			if (p_adata->state == NODE_UP) {
				foreach_link(p_adata, p_alink, iter2) {
					if (link_is_outgoing(p_adata, p_alink)) {
						if (asym_link_is_isolated(p_asym, p_alink)) {
							p_alink->state = LINK_NA;
						} else {
							p_alink->state = LINK_UP;
							p_alink->lw = W(1);
							count++;
						}
					}
				}
			}
		}

		return count;
	}

	/*
	 * Set nodes and links UP for determining paths.
	 */
	foreach_layer_switch(p_layer, p_sw, iter1) {
		/*
		 * Assume the node is NA.
		 */
		p_adata = switch_to_adata(p_sw);
		p_adata->nw = W(0);
		p_adata->state = NODE_NA;
		p_adata->balanced = FALSE;

		/*
		 * If the incoming links are from nodes that are NA, then
		 * set the link NA.  Otherwise, set the link UP.
		 */
		incoming_count = 0;
		foreach_link(p_adata, p_alink, iter2) {
			if (link_is_incoming(p_adata, p_alink)) {
				p_rdata = alink_remote_adata(p_adata, p_alink);

				if (asym_link_is_isolated(p_asym, p_alink)) {
					p_alink->state = LINK_NA;
				} else if (node_is_na(p_rdata)) {
					p_alink->state = LINK_NA;
				} else {
					incoming_count++;
					p_alink->state = LINK_UP;
					p_alink->lw = W(0);
				}
			}
		}

		/*
		 * If this node has incoming links, it is connected, so set it UP.
		 */
		if (incoming_count != 0) {
			p_adata->state = NODE_UP;
		}

		/*
		 * If this node is UP, then turn its outgoing links UP.
		 */
		if (node_is_up(p_adata)) {
			foreach_link(p_adata, p_alink, iter2) {
				if (link_is_outgoing(p_adata, p_alink)) {
					p_alink->state = LINK_UP;
					p_alink->lw = W(0);
				}
			}
		}
	}

	return 1;
}

static void asym_layer_calculate_link_weights(IN asym_algorithm_t *p_algo, IN asym_set_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_node_calculate_link_weight(p_algo, p_adata);
	}
}

static void asym_layer_calculate_node_weights(IN asym_algorithm_t *p_algo, IN asym_set_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_node_calculate_weight(p_algo, p_adata);
	}
}

static void asym_layer_calculate_all_weights(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{

	/*
	 * Recalculate the node and link weights for this layer.
	 */
	asym_layer_calculate_node_weights(p_algo, curr_layer);
	asym_layer_calculate_link_weights(p_algo, curr_layer);

	/*
	 * The links between this layer and next layer have possibly changed.  Therefore, we need
	 * to recalculate the node weights for the next layer.
	 */
	asym_layer_calculate_node_weights(p_algo, next_layer);
}

static void asym_layer_set_counts_and_weight(IN asym_algorithm_t *p_algo, IN asym_layer_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_node_set_counts_and_weight(p_algo, p_adata);
	}
}

static boolean_t asym_layer_find_unbalanced(IN asym_algorithm_t *p_algo, IN asym_set_t *p_layer, OUT asym_set_t *p_unbalanced)
{
	asym_iter_t		iter1;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	/*
	 * Find all of the unbalanced nodes.
	 */
	asym_set_remove_all(p_unbalanced);

	foreach_layer_switch(p_layer, p_sw, iter1) {
		p_adata = switch_to_adata(p_sw);
		p_adata->balanced = asym_node_is_balanced(p_algo, p_adata);

		if (!p_adata->balanced) {
			asym_set_insert(p_unbalanced, p_adata);
		}
	}

	/*
	 * Sort them in reverse node weight order.
	 */
	asym_set_sort(p_unbalanced, asym_highest_weight_cmp);

	return asym_set_is_not_empty(p_unbalanced);
}

static boolean_t asym_layer_rebalance(IN asym_algorithm_t *p_algo, IN asym_set_t *p_unbalanced)
{
	int			count;
	asym_adata_t		*p_adata;
	asym_adata_t		*p_rdata;
	asym_adata_t		*p_cdata;
	asym_alink_t		*p_alink;
	asym_alink_t		*p_rlink;
	asym_alink_t		*p_clink;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_iter_t		iter3;

	/*
	 * We need to find the node with a link we can remove.  We can remove a link if the node
	 * is unbalanced (already true) and if it has a link which is NOT the only link between
	 * a node on the curr_layer and this node.
	 */
	p_cdata = NULL;
	p_clink = NULL;

	foreach_set_adata(p_unbalanced, p_adata, iter1) {
		foreach_incoming_link(p_adata, p_alink, iter2) {
			count = 0;
			p_rdata = alink_remote_adata(p_adata, p_alink);
			foreach_outgoing_link(p_rdata, p_rlink, iter3) {
				count++;
			}

			if (count < 2) {	// can't remove only link between these nodes.
				continue;
			}

			if (p_clink == NULL) {
				p_clink = p_alink;
				p_cdata = p_rdata;
			} else if (W_lt(p_alink->lw, p_clink->lw)) {
				p_clink = p_alink;
				p_cdata = p_rdata;
			} else if (W_eq(p_alink->lw, p_clink->lw)) {
				p_clink = link_choose(p_algo, p_adata, p_alink, p_clink);
				p_cdata = (p_clink == p_alink) ? p_rdata : p_cdata;
			}
		}

		if (p_clink != NULL) {
			asym_link_remove(p_algo, p_clink, "balance");
			return TRUE;
		}
	}

	return FALSE;
}

static void asym_layer_fini_spines(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	int			i;
	int			links_to_remove;
	asym_iter_t		iter1;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	OSM_LOG_ENTER(p_algo->p_asym->p_log);

	/*
	 * Calculate the weights.
	 */
	asym_layer_calculate_all_weights(p_algo, curr_layer, next_layer);

	/*
	 * Sort the spines in order of node weights (greatest first)
	 */
	asym_set_remove_all(&p_algo->scratch);

	foreach_layer_switch(curr_layer, p_sw, iter1) {
		p_adata = switch_to_adata(p_sw);
		if (node_is_up(p_adata)) {
			asym_set_insert(&p_algo->scratch, p_adata);
		}
	}

	asym_set_sort(&p_algo->scratch, asym_fini_cmp);

	/*
	 * Remove links from nodes with more incoming links than outgoing links.
	 */
	foreach_set_adata(&p_algo->scratch, p_adata, iter1) {
		asym_layer_set_counts_and_weight(p_algo, curr_layer);
		asym_layer_set_counts_and_weight(p_algo, next_layer);

		links_to_remove = p_adata->out_count - p_adata->in_count;
		if (links_to_remove > 0) {
			for (i = 0; i < links_to_remove; i++) {
				asym_link_fini_remove_from_spine(p_algo, curr_layer, next_layer, p_adata);
				asym_layer_calculate_all_weights(p_algo, curr_layer, next_layer);
			}
		}
	}

	OSM_LOG_EXIT(p_algo->p_asym->p_log);
}

static void asym_layer_fini_leaves(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	int			i;
	int			quasi;
	int			quasi_in_count;
	int			quasi_weight;
	int			links_to_remove;
	weight_t		incoming_weight;
	weight_t		quasi_incoming_weight;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_node_t		*p_anode;
	asym_adata_t		*p_adata;
	asym_alink_t		*p_alink;
	osm_switch_t		*p_sw;

	p_asym = p_algo->p_asym;
	OSM_LOG_ENTER(p_algo->p_asym->p_log);

	/*
	 * Remove links from nodes with more outgoing links than incoming layer weight.
	 */
	for (quasi = 1; quasi <= p_asym->num_quasis; quasi++) {
		quasi_incoming_weight = W(0);
		foreach_layer_switch(curr_layer, p_sw, iter1) {
			p_anode = switch_to_anode(p_sw);
			if (p_anode->quasi != quasi) {
				continue;
			}

			p_adata = switch_to_adata(p_sw);

			incoming_weight = asym_link_weight(p_algo, p_adata);
			quasi_incoming_weight = W_add(quasi_incoming_weight, incoming_weight);
		}

		quasi_weight = W_ceil(quasi_incoming_weight);
		foreach_set_switch(next_layer, p_sw, iter1) {
			p_adata = switch_to_adata(p_sw);
			asym_layer_set_counts_and_weight(p_algo, curr_layer);

			quasi_in_count = 0;
			foreach_incoming_link(p_adata, p_alink, iter2) {
				p_anode = alink_remote_anode(p_adata, p_alink);
				if (p_anode->quasi == quasi) {
					quasi_in_count++;
				}
			}

			links_to_remove = quasi_in_count - quasi_weight;

			if (links_to_remove > 0) {
				for (i = 0; i < links_to_remove; i++) {
					asym_layer_set_counts_and_weight(p_algo, curr_layer);
					asym_layer_set_counts_and_weight(p_algo, next_layer);
					asym_link_fini_remove_from_leaf(p_algo, curr_layer, next_layer, p_adata, quasi);
					asym_layer_calculate_all_weights(p_algo, curr_layer, next_layer);
				}
			}
		}
	}

	OSM_LOG_EXIT(p_algo->p_asym->p_log);
}

static void asym_layer_calculate_node_thresholds(IN asym_algorithm_t *p_algo, IN asym_layer_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_node_calculate_threshold(p_algo, p_adata);
	}
}

/* --------------------------------------------------------------------------------------------------------------------- */

static void asym_threshold(IN asym_algorithm_t *p_algo, IN asym_set_t *p_removals)
{
	int			i;
	int			count;
	int			count_limit;
	int			min_distance;
	uint16_t		dst_lid;
	asym_t			*p_asym;
	asym_set_t		*p_src_spines;
	asym_iter_t		iter;
	asym_iter_t		iter1;
	asym_adata_t		*p_adata;
	asym_layer_t		*p_leaf_layer;
	asym_layer_t		*p_spine_layer;
	threshold_t		max_threshold;
	threshold_t		subn_threshold;
	threshold_t		cutoff_threshold;
	osm_switch_t		*p_sw;
	osm_subn_t		*p_subn;

	p_asym = p_algo->p_asym;
	p_subn = p_asym->p_subn;
	if (p_subn->opt.ar_tree_asymmetric_flow_threshold == 0) {
		return;
	}

	if (p_algo->p_src_pod->leaf_rank <= 1) {
		OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "thresholding is disabled on pods with leaves at rank 1\n");
		return;
	}

	p_src_spines = &p_algo->p_src_pod->spines;

	min_distance = OSM_NO_PATH + 1;
	dst_lid = osm_node_get_base_lid_ho(p_algo->p_dst_sw->p_node, 0);

	subn_threshold = T(100 - p_subn->opt.ar_tree_asymmetric_flow_threshold, 100);
	count_limit = p_subn->opt.ar_tree_asymmetric_flow_threshold_limit;

	/*
	 * Setup the link counts.
	 */
	p_spine_layer = index_to_layer(p_algo->num_layers - 2);
	p_leaf_layer = index_to_layer(p_algo->num_layers - 1);

	asym_layer_set_counts_and_weight(p_algo, p_spine_layer);
	asym_layer_set_counts_and_weight(p_algo, p_leaf_layer);

	foreach_pod_spine(p_algo->p_src_pod, p_sw, iter1) {
		min_distance = asym_min(min_distance, osm_switch_get_least_hops(p_sw, dst_lid));
	}

	/*
	 * Find the max threshold and sort the nodes by their thresholds.
	 */
	asym_layer_calculate_node_thresholds(p_algo, p_src_spines);
	asym_set_remove_all(&p_algo->scratch);

	max_threshold = T(0, 100);
	foreach_pod_spine(p_algo->p_src_pod, p_sw, iter) {
		if (osm_switch_get_least_hops(p_sw, dst_lid) == min_distance) {
			p_adata = switch_to_adata(p_sw);
			asym_set_insert(&p_algo->scratch, p_adata);

			if (T_gt(p_adata->threshold, max_threshold)) {
				max_threshold = p_adata->threshold;
			}
		}
	}

	asym_set_sort(&p_algo->scratch, asym_threshold_cmp);

	/*
	 * Find the nodes which don't exceed the cutoff.
	 */
	cutoff_threshold = T_mul(max_threshold, subn_threshold);

	count = 0;
	foreach_set_adata(&p_algo->scratch, p_adata, iter) {
		if (T_lt(p_adata->threshold, cutoff_threshold)) {
			count++;
		}
	}

	if (count > count_limit) {
		OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "no nodes excluded\n");
		return;
	}

	for (i = 1; i <= count; i++) {
		p_adata = asym_set_get(&p_algo->scratch, i);
		asym_node_exclude(p_algo, p_adata);
	}
}

/* --------------------------------------------------------------------------------------------------------------------- */

static void asym_layer_init(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	osm_switch_t		*p_sw;
	osm_switch_t		*p_spine;
	asym_adata_t		*p_adata;
	asym_alink_t		*p_alink;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	OSM_LOG_ENTER(p_asym->p_log);

	/*
	 * Initialize spines connected to this leaf.
	 */
	foreach_layer_switch(next_layer, p_spine, iter1) {
		p_adata = switch_to_adata(p_spine);

		foreach_incoming_link(p_adata, p_alink, iter2) {
			p_sw = alink_remote_switch(p_adata, p_alink);
			if (p_sw == p_algo->p_dst_sw) {
				p_alink->lw = W(1);
			}
		}
	}

	asym_layer_calculate_node_weights(p_algo, next_layer);

	OSM_LOG_EXIT(p_asym->p_log);
}

static void asym_layer_balance(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	OSM_LOG_ENTER(p_algo->p_asym->p_log);

	/*
	 * Calculate the weights.
	 */
	asym_layer_calculate_all_weights(p_algo, curr_layer, next_layer);

	/*
	 * Balance the nodes.
	 */
	while (asym_layer_find_unbalanced(p_algo, next_layer, &p_algo->scratch)) {
		if (!asym_layer_rebalance(p_algo, &p_algo->scratch)) {
			break;
		}

		asym_layer_calculate_all_weights(p_algo, curr_layer, next_layer);
	}

	OSM_LOG_EXIT(p_algo->p_asym->p_log);
}

static void asym_layer_bottleneck(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	int			i;
	int			incoming_count;
	int			outgoing_count;
	int			links_to_remove;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_adata_t		*p_adata;
	asym_adata_t		*p_rdata;
	asym_adata_t		*p_acand;
	asym_alink_t		*p_alink;
	asym_alink_t		*p_lcand;
	osm_switch_t		*p_sw;

	OSM_LOG_ENTER(p_algo->p_asym->p_log);

	/*
	 * Calculate the weights.
	 */
	asym_layer_calculate_all_weights(p_algo, curr_layer, next_layer);

	/*
	 * Look at each node to determine if it has bottlenecks.  We then remove
	 * links to resolve the imbalance.
	 */
	asym_set_sort(curr_layer, asym_sw_name_cmp);
	foreach_layer_switch(curr_layer, p_sw, iter1) {
		p_adata = switch_to_adata(p_sw);
		if (node_is_na(p_adata)) {
			continue;
		}

		asym_node_link_counts(p_algo, p_adata, &incoming_count, &outgoing_count);

		if (incoming_count < outgoing_count) {
			links_to_remove = outgoing_count - incoming_count;

			for (i = 0; i < links_to_remove; i++) {
				p_lcand = NULL;
				p_acand = NULL;

				foreach_outgoing_link(p_adata, p_alink, iter2) {
					p_rdata = alink_remote_adata(p_adata, p_alink);

					if (p_acand == NULL) {
						p_lcand = p_alink;
						p_acand = p_rdata;
					} else if (W_gt(p_rdata->nw, p_acand->nw)) {
						p_lcand = p_alink;
						p_acand = p_rdata;
					} else if (W_eq(p_rdata->nw, p_acand->nw)) {
						p_lcand = link_choose(p_algo, p_adata, p_lcand, p_alink);
						p_acand = alink_remote_adata(p_adata, p_lcand);
					}
				}

				if (p_lcand != NULL) {
					asym_link_remove(p_algo, p_lcand, "bottleneck");
					asym_layer_calculate_all_weights(p_algo, curr_layer, next_layer);
				}
			}
		}
	}

	OSM_LOG_EXIT(p_algo->p_asym->p_log);
}

static void asym_layer_both(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	OSM_LOG_ENTER(p_algo->p_asym->p_log);
	asym_layer_balance(p_algo, curr_layer, next_layer);
	asym_layer_bottleneck(p_algo, curr_layer, next_layer);
	OSM_LOG_EXIT(p_algo->p_asym->p_log);
}

static void asym_layer_fini(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	OSM_LOG_ENTER(p_algo->p_asym->p_log);

	asym_layer_fini_spines(p_algo, curr_layer, next_layer);
	asym_layer_fini_leaves(p_algo, curr_layer, next_layer);

	OSM_LOG_EXIT(p_algo->p_asym->p_log);
}

static void asym_layer_null(IN asym_algorithm_t *p_algo, IN asym_set_t *curr_layer, IN asym_set_t *next_layer)
{
	return;
}

/* ==============================================================================================================
 *
 * Algorithm section.
 */

static void asym_algorithm_free(IN void *context)
{
	asym_adata_t		*p_adata;
	asym_algorithm_t	*p_algo;
	asym_layer_t		*p_layer;
	asym_node_t		*p_anode;
	asym_iter_t		iter;
	asym_t			*p_asym;

	p_algo = (asym_algorithm_t *)context;

	if (p_algo != NULL) {
		OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_VERBOSE, "freeing asym_algorithm\n");
		p_asym = p_algo->p_asym;

		if (p_algo->p_adata != NULL) {
			foreach_anode(p_asym, p_anode) {
				p_adata = index_to_adata(p_anode->index);
				asym_set_destroy(&p_adata->links);
			}
		}

		asym_set_destroy(&p_algo->src_removals);
		asym_set_destroy(&p_algo->ar_removals);
		asym_set_destroy(&p_algo->fc_removals);
		asym_set_destroy(&p_algo->removals);
		asym_set_destroy(&p_algo->scratch);
		asym_set_destroy(&p_algo->excluded_spines);

		p_algo->num_layers = MAX_LAYERS;
		foreach_layer(p_layer, iter) {
			asym_set_destroy(p_layer);
		}

		asym_free(p_algo->p_alink);
		asym_free(p_algo->p_adata);
		asym_free(p_algo->distances);
		asym_free(p_algo->visited);
		asym_free(p_algo->queue);

		memset(p_algo, 0, sizeof(asym_algorithm_t));
		asym_free(p_algo);
	}
}

static void *asym_algorithm_init(IN asym_t *p_asym)
{
	int			status = -1;
	asym_iter_t		iter1;
	asym_layer_t		*p_layer = NULL;
	asym_node_t		*p_anode = NULL;
	asym_alink_t		*p_alink = NULL;
	asym_adata_t		*p_adata = NULL;
	asym_adata_t		*p_bdata = NULL;
	asym_algorithm_t	*p_algo = NULL;
	flow_edge_t		*p_edge = NULL;
	osm_switch_t		*p_sw1 = NULL;
	osm_switch_t		*p_sw2 = NULL;

	/*
	 * Allocate the main data structure.
	 */
	if ((p_algo = asym_calloc(1, sizeof(asym_algorithm_t))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B106: allocation failure for p_algo\n");
		goto exit;
	}

	p_algo->p_asym = p_asym;

	/*
	 * Create the sets for the algorithm.
	 */
	if (!asym_set_create(&p_algo->src_removals)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B108: allocation failure for src_removals set\n");
		goto exit;
	}

	if (!asym_set_create(&p_algo->ar_removals)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B11B: allocation failure for ar_removals set\n");
		goto exit;
	}

	if (!asym_set_create(&p_algo->fc_removals)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B109: allocation failure for fc_removals set\n");
		goto exit;
	}

	if (!asym_set_create(&p_algo->removals)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B10A: allocation failure for removals set\n");
		goto exit;
	}

	if (!asym_set_create(&p_algo->scratch)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B10B: allocation failure for scratch set\n");
		goto exit;
	}

	if (!asym_set_create(&p_algo->excluded_spines)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B10C: allocation failure for excluded spines set\n");
		goto exit;
	}

	/*
	 * BFS data structs.
	 */
	p_algo->distances_size = (p_asym->num_switches + 1) * sizeof(uint8_t);
	p_algo->visited_size   = (p_asym->num_switches + 1) * sizeof(uint8_t);
	p_algo->queue_size     = (p_asym->num_switches + 1) * sizeof(uint16_t);

	p_algo->distances = asym_calloc(1, p_algo->distances_size);
	p_algo->visited   = asym_calloc(1, p_algo->visited_size);
	p_algo->queue     = asym_calloc(1, p_algo->queue_size);

	if (any_null(3, p_algo->distances, p_algo->visited, p_algo->queue)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B10D: allocation failure for BFS data\n");
		goto exit;
	}

	/*
	 * Algorithm structs.
	 */
	p_algo->alink_size = (p_asym->num_links + 1) * sizeof(asym_alink_t);
	p_algo->adata_size = (p_asym->num_switches + 1) * sizeof(asym_adata_t);
	p_algo->layer_size = (p_asym->max_layers + 2) * sizeof(asym_layer_t);

	p_algo->p_alink  = asym_calloc(1, p_algo->alink_size);
	p_algo->p_adata  = asym_calloc(1, p_algo->adata_size);

	if (any_null(2, p_algo->p_alink, p_algo->p_adata)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B10E: allocation failure for algorithm data\n");
		goto exit;
	}

	p_algo->p_asym = p_asym;
	p_algo->max_layers = p_asym->max_layers;

	/*
	 * Init the data structs.
	 */
	foreach_anode(p_asym, p_anode) {
		p_adata = index_to_adata(p_anode->index);
		p_adata->index = p_anode->index;
		p_adata->p_sw = p_anode->p_sw;
		p_adata->nw = W(0);

		if (!asym_set_create(&p_adata->links)) {
			OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B10F: node set creation failed\n");
			goto exit;
		}

		p_algo->num_adata++;
	}

	/*
	 * Init the link structs.
	 */
	p_alink = p_algo->p_alink;
	foreach_edge(p_asym->p_graph, p_edge) {
		p_sw1 = p_asym->p_graph->vertices[p_edge->index].p_sw;
		p_sw2 = p_asym->p_graph->vertices[p_edge->adj_index].p_sw;

		if (asym_switch_is_isolated(p_asym->p_subn, p_sw1) || asym_switch_is_isolated(p_asym->p_subn, p_sw2)) {
			continue;
		}

		if (lookup_isolated_port(&p_asym->p_osm->sm, p_sw1->p_node, p_edge->port_num)) {
			continue;
		}

		if (lookup_isolated_port(&p_asym->p_osm->sm, p_sw2->p_node, p_edge->adj_port_num)) {
			continue;
		}

		if (p_edge->index < p_edge->adj_index) {
			p_alink->p_edge = p_edge;
			p_alink->p_algo = p_algo;
			p_alink->lw = W(0);
			p_alink++;

			p_algo->num_alink++;
		}
	}

	/*
	 * Map the data -> links.
	 */
	foreach_alink(p_algo, p_alink) {
		alink_to_adata(p_alink, p_adata, p_bdata);
		asym_set_insert(&p_adata->links, p_alink);
		asym_set_insert(&p_bdata->links, p_alink);
	}

	/*
	 * Create the layers.
	 */
	p_algo->num_layers = MAX_LAYERS;
	foreach_layer(p_layer, iter1) {
		if (!asym_set_create(p_layer)) {
			OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B110: layer set creation failure.\n");
			goto exit;
		}
	}

	status = 0;

exit:
	/*
	 * Cleanup if necessary.
	 */
	if (status < 0) {
		asym_algorithm_free(p_algo);
		p_algo = NULL;
	}

	return p_algo;
}

static void asym_algorithm_reset(IN void *context)
{
	osm_switch_t		*p_sw;
	asym_algorithm_t	*p_algo;
	asym_layer_t		*p_layer;
	asym_adata_t		*p_adata;
	asym_iter_t		iter1;

	/*
	 * Clean up the previous run.
	 */
	p_algo = (asym_algorithm_t *)context;

	if (p_algo != NULL) {
		foreach_layer(p_layer, iter1) {
			asym_set_remove_all(p_layer);
		}

		asym_set_remove_all(&p_algo->src_removals);
		asym_set_remove_all(&p_algo->ar_removals);
		asym_set_remove_all(&p_algo->fc_removals);
		asym_set_remove_all(&p_algo->removals);

		p_algo->p_src_sw = NULL;
		p_algo->p_dst_sw = NULL;

		p_algo->layer_number = 0;
		p_algo->retries = 0;

		p_algo->fully_connected_done = FALSE;

		foreach_active_switch(p_algo->p_asym->p_subn, p_sw) {
			p_adata = switch_to_adata(p_sw);
			p_adata->layer = NO_LAYER;
		}
	}
}

static int asym_algorithm_setup(IN asym_algorithm_t *p_algo, IN osm_switch_t *p_dst_sw, IN asym_pod_t *p_src_pod)
{
	asym_layer_t		*p_layer;
	asym_iter_t		iter;

	/*
	 * Initialize the layers.
	 */
	foreach_active_layer(p_layer, iter) {
		if (asym_layer_initialize(p_algo, p_layer) == 0) {
			return -1;
		}
	}

	return 0;
}

/* ==============================================================================================================
 *
 * Long section.  This code handles the path from LEAF -> SPINE -> CORE -> SPINE -> LEAF.
 */

static int asym_algorithm_pod_to_leaf_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN osm_switch_t *p_dst_sw)
{
	asym_set_t		*curr_layer;
	asym_set_t		*next_layer;
	asym_iter_t		iter1;
	asym_alink_t		*p_alink;
	asym_layer_function_t	p_function;

	p_algo->p_dst_sw = p_dst_sw;

	asym_set_remove_all(&p_algo->src_removals);
	asym_set_remove_all(&p_algo->removals);
	asym_links_reset(p_algo);

	/*
	 * We proceed layer by layer.
	 * At the end of the run, we update the AR groups for this destination switch.
	 */
	if (asym_algorithm_setup(p_algo, p_dst_sw, p_src_pod) != 0) {
		return -1;
	}

	foreach_layer_step(p_algo, curr_layer, next_layer, p_function, iter1) {
		p_function(p_algo, curr_layer, next_layer);
	}

	/*
	 * Add the source specific removals to the general removals.
	 */
	foreach_link_removal(&p_algo->src_removals, p_alink, iter1) {
		if (!asym_set_contains(&p_algo->removals, p_alink)) {
			asym_set_insert(&p_algo->removals, p_alink);
		}
	}

	return 0;
}

static int asym_algorithm_fc_pod_to_leaf_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN osm_switch_t *p_dst_sw)
{
	OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "FC_LONG: POD %d -> %s\n", p_src_pod->index, sw_name(p_dst_sw));

	if (!p_algo->fully_connected_done) {
		if (asym_algorithm_pod_to_leaf_long(p_algo, p_src_pod, p_dst_sw) != 0) {
			return -1;
		}

		p_algo->fully_connected_done = TRUE;

		asym_set_remove_all(&p_algo->fc_removals);
		asym_set_union(&p_algo->fc_removals, &p_algo->removals);

		/*
		 * Look for spine removals.  The removals will be in two sets:
		 *	removals     - those links not attached to a src node
		 *	src_removals - those links attached to a src node
		 *
		 * Update the fc_removals to sweep these up.
		 */
		asym_threshold(p_algo, &p_algo->fc_removals);
		asym_set_union(&p_algo->fc_removals, &p_algo->removals);
		asym_set_union(&p_algo->fc_removals, &p_algo->src_removals);
	} else {
		p_algo->p_dst_sw = p_dst_sw;
	}

	return 0;
}

static int asym_algorithm_pc_pod_to_leaf_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN osm_switch_t *p_dst_sw)
{
	OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "PC_LONG: POD %d -> %s\n", p_src_pod->index, sw_name(p_dst_sw));

	asym_algorithm_pod_to_leaf_long(p_algo, p_src_pod, p_dst_sw);

	/*
	 * Look for spine removals.  The removals will be in two sets:
	 *	removals     - those links not attached to a src node
	 *	src_removals - those links attached to a src node
	 *
	 * Update the removals with the src_removals.
	 */
	asym_threshold(p_algo, &p_algo->removals);
	asym_set_union(&p_algo->removals, &p_algo->src_removals);
	return 0;
}

/* ==============================================================================================================
 *
 * Short section.  This code handles the path from LEAF -> SPINE -> LEAF.
 */

static void asym_short_link_remove(IN asym_algorithm_t *p_algo, IN uint8_t port_num)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	asym_alink_t		*p_alink;

	p_adata = switch_to_adata(p_algo->p_src_sw);

	foreach_link(p_adata, p_alink, iter) {
		if (alink_local_port_num(p_adata, p_alink) == port_num) {
			asym_link_remove(p_algo, p_alink, "short");
			break;
		}
	}
}

static void asym_algorithm_spine_short(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_pod, IN osm_switch_t *p_spine)
{
	uint16_t                dst_lid;
	osm_switch_t		*p_dst_sw;
	osm_ar_subgroup_t       dst_ar_group;
	asym_t                  *p_asym;
	asym_iter_t             iter1;
	asym_iter_t             iter2;
	asym_node_t             *p_dst_anode;
	asym_port_t             *p_dst_aport;

	p_asym = p_algo->p_asym;

	foreach_pod_leaf(p_pod, p_dst_sw, iter1) {
		p_dst_anode = switch_to_anode(p_dst_sw);
		dst_lid = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);

		if (osm_switch_get_least_hops(p_spine, dst_lid) == 1) {
			/*
			 * If this is just one hop, then use all of the links to the destination.
			 */

			osm_switch_clear_and_get_minhop_ports(p_spine, dst_lid, &dst_ar_group, NULL);
			asym_set_routes(p_asym, p_spine, p_dst_sw, &dst_ar_group);
		} else {
			asym_set_lft_entry_port0(p_asym, p_spine, p_dst_sw, NULL, TRUE);

			foreach_asym_port(p_dst_anode, p_dst_aport, iter2) {
				asym_set_static_lft_entry_by_port(p_asym, p_spine, p_dst_aport);
			}
		}
	}

	return;
}

static void asym_algorithm_leaf_short(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_pod, IN osm_switch_t *p_src_sw)
{
	int			src_count;
	int			dst_count;
	int			links_to_use;
	int			links_to_remove;
	int			remove_index;
	int			link_index[256];
	uint8_t			dst_hops;
	uint8_t			src_hops;
	uint8_t			port_num;
	uint8_t			port_nums[256];
	uint16_t		dst_lid;
	uint16_t		spine_lid;
	osm_switch_t		*p_dst_sw;
	osm_switch_t		*p_spine_sw;
	osm_ar_subgroup_t       dst_ar_group;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_iter_t		iter2;

	p_asym = p_algo->p_asym;
	p_algo->p_src_sw = p_src_sw;

	memset(link_index, 0, sizeof(link_index));

	foreach_pod_leaf(p_pod, p_dst_sw, iter1) {
		if (p_dst_sw == p_src_sw) {
			continue;
		}               

		dst_lid = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);
		if (osm_switch_get_least_hops(p_src_sw, dst_lid) != 2) {	// no src -> spine -> dst path, skip
			continue;
		}

		/*
		 * Setup the src -> dst AR group.  Any spine switch which is 1 hop away from the source
		 * and destination is in the AR group.  We balance the number of links from the source
		 * to the spine with the number of links from the spine to the destination.
		 */
		p_algo->p_dst_sw = p_dst_sw;
		osm_switch_clear_and_get_minhop_ports(p_src_sw, dst_lid, &dst_ar_group, NULL);

		foreach_pod_spine(p_pod, p_spine_sw, iter2) {
			spine_lid = osm_node_get_base_lid_ho(p_spine_sw->p_node, 0);

			/*
			 * If this spine isn't an up and down path, skip it.
			 */
			src_hops = osm_switch_get_least_hops(p_src_sw, spine_lid);
			dst_hops = osm_switch_get_least_hops(p_spine_sw, dst_lid);

			if ((src_hops != 1) || (dst_hops != 1)) {
				continue;
			}

			/*
			 * If there are more links going from the src leaf to the spine than there are from the
			 * spine to the dst leaf, we need to remove some to balance out both sides of the spine.
			 * Otherwise, use all of the links from the src leaf to the spine.
			 */
			src_count = (int)osm_switch_get_minhop_port_nums(p_src_sw, spine_lid, port_nums);
			dst_count = (int)osm_switch_get_minhop_port_count(p_spine_sw, dst_lid);

			links_to_use = asym_min(src_count, dst_count);
			links_to_remove = src_count - links_to_use;

			/*
			 * Skip over the links we going to use and then remove from that index.  The next time we use
			 * this spine (for another destination), we will pick up where we left off with the links to
			 * use.  In this manner, we will load level the src -> spine links across all destinations.
			 */
			link_index[iter2] += links_to_use;
			remove_index = link_index[iter2];

			/*
			 * Remove the links.
			 */
			while (links_to_remove > 0) {
				links_to_remove--;

				port_num = port_nums[remove_index % src_count];
				remove_index++;

				asym_short_link_remove(p_algo, port_num);
				osm_ar_subgroup_remove_port(&dst_ar_group, port_num);
			}
		}

		asym_set_routes(p_asym, p_src_sw, p_dst_sw, &dst_ar_group);

		/*
		 * Dump the removals to the log.
		 */
		asym_set_sort(&p_algo->removals, asym_removal_cmp);
		asym_log_removals(p_algo, &p_algo->removals);
		asym_set_remove_all(&p_algo->removals);
	}

	return;
}

/* ==============================================================================================================
 *
 * Quasi section.
 */

__attribute__((unused))
static void asym_quasi_dump(IN asym_t *p_asym, IN asym_quasi_t *p_quasi)
{
	size_t			count;
	asym_layer_t		*p_layer;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	osm_switch_t		*p_sw;
	char			*buffer;

	if (!osm_log_is_active(p_asym->p_log, OSM_LOG_DEBUG)) {
		return;
	}

	count = p_quasi->num_layers + 2;
	foreach_quasi_layer(p_quasi, p_layer, iter1) {
		count += asym_set_get_size(p_layer);
	}

	if ((buffer = asym_calloc(1, 128 * count)) != NULL) {
		sprintf(buffer, "QUASI %d\n", p_quasi->index);
		foreach_quasi_layer(p_quasi, p_layer, iter1) {
			if (asym_set_get_size(p_layer) > 1) {
				sprintf(buffer+strlen(buffer), "\tLAYER %ld\n", iter1);
				foreach_layer_switch(p_layer, p_sw, iter2) {
					sprintf(buffer+strlen(buffer), "\t\t%s\n", sw_name(p_sw));
				}
			}
		}

		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "%s\n", buffer);
		asym_free(buffer);
	}
}

static boolean_t asym_quasi_populate(IN asym_t *p_asym, IN asym_quasi_t *p_quasi, IN asym_set_t *p_layer, IN int index)
{
	int			i;
	boolean_t		found;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	flow_vertex_t		*p_vertex;
	flow_vertex_t		*p_remote_vertex;
	flow_edge_t		*p_edge;
	asym_node_t		*p_remote_node;

	found = FALSE;

	for (i = index; i < asym_set_get_size(p_layer); i++) {
		p_sw = (osm_switch_t *)asym_set_get(p_layer, i);
		if (asym_switch_is_isolated(p_asym->p_subn, p_sw)) {
			continue;
		}

		p_vertex = switch_to_vertex(p_sw);

		foreach_vertex_edge(p_asym->p_graph, p_vertex, p_edge) {
			p_remote_vertex = index_to_vertex(p_edge->adj_index);
			p_remote_sw = p_remote_vertex->p_sw;
			if (asym_switch_is_isolated(p_asym->p_subn, p_remote_sw)) {
				continue;
			}

			p_remote_node = switch_to_anode(p_remote_sw);
			if ((p_remote_sw->rank != p_asym->max_rank) && (p_remote_node->quasi == 0)) {
				asym_set_insert(&p_quasi->layers[p_asym->max_rank - p_remote_sw->rank], p_remote_sw);
				p_remote_node->quasi = p_quasi->index;
				found = TRUE;
			}
		}
	}

	return found;
}

static int asym_quasi_create(IN asym_t *p_asym, IN osm_switch_t *p_sw, IN int index)
{
	int			status;
	boolean_t		found;
	asym_quasi_t		*p_quasi;
	asym_node_t		*p_anode;
	asym_set_t		*p_layer;
	asym_set_t		*p_layers;
	asym_iter_t		iter;

	status = -1;
	p_quasi = NULL;
	p_layers = NULL;

	/*
	 * Create the quasi.
	 */
	p_quasi  = asym_calloc(1, sizeof(asym_quasi_t));
	p_layers = asym_calloc(1, (p_asym->max_layers + 2) * sizeof(asym_layer_t));
	if (any_null(2, p_quasi, p_layers)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B101: allocation failed for quasis\n");
		goto exit;
	}

	p_quasi->index = index;
	p_quasi->num_layers = p_asym->max_layers + 2;
	p_quasi->layers = p_layers;

	/*
	 * Seed the quasi with the switch.
	 */
	foreach_quasi_layer(p_quasi, p_layer, iter) {
		if (!asym_set_create(p_layer)) {
			OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B102: quasi set creation failed\n");
			goto exit;
		}
	}

	/*
	 * Good set creation.  Populate it.
	 */
	asym_set_insert(&p_quasi->layers[p_asym->max_rank - p_sw->rank], p_sw);
	p_anode = switch_to_anode(p_sw);
	p_anode->quasi = p_quasi->index;

	/*
	 * Build the quasi starting at the first core node.
	 */
	do {
		found = FALSE;
		foreach_quasi_layer(p_quasi, p_layer, iter) {
			for (index = 1; index < asym_set_get_size(p_layer); index++) {
				found |= asym_quasi_populate(p_asym, p_quasi, p_layer, index);
			}
		}
	} while (found);

	foreach_quasi_layer(p_quasi, p_layer, iter) {
		if (asym_set_is_not_empty(p_layer)) {
			asym_set_sort(p_layer, asym_sw_name_cmp);
		}
	}

	status = 0;
	p_asym->num_quasis = asym_max(p_asym->num_quasis, p_quasi->index);
	asym_quasi_dump(p_asym, p_quasi);

exit:
	if (all_non_null(2, p_quasi, p_layers)) {
		foreach_quasi_layer(p_quasi, p_layer, iter) {
			asym_set_destroy(p_layer);
		}
	}

	asym_free(p_quasi);
	asym_free(p_layers);

	return status;
}

static int asym_quasis_create(IN asym_t *p_asym)
{
	int			index;
	int			status;
	osm_switch_t		*p_sw;
	flow_vertex_t		*p_vertex;
	asym_node_t		*p_anode;

	status = 0;

	index = 0;
	foreach_active_vertex(p_asym, p_vertex) {
		p_sw = p_vertex->p_sw;
		p_anode = switch_to_anode(p_sw);
		if ((p_sw->rank == p_asym->core_rank) && (p_anode->quasi == 0)) {
			index++;
			if (asym_quasi_create(p_asym, p_sw, index) < 0) {
				status = -1;
				break;
			}
		}
	}

	return status;
}

/* ==============================================================================================================
 *
 * Pod section.
 */

__attribute__((unused))
static int asym_algorithm_pod_to_pod_short(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_pod)
{
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_spine_sw;

	p_asym = p_algo->p_asym;
	asym_algorithm_reset(p_algo);
	OSM_LOG(p_asym->p_log, OSM_LOG_VERBOSE, "SHORT: %d\n", p_pod->index);

	/*
	 * Clean up the layers for all switches.
	 */
	foreach_active_switch(p_asym->p_subn, p_sw) {
		p_adata = switch_to_adata(p_sw);
		p_adata->layer = NO_LAYER;
	}

	/*
	 * Assign layers so that the removal code knows which endpoint to remove.
	 */
	foreach_pod_leaf(p_pod, p_src_sw, iter1) {
		p_adata = switch_to_adata(p_src_sw);
		p_adata->layer = 2;
	}

	foreach_pod_spine(p_pod, p_spine_sw, iter1) {
		p_adata = switch_to_adata(p_spine_sw);
		p_adata->layer = 1;
	}

	/*
	 * Leaf as src.
	 */
	foreach_pod_leaf(p_pod, p_src_sw, iter1) {
		asym_algorithm_leaf_short(p_algo, p_pod, p_src_sw);
	}

	/*
	 * Spine as src.
	 */
	foreach_pod_spine(p_pod, p_spine_sw, iter1) {
		asym_algorithm_spine_short(p_algo, p_pod, p_spine_sw);
	}

	return 0;
}

__attribute__((unused))
static int asym_algorithm_pod_to_pod_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN asym_pod_t *p_dst_pod)
{
	int			i;
	int			q;
	int			d;
	int			status;
	uint8_t			*distances;
	uint16_t		*queue;
	asym_t			*p_asym;
	asym_set_t		*p_removals;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_iter_t		iter3;
	asym_node_t		*p_anode;
	asym_alink_t		*p_alink;
	asym_adata_t		*p_adata;
	asym_layer_t		*p_layer;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_dst_sw;

	OSM_LOG_ENTER(p_algo->p_asym->p_log);
	OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_VERBOSE, "LONG: %2d -> %2d\n", p_src_pod->index, p_dst_pod->index);

	status = 0;

	asym_algorithm_reset(p_algo);
	p_algo->p_src_pod = p_src_pod;
	p_algo->p_dst_pod = p_dst_pod;

	p_asym = p_algo->p_asym;

	/*
	 * Get the minhop table.
	 */
	asym_ar_bfs(p_algo);

	distances = p_algo->distances;
	queue     = p_algo->queue;

	q = queue[p_algo->q_max-1];

	foreach_active_switch(p_algo->p_asym->p_subn, p_sw) {
		p_adata = switch_to_adata(p_sw);
		p_adata->nw = W(0);
		p_adata->state = NODE_NA;
		p_adata->layer = NO_LAYER;
	}

	/*
	 * Populate the layers for the algorithm.
	 */
	foreach_layer(p_layer, iter2) {
		asym_set_remove_all(p_layer);
	}

	/*
	 * Fill in all of the intermediate layers.
	 */
	for (i = 0; i < p_algo->q_max; i++) {
		q = queue[i];
		d = distances[q];
		p_sw = index_to_switch(q);
		p_anode = index_to_anode(q);
		p_adata = index_to_adata(q);

		p_layer = index_to_layer(d);
		asym_set_insert(p_layer, p_sw);
		p_adata->layer = d;
	}

	/*
	 * Fill in the correct layer for nodes marked N/A.
	 */
	foreach_active_switch(p_algo->p_asym->p_subn, p_sw) {
		p_adata = switch_to_adata(p_sw);
		p_anode = switch_to_anode(p_sw);

		if (p_adata->layer == NO_LAYER) {
			uint16_t                sw_lid = osm_node_get_base_lid_ho(p_sw->p_node, 0);
			uint8_t                 dst_rank = OSM_SW_NO_RANK;
			uint8_t                 src_rank = OSM_SW_NO_RANK;
			uint8_t                 dst_hops = OSM_NO_PATH;
			uint8_t                 src_hops = OSM_NO_PATH;
			uint8_t                 max_layers = 0;
			osm_switch_t            *p_tmp_sw;

			foreach_pod_leaf(p_src_pod, p_tmp_sw, iter3) {
				src_rank = p_tmp_sw->rank;
				src_hops = asym_min(src_hops, osm_switch_get_least_hops(p_tmp_sw, sw_lid));
			}

			foreach_pod_leaf(p_dst_pod, p_tmp_sw, iter3) {
				dst_rank = p_tmp_sw->rank;
				dst_hops = asym_min(dst_hops, osm_switch_get_least_hops(p_tmp_sw, sw_lid));
			}

			max_layers = src_rank + dst_rank;

			if ((dst_hops == OSM_NO_PATH) && (src_hops == OSM_NO_PATH)) {
				// not possible
			} else if ((dst_hops != OSM_NO_PATH) && (src_hops == OSM_NO_PATH)) {
				p_adata->layer = -dst_hops;
			} else if ((dst_hops == OSM_NO_PATH) && (src_hops != OSM_NO_PATH)) {
				p_adata->layer = -(max_layers - src_hops);
			} else if ((dst_hops != OSM_NO_PATH) && (src_hops != OSM_NO_PATH)) {
				// not possible
			}

			foreach_link(p_adata, p_alink, iter1) {
				if (link_is_incoming(p_adata, p_alink)) {
					p_alink->state = LINK_NA;
				}
			}
		}
	}

	/*
	 * For each pod leaf, run the algorithm.
	 */
	foreach_pod_leaf(p_dst_pod, p_dst_sw, iter1) {
		p_anode = switch_to_anode(p_dst_sw);

		if (p_anode->is_fully_connected) {
			status = asym_algorithm_fc_pod_to_leaf_long(p_algo, p_src_pod, p_dst_sw);
			p_removals = &p_algo->fc_removals;
		} else {
			status = asym_algorithm_pc_pod_to_leaf_long(p_algo, p_src_pod, p_dst_sw);
			p_removals = &p_algo->removals;
		}

		if (status < 0) {
			goto exit;
		}

		asym_set_ar_groups(p_algo, p_removals);
	}

exit:
	OSM_LOG_EXIT(p_algo->p_asym->p_log);
	return status;
}

static int asym_algorithm_pod_to_pod(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN asym_pod_t *p_dst_pod)
{
	int			i = 0;
	uint8_t			hops;
	uint16_t		dst_lid;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;

	/*
	 * Special case - a pod routing to itself.
	 */
	if (p_src_pod == p_dst_pod) {
		return asym_algorithm_pod_to_pod_short(p_algo, p_src_pod);
	}

	/*
	 * Determine the number of layers needed.   It equals the number of hops between the leaves in the pods + 1
	 */
	p_asym = p_algo->p_asym;

	p_algo->num_layers = 0;

	foreach_pod_leaf(p_src_pod, p_src_sw, iter1) {
		foreach_pod_leaf(p_dst_pod, p_dst_sw, iter2) {
			dst_lid = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);
			hops = osm_switch_get_least_hops(p_src_sw, dst_lid);
			if ((0 < hops) && (hops < OSM_NO_PATH)) {
				p_algo->num_layers = hops + 1;
				goto process;
			}
		}
	}

process:
	if (p_algo->num_layers == 0) {
		return 0;
	}

	/*
	 * Poplulate the layer functions for this pair of pods.  The case where the number of layers
	 * equals 3 is handled by the short() functions.
	 */
	for (i = 0; i < MAX_LAYERS; i++) {
		p_algo->layer_functions[i] = asym_layer_null;
	}

	if (p_algo->num_layers == 4) {
		p_algo->layer_functions[0] = asym_layer_init;
		p_algo->layer_functions[1] = asym_layer_balance;
		p_algo->layer_functions[2] = asym_layer_fini;
		p_algo->layer_functions[3] = asym_layer_null;
	} else {
		p_algo->layer_functions[0] = asym_layer_init;
		p_algo->layer_functions[1] = asym_layer_balance;

		for (i = 2; i < p_algo->num_layers-3; i++) {
			p_algo->layer_functions[i] = asym_layer_both;
		}

		p_algo->layer_functions[p_algo->num_layers-3] = asym_layer_bottleneck;
		p_algo->layer_functions[p_algo->num_layers-2] = asym_layer_fini;
		p_algo->layer_functions[p_algo->num_layers-1] = asym_layer_null;
	}

	return asym_algorithm_pod_to_pod_long(p_algo, p_src_pod, p_dst_pod);
}

/* ==============================================================================================================
 *
 * Asym section.
 */

static int asym_validate_tree(IN osm_subn_t *p_subn, IN flow_graph_t *p_graph)
{
	int			status;
	int			max_rank;
	int			num_nodes;
	int			min_num_groups;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	flow_vertex_t		*p_vertex;
	flow_edge_t		*p_edge;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	status = -1;

	/*
	 * Find the max rank.
	 */
	max_rank = 0;
	foreach_active_switch(p_subn, p_sw) {
		if (p_sw->rank > max_rank) {
			max_rank = p_sw->rank;
		}
	}

	if (max_rank == OSM_SW_NO_RANK) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "unranked switch detected\n");
		goto exit;
	}

	/*
	 * Verify the following assumptions:
	 *	1) No switch to switch links of the same rank.
	 *	2) No switch to switch links which skip tree layers.
	 */
	foreach_vertex(p_graph, p_vertex) {
		p_sw = p_vertex->p_sw;
		if (asym_switch_is_isolated(p_subn, p_sw)) {
			continue;
		}

		foreach_vertex_edge(p_graph, p_vertex, p_edge) {
			p_remote_sw = p_graph->vertices[p_edge->adj_index].p_sw;

			if (p_remote_sw == p_sw) {
				continue;
			}

			if (p_sw->rank == p_remote_sw->rank) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "equal ranked connected switches detected\n");
				goto exit;
			}
			if (abs(p_sw->rank - p_remote_sw->rank) != 1) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "connected switches crossing layers detected\n");
				goto exit;
			}
		}
	}

	/*
	 * Find the maximum rank.  It must not be greater than 2.  We previously checked for leaf switches
	 * with rank > 2.  If we encounter a switch with rank > 2 now, it is NOT a leaf switch.
	 *
	 * Find the switch which has the least number of AR groups configured.  We compare this to the
	 * number of leaves in the fabric.  If it is less than the number of leaves, then we do not have
	 * the capabilities to be compliant.
	 */
	num_nodes = 0;
	min_num_groups = AR_INVALID_GROUP_ID;
	foreach_active_switch(p_subn, p_sw) {
		num_nodes++;

		if (p_sw->ar_data.num_groups < min_num_groups) {
			min_num_groups = p_sw->ar_data.num_groups;
		}
	}

	if (min_num_groups < num_nodes) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "number of leaves exceeds switch capabilites"
			" (num leaves = %d, minimal group cap = %d)\n",
			num_nodes, min_num_groups);
		goto exit;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG, "tree is asymtree compliant\n");
	status = max_rank;

exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

/* ----------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static int asym_resolve_attached_port(IN asym_t *p_asym, IN osm_switch_t *p_sw)
{
	uint16_t		min_lid;
	uint16_t		max_lid;
	asym_node_t		*p_anode;
	asym_port_t		*p_aport;
	asym_iter_t		iter;
	osm_subn_t		*p_subn;
	osm_port_t		*p_port0;

	p_subn = p_asym->p_subn;
	p_anode = switch_to_anode(p_sw);

	/*
	 * Is this switch AR capable?
	 */
	if (!osm_switch_is_ar_supported(p_sw)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "%s is not AR enabled\n", sw_name(p_sw));
		return 0;
	}

	if (asym_switch_is_isolated(p_subn, p_sw)) {
		return 0;
	}

	/*
	 * Set a static route to port 0.
	 */
	p_port0 = osm_switch_get_port0(p_sw, p_subn);
	osm_port_get_lid_range_ho(p_port0, &min_lid, &max_lid);
	asym_set_static_lft_entry_by_lid_range(p_asym, p_sw, min_lid, max_lid);

	/*
	 * Set a static route to attached CAs.
	 */
	foreach_asym_port(p_anode, p_aport, iter) {
		asym_set_static_lft_entry_by_port(p_asym, p_sw, p_aport);
	}

	return 0;
}

__attribute__((unused))
static int asym_resolve_attached_ports(IN void **tasks,
				       IN unsigned task_len,
				       IN void *context,
				       IN unsigned int thread_id)
{
	int			i;
	int			status;
	asym_t			*p_asym;

	status = 0;
	p_asym = (asym_t *)context;

	for (i = 0; i < task_len; i++) {
		status += asym_resolve_attached_port(p_asym, (osm_switch_t *)tasks[i]);
	}

	return status;
}

/* ----------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static int asym_resolve_non_leaf(IN asym_t *p_asym, IN osm_switch_t *p_dst_sw)
{
	int			status;
	osm_switch_t		*p_src_sw;

	OSM_LOG_ENTER(p_asym->p_log);

	status = 0;

	foreach_active_switch(p_asym->p_subn, p_src_sw) {
		if (p_src_sw != p_dst_sw) {
			asym_set_routes(p_asym, p_src_sw, p_dst_sw, NULL);
		}
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return status;
}

__attribute__((unused))
static int asym_resolve_non_leaves(IN void **tasks,
				  IN unsigned task_len,
				  IN void *context,
				  IN unsigned int thread_id)
{
	int			i;
	int			status;
	asym_t			*p_asym;
	osm_switch_t		*p_sw;

	p_asym = (asym_t *)context;
	OSM_LOG_ENTER(p_asym->p_log);

	status = 0;
	for (i = 0; i < task_len; i++) {
		p_sw = (osm_switch_t *)tasks[i];
		status += asym_resolve_non_leaf(p_asym, p_sw);
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return status;
}

/* ----------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static int asym_resolve_non_pod_leaf(IN asym_t *p_asym, IN osm_switch_t *p_dst_sw)
{
	int			status;
	osm_switch_t		*p_src_sw;

	OSM_LOG_ENTER(p_asym->p_log);

	status = 0;

	foreach_active_switch(p_asym->p_subn, p_src_sw) {
		if (p_src_sw != p_dst_sw) {
			asym_set_routes(p_asym, p_src_sw, p_dst_sw, NULL);
		}
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return status;
}

__attribute__((unused))
static int asym_resolve_non_pod_leaves(IN void **tasks,
				       IN unsigned task_len,
				       IN void *context,
				       IN unsigned int thread_id)
{
	int			i;
	int			status;
	asym_t			*p_asym;
	osm_switch_t		*p_sw;

	p_asym = (asym_t *)context;
	OSM_LOG_ENTER(p_asym->p_log);

	status = 0;
	for (i = 0; i < task_len; i++) {
		p_sw = (osm_switch_t *)tasks[i];
		status += asym_resolve_non_pod_leaf(p_asym, p_sw);
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return status;
}

/* ----------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static int asym_resolve_pod(IN asym_t *p_asym, IN asym_pod_t *p_dst_pod, IN void *context)
{
	int			status = 0;
	asym_pod_t		*p_src_pod = NULL;

	OSM_LOG_ENTER(p_asym->p_log);

	foreach_asym_pod(p_asym, p_src_pod) {
		status += asym_algorithm_pod_to_pod(context, p_src_pod, p_dst_pod);
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return status;
}

__attribute__((unused))
static int asym_resolve_pods(IN void **tasks,
				IN unsigned task_len,
				IN void *context,
				IN unsigned int thread_id)
{
	int			i;
	int			status;
	void			*p_algo;
	asym_t			*p_asym;

	status = -1;
	p_asym = (asym_t *)context;
	OSM_LOG_ENTER(p_asym->p_log);

	if ((p_algo = asym_algorithm_init(p_asym)) != NULL) {
		status = 0;
		for (i = 0; i < task_len; i++) {
			status += asym_resolve_pod(p_asym, tasks[i], p_algo);
		}

		asym_update_excluded_spines(p_algo);
		asym_algorithm_free(p_algo);
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return status;
}

/* ----------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static int asym_set_missing_switch_routes(IN asym_t *p_asym, IN osm_switch_t *p_sw)
{
	uint8_t			port_num;
	uint8_t			state;
	uint16_t		lid;
	uint16_t		group_id;

	OSM_LOG_ENTER(p_asym->p_log);

	for (lid = 1; lid <= p_sw->max_lid_ho; lid++) {
		osm_ar_lft_entry_get(p_sw, 0, lid, &port_num, &state, &group_id);
		if (port_num != OSM_NO_PATH) {
			continue;
		}

		port_num = osm_switch_get_port_by_lid(p_sw, lid, OSM_NEW_LFT);
		if (port_num == OSM_NO_PATH) {
			continue;
		}

		osm_ar_lft_set(&p_sw->ar_data, lid, 0, port_num, OSM_AR_IB_LID_STATE_STATIC, 0);
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return 0;
}

__attribute__((unused))
static int asym_set_missing_routes(IN void **tasks,
				   IN unsigned task_len,
				   IN void *context,
				   IN unsigned int thread_id)
{
	int			i;
	asym_t			*p_asym;

	p_asym = (asym_t *)context;
	OSM_LOG_ENTER(p_asym->p_log);

	for (i = 0; i < task_len; i++) {
		asym_set_missing_switch_routes(p_asym, (osm_switch_t *)tasks[i]);
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return 0;
}

/* ----------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static int asym_ar_set_all_group0(IN void **tasks,
				  IN unsigned task_len,
				  IN void *context,
				  IN unsigned int thread_id)
{
	int			i;

	for (i = 0; i < task_len; i++) {
		asym_clear_group0(tasks[i]);
	}

	return 0;
}

/* ----------------------------------------------------------------------------------------------------------- */

static int asym_execute_threaded(IN asym_t *p_asym, IN void *function, IN int task_type)
{
	int			count;
	int			status;
	size_t			task_len;
	asym_pod_t		*p_pod;
	osm_subn_t		*p_subn;
	osm_ucast_mgr_t		*p_mgr;
	osm_switch_t		*p_sw;
	asym_node_t		*p_anode;
	void			**tasks;

	OSM_LOG_ENTER(p_asym->p_log);

	status = -1;
	p_subn = p_asym->p_subn;
	p_mgr = &p_subn->p_osm->sm.ucast_mgr;

	/*
	 * Allocate and load the task list.
	 */
	switch (task_type) {
	case ASYM_TASK_SWITCHES:
		task_len = p_asym->num_switches;
		break;
	case ASYM_TASK_PODS:
		task_len = p_asym->num_pods;
		break;
	case ASYM_TASK_NON_LEAVES:
		task_len = p_asym->num_switches;
		break;
	case ASYM_TASK_NON_POD_LEAVES:
		task_len = p_asym->num_switches;
		break;
	default:
		goto exit;
		break;
	}

	if ((tasks = (void **)asym_calloc(task_len, sizeof(void *))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B112: allocation failed for tasks\n");
		goto exit;
	}

	count = 0;

	if (task_type == ASYM_TASK_SWITCHES) {
		foreach_active_switch(p_subn, p_sw) {
			tasks[count++] = (void *)p_sw;
		}
	} else if (task_type == ASYM_TASK_PODS) {
		foreach_asym_pod(p_asym, p_pod) {
			tasks[count++] = (void *)p_pod;
		}
	} else if (task_type == ASYM_TASK_NON_LEAVES) {
		foreach_active_switch(p_subn, p_sw) {
			p_anode = switch_to_anode(p_sw);
			if (!p_anode->has_hcas) {
				tasks[count++] = (void *)p_sw;
			}
		}
	} else if (task_type == ASYM_TASK_NON_POD_LEAVES) {
		foreach_active_switch(p_subn, p_sw) {
			p_anode = switch_to_anode(p_sw);
			if ((p_anode->has_hcas) && (p_anode->pod == 0)) {
				tasks[count++] = (void *)p_sw;
			}
		}
	}

	/*
	 * Execute the function.
	 */
	status = cl_mp_thread_pool_array_apply(&p_mgr->thread_pool,
					       tasks,
					       count,
					       CL_MP_THREAD_POOL_BATCH_PER_THREAD,
					       (cl_pfn_mp_thread_pool_apply_t)function,
					       (void *)p_asym);

	asym_free(tasks);

exit:
	OSM_LOG_EXIT(p_asym->p_log);
	return status;
}

/* ----------------------------------------------------------------------------------------------------------- */

static int asym_nodes_init(IN asym_t *p_asym)
{
	int			status;
	flow_vertex_t		*p_vertex;
	asym_node_t		*p_anode;
	osm_switch_t		*p_sw;

	if ((p_asym->nodes = (asym_node_t *)asym_calloc(p_asym->num_switches, sizeof(asym_node_t))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B113: allocation failed for asymtree\n");
		return -1;
	}

	/*
	 * Label the switches with their indices.
	 */
	foreach_vertex(p_asym->p_graph, p_vertex) {
		p_sw = p_vertex->p_sw;
		p_sw->re_index = p_vertex->index;
		p_sw->re_priv = p_asym;

		if (!asym_switch_is_isolated(p_asym->p_subn, p_sw)) {
			if (p_sw->rank > p_asym->max_rank) {
				p_asym->max_rank = p_sw->rank;
			}
		}
	}

	/*
	 * Initialize the nodes.
	 */
	status = 0;

	foreach_active_vertex(p_asym, p_vertex) {
		p_anode = index_to_anode(p_vertex->index);

		p_anode->p_sw = p_vertex->p_sw;
		p_anode->index = p_vertex->index;
		p_anode->is_fully_connected = FALSE;
		p_anode->has_hcas = FALSE;

		if (!asym_set_create(&p_anode->ports)) {
			OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B114: anode set creation failed\n");
			status = -1;
			break;
		}
	}

	return status;
}

static void asym_nodes_free(IN asym_t *p_asym)
{
	asym_node_t		*p_anode;
	osm_switch_t		*p_sw;

	if (p_asym->nodes != NULL) {
		foreach_anode(p_asym, p_anode) {
			asym_set_destroy(&p_anode->ports);
			p_sw = anode_to_switch(p_anode);
			if (p_sw != NULL) {
				p_sw->re_index = 0;
				p_sw->re_priv = NULL;
			}
		}

		asym_free(p_asym->nodes);
	}

	return;
}

static int asym_ports_init(IN asym_t *p_asym)
{
	uint16_t		min_lid;
	uint16_t		max_lid;
	ib_special_port_type_t	port_type;
	asym_node_t		*p_remote_anode;
	asym_port_t		*p_aport;
	osm_switch_t		*p_remote_sw;
	osm_port_t		*p_port;

	if ((p_asym->ports = (asym_port_t *)asym_calloc(p_asym->num_ports, sizeof(asym_port_t))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B115: allocation failed for asymtree\n");
		return -1;
	}

	/*
	 * Find the ports for each switch.
	 */
	p_aport = p_asym->ports;
	foreach_port(p_asym->p_subn, p_port) {
		if ((port_type = asym_get_special_port_type(p_port)) == IB_SPECIAL_PORT_TYPE_MAX) {
			continue;
		}

		/*
		 * The port in question is either a CA port or a special port.
		 * CA ports get an AR group, while special ports do not.
		 */
		osm_port_get_lid_range_ho(p_port, &min_lid, &max_lid);

		if (min_lid == 0) {
			continue;
		}

		p_aport->type = port_type;
		p_aport->min_lid = min_lid;
		p_aport->max_lid = max_lid;

		p_remote_sw = asym_get_switch_by_port(p_port);

		if (asym_switch_is_isolated(p_asym->p_subn, p_remote_sw)) {
			continue;
		}

		p_remote_anode = switch_to_anode(p_remote_sw);
		asym_set_insert(&p_remote_anode->ports, p_aport);

		if (port_type == IB_SPECIAL_PORT_TYPE_RESERVED) {
			p_remote_anode->has_hcas = TRUE;
		}

		p_aport++;
	}

	return 0;
}

static void asym_ports_free(IN asym_t *p_asym)
{
	asym_free(p_asym->ports);
	return;
}

static int asym_links_init(IN asym_t *p_asym)
{
	flow_edge_t		*p_edge;
	asym_node_t		*p_anode;
	asym_node_t		*p_rnode;
	osm_switch_t		*p_sw1;
	osm_switch_t		*p_sw2;

	/*
	 * Find the links between switches.
	 */
	foreach_edge(p_asym->p_graph, p_edge) {
		if (p_edge->index < p_edge->adj_index) {
			p_sw1 = p_asym->p_graph->vertices[p_edge->index].p_sw;
			p_sw2 = p_asym->p_graph->vertices[p_edge->adj_index].p_sw;

			if (asym_switch_is_isolated(p_asym->p_subn, p_sw1) || asym_switch_is_isolated(p_asym->p_subn, p_sw2)) {
				continue;
			}

			p_anode = index_to_anode(p_edge->index);
			p_anode->num_links++;

			p_rnode = index_to_anode(p_edge->adj_index);
			p_rnode->num_links++;
		}

		p_asym->num_links++;
	}

	return 0;
}

/* ==============================================================================================================
 *
 * API functions.
 */
int osm_asymtree_ar_process_switch(IN osm_subn_t *p_subn,
				   IN osm_switch_t *p_sw,
				   IN void *context)
{
	asym_t			*p_asym;
	asym_node_t		*p_anode;

	p_asym = (asym_t *)context;
	p_anode = switch_to_anode(p_sw);

	return p_anode->max_group_id + 1;
}

void *osm_asymtree_ar_process_subnet(IN osm_subn_t *p_subn,
				     IN int ar_enable,
				     IN int ar_mode,
				     IN void *context)
{
	uint16_t		sw_lid;
	asym_t			*p_asym;
	asym_node_t		*p_anode;
	struct timeval		start, stop;
	double			elapsed_time;

	p_asym = (asym_t *)context;

	OSM_LOG_ENTER(p_asym->p_log);

	p_asym->ar_enable = ar_enable;
	p_asym->ar_mode = ar_mode;
	p_asym->ar_state = ar_enable ? ar_mode : OSM_AR_IB_LID_STATE_STATIC;

	/*
	 * Set the group ids.  This is done here because locking isn't allowed in the routing code.
	 * If we tried to do it later, it might be in the parallel threading model.
	 */
	foreach_asym_node(p_asym, p_anode) {
		if (anode_has_attached_hcas(p_anode)) {
			sw_lid = osm_node_get_base_lid_ho(p_anode->p_sw->p_node, 0);
			p_anode->group_id = osm_ucast_mgr_make_groupid(&p_subn->p_osm->sm.ucast_mgr, sw_lid, 0);
		} else {
			p_anode->group_id = UCAST_MGR_INVALID_GROUP_ID;
		}
	}

	gettimeofday(&start, NULL);

	/*
	 * Do the locally attached ports.
	 */
	if (asym_execute_threaded(p_asym, asym_resolve_attached_ports, ASYM_TASK_SWITCHES) != 0) {
		OSM_LOG(p_asym->p_log, OSM_LOG_INFO, "asym_resolve_attached_ports failed\n");
	}

	/*
	 * Do the non-leafs.
	 */
	if (asym_execute_threaded(p_asym, asym_resolve_non_leaves, ASYM_TASK_NON_LEAVES) != 0) {
		OSM_LOG(p_asym->p_log, OSM_LOG_INFO, "asym_resolve_non_leaves failed\n");
	}

	/*
	 * Do the non-pod leaves.
	 */
	if (asym_execute_threaded(p_asym, asym_resolve_non_pod_leaves, ASYM_TASK_NON_POD_LEAVES) != 0) {
		OSM_LOG(p_asym->p_log, OSM_LOG_INFO, "resolve_non_pod_leaves failed\n");
	}

	/*
	 * Run the algorithm and set the remote AR groups.
	 */
	if (asym_execute_threaded(p_asym, asym_resolve_pods, ASYM_TASK_PODS) != 0) {
		OSM_LOG(p_asym->p_log, OSM_LOG_INFO, "asym_resolve_pods failed\n");
	}

	/*
	 * Clear all of the group 0 ports.
	 */
	if (asym_execute_threaded(p_asym, asym_ar_set_all_group0, ASYM_TASK_SWITCHES) != 0) {
		OSM_LOG(p_asym->p_log, OSM_LOG_INFO, "asym_ar_set_all_group0 failed\n");
	}

	/*
	 * Set missing routes.
	 */
	if (asym_execute_threaded(p_asym, asym_set_missing_routes, ASYM_TASK_SWITCHES) != 0) {
		OSM_LOG(p_asym->p_log, OSM_LOG_INFO, "asym_set_missing_routes failed\n");
	}

	gettimeofday(&stop, NULL);
	elapsed_time = (double)(stop.tv_usec - start.tv_usec) / 1000000 + (double)(stop.tv_sec - start.tv_sec);
	OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "time to solution = %11.8f secs\n", elapsed_time);

	/*
	 * Process excluded spines (if necessary).
	 */
	asym_process_excluded_spines(p_asym);

	OSM_LOG_EXIT(p_asym->p_log);
	return p_asym->re_context;
}

void osm_asymtree_fini(IN void *context)
{
	if (context != NULL) {
		asym_set_destroy(context);
		asym_free(context);
	}
}

void osm_asymtree_free(IN void *context)
{
	asym_t			*p_asym;

	p_asym = (asym_t *)context;

	if (p_asym != NULL) {
		asym_pods_destroy(p_asym);
		asym_ports_free(p_asym);
		asym_nodes_free(p_asym);

		if (p_asym->p_graph != NULL) {
			osm_flow_destroy_graph(&p_asym->p_graph);
		}

		asym_set_destroy(&p_asym->excluded_spines);
		cl_plock_destroy(&p_asym->lock);

		memset(p_asym, 0, sizeof(asym_t));
		asym_free(p_asym);
	}
}

void *osm_asymtree_init(IN osm_subn_t *p_subn, IN void *context)
{
	int			status;
	int			max_rank;
	size_t			num_switches;
	void			*re_context;
	flow_graph_t		*p_graph;
	asym_t			*p_asym;
	osm_log_t		*p_log;

	status = -1;
	p_asym = NULL;
	p_graph = NULL;
	re_context = context;
	p_log = &p_subn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	/*
	 * Validate the tree.
	 */
	if ((p_graph = osm_flow_build_graph(p_subn)) == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR B116: allocation failed for asymtree\n");
		goto exit;
	}

	if ((max_rank = asym_validate_tree(p_subn, p_graph)) < 0) {
		goto exit;
	}

	/*
	 * If the re_context is NULL, we need to create it.
	 */
	if (re_context == NULL) {
		re_context = asym_calloc(1, sizeof(asym_set_t));
		if (re_context == NULL) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR B117: allocation failed for re_context\n");
			goto exit;
		}

		if (!asym_set_create(re_context)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR B118: re_context set creation failed\n");
			asym_free(re_context);
			re_context = NULL;
			goto exit;
		}
	}

	/*
	 * The tree is valid.  Allocate our data struct.
	 */
	if ((p_asym = (asym_t *)asym_calloc(1, sizeof(asym_t))) == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR B119: allocation failed for asymtree\n");
		goto exit;
	}

	/*
	 * Populate the data structs.
	 */
	cl_qmap_init(&p_asym->pod_map);
	cl_plock_construct(&p_asym->lock);

	num_switches = p_graph->vertices_num;

	p_asym->p_osm = p_subn->p_osm;
	p_asym->p_subn = p_subn;
	p_asym->p_log = p_log;
	p_asym->p_mgr = &p_subn->p_osm->sm.ucast_mgr;
	p_asym->re_context = re_context;
	p_asym->p_graph = p_graph;
	p_asym->threshold = T(p_subn->opt.ar_tree_asymmetric_flow_threshold, 100);
	p_asym->num_switches = num_switches;
	p_asym->num_ports = cl_qmap_count(&p_subn->port_guid_tbl);
	p_asym->max_rank = max_rank;
	p_asym->core_rank = 0;
	p_asym->max_layers = 2*max_rank + 1;

	/*
	 * Init the nodes, port and links.
	 */
	if (asym_nodes_init(p_asym) < 0) {
		goto exit;
	}

	if (asym_ports_init(p_asym) < 0) {
		goto exit;
	}

	if (asym_links_init(p_asym) < 0) {
		goto exit;
	}

	/*
	 * Create the pods and quasis.
	 */
	if (asym_pods_create(p_asym) < 0) {
		goto exit;
	}

	if (asym_quasis_create(p_asym) < 0) {
		goto exit;
	}

	/*
	 * Create the set and mutex for excluded spines reporting to UFM.
	 */
	if (!asym_set_create(&p_asym->excluded_spines)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B11A: excluded spines set creation failed\n");
		goto exit;
	}

	if (cl_plock_init(&p_asym->lock) != CL_SUCCESS) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B11C: asym lock initialization failed\n");
		goto exit;
	}

	/*
	 * Everything is setup correctly.  Set the status.
	 */
	status = 0;

exit:

	if (status < 0) {
		if (p_asym != NULL) {
			osm_asymtree_free(p_asym);
			p_asym = NULL;
		} else {
			if ((context == NULL) && (re_context != NULL)) {
				asym_set_destroy(re_context);
				asym_free(re_context);
			}

			if (p_graph != NULL) {
				osm_flow_destroy_graph(&p_graph);
			}
		}

	}

	OSM_LOG_EXIT(p_log);
	return p_asym;
}
