/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2009 Sun Microsystems, Inc. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2010 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_vl15_t.
 * This object represents the VL15 Interface object.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_thread.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_VL15INTF_C
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_vl15intf.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_log.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>

static void vl15_statistic_notice_counter_increase(uint16_t trap_num, osm_subn_t *p_subn)
{
	/* Notices are currently sent only on the main port */
	uint64_t *stats_counters = p_subn->p_osm->stats_counters[0];

	switch(trap_num) {
	case 128:
		stats_counters[NOTICE_128_SEND_COUNT]++;
		break;
	case 130:
		stats_counters[NOTICE_130_SEND_COUNT]++;
		break;
	case 131:
		stats_counters[NOTICE_131_SEND_COUNT]++;
		break;
	case 144:
		stats_counters[NOTICE_144_SEND_COUNT]++;
		break;
	case 145:
		stats_counters[NOTICE_145_SEND_COUNT]++;
		break;
	case 256:
		stats_counters[NOTICE_256_SEND_COUNT]++;
		break;
	case 257:
		stats_counters[NOTICE_257_SEND_COUNT]++;
		break;
	case 258:
		stats_counters[NOTICE_258_SEND_COUNT]++;
		break;
	case 259:
		stats_counters[NOTICE_259_SEND_COUNT]++;
		break;
	case 1144:
		stats_counters[NOTICE_1144_SEND_COUNT]++;
		break;
	case 1146:
		stats_counters[NOTICE_1146_SEND_COUNT]++;
		break;
	case 1601:
		stats_counters[NOTICE_1601_SEND_COUNT]++;
		break;
	case 1605:
		stats_counters[NOTICE_1605_SEND_COUNT]++;
		break;
	case 1606:
		stats_counters[NOTICE_1606_SEND_COUNT]++;
		break;
	case 1609:
		stats_counters[NOTICE_1609_SEND_COUNT]++;
		break;
	case 1610:
		stats_counters[NOTICE_1610_SEND_COUNT]++;
		break;
	default:
		break;
	}
}

static void vl15_statistic_counter_increase(ib_net16_t attr_id, osm_subn_t *p_subn,
					    uint16_t trap_num, uint8_t sm_binding_port_index) {
	uint64_t *stats_counters = p_subn->p_osm->stats_counters[sm_binding_port_index];

	switch (attr_id) {
	case IB_MAD_ATTR_NODE_DESC:
		stats_counters[NODE_DESC_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_NODE_INFO:
		stats_counters[NODE_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_GUID_INFO:
		stats_counters[GUID_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VIRTUALIZATION_INFO:
		stats_counters[VIRTUALIZATION_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VPORT_STATE:
		stats_counters[VPORT_PORT_STATE_TBL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VNODE_INFO:
		stats_counters[VPORT_NODE_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VPORT_INFO:
		stats_counters[VPORT_PORT_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_V_PKEY_TABLE:
		stats_counters[VPORT_PKEY_TABLE_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VNODE_DESC:
		stats_counters[VPORT_NODE_DESCRIPTION_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VPORT_GUID_INFO:
		stats_counters[VPORT_GUID_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_SWITCH_INFO:
		stats_counters[SWITCH_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_EXTENDED_SWITCH_INFO:
		stats_counters[EXTENDED_SWITCH_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PORT_INFO:
		stats_counters[PORT_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_SWITCH_PORT_STATE_TBL:
		stats_counters[SPST_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_LIN_FWD_TBL:
		stats_counters[LFT_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_ADAPTIVE_ROUTING_INFO:
		stats_counters[AR_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_MLNX_EXTENDED_NODE_INFO:
		stats_counters[EXTENDED_NODE_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE:
		stats_counters[AR_GROUP_TABLE_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_ADAPTIVE_ROUTING_GROUP_TABLE_COPY:
		stats_counters[AR_GROUP_TABLE_COPY_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_ADAPTIVE_ROUTING_LFT:
		stats_counters[AR_LFT_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PLFT_INFO:
		stats_counters[PLFT_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PLFT_DEF:
		stats_counters[PLFT_DEF_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PORT_SL_TO_PLFT_MAP:
		stats_counters[PORT_SL_TO_PLFT_MAP_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PLFT_MAP:
		stats_counters[PLFT_MAP_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RN_SUB_GROUP_DIRECTION_TABLE:
		stats_counters[RN_SUB_GROUP_DIRECTION_TABLE_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RN_GEN_STRING_TABLE:
		stats_counters[RN_GEN_STRING_TABLE_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RN_RCV_STRING_TABLE:
		stats_counters[RN_RCV_STRING_TABLE_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RN_XMIT_PORT_MASK:
		stats_counters[RN_XMIT_PORT_MASK_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RN_GEN_BY_SUB_GROUP_PRIORITY:
		stats_counters[RN_GEN_BY_SUB_GROUP_PRIORITY_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_MCAST_FWD_TBL:
		stats_counters[MFT_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_HIERARCHY_INFO:
		stats_counters[HIERARCHY_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CHASSIS_INFO:
		stats_counters[CHASSIS_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_SM_INFO:
		stats_counters[SM_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_SLVL_TABLE:
		stats_counters[SL_VL_MAP_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_VL_ARBITRATION:
		stats_counters[VL_ARBIT_TBL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_GENERAL_INFO:
		stats_counters[GENERAL_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_P_KEY_TABLE:
		stats_counters[P_KEY_TBL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_MLNX_EXTENDED_PORT_INFO:
		stats_counters[MEPI_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_ROUTER_INFO:
		stats_counters[ROUTER_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RTR_ADJ_SITE_LOCAL_TBL:
		stats_counters[RTR_ADJ_SITE_LOCAL_TBL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RTR_NEXT_HOP_TBL:
		stats_counters[RTR_NEXT_HOP_TBL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RTR_ADJ_ROUTER_LID_INFO:
		stats_counters[RTR_ADJ_ROUTER_LID_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RTR_ROUTER_LID_TBL:
		stats_counters[RTR_ROUTER_LID_TBL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RTR_AR_GROUP_TO_LID_TBL:
		stats_counters[RTR_ROUTER_AR_GROUP_TO_LID_TBL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_QOS_CONFIG_SL:
		stats_counters[QOS_CONFIG_SL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_QOS_CONFIG_VL:
		stats_counters[QOS_CONFIG_VL_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PFRN_CONFIG:
		stats_counters[PFRN_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_NOTICE:
		vl15_statistic_notice_counter_increase(trap_num, p_subn);
		break;
	case IB_MAD_ATTR_HBF_CONFIG:
		stats_counters[HBF_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_WHBF_CONFIG:
		stats_counters[WHBF_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_NVL_HBF_CONFIG:
		stats_counters[NVL_HBF_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_PROFILES_CONFIG:
		stats_counters[PROFILES_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CREDIT_WATCHDOG_CONFIG:
		stats_counters[CREDIT_WATCHDOG_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_BER_CONFIG:
		stats_counters[BER_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_ALID_INFO:
		stats_counters[ALID_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_RAIL_FILTER_CONFIG:
		stats_counters[RAIL_FILTER_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_LFT_SPLIT:
		stats_counters[LFT_SPLIT_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_END_PORT_PLANE_FILTER_CONFIG:
		stats_counters[END_PORT_PLANE_FILTER_CONFIG_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CONTAIN_AND_DRAIN_INFO:
		stats_counters[CND_INFO_SEND_COUNT]++;
		break;
	case IB_MAD_ATTR_CONTAIN_AND_DRAIN_PORT_STATE:
		stats_counters[CND_PORT_STATE_SEND_COUNT]++;
		break;

	default:
		break;
	}
}

static void vl15_send_mad(osm_vl15_t * p_vl, osm_madw_t * p_madw)
{
	ib_api_status_t status;
	boolean_t resp_expected = p_madw->resp_expected;
	ib_smp_t * p_smp;
	ib_mad_notice_attr_t *p_ntci;
	ib_net16_t attr_id;
	uint16_t trap_num = 0;
	uint8_t method;

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_ntci = (ib_mad_notice_attr_t *) p_smp->data;
	method = p_smp->method;
	attr_id = p_smp->attr_id;
	/*
	   Non-response-expected mads are not throttled on the wire
	   since we can have no confirmation that they arrived
	   at their destination.
	 */
	if (resp_expected)
		/*
		   Note that other threads may not see the response MAD
		   arrive before send() even returns.
		   In that case, the wire count would temporarily go negative.
		   To avoid this confusion, preincrement the counts on the
		   assumption that send() will succeed.
		 */
		cl_atomic_inc(&p_vl->p_stats->qp0_mads_outstanding_on_wire);
	else
		cl_atomic_inc(&p_vl->p_stats->qp0_unicasts_sent);

	cl_atomic_inc(&p_vl->p_stats->qp0_mads_sent);

	if (attr_id == IB_MAD_ATTR_NOTICE) {
		if (ib_notice_is_generic(p_ntci))
			trap_num = cl_ntoh16(p_ntci->g_or_v.generic.trap_num);
		else
			trap_num = cl_ntoh16(p_ntci->data_details.vendor_data_details.trap_num);
	}

	status = osm_vendor_send(osm_madw_get_bind_handle(p_madw),
				 p_madw, p_madw->resp_expected);

	if (status == IB_SUCCESS) {
		vl15_statistic_counter_increase(attr_id, p_vl->p_subn, trap_num, p_vl->port_index);
		OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
			"%u QP0 MADs on wire, %u outstanding, "
			"%u unicasts sent, %u total sent\n",
			p_vl->p_stats->qp0_mads_outstanding_on_wire,
			p_vl->p_stats->qp0_mads_outstanding,
			p_vl->p_stats->qp0_unicasts_sent,
			p_vl->p_stats->qp0_mads_sent);
		return;
	}

	OSM_LOG(p_vl->p_log, OSM_LOG_ERROR, "ERR 3E03: "
		"MAD send failed (%s)\n", ib_get_err_str(status));

	/*
	   The MAD was never successfully sent, so
	   fix up the pre-incremented count values.
	 */

	/* Decrement qp0_mads_sent that were incremented in the code above.
	   qp0_mads_outstanding will be decremented by send error callback
	   (called by osm_vendor_send() */
	cl_atomic_dec(&p_vl->p_stats->qp0_mads_sent);
	if (!resp_expected) {
		cl_atomic_dec(&p_vl->p_stats->qp0_unicasts_sent);
		return;
	}

	/* need to cause heavy-sweep if resp_expected MAD sending failed */
	OSM_LOG(p_vl->p_log, OSM_LOG_ERROR, "ERR 3E04: "
		"%s method failed for attribute 0x%X (%s)\n",
		method == IB_MAD_METHOD_SET ? "SET" : "GET",
		cl_ntoh16(attr_id),
		ib_get_sm_attr_str(attr_id));

	p_vl->p_subn->subnet_initialization_error = TRUE;

}

static void vl15_poller(IN void *p_ptr)
{
	ib_api_status_t status;
	osm_madw_t *p_madw;
	osm_vl15_t *p_vl = p_ptr;
	cl_qlist_t *p_fifo;
	int32_t max_smps = p_vl->max_wire_smps;
	int32_t max_smps2 = p_vl->max_wire_smps2;
	ib_net64_t port_guid;

	OSM_LOG_ENTER(p_vl->p_log);

	port_guid = p_vl->port_index == p_vl->p_subn->sm_port_idx ?
		p_vl->p_subn->sm_port_guid :
		p_vl->p_subn->opt.guid[p_vl->port_index];
	if (p_vl->thread_state == OSM_THREAD_STATE_NONE)
		p_vl->thread_state = OSM_THREAD_STATE_RUN;

	while (p_vl->thread_state == OSM_THREAD_STATE_RUN) {
		/*
		   Start servicing the FIFOs by pulling off MAD wrappers
		   and passing them to the transport interface.
		   There are lots of corner cases here so tread carefully.

		   The unicast FIFO has priority, since somebody is waiting
		   for a timely response.
		 */
		CL_SPINLOCK_ACQUIRE(&p_vl->lock);

		if (cl_qlist_count(&p_vl->hp_ufifo) != 0)
			p_fifo = &p_vl->hp_ufifo;
		else if (cl_qlist_count(&p_vl->ufifo) != 0)
			p_fifo = &p_vl->ufifo;
		else
			p_fifo = &p_vl->rfifo;

		p_madw = (osm_madw_t *) cl_qlist_remove_head(p_fifo);

		CL_SPINLOCK_RELEASE(&p_vl->lock);

		if (p_madw != (osm_madw_t *) cl_qlist_end(p_fifo)) {
			OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
				"Servicing p_madw = %p on port 0x%" PRIx64 "\n",
				p_madw, cl_ntoh64(port_guid));
			if (OSM_LOG_IS_ACTIVE_V2(p_vl->p_log, OSM_LOG_FRAMES))
				osm_dump_dr_smp_v2(p_vl->p_log,
						   osm_madw_get_smp_ptr(p_madw),
						   port_guid,
						   p_madw->dest_guid,
						   FILE_ID, OSM_LOG_FRAMES);

			vl15_send_mad(p_vl, p_madw);
		} else {
			/*
			   The VL15 FIFO is empty, so we have nothing left to do.
			 */
			status = cl_event_wait_on(&p_vl->signal,
						  EVENT_NO_TIMEOUT, TRUE);
			if (status != CL_SUCCESS) {
				OSM_LOG(p_vl->p_log, OSM_LOG_ERROR, "ERR 3E01: "
					"Event wait failed (%s)\n",
					CL_STATUS_MSG(status));
				continue;
			}
		}

		while (p_vl->p_stats->qp0_mads_outstanding_on_wire >= max_smps &&
		       p_vl->thread_state == OSM_THREAD_STATE_RUN) {

			CL_SPINLOCK_ACQUIRE(&p_vl->lock);
			if (cl_qlist_count(&p_vl->hp_ufifo) != 0 ||
			    cl_qlist_count(&p_vl->ufifo) != 0) {
				CL_SPINLOCK_RELEASE(&p_vl->lock);
				break;
			}
			CL_SPINLOCK_RELEASE(&p_vl->lock);

			status = cl_event_wait_on(&p_vl->signal,
						  p_vl->max_smps_timeout,
						  TRUE);
			if (status == CL_TIMEOUT) {
				if (max_smps < max_smps2)
					max_smps++;
				break;
			} else if (status != CL_SUCCESS) {
				OSM_LOG(p_vl->p_log, OSM_LOG_ERROR, "ERR 3E02: "
					"Event wait failed (%s)\n",
					CL_STATUS_MSG(status));
				break;
			}
			max_smps = p_vl->max_wire_smps;
			if (p_vl->p_stats->qp0_mads_outstanding_on_wire < max_smps)
				max_smps = p_vl->max_wire_smps;
		}
	}

	/*
	   since we abort immediately when the state != OSM_THREAD_STATE_RUN
	   we might have some mads on the queues. After the thread exits
	   the vl15 destroy routine should put these mads back...
	 */

	OSM_LOG_EXIT(p_vl->p_log);
}

void osm_vl15_construct(IN osm_vl15_t * p_vl)
{
	memset(p_vl, 0, sizeof(*p_vl));
	p_vl->state = OSM_VL15_STATE_INIT;
	p_vl->thread_state = OSM_THREAD_STATE_NONE;
	cl_event_construct(&p_vl->signal);
	cl_spinlock_construct(&p_vl->lock);
	cl_qlist_init(&p_vl->rfifo);
	cl_qlist_init(&p_vl->ufifo);
	cl_qlist_init(&p_vl->hp_ufifo);
	cl_thread_construct(&p_vl->poller);
}

void osm_super_vl15_construct(IN OUT osm_super_vl15_t * p_super_vl15,
			      IN osm_subn_opt_t * p_opt)
{
	uint8_t port_idx;

	OSM_ASSERT(p_opt);

	p_super_vl15->num_ports = p_opt->num_guids;

	for (port_idx = 0; port_idx < p_opt->num_guids; port_idx++)
		osm_vl15_construct(&p_super_vl15->vl15[port_idx]);
}

void osm_vl15_destroy(IN osm_vl15_t * p_vl, IN struct osm_mad_pool *p_pool)
{
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(p_vl->p_log);

	/*
	   Signal our threads that we're leaving.
	 */
	p_vl->thread_state = OSM_THREAD_STATE_EXIT;

	/*
	   Don't trigger unless event has been initialized.
	   Destroy the thread before we tear down the other objects.
	 */
	if (p_vl->state != OSM_VL15_STATE_INIT)
		cl_event_signal(&p_vl->signal);

	cl_thread_destroy(&p_vl->poller);

	/*
	   Return the outstanding messages to the pool
	 */

	CL_SPINLOCK_ACQUIRE(&p_vl->lock);

	while (!cl_is_qlist_empty(&p_vl->rfifo)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->rfifo);
		osm_mad_pool_put(p_pool, p_madw);
	}
	while (!cl_is_qlist_empty(&p_vl->ufifo)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->ufifo);
		osm_mad_pool_put(p_pool, p_madw);
	}
	while (!cl_is_qlist_empty(&p_vl->hp_ufifo)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->hp_ufifo);
		osm_mad_pool_put(p_pool, p_madw);
	}

	CL_SPINLOCK_RELEASE(&p_vl->lock);

	cl_event_destroy(&p_vl->signal);
	p_vl->state = OSM_VL15_STATE_INIT;
	cl_spinlock_destroy(&p_vl->lock);

	OSM_LOG_EXIT(p_vl->p_log);
}

void osm_super_vl15_destroy(IN osm_super_vl15_t * p_super_vl,
			    IN struct osm_mad_pool *p_pool)
{
	uint8_t port_idx;

	for (port_idx = 0; port_idx < p_super_vl->num_ports; port_idx++)
		osm_vl15_destroy(&p_super_vl->vl15[port_idx], p_pool);
}

ib_api_status_t osm_vl15_init(IN osm_vl15_t * p_vl, IN osm_vendor_t * p_vend,
			      IN osm_log_t * p_log, IN osm_stats_t * p_stats,
			      IN osm_subn_t * p_subn,
			      IN int32_t max_wire_smps,
			      IN int32_t max_wire_smps2,
			      IN uint32_t max_smps_timeout,
			      IN uint8_t port_index)
{
	cl_status_t status = CL_SUCCESS;

	OSM_LOG_ENTER(p_log);

	p_vl->p_vend = p_vend;
	p_vl->p_subn = p_subn;
	p_vl->p_log = p_log;
	p_vl->p_stats = p_stats;
	p_vl->max_wire_smps = max_wire_smps;
	p_vl->max_wire_smps2 = max_wire_smps2;
	p_vl->max_smps_timeout = max_wire_smps < max_wire_smps2 ?
				 max_smps_timeout : EVENT_NO_TIMEOUT;
	p_vl->port_index = port_index;

	status = cl_event_init(&p_vl->signal, FALSE);
	if (status != CL_SUCCESS)
		goto Exit;

	p_vl->state = OSM_VL15_STATE_READY;

	status = cl_spinlock_init(&p_vl->lock);
	if (status != CL_SUCCESS)
		goto Exit;

	/*
	   Initialize the thread after all other dependent objects
	   have been initialized.
	 */
	status = cl_thread_init(&p_vl->poller, vl15_poller, p_vl,
				"opensm poller");
Exit:
	OSM_LOG_EXIT(p_log);
	return (status == CL_SUCCESS) ? IB_SUCCESS : IB_UNKNOWN_ERROR;
}

ib_api_status_t osm_super_vl15_init(IN OUT osm_super_vl15_t * p_super_vl15,
				    IN osm_vendor_t * p_vend[],
				    IN osm_log_t * p_log,
				    IN osm_stats_t p_stats[],
				    IN osm_subn_t * p_subn,
				    IN const osm_subn_opt_t * p_opt)
{
	uint8_t port_idx;
	ib_api_status_t status = IB_SUCCESS;

	p_super_vl15->p_osm = p_subn->p_osm;

	for (port_idx = 0; port_idx < p_super_vl15->num_ports; port_idx++)
	{
		status = osm_vl15_init(&p_super_vl15->vl15[port_idx],
				       p_vend[port_idx],
				       p_log, &p_stats[port_idx],
				       p_subn, p_opt->max_wire_smps,
				       p_opt->max_wire_smps2,
				       p_opt->max_smps_timeout,
				       port_idx);
		if (status != IB_SUCCESS) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 3E05: "
				"VL15 initialization failed for port index %u "
				"port GUID 0x%" PRIx64 "\n",
				port_idx, cl_ntoh64(p_subn->opt.guid[port_idx]));
			break;
		}
	}
	return status;
}

void osm_vl15_poll(IN osm_vl15_t * p_vl, IN boolean_t force_flag)
{
	OSM_LOG_ENTER(p_vl->p_log);

	OSM_ASSERT(p_vl->state == OSM_VL15_STATE_READY);

	/*
	   If we have room for more VL15 MADs on the wire,
	   then signal the poller thread.

	   This is not an airtight check, since the poller thread
	   could be just about to send another MAD as we signal
	   the event here.  To cover this rare case, the poller
	   thread checks for a spurious wake-up.
	 */
	if (force_flag || (p_vl->p_stats->qp0_mads_outstanding_on_wire <
	    (int32_t) p_vl->max_wire_smps)) {
		OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
			"Signalling poller thread\n");
		cl_event_signal(&p_vl->signal);
	}

	OSM_LOG_EXIT(p_vl->p_log);
}

void osm_super_vl15_poll(IN osm_super_vl15_t * p_super_vl,
			 IN boolean_t force_flag,
			 uint8_t port_index)
{
	osm_vl15_poll(&p_super_vl->vl15[port_index], force_flag);
}

void osm_vl15_post_req_batch(IN osm_vl15_t * p_vl, IN cl_qlist_t * p_list)
{
	int32_t size;

	OSM_LOG_ENTER(p_vl->p_log);
	OSM_ASSERT(p_vl->state == OSM_VL15_STATE_READY);
	OSM_ASSERT(p_list);

	size = cl_qlist_count(p_list);

	if (!size)
		goto Exit;

	OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG, "Posting list of MADs - size %d \n", size);

	CL_SPINLOCK_ACQUIRE(&p_vl->lock);

	cl_qlist_insert_list_tail(&p_vl->rfifo, p_list);

	osm_stats_add_qp0_outstanding(p_vl->p_stats, size);

	CL_SPINLOCK_RELEASE(&p_vl->lock);

	osm_vl15_poll(p_vl, FALSE);
Exit:
	OSM_LOG_EXIT(p_vl->p_log);
}

void osm_super_vl15_post_req_batch(IN osm_super_vl15_t * p_super_vl,
				   IN cl_qlist_t * p_list)
{
	osm_vl15_post_req_batch(&p_super_vl->vl15[p_super_vl->p_osm->subn.sm_port_idx], p_list);
}

void osm_vl15_post(IN osm_vl15_t * p_vl, IN osm_madw_t * p_madw, boolean_t force_flag)
{
	OSM_LOG_ENTER(p_vl->p_log);

	OSM_ASSERT(p_vl->state == OSM_VL15_STATE_READY);

	OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG, "Posting p_madw = %p\n", p_madw);

	/*
	   Determine in which fifo to place the pending madw.
	 */
	CL_SPINLOCK_ACQUIRE(&p_vl->lock);
	if (p_madw->resp_expected == TRUE) {
		cl_qlist_insert_tail(&p_vl->rfifo, &p_madw->list_item);
		osm_stats_inc_qp0_outstanding(p_vl->p_stats);
	} else
		cl_qlist_insert_tail(force_flag ? &p_vl->hp_ufifo : &p_vl->ufifo,
				     &p_madw->list_item);
	CL_SPINLOCK_RELEASE(&p_vl->lock);

	OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
		"%u QP0 MADs on wire, %u QP0 MADs outstanding\n",
		p_vl->p_stats->qp0_mads_outstanding_on_wire,
		p_vl->p_stats->qp0_mads_outstanding);

	osm_vl15_poll(p_vl, force_flag);

	OSM_LOG_EXIT(p_vl->p_log);
}

static uint32_t vl15_get_load(IN osm_vl15_t * p_vl)
{
	/* currently we look only on rfifo */
	return cl_qlist_count(&p_vl->rfifo);
}

void osm_super_vl15_post(IN osm_super_vl15_t * p_super_vl,
			 IN osm_madw_t * p_madw,
			 boolean_t force_flag,
			 IN const osm_physp_t *p_physp,
			 IN boolean_t allow_load_balance)
{
	osm_subn_t *p_subn = &p_super_vl->p_osm->subn;
	uint32_t min_load = 0xFFFFFFFF, load;
	uint8_t port_idx, min_load_port_idx = p_subn->sm_port_idx;
	ib_smp_t *p_smp;
	osm_mad_addr_t *p_addr;
	boolean_t is_port_active;

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_addr = osm_madw_get_mad_addr_ptr(p_madw);
	if (p_physp && allow_load_balance) {
		ib_net64_t guid = osm_physp_get_port_guid(p_physp);
		for (port_idx = 0; port_idx < p_super_vl->num_ports; port_idx++) {
			load = vl15_get_load(&p_super_vl->vl15[port_idx]);
			is_port_active = p_subn->active_sm_guid[port_idx];
			/*
			 * check that the path is valid, i.e hop count > 0
			 * and also the port was binded and active
			 */
			if (!is_port_active)
				continue;

			/* if the destination is one of SM ports, use that port */
			if (guid == p_subn->opt.guid[port_idx] && guid) {
				min_load_port_idx = port_idx;
				break;
			}

			if (load < min_load &&
			    p_physp->dr_path[port_idx].hop_count) {
				min_load = load;
				min_load_port_idx = port_idx;
			}
		}
		/*
		 * if the chosen port is different than the main port,
		 * update the path and the bind
		 */
		if (min_load_port_idx != p_subn->sm_port_idx) {
			p_smp->hop_count = p_physp->dr_path[min_load_port_idx].hop_count;
			memcpy(&p_smp->initial_path,
			       p_physp->dr_path[min_load_port_idx].path,
			       sizeof(p_smp->initial_path));
			p_madw->h_bind =
				p_super_vl->p_osm->sm.mad_ctrl.h_bind[min_load_port_idx];
			p_madw->vend_wrap.h_bind =
				p_super_vl->p_osm->sm.mad_ctrl.h_bind[min_load_port_idx];

			/*
			 * If this is a planarized port, need to decide which port to use inside it.
			 * Physical port number is used by bind to specify the port.
			 */
			p_addr->addr_type.smi.port_num = p_physp->dr_path[min_load_port_idx].exit_port_num;
		}
	}
	
	/* post to the least loaded port, by default main port */
	osm_vl15_post(&p_super_vl->vl15[min_load_port_idx], p_madw, force_flag);
}

void osm_vl15_shutdown(IN osm_vl15_t * p_vl, IN osm_mad_pool_t * p_mad_pool)
{
	osm_madw_t *p_madw;

	OSM_LOG_ENTER(p_vl->p_log);

	/* we only should get here after the VL15 interface was initialized */
	OSM_ASSERT(p_vl->state == OSM_VL15_STATE_READY);

	/* grab a lock on the object */
	CL_SPINLOCK_ACQUIRE(&p_vl->lock);

	/* go over all outstanding MADs and retire their transactions */

	/* first we handle the list of response MADs */
	p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->hp_ufifo);
	while (p_madw != (osm_madw_t *) cl_qlist_end(&p_vl->hp_ufifo)) {
		OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
			"Releasing High Priority Response p_madw = %p\n", p_madw);

		osm_mad_pool_put(p_mad_pool, p_madw);

		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->hp_ufifo);
	}

	p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->ufifo);
	while (p_madw != (osm_madw_t *) cl_qlist_end(&p_vl->ufifo)) {
		OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
			"Releasing Response p_madw = %p\n", p_madw);

		osm_mad_pool_put(p_mad_pool, p_madw);

		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->ufifo);
	}

	/* Request MADs we send out */
	p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->rfifo);
	while (p_madw != (osm_madw_t *) cl_qlist_end(&p_vl->rfifo)) {
		OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
			"Releasing Request p_madw = %p\n", p_madw);

		osm_mad_pool_put(p_mad_pool, p_madw);
		osm_stats_dec_qp0_outstanding(p_vl->p_stats);

		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_vl->rfifo);
	}

	/* free the lock */
	CL_SPINLOCK_RELEASE(&p_vl->lock);

	OSM_LOG_EXIT(p_vl->p_log);
}

void osm_super_vl15_shutdown(IN osm_super_vl15_t * p_super_vl,
			     IN osm_mad_pool_t * p_mad_pool)
{
	uint8_t port_idx;

	for (port_idx = 0; port_idx < p_super_vl->num_ports; port_idx++)
		osm_vl15_shutdown(&p_super_vl->vl15[port_idx], p_mad_pool);
}

void osm_vl15_cancel_dr_request(IN osm_vl15_t * p_vl,
				IN osm_mad_pool_t * p_mad_pool,
				IN uint8_t hop_count,
				IN const uint8_t *initial_path)
{
	char str[IB_SUBNET_PATH_HOPS_MAX * 4] = {0};
	cl_list_item_t *p_item, *p_next_item;
	osm_madw_t *p_madw;
	ib_smp_t *p_smp;

	OSM_LOG_ENTER(p_vl->p_log);

	/* we only should get here after the VL15 interface was initialized */
	OSM_ASSERT(p_vl->state == OSM_VL15_STATE_READY);

	if (OSM_LOG_IS_ACTIVE_V2(p_vl->p_log, OSM_LOG_VERBOSE)) {
		sprint_uint8_arr(str, sizeof(str), initial_path, hop_count + 1);
		OSM_LOG(p_vl->p_log, OSM_LOG_VERBOSE,
			"Canceling queued requests to destination DR:%s, "
			"on port index %u\n",
			str, p_vl->port_index);
	}

	/* grab a lock on the object */
	CL_SPINLOCK_ACQUIRE(&p_vl->lock);

	p_next_item = cl_qlist_head(&p_vl->rfifo);
	while (p_next_item != cl_qlist_end(&p_vl->rfifo)) {
		p_item = p_next_item;
		p_next_item = cl_qlist_next(p_item);

		p_madw = (osm_madw_t *)p_item;
		p_smp = osm_madw_get_smp_ptr(p_madw);

		/* Cancel only SubnSet MADs */
		if (!p_smp ||
		    (p_smp->mgmt_class != IB_MCLASS_SUBN_DIR) ||
		    (p_smp->hop_count != hop_count) ||
		    (memcmp(p_smp->initial_path, initial_path, hop_count + 1)) ||
		    !(p_smp->method == IB_MAD_METHOD_SET &&
		      osm_sm_mad_is_subnet_mad(p_smp->attr_id)))
			continue;

		OSM_LOG(p_vl->p_log, OSM_LOG_DEBUG,
			"Canceling request p_madw = %p\n", p_madw);

		cl_qlist_remove_item(&p_vl->rfifo, p_item);
		osm_mad_pool_put(p_mad_pool, p_madw);
		osm_stats_dec_qp0_outstanding(p_vl->p_stats);
	}

	/* free the lock */
	CL_SPINLOCK_RELEASE(&p_vl->lock);

	OSM_LOG_EXIT(p_vl->p_log);
}

void osm_super_vl15_cancel_dr_request(IN osm_super_vl15_t * p_super_vl,
				      IN osm_mad_pool_t * p_mad_pool,
				      IN uint8_t hop_count,
				      IN const uint8_t *initial_path,
				      IN uint8_t port_index)
{
	osm_vl15_cancel_dr_request(&p_super_vl->vl15[port_index], p_mad_pool, hop_count,
				   initial_path);
}

